/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2012 by Chuck Chen <ycchen@realtek.com>
 *
 */

#include <common.h>
#include <asm/armv7.h>
#include <asm/sizes.h>
#include <config.h>
#include <asm/errno.h>
#include <asm/io.h>
#include <rbus/sys_reg_reg.h>
#include <rbus/pll_reg_reg.h>
#include <rbus/stb_reg.h>
#include <rbus/topbist_reg.h>
#include <rbus/vde_reg.h>
#include <rbus/scpu_core_reg.h>
#include <rbus/sb2_reg.h>
#include <rbus/emcu_reg.h>
#include <rbus/dc_sys_reg.h>
#include <asm/arch/system.h>
#include <asm/arch/fw_info.h>
#include <asm/arch/bootparam.h>

extern void enable_console(void);
extern void disable_console(void);
extern uint bIsSTR;
extern int boot_power_on_mode;
DECLARE_GLOBAL_DATA_PTR;

#ifndef CONFIG_BSP_REALTEK // mark unused code
unsigned int cortex_rev(void)
{

    unsigned int rev;

    /* Read Main ID Register (MIDR) */
    asm ("mrc p15, 0, %0, c0, c0, 0" : "=r" (rev));

    return rev;
}
#endif

unsigned int get_ic_version(void)
{
#define GET_IC_VERSION()	rtd_inl(STB_SC_VerID_reg)
#define VERSION_A_REG_CODE	0x65570000
#define VERSION_B_REG_CODE	0x65570010
#define VERSION_C_REG_CODE	0x65570020

	if (GET_IC_VERSION() == VERSION_A_REG_CODE)
	{
		return VERSION_A;
	}
	else if (GET_IC_VERSION() == VERSION_B_REG_CODE)
	{
		return VERSION_B;
	}
	else
	{
		return VERSION_C;
	}
}

/*
 *
 */
#define MAX_MDIO_TIMEOUT 		100
#define MIIAR           		0xb801605c
#define MDIO_WRITE			0x80000000
#define SET_FUSE_RDY_CHANGE_PAGE_A46	0x041f0a46
#define SET_FUSE_RDY_INIT_DONE_FLAG	0x04140003

unsigned int mdio_read_complete(unsigned int reg) {
	unsigned int iTimeoutCounter;
	unsigned int iRegVal;
	unsigned int value = 0;

	rtd_outl(MIIAR, (reg & ~MDIO_WRITE));
	for( iTimeoutCounter = 0; iTimeoutCounter < MAX_MDIO_TIMEOUT; iTimeoutCounter++ ) {
		iRegVal = rtd_inl(MIIAR);
		//printf("cnt=%02d : Reg(MIIAR)=0x%08x\n", iTimeoutCounter, iRegVal);
		if( (reg | MDIO_WRITE) == (iRegVal & 0xffff0000))  {
			value = iRegVal & 0x0000ffff;
			break;
		}
		udelay(10);
	}
	if( iTimeoutCounter == MAX_MDIO_TIMEOUT ) {
		printf("MISC : read MIIAR with %08x fail\n", reg);
	}

	return value;
}

void mdio_write_complete(unsigned int reg) {
	unsigned int iTimeoutCounter;
	unsigned int iRegVal;

	rtd_outl(MIIAR, (reg|MDIO_WRITE));
	for( iTimeoutCounter = 0; iTimeoutCounter < MAX_MDIO_TIMEOUT; iTimeoutCounter++ ) {
		iRegVal = rtd_inl(MIIAR);
		//printf("cnt=%02d : Reg(MIIAR)=0x%08x\n", iTimeoutCounter, iRegVal);
		if( iRegVal == (reg & ~MDIO_WRITE) ) {
			break;
		}
		udelay(10);
	}
	if( iTimeoutCounter == MAX_MDIO_TIMEOUT ) {
		printf("MISC : update MIIAR with %08x fail\n", reg);
	}
}

void getn_load_efusedata2DOUT(void){

	unsigned int val, val1, val2;
	unsigned int R_cal_value;
	unsigned int ib_up, ib_dn, amp_dn_fine;
#ifdef DBG
	unsigned int RC_cal_value, amp_dn;
#endif

/****************************************************************************************
	amp_cal value[23:16]= amp_dn_fine [1:0]port A ...[7:6]port D page bcc/reg20 [7:0]
	amp_cal value[15:8]= gdac_ib_dn page bcc/reg18 [15:8]
	amp_cal value[7:0]= gdac_ib_up page bcc/reg18 [7:0]
*******************************************************************************************/

	val1 = rtd_inl(0xb80381e4);
	val2 = rtd_inl(0xb80381e8);

	if ((val1&_BIT31)&&(val2&_BIT31)){//efuse have burned
		//R,RC can be overwrite
		mdio_write_complete(0x841f0a42);// change Page to a42
		mdio_write_complete(0x841b8013);//write reg27 = 0x8013
		mdio_write_complete(0x841c0f08);//wrtie reg28 bit[15]=0,let R/RC can be written over

		R_cal_value = ((val2>>15)&0xf);
		//RC_cal_value = ((val2>>19)&0xf);
#ifdef DBG
		amp_dn =  ((val2>>19)&0x7);
#endif
		amp_dn_fine = (val1&0xff);//amp_cal value[23:16]
		ib_dn = ((val1>>17)&0xff);//amp_cal value[15:8]
		ib_up = ((val1>>9)&0xff);//amp_cal value[7:0]
#ifdef DBG
		printf("R_cal_NOR value =%x\n",R_cal_value);
		//printf("RC_cal_NOR value =%x\n",RC_cal_value);
		printf("amp_dn_NOR =%x\n",amp_dn);
		printf("amp_dn_fine_NOR =%x\n",amp_dn_fine);
		printf("ib_dn_NOR =%x\n",ib_dn);
		printf("ib_up_NOR =%x\n",ib_up);
#endif
		/* write DOUT data */
		mdio_write_complete(0x841f0a46);// change Page to a46
		val = (((ib_dn&0x7f)<<9)|ib_up);
		mdio_write_complete((0x84100100)|val);// bit[15:9]=Fuse_DOUT[14:8](amp_cal value[14:8]),bit[7:0]=Fuse_DOUT[7:0](amp_cal value[7:0])
		mdio_read_complete(0x04100000);
		//printf("reg 16 =%x\n",rtd_inl(MIIAR));
		val = (amp_dn_fine<<1)|(ib_dn>>7);
		mdio_write_complete(0x8411fe00|val);// bit[8:1]=Fuse_DOUT[24:17](amp_cal value[23:16]),bit[0]=Fuse_DOUT[16](amp_cal value[15])
		mdio_read_complete(0x04110000);
		//printf("reg 17 =%x\n",rtd_inl(MIIAR));
		//printf("reg 17 =%x\n",rtd_inl(MIIAR));
		val = ((R_cal_value&0x1)<<15);
		mdio_write_complete(0x84127fff|val);// bit[15]=Fuse_DOUT[47](R value [0])
		mdio_read_complete(0x04120000);
		//printf("reg 18 =%x\n",rtd_inl(MIIAR));
		//val = (RC_cal_value<<3)|(R_cal_value>>1);
		//mdio_write_complete(0x8413ff80|val);// bit[6:3]=Fuse_DOUT[54:51](RC value),bit[2:0]=Fuse_DOUT[50:48](R value [3:1])
		val = (R_cal_value>>1);
		mdio_write_complete(0x8413fff8|val);// bit[6:3]=Fuse_DOUT[54:51](RC value),bit[2:0]=Fuse_DOUT[50:48](R value [3:1])
		mdio_read_complete(0x04130000);
		//printf("reg 19 =%x\n",rtd_inl(MIIAR));
	}
	else{
		mdio_write_complete(0x841f0a46);// change Page to a46
		ib_up =  0xaa;
		mdio_write_complete((0x8410ff00|ib_up));// bit[7:0]=Fuse_DOUT[7:0](amp_cal value[7:0])
	}

}

void greentable_modify(void){
	unsigned int val, val1, val2;
	unsigned int amp_dn;

	/* adjust common mode voltage */
	mdio_write_complete(0x841f0bc0);// change Page to bc0
	mdio_write_complete(0x84170001);// bit[1:0]:common mode voltage => 0x2->0x1

	val1 = rtd_inl(0xb80381e4);
	val2 = rtd_inl(0xb80381e8);

	if ((val1&_BIT31)&&(val2&_BIT31)){//efuse have burned
		amp_dn = ((val2>>19)&0x7);//from efuse [53:51]
		amp_dn = (amp_dn^0x7);
		amp_dn = (amp_dn<<9);
		printf("MISC: efuse had been written\n");
	}else{
		amp_dn = (0x2<<9); // bit[2:0]:amp_dn => 00:100% 01:90% 10:80%
		printf("MISC: efuse had NOT been written\n");
	}

	//printf("amp_dn = %x\n",(amp_dn>>9));
	/* grntbl_0_03 */
	mdio_write_complete(0x841b8079);
	val = mdio_read_complete(0x041c0000);
	val = ((val&0xf1ff)|amp_dn);		//amp_dn = bit[11:9]
	mdio_write_complete(0x841c0000|val);
	mdio_read_complete(0x041c0000);
	//printf("grntbl_0_03 =%x\n",rtd_inl(MIIAR));

	/* grntbl_1_03 */
	mdio_write_complete(0x841b8082);
	val = mdio_read_complete(0x041c0000);
	val = ((val&0xf1ff)|amp_dn);		//amp_dn = bit[11:9]
	mdio_write_complete(0x841c0000|val);
	mdio_read_complete(0x041c0000);
	//printf("grntbl_1_03 =%x\n",rtd_inl(MIIAR));

	/* grntbl_2_03 */
	mdio_write_complete(0x841b808B);
	val = mdio_read_complete(0x041c0000);
	val = ((val&0xf1ff)|amp_dn);		//amp_dn = bit[11:9]
	mdio_write_complete(0x841c0000|val);
	mdio_read_complete(0x041c0000);
	//printf("grntbl_2_03 =%x\n",rtd_inl(MIIAR));

	/* grntbl_3_03 */
	mdio_write_complete(0x841b8094);
	val = mdio_read_complete(0x041c0000);
	val = ((val&0xf1ff)|amp_dn);		//amp_dn = bit[11:9]
	mdio_write_complete(0x841c0000|val);
	mdio_read_complete(0x041c0000);
	//printf("grntbl_3_03 =%x\n",rtd_inl(MIIAR));

	/* grntbl_4_03 */
	mdio_write_complete(0x841b809d);
	val = mdio_read_complete(0x041c0000);
	val = ((val&0xf1ff)|amp_dn);		//amp_dn = bit[11:9]
	mdio_write_complete(0x841c0000|val);
	mdio_read_complete(0x041c0000);
	//printf("grntbl_4_03 =%x\n",rtd_inl(MIIAR));

	/* grntbl_5_03 */
	mdio_write_complete(0x841b80A6);
	val = mdio_read_complete(0x041c0000);
	val = ((val&0xf1ff)|amp_dn);		//amp_dn = bit[11:9]
	mdio_write_complete(0x841c0000|val);
	mdio_read_complete(0x041c0000);
	//printf("grntbl_5_03 =%x\n",rtd_inl(MIIAR));

	/* grntbl_6_03 */
	mdio_write_complete(0x841b80AF);
	val = mdio_read_complete(0x041c0000);
	val = ((val&0xf1ff)|amp_dn);		//amp_dn = bit[11:9]
	mdio_write_complete(0x841c0000|val);
	mdio_read_complete(0x041c0000);
	//printf("grntbl_6_03 =%x\n",rtd_inl(MIIAR));

#if 0 // get status later that can reduce about 60ms delay time
	mdelay(60);

	greentable_status();
#endif

}

void greentable_status(void)
{
#ifdef DBG
	unsigned int val;

	mdio_write_complete(0x841f0bcc);// change Page to bcc
	val = mdio_read_complete(0x04100000);
	printf(" amp_dn = %x\n",val&0x7);

	val = mdio_read_complete(0x04140000);
	printf(" amp_dn_fine = %x\n",val&0xff);

	val = mdio_read_complete(0x04120000);
	printf(" ib_up = %x\n",val&0xff);//bit[1:0]:up  => 00:100% 01:102.5% 10:105%
	printf(" ib_dn = %x\n",(val>>8)&0xff);

	mdio_write_complete(0x841f0bce);// change Page to bce
	val = mdio_read_complete(0x04100000);
	printf(" R = %x\n",val&0xffff);

	mdio_write_complete(0x841f0bcd);// change Page to bcd
	val = mdio_read_complete(0x04160000);
	printf(" RC = %x\n",val&0xffff);
#endif
}

#ifdef CONFIG_ARMV7
#define local_abt_enable()  __asm__("cpsie a	@ __sta" : : : "memory", "cc")
#define local_abt_disable() __asm__("cpsid a	@ __cla" : : : "memory", "cc")
#else
#define local_abt_enable()	do { } while (0)
#define local_abt_disable()	do { } while (0)
#endif

int rtk_rbus_timeout_thr_init(void)
{
             local_abt_enable(); 
	// use the maximum timeout value
	rtd_setbits(SB2_DEBUG_REG_reg, _BIT8 | _BIT7 | _BIT6 | _BIT5 | _BIT4 | _BIT3 | _BIT2 | _BIT1 | _BIT0);
	//clear Sciva_int
	rtd_outl(SB2_INV_INTSTAT_reg, _BIT1);

	//Enable scivairq_en, To_en, To_en1, To_end(disp-relative)
	rtd_outl(SB2_INV_INTEN_reg, _BIT8 | _BIT5 | _BIT4 | _BIT1 | _BIT0);

	//rtd_setbits(0xb8060400, _BIT6 | _BIT5 | _BIT2);

	//Enable EMCU rbus timeout, and TO cycle = 32768T (bit[7..5] = 0x6).
	/* NOTICE, EMCU TO cycles should be shorter than SB2, 
	 * otherwise SB2 could be disturbed by occurrence of both faked bus ACK. 
	 */
	rtd_setbits(EMCU_wrapper_8051_reg, _BIT7 | _BIT6 | _BIT2);

	printf("@sb2_intr_setup:    enable sb2/emcu interrupt@uboot\n");
    return 0;
}

void rtk_rbus_timeout_show(void)
{
        if(rtd_inl(SB2_INV_INTSTAT_reg ) & _BIT1)
                printf("You have accessed an reserved register address: %x\n", rtd_inl(SB2_INV_ADDR_reg));

        return;
}

int rtk_misc_init(void)
{
//	unsigned int iTimeoutCounter;
	unsigned int iRegVal;
#ifdef IS_FPGA
	printf("fpga skip\n");
	return 0;
#endif
	getn_load_efusedata2DOUT();

	iRegVal = rtd_inl(MIIAR);
	if( iRegVal == SET_FUSE_RDY_INIT_DONE_FLAG ) {
		return 0;
	}

	printf("MISC : set FUSE_RDY\n");

	//set_fuse_rdy.config
	mdio_write_complete(SET_FUSE_RDY_CHANGE_PAGE_A46); // change Page to a46
	mdio_write_complete(SET_FUSE_RDY_INIT_DONE_FLAG); // set fuse_rdy = 1 , set ext_ini_done = 1, to skip ext_ini

	greentable_modify();

	return 0;
}

#ifdef ENABLE_VE_BISR
#define ve_bisr_repair_addr     0x1800fff0
#define ve_repair_check_bits    (_BIT31)
#define ve_VCPU_bisr_fail_g_reg 0x1800ff2c

int do_ve_bisr(void)
{
        unsigned int iMaxCount = 1000;
        unsigned bit_flags;
        int ret = 0;

        printf("VE: BISR testing ....\n");

	rtd_setbits(SYS_REG_SYS_CLKSEL_reg, _BIT28);
        bit_flags = _BIT0;
        rtd_setbits(TOPBIST_ve_bisr_rstn_reg, bit_flags); // release bisr rstn
        while((rtd_inl(TOPBIST_ve_bisr_done_reg) & bit_flags) != bit_flags) { //wait bist done
                iMaxCount--;
                udelay(100);
                if( iMaxCount == 0 ) {
                        printf("VE: BISR is no done ....\n");
                        ret = -1;
                }
        }
        rtd_setbits(TOPBIST_ve_bisr_remap_reg, bit_flags); // assert hold remap

        if((rtd_inl(ve_bisr_repair_addr) & ve_repair_check_bits) != 0) {
                printf("VE: Need to repair\n");
        }

        bit_flags = _BIT0;

        if (rtd_inl(ve_VCPU_bisr_fail_g_reg) & bit_flags)
        {
                printf("VE BISR status: failed %x\n", rtd_inl(ve_VCPU_bisr_fail_g_reg) & bit_flags);
                ret = -1;
        }
        else
                printf("VE BISR status: ok\n");

        return ret;
}
#endif


#ifdef ENABLE_GPU_BISR
#define gpu_force_all_sram_clken 0x1810E090
#define gpu_bisr_repair_addr     0x1810E0A8
#define gpu_repair_check_bits    (_BIT31)
#define gpu_bisr_fail_g_reg      0x1810E0B0
#define SYS_CLKEN1_reg    0xb8000114
#define SYS_SRST1_reg     0xb8000104
#define REGS_GPU_WRAPPER_RESET  (0xb810F000)


int rtk_gpu_bisr(void)
{

  unsigned int iMaxCount = 1000;
  unsigned bit_flags;

        printf("GPU: BISR testing ....\n");

	//bit 0 is write data
	//CLKEN = 0
	rtd_outl(SYS_CLKEN1_reg, 0x1<<3 | 0x0); 
	udelay(100);
	//RSTN = 0
	rtd_outl(SYS_SRST1_reg,  0x1<<3 | 0x0); 
	udelay(100);
	//RSTN = 1
	rtd_outl(SYS_SRST1_reg,  0x1<<3 | 0x1); 
	udelay(100);
	//CLKEN = 1
	rtd_outl(SYS_CLKEN1_reg, 0x1<<3 | 0x1); 
	udelay(100);
	// Release GPU core reset	
	rtd_outl(REGS_GPU_WRAPPER_RESET,0x1);
		
    //force enable clk to gpu sram
    rtd_setbits(gpu_force_all_sram_clken, _BIT31);

        bit_flags = _BIT0;
        rtd_setbits(TOPBIST_cpu_bisr_rstn_reg, bit_flags); // release bisr rstn
        while((rtd_inl(TOPBIST_cpu_bisr_done_reg) & bit_flags) != bit_flags) { //wait bist done
                iMaxCount--;
                udelay(100);
                if( iMaxCount == 0 ) {
                        break;
                }
        }
        rtd_setbits(TOPBIST_cpu_bisr_remap_reg, bit_flags); // assert hold remap

        if((rtd_inl(gpu_bisr_repair_addr) & gpu_repair_check_bits) != 0) {
                printf("GPU: Need to repair\n");
        }

        bit_flags = _BIT0 | _BIT1 | _BIT2 | _BIT3;

        if (rtd_inl(gpu_bisr_fail_g_reg) & bit_flags)
                printf("GPU BISR status: failed %x\n", rtd_inl(gpu_bisr_fail_g_reg) & bit_flags);
        else
                printf("GPU BISR status: ok\n");

      //force disable clk to gpu sram
      rtd_clearbits(gpu_force_all_sram_clken, _BIT31);

      //CLKEN = 0
      //rtd_outl(SYS_CLKEN1_reg, 0x1<<3 | 0x0); 
  
        return 0;

}
#endif

#ifdef ENABLE_DISPI_BISR
int rtk_dispi_bisr(void)
{
#define dispi_rstn_bits 	(_BIT0)
#define dispi_remap_bits 	(_BIT0)
#define dispi_done_bits 	(_BIT0)

#define VODMA_VODMA_CLKGEN_reg                               		(0xB800560c)
//#define VODMA2_VODMA2_CLKGEN_reg                               		(0xB8005e0c)
#define VGIP_VGIP_CHN1_CTRL_reg                                		(0xB8022210)
#define VGIP2_MAIN_2_VGIP_CHN1_CTRL_VADDR          			(0xB8022710)
#define  VGIP_Data_Path_Select_reg                                             (0xB8022284)

#define dispi_bisr_repair_addr (0xb8025148)
#define dispi_repair_check_bits (_BIT31)
#define dispi_bisr_fail_addr 	(0xb8025f24)
#define dispi_fail_bits 	(_BIT0)


	unsigned int iMaxCount = 100;
	unsigned rstn_bits = dispi_rstn_bits;
	unsigned done_bits = dispi_done_bits;
	unsigned remap_bits = dispi_remap_bits;
	unsigned repair_check_bits = dispi_repair_check_bits;
	unsigned fail_bits = dispi_fail_bits;

	printf("\n DISPI: BISR testing ....\n");
		//modules reset de-assert
	rtd_outl(PLL_REG_SYS_PLL_VODMA2_reg,0x00000004); // PLLVODMA power down, hold reset, output disable
	rtd_outl(PLL_REG_SYS_PLL_VODMA2_reg,0x00000005); // PLLVODMA power on
	rtd_outl(PLL_REG_SYS_PLL_VODMA2_reg,0x00000007); // release PLLVODMA reset
	udelay(10); // at least > 150us
	rtd_outl(PLL_REG_SYS_PLL_VODMA2_reg,0x00000003); // PLLVODMA output enable
	//"CLKEN=0 -> RSTN=0 -> CLKEN=1 -> CLKEN=0 -> RSTN=1 -> CLKEN=1"
	rtd_outl(SYS_REG_SYS_SRST1_reg, _BIT2);	  // hold VODMA reset
	rtd_outl(SYS_REG_SYS_CLKEN1_reg,  _BIT2 | _BIT0); 	   // VODMA clk enable
	rtd_outl(SYS_REG_SYS_CLKEN1_reg, _BIT2);	  // VODMA clk disable
	rtd_outl(SYS_REG_SYS_SRST1_reg, _BIT2| _BIT0); 	   // release VODMA reset
	rtd_outl(SYS_REG_SYS_CLKEN1_reg, _BIT2 | _BIT0); 	   // VODMA clk enable

	//I domain clk setting
	rtd_maskl(VODMA_VODMA_CLKGEN_reg, ~_BIT31, _BIT31); // [31]: VODMA CLKEN, MAIN
	//rtd_maskl(VODMA2_VODMA2_CLKGEN_reg, ~_BIT31, _BIT31); // [31]: VODMA CLKEN, MAIN
	rtd_maskl(VGIP_VGIP_CHN1_CTRL_reg, ~(_BIT30|_BIT0), (_BIT30|_BIT0)); // vgip ch1 iclk enable, select to VO1:4
	rtd_maskl(VGIP2_MAIN_2_VGIP_CHN1_CTRL_VADDR, ~(_BIT30|_BIT0), (_BIT30|_BIT0)); // vgip ch2 iclk enable, select to VO2:5
	rtd_maskl(VGIP_Data_Path_Select_reg, ~(_BIT13|_BIT12), (_BIT13|_BIT12)); // Enable main/sub SDNR clock
	
	rtd_setbits(TOPBIST_dispi_bisr_rstn_reg, rstn_bits);
	udelay(100);
	while((rtd_inl(TOPBIST_dispi_bisr_done_reg) & done_bits) != done_bits) {
		iMaxCount--;
		udelay(100);
		if( iMaxCount == 0 ) {
			if(bIsSTR)
			{
				enable_console();
			}
			else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
			{
				enable_console();
			}

			printf("\n\033[31m[ERR] DISPI BISR DOING FAIL!!!!! \033[0m\n");

			if(bIsSTR)
			{
				disable_console();
			}
			else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
			{
				disable_console();
			}
			break;
		}
	}

	rtd_setbits(TOPBIST_dispi_bisr_remap_reg, remap_bits);
	if(((rtd_inl(dispi_bisr_repair_addr) & repair_check_bits) != 0)) {
		printf("DISPI: Need to repair\n");
	}
	if((rtd_inl(dispi_bisr_fail_addr) & fail_bits) != 0) {
		printf("DISPI: BISR fail!\n");
	}

	printf("DISPI: Finish BISR testing %s\n\n", (iMaxCount ? "done" : "fail" ) );

	if( iMaxCount == 0) {
		return -1;
	}

	return 0;
}
#endif

#ifdef ENABLE_DISPD_BISR

int rtk_dispd_bisr(void)
{
#define dispd_rstn_bits 	(_BIT0)
#define dispd_remap_bits 	(_BIT0)
#define dispd_done_bits 	(_BIT0)
#define dispd2_rstn_bits 	(_BIT6)
#define dispd2_remap_bits (_BIT6)
#define dispd2_done_bits 	(_BIT6)
#define dispd_bisr_repair_addr (0xb802A7DC)
#define dispd_repair_check_bits (_BIT31)
#define dispd_bisr_repair_addr2 (0xb802d9c8)
#define dispd_repair_check_bits2 (_BIT31)
//#define dispd_bisr_repair_addr3 (0xb802B94C)
//#define dispd_repair_check_bits3 (_BIT0)
//#define dispd_bisr_repair_addr4 (0xb802A7DC)
//#define dispd_repair_check_bits4 (_BIT31|_BIT30)
#define dispd_bisr_fail_addr 	(0xb8029fe4)
#define dispd_fail_bits 	(_BIT0)
#define dispd_bisr_fail_addr2 	(0xb802d9c4)
#define dispd_fail_bits2 	(_BIT0)
//#define dispd_bisr_fail_addr3 	(0xb8029f44)
//#define dispd_fail_bits3 	(_BIT0)
//#define dispd_bisr_fail_addr4 	(0xb8029fe4)
//#define dispd_fail_bits4 	(_BIT1|_BIT0)


	unsigned int iMaxCount = 100;
	unsigned rstn_bits = dispd_rstn_bits;
	unsigned done_bits = dispd_done_bits;
	unsigned remap_bits = dispd_remap_bits;
	unsigned repair_check_bits = dispd_repair_check_bits;
	unsigned repair_check_bits2 = dispd_repair_check_bits2;
	//unsigned repair_check_bits3 = dispd_repair_check_bits;
	//unsigned repair_check_bits4 = dispd_repair_check_bits2;
	unsigned fail_bits = dispd_fail_bits;
	unsigned fail_bits2 = dispd_fail_bits2;
	//unsigned fail_bits3 = dispd_fail_bits3;
	//unsigned fail_bits4 = dispd_fail_bits4;

	printf("\n DISPD: BISR testing ....\n");
	//enable Display Domain SFG Path Enable, Display Domain GDMA Path Enable, Display Domain STAGE1  Path Enable, Display Domain OSD Enable 
	rtd_maskl(SYS_REG_SYS_DISPCLKSEL_reg, ~(_BIT4|_BIT3|_BIT2|_BIT1), (_BIT4|_BIT3|_BIT2|_BIT1));

	rtd_outl(0xb800011c,0x00404001);// enable bist clk ,lg pod2/lg_hcic
	rtd_outl(0xb800011c,0x00404000);// disable bist clk ,lg pod2/lg_hcic
	rtd_outl(0xb800010c,0x00404001);// release bist rst ,lg pod2/lg_hcic
	rtd_outl(0xb800011c,0x00404001);// enable bist clk ,lg pod2/lg_hcic
	
	rtd_clearbits(0xb8028028, _BIT10);//clear double buffer reg
	rtd_maskl(0xb8028004, 0xFF7FFFFF, 0x800000); //[23]=1 POD2 enable

	rtd_setbits(TOPBIST_fcic_bisr_rstn_reg, rstn_bits);
	rtd_setbits(TOPBIST_dispd_bisr_rstn_reg, dispd2_rstn_bits);
	udelay(100);
	while(((rtd_inl(TOPBIST_fcic_bisr_done_reg) & done_bits) != done_bits)||((rtd_inl(TOPBIST_dispd_bisr_done_reg) & dispd2_done_bits) != dispd2_done_bits)) {
		iMaxCount--;
		udelay(100);
		if( iMaxCount == 0 ) {
			if(bIsSTR)
			{
				enable_console();
			}
			else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
			{
				enable_console();
			}

			printf("\n\033[31m[ERR] DISPD BISR DOING FAIL!!!!! \033[0m\n");

			if(bIsSTR)
			{
				disable_console();
			}
			else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
			{
				disable_console();
			}
			break;
		}
	}
	
	rtd_setbits(TOPBIST_fcic_bisr_remap_reg, remap_bits);
	rtd_setbits(TOPBIST_dispd_bisr_remap_reg, dispd2_remap_bits);

	if(((rtd_inl(dispd_bisr_repair_addr) & repair_check_bits) != 0)||((rtd_inl(dispd_bisr_repair_addr2) & repair_check_bits2) != 0)
		 /*||((rtd_inl(dispd_bisr_repair_addr3) & repair_check_bits3) != 0)||((rtd_inl(dispd_bisr_repair_addr4) & repair_check_bits4) != 0)*/) {
		printf("DISPD: Need to repair\n");
	}

	if(((rtd_inl(dispd_bisr_fail_addr) & fail_bits) != 0)||((rtd_inl(dispd_bisr_fail_addr2) & fail_bits2) != 0)/* ||
		((rtd_inl(dispd_bisr_fail_addr3) & fail_bits3) != 0)||((rtd_inl(dispd_bisr_fail_addr4) & fail_bits4) != 0)*/) {
		printf("DISPD: BISR fail!\n");
	}

	printf("DISPD: Finish BISR testing %s\n\n", (iMaxCount ? "done" : "fail" ) );

	if( iMaxCount == 0) {
		return -1;
	}

	return 0;
}
#endif

#ifdef ENABLE_KME_BISR

int rtk_kme_bisr(void)
{
#if 0
#define kme_rstn_bits 	(_BIT0)
#define kme_remap_bits 	(_BIT0)
#define kme_done_bits 	(_BIT0)
#if 1
#define kme_bisr_repair_addr (0xb802A7DC)
#define kme_repair_check_bits (_BIT31)
#endif
#define kme_bisr_fail_addr 	(0xb8029FE4)
#define kme_fail_bits 	(_BIT0)

			unsigned int iMaxCount = 100;
			unsigned rstn_bits = kme_rstn_bits;
			unsigned done_bits = kme_done_bits;
			unsigned remap_bits = kme_remap_bits;
//			unsigned repair_check_bits = kme_repair_check_bits;
			unsigned fail_bits = kme_fail_bits;

			printf("\n KME: BISR testing ....\n");
			rtd_outl(SYS_REG_SYS_SRST3_reg, _BIT29 | _BIT0);	//enable RBUS clk of MEMC reset reg
			rtd_outl(SYS_REG_SYS_CLKEN3_reg, _BIT29 | _BIT0);	  //enable RBUS clk of MEMC
			rtd_outl(SYS_REG_SYS_CLKEN3_reg, _BIT31 | _BIT30);	  //disable CLKEN_MEMC/CLKEN_MEME_ME
			rtd_outl(SYS_REG_SYS_SRST3_reg, _BIT31 | _BIT30);//reset of memc
			rtd_outl(SYS_REG_SYS_SRST3_reg, _BIT31 | _BIT30 | _BIT0);//release reset of memc
			rtd_outl(SYS_REG_SYS_CLKEN3_reg, _BIT31 | _BIT30 | _BIT0);	  //enable CLKEN_MEMC/CLKEN_MEME_ME
			rtd_maskl(SYS_REG_SYS_DISPCLKSEL_reg, ~(_BIT28|_BIT27|_BIT26), 0);	  //dispd-clock div 2

			rtd_setbits(TOPBIST_fcic_bisr_rstn_reg, rstn_bits);
			udelay(100);
			while((rtd_inl(TOPBIST_fcic_bisr_done_reg) & done_bits) != done_bits) {
				iMaxCount--;
				udelay(100);
				if( iMaxCount == 0 ) {
					break;
				}
			}
			rtd_setbits(TOPBIST_fcic_bisr_remap_reg, remap_bits);
#if 1
			if((rtd_inl(kme_bisr_repair_addr) & kme_repair_check_bits) != 0) {
				printf("KME: Need to repair\n");
			}
#endif

			if((rtd_inl(kme_bisr_fail_addr) & fail_bits) != 0) {
				printf("KME: BISR fail!\n");
			}

			printf("KME: Finish BISR testing %s\n\n", (iMaxCount ? "done" : "fail" ) );

			if( iMaxCount == 0) {
				return -1;
			}

			return 0;
#else
	return 0; //merlin4 no Kme bisr
#endif

}
#endif


#ifdef ENABLE_KMC_BISR
int rtk_kmc_bisr(void)
{
#define kmc_rstn_bits 	(_BIT21)
#define kmc_remap_bits 	(_BIT0)
#define kmc_done_bits 	(_BIT0)
#define kmc_bisr_repair_addr (0xb8099098)
#define kmc_repair_check_bits (_BIT0)
#define kmc_bisr_fail_addr 	(0xb8099098)
#define kmc_fail_bits 	(_BIT0)
#define kmc_bisr_retn_reg		(0xb809f100)

#define kmc_dma_w1_reg		(0xb8099460)
#define kmc_dma_w2_reg		(0xb80994dc)

#define kmc_dma_r1_reg		(0xb809ae1c)
#define kmc_dma_r2_reg		(0xb809ae3c)
#define kmc_dma_r3_reg		(0xb809ae5c)
#define kmc_dma_r4_reg		(0xb809ae7c)
#define kmc_DMA_Disable_bits 	(0x00000300)

		unsigned int iMaxCount = 100;
		unsigned rstn_bits = kmc_rstn_bits;
		unsigned done_bits = kmc_done_bits;
		unsigned remap_bits = kmc_remap_bits;
//		unsigned repair_check_bits = kmc_repair_check_bits;
		unsigned fail_bits = kmc_fail_bits;



		printf("\n KMC: BISR testing ....\n");

		rtd_maskl(SYS_REG_SYS_DISPCLKSEL_reg, ~(_BIT6|_BIT5), (_BIT6|_BIT5));	  //enable memc input/output clock
		rtd_outl(SYS_REG_SYS_SRST3_reg, _BIT29 | _BIT0);	//enable RBUS clk of MEMC reset reg
		rtd_outl(SYS_REG_SYS_CLKEN3_reg, _BIT29 | _BIT0);	  //enable RBUS clk of MEMC

		rtd_outl(kmc_dma_w1_reg,kmc_DMA_Disable_bits); //dma_off mc w
		rtd_outl(kmc_dma_w2_reg,kmc_DMA_Disable_bits); //dma_off mc w
		rtd_outl(kmc_dma_r1_reg,kmc_DMA_Disable_bits); //dma_off mc r
		rtd_outl(kmc_dma_r2_reg,kmc_DMA_Disable_bits); //dma_off mc r
		rtd_outl(kmc_dma_r3_reg,kmc_DMA_Disable_bits); //dma_off mc r
		rtd_outl(kmc_dma_r4_reg,kmc_DMA_Disable_bits); //dma_off mc r
		
		rtd_outl(SYS_REG_SYS_CLKEN3_reg, _BIT31 | _BIT30);	  //disable CLKEN_MEMC/CLKEN_MEME_ME
		rtd_outl(SYS_REG_SYS_SRST3_reg, _BIT31 | _BIT30);//reset of memc
		rtd_outl(SYS_REG_SYS_SRST3_reg, _BIT31 | _BIT30 | _BIT0);//release reset of memc
		rtd_outl(SYS_REG_SYS_CLKEN3_reg, _BIT31 | _BIT30 | _BIT0);	  //enable CLKEN_MEMC/CLKEN_MEME_ME
		
		rtd_maskl(SYS_REG_SYS_DISPCLKSEL_reg, ~(_BIT28|_BIT27|_BIT26), 0);	  //dispd-clock div 2

		rtd_setbits(kmc_bisr_retn_reg, rstn_bits);
		udelay(100);
		while((rtd_inl(TOPBIST_memc_bisr_done_reg) & done_bits) != done_bits) {
			iMaxCount--;
			udelay(100);
			if( iMaxCount == 0 ) {
				if(bIsSTR)
				{
					enable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					enable_console();
				}

				printf("\n\033[31m[ERR] KMC BISR DOING FAIL!!!!! \033[0m\n");

				if(bIsSTR)
				{
					disable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					disable_console();
				}
				break;
			}
		}
		rtd_setbits(TOPBIST_memc_bisr_remap_reg, remap_bits);
		if((rtd_inl(kmc_bisr_repair_addr) & kmc_repair_check_bits) != 0) {
			printf("KMC: Need to repair\n");
		}

		if((rtd_inl(kmc_bisr_fail_addr) & fail_bits) != 0) {
			printf("KMC: BISR fail!\n");
		}

		printf("KMC: Finish BISR testing %s\n\n", (iMaxCount ? "done" : "fail" ) );

		if( iMaxCount == 0) {
			return -1;
		}

		return 0;

}

#endif

#ifdef ENABLE_SCPU1_BISR
int rtk_scpu1_bisr(void)
{
#if 0
	if(rtd_inl(TOPBIST_cpu_bisr_remap_reg) & _BIT0)
		printf("SCPU1: BISR hold remap\n");

	// SCPU2 MBISR
	if (rtd_inl(0xb805b464) & 0x003FC000) // bit 14~21
		printf("SCPU1: BISR failed %x\n", rtd_inl(0xb805b464) & 0x003FC000);
	else
#endif
		printf("SCPU1: BISR ok\n");
	return 0;
}
#endif	//ENABLE_SCPU1_BISR

#ifdef ENABLE_SCPU2_BISR
int rtk_scpu2_bisr(void)
{
#if 1
#if 0
	unsigned int status;

	if(rtd_inl(TOPBIST_cpu_bisr_remap_reg) & _BIT5)
		printf("SCPU2: BISR hold remap\n");

	status = rtd_inl(0xb805b460);
	// SCPU2 MBISR
	if (status & 0x003FC000) {	// bit 14~21
		printf("SCPU2: BISR failed %x\n", status & 0x003FC000);
		printf("System stop booting because BISR failure!\n\n");
		while(1);	//stop booting
	}
	else if (status & _BIT13)
		printf("SCPU2: BISR ok, but repaired! %x\n\n", status);
	else

		printf("SCPU2: BISR ok %x\n", status);

#endif
#else
	unsigned int iMaxCount = 1000;
	unsigned bit_flags;
	unsigned int run_times;

	printf("SCPU2: BISR testing\n");

	bit_flags = (_BIT5); // A53
	rtd_setbits( TOP_BIST_cpu_bisr_rstn_reg, bit_flags); // release bisr rstn
	while((rtd_inl(TOP_BIST_cpu_bisr_done_reg) & bit_flags) != bit_flags){ //wait bist done
		iMaxCount--;
		udelay(100);
		if( iMaxCount == 0 ) {
			printf("SCPU2: BISR timeout %x\n", rtd_inl(TOP_BIST_cpu_bisr_done_reg));
			break;
		}
	}

	if (iMaxCount != 0) // BISR timeout cannot assert hold remap
	{
		rtd_setbits(TOP_BIST_cpu_bisr_remap_reg, bit_flags); // assert hold remap
		printf("SCPU2: BISR hold remap\n");
	}

	// SCPU2 MBISR
	if (rtd_inl(0xb805b460) & 0x003FC000) // bit 14~21
		printf("SCPU2: BISR failed %x\n", rtd_inl(0xb805b460) & 0x003FC000);
	else
		printf("SCPU2: BISR ok\n");
#endif
	return 0;
}
#endif	//ENABLE_SCPU2_BISR

#ifdef ENABLE_VCPU_BISR
#define vcpu1_bisr_repair_addr     0x18001130
#ifdef CONFIG_VCPU2_SUPPORT
#define vcpu2_bisr_repair_addr     0x18037130
#endif
#define vcpu_repair_check_bits    (_BIT31)

/*
 * cpuid:
 *      bit0 -> vcpu1
 *      bit1 -> vcpu2
 */
int do_vcpu_bisr(unsigned int vcpuid)
{
	unsigned int iMaxCount = 10000;
	unsigned bit_flags;
//	unsigned int run_times;

	if (vcpuid & _BIT0) // VCPU1
	{
		printf("VCPU1: BISR testing ....\n");

		rtd_setbits(TOPBIST_cpu_bisr_rstn_reg, _BIT16);
		bit_flags = _BIT1;
		rtd_setbits(TOPBIST_cpu_bisr_rstn_reg, bit_flags); // release bisr rstn
		while((rtd_inl(TOPBIST_cpu_bisr_done_reg) & bit_flags) != bit_flags) { //wait bist done
			iMaxCount--;
			udelay(100);
			if( iMaxCount == 0 ) 
			{
				if(bIsSTR)
				{
					enable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					enable_console();
				}

				printf("\n\033[31m[ERR] VCPU1 BISR DOING FAIL!!!!! \033[0m\n");

				if(bIsSTR)
				{
					disable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					disable_console();
				}
			}
		}
		rtd_setbits(TOPBIST_cpu_bisr_remap_reg, bit_flags); // assert hold remap

                if((rtd_inl(vcpu1_bisr_repair_addr) & vcpu_repair_check_bits) != 0) {
                    printf("VCPU1: Need to repair\n");
                }

		bit_flags = _BIT0;

		if (rtd_inl(VDE_VCPU_bisr_fail_g_reg) & bit_flags)
			printf("VCPU1 BISR status: failed %x\n", rtd_inl(VDE_VCPU_bisr_fail_g_reg) & bit_flags);
		else
			printf("VCPU1 BISR status: ok\n");
	}

#ifdef CONFIG_VCPU2_SUPPORT
	if (vcpuid & _BIT1) // VCPU2
	{
		printf("VCPU2: BISR testing ....\n");

		iMaxCount = 10000;
		bit_flags = _BIT2;
		rtd_setbits(TOPBIST_cpu_bisr_rstn_reg, bit_flags); // release bisr rstn
		while((rtd_inl(TOPBIST_cpu_bisr_done_reg) & bit_flags) != bit_flags) { //wait bist done
			iMaxCount--;
			udelay(100);
			if( iMaxCount == 0 ) 
			{
				if(bIsSTR)
				{
					enable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					enable_console();
				}

				printf("\n\033[31m[ERR] VCPU2 BISR DOING FAIL!!!!! \033[0m\n");

				if(bIsSTR)
				{
					disable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					disable_console();
				}
			}
		}
		rtd_setbits(TOPBIST_cpu_bisr_remap_reg, bit_flags); // assert hold remap
                if((rtd_inl(vcpu2_bisr_repair_addr) & vcpu_repair_check_bits) != 0) {
                    printf("VCPU2: Need to repair\n");
                }

		bit_flags = _BIT0;

		if (rtd_inl(0xb8037114) & bit_flags)
			printf("VCPU2 BISR status: failed %x\n", rtd_inl(0xb8037114) & bit_flags);
		else
			printf("VCPU2 BISR status: ok\n");
	}
#endif

	return 0;
}
#endif	//ENABLE_VCPU_BISR

#ifdef ENABLE_ACPU_BISR
#define acpu1_bisr_repair_addr    0x18002258
#ifdef CONFIG_ACPU2_SUPPORT
#define acpu2_bisr_repair_addr    0x1803A258
#endif
#define acpu_repair_check_bits   (_BIT31|_BIT30|_BIT29|_BIT28)
/*
 * cpuid:
 *      bit0 -> acpu1
 *      bit1 -> acpu2
 */
int do_acpu_bisr(unsigned int acpuid)
{
	unsigned int iMaxCount = 10000;
	unsigned bit_flags;
//	unsigned int run_times;

	if (acpuid & _BIT0) //ACPU1
	{
		printf("ACPU1: BISR testing ....\n");

		rtd_setbits(TOPBIST_cpu_bisr_rstn_reg, _BIT17 );
		bit_flags = _BIT13 | _BIT14 | _BIT15;
		rtd_setbits(TOPBIST_cpu_bisr_rstn_reg, bit_flags); // release bisr rstn
		while((rtd_inl(TOPBIST_cpu_bisr_done_reg) & bit_flags) != bit_flags) { //wait bist done
			iMaxCount--;
			udelay(100);
			if( iMaxCount == 0 ) 
			{
				if(bIsSTR)
				{
					enable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					enable_console();
				}

				printf("\n\033[31m[ERR] ACPU1 BISR DOING FAIL!!!!! \033[0m\n");

				if(bIsSTR)
				{
					disable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					disable_console();
				}
			}
		}
		rtd_setbits(TOPBIST_cpu_bisr_remap_reg, bit_flags); // assert hold remap

                if((rtd_inl(acpu1_bisr_repair_addr) & acpu_repair_check_bits) != 0) {
                    printf("ACPU1: Need to repair\n");
                }
		bit_flags = _BIT1 | _BIT2 | _BIT3;

		if (rtd_inl(0xb800223c) & bit_flags)
			printf("ACPU1 BISR status: failed %x\n", rtd_inl(0xb800223c) & bit_flags);
		else
			printf("ACPU1 BISR status: ok\n");
	}

#ifdef CONFIG_ACPU2_SUPPORT
	if (acpuid & _BIT1) //ACPU2
	{
		printf("ACPU2: BISR testing ....\n");

		iMaxCount = 10000;
		bit_flags = _BIT16 | _BIT17 | _BIT18 | _BIT19;
		rtd_setbits(TOPBIST_cpu_bisr_rstn_reg, bit_flags); // release bisr rstn
		while((rtd_inl(TOPBIST_cpu_bisr_done_reg) & bit_flags) != bit_flags) { //wait bist done
			iMaxCount--;
			udelay(100);
			if( iMaxCount == 0 ) 
			{
				if(bIsSTR)
				{
					enable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					enable_console();
				}

				printf("\n\033[31m[ERR] ACPU2 BISR DOING FAIL!!!!! \033[0m\n");

				if(bIsSTR)
				{
					disable_console();
				}
				else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
				{
					disable_console();
				}
			}
		}
		rtd_setbits(TOPBIST_cpu_bisr_remap_reg, bit_flags); // assert hold remap

                if((rtd_inl(acpu2_bisr_repair_addr) & acpu_repair_check_bits) != 0) {
                    printf("ACPU2: Need to repair\n");
                }

		bit_flags = _BIT0 | _BIT1 | _BIT2;

		if (rtd_inl(0xb803a23c) & bit_flags)
			printf("ACPU2 BISR status: failed %x\n", rtd_inl(0xb803a23c) & bit_flags);
		else
			printf("ACPU2 BISR status: ok\n");
	}
#endif 
	return 0;
}
#endif	//ENABLE_ACPU_BISR

/*
 * Print CPU information
 */
int print_cpuinfo(void)
{
	unsigned int regValue;
	unsigned int cores = 0;

	regValue = rtd_inl(0xb805b054);

	if ((regValue & _BIT10) == _BIT10)
		cores++;

	if ((regValue & _BIT9) == _BIT9)
		cores++;

	if ((regValue & _BIT4) == _BIT4)
		cores++;

	if ((regValue & _BIT3) == _BIT3)
		cores++;

	printf("CPU: Cortex-A53x%d\n", cores);

    return 0;
}

#ifndef CONFIG_SYS_DCACHE_OFF
void enable_caches(void)
{
    /* Enable D-cache. I-cache is already enabled in start.S */
    dcache_enable();
}
#endif

#ifndef CONFIG_SYS_L2CACHE_OFF
void v7_outer_cache_enable(void)
{
	//set_pl310_ctrl_reg(1);
}

void v7_outer_cache_disable(void)
{
	//set_pl310_ctrl_reg(0);
}
#endif

#ifdef CONFIG_USE_RTK_RTC
int rtc_init(void)
{
	uint rtcstop_regValue = 0xffffffff;

	printf("RTC: ");

	if (rtd_inl(SYS_REG_WDOG_DATA13_VADDR) & _BIT0)
		rtcstop_regValue = rtd_inl(D_RTCSTOP_reg);

	// check RTC status
	if ((rtcstop_regValue & _BIT0)== 0x0) // status: running
	{
		// RTC: Already running ... do nothing ...
		printf("Running\n");
	}
	else // status: stopped
	{
		printf("Init...");

		// enable DRTC clock
		rtd_clearbits(SYS_REG_WDOG_DATA13_VADDR, _BIT0);
		rtd_clearbits(SYS_REG_WDOG_DATA13_VADDR, _BIT2);
		rtd_setbits(SYS_REG_WDOG_DATA13_VADDR, _BIT2);
		rtd_setbits(SYS_REG_WDOG_DATA13_VADDR, _BIT0);

		// initialize RTC
		rtd_outl(D_RTCSTOP_reg, D_RTCSTOP_27Mclk_en(0x1)); // clk select

		// reset time info
		rtd_outl(D_RTCSEC_reg, D_RTCSEC_D_RTCSEC(0x0)); // sec
		rtd_outl(D_RTCMIN_reg, D_RTCMIN_D_RTCMIN(0x0)); // min
		rtd_outl(D_RTCHR_reg, D_RTCHR_D_RTCHR(0x0)); // hour
		rtd_outl(D_RTCDATE1_reg, D_RTCDATE1_D_RTCDATE1(0x0)); // date LSB
		rtd_outl(D_RTCDATE2_reg, D_RTCDATE2_D_RTCDATE2(0x0)); // date MSB

		// enable RTC
		rtd_outl(D_RTCEN_reg, D_RTCEN_D_RTCEN(0x5A));

		// check RTC status
		rtcstop_regValue = rtd_inl(D_RTCSTOP_reg);
		if ((rtcstop_regValue & _BIT0) == 0x0)
		{
			printf("Running\n");	// RTC: Running
		}
		else
		{
			printf("Stopped\n");	// RTC: Stop
		}
	}

	return 0;
}
#endif	//CONFIG_USE_RTK_RTC


int rtk_do_ip_bisr(void)
{

#ifdef ENABLE_KME_BISR
        rtk_kme_bisr();
#endif

#ifdef ENABLE_KMC_BISR
        rtk_kmc_bisr();
#endif

#ifdef ENABLE_DISPD_BISR
        rtk_dispd_bisr();
#endif

#ifdef ENABLE_DISPI_BISR
        rtk_dispi_bisr();
#endif

#ifdef ENABLE_GPU_BISR
        rtk_gpu_bisr();
#endif
	return 0;
}

#ifdef ENABLE_BISR_CHECK

int rtk_check_scpu_bisr(void)
{
        const unsigned int bisr_remap_bits = (_BIT2);
        const unsigned int bisr_done_bits = (_BIT2);
        if ( ((rtd_inl(TOPBIST_cpu_bisr_done_reg) & bisr_done_bits) != bisr_done_bits ) &&
                (rtd_inl(TOPBIST_cpu_bisr_remap_reg & bisr_remap_bits) != bisr_remap_bits ) )
        {
                printf("\n\033[31m[ERR] NO SCPU BISR FLOW!\033[0m\n");
                return -1;
        }
        return 0;
}


int rtk_check_dispd_bisr(void)
{
	const unsigned int bisr_remap_bits = (_BIT0);
	const unsigned int bisr_done_bits = (_BIT0);

        if( ((rtd_inl(TOPBIST_fcic_bisr_remap_reg) & bisr_remap_bits) != bisr_remap_bits) &&
                ((rtd_inl(TOPBIST_fcic_bisr_done_reg) & bisr_done_bits) != bisr_done_bits) )
	{
		printf("\n\033[31m[ERR] NO DISPD BISR FLOW!\033[0m\n");
		return -1;
	}

	return 0;
}
int rtk_check_dispi_bisr(void)
{
        const unsigned int bisr_remap_bits = (_BIT0);
        const unsigned int bisr_done_bits = (_BIT0);
        if( ((rtd_inl(TOPBIST_dispi_bisr_remap_reg) & bisr_remap_bits) != bisr_remap_bits) &&
                ((rtd_inl(TOPBIST_dispi_bisr_done_reg) & bisr_done_bits) != bisr_done_bits) )
	{
		printf("\n\033[31m[ERR] NO DISPI BISR FLOW!\033[0m\n");
		return -1;
	}

	return 0;
}

int rtk_check_kme_bisr(void)
{
#if 0
        const unsigned int bisr_remap_bits = (_BIT13);
        const unsigned int bisr_done_bits = (_BIT13);
        if( ((rtd_inl(TOPBIST_dispi_bisr_remap_reg) & bisr_remap_bits) != bisr_remap_bits) &&
                ((rtd_inl(TOPBIST_dispi_bisr_done_reg) & bisr_done_bits) != bisr_done_bits) )
        {
                printf("\n\033[31m[ERR] NO KME BISR FLOW!\033[0m\n");
                return -1;
        }
#endif
        return 0;
}

int rtk_check_kmc_bisr(void)
{
        const unsigned int bisr_remap_bits = (_BIT0);
        const unsigned int bisr_done_bits = (_BIT0);
        if( ((rtd_inl(TOPBIST_dispi_bisr_remap_reg) & bisr_remap_bits) != bisr_remap_bits) &&
                ((rtd_inl(TOPBIST_dispi_bisr_done_reg) & bisr_done_bits) != bisr_done_bits) )
        {
                printf("\n\033[31m[ERR] NO KMC BISR FLOW!\033[0m\n");
                return -1;
        }

        return 0;
}

int rtk_check_acpu_bisr(void)
{
	const unsigned int bisr_remap_bits = (_BIT15|_BIT14|_BIT13);
	const unsigned int bisr_done_bits = (_BIT13 | _BIT14 | _BIT15);
	if( ((rtd_inl(TOPBIST_cpu_bisr_remap_reg) & bisr_remap_bits) != bisr_remap_bits) &&
		((rtd_inl(TOPBIST_cpu_bisr_done_reg) & bisr_done_bits) != bisr_done_bits) )
	{
		printf("\n\033[31m[ERR] NO ACPU BISR FLOW!\033[0m\n");
		return -1;
	}

	return 0;
}
int rtk_check_vcpu_bisr(void)
{
	const unsigned int bisr_remap_bits = (_BIT1);
	const unsigned int bisr_done_bits = (_BIT1);
        if ( ((rtd_inl(TOPBIST_cpu_bisr_done_reg) & bisr_done_bits) != bisr_done_bits ) &&
                (rtd_inl(TOPBIST_cpu_bisr_remap_reg & bisr_remap_bits) != bisr_remap_bits) )
	{
		printf("\n\033[31m[ERR] NO VCPU BISR FLOW!\033[0m\n");
		return -1;
	}

	return 0;
}

int rtk_check_ve_bisr(void)
{
        const unsigned int bisr_remap_bits = (_BIT0);
        const unsigned int bisr_done_bits = (_BIT0);
        if ( ((rtd_inl(TOPBIST_ve_bisr_done_reg) & bisr_done_bits) != bisr_done_bits ) &&
                (rtd_inl(TOPBIST_ve_bisr_remap_reg & bisr_remap_bits) != bisr_remap_bits ) )
        {
                printf("\n\033[31m[ERR] NO VE BISR FLOW!\033[0m\n");
                return -1;
        }
        return 0;
}

int rtk_check_sfg_bisr(void)
{
        #if 1
        const unsigned int bisr_remap_bits = (_BIT6);
        const unsigned int bisr_done_bits = (_BIT6);

        if ( ((rtd_inl(TOPBIST_dispd_bisr_done_reg) & bisr_done_bits) != bisr_done_bits) &&
                ((rtd_inl(TOPBIST_dispd_bisr_remap_reg) & bisr_remap_bits) != bisr_remap_bits) )
        {
                printf("\n\033[31m[ERR] NO sfg BISR FLOW!\033[0m\n");
                return -1;
        }
        #endif
        return 0;
}


int rtk_check_gpu_bisr(void)
{
        #if 1
        const unsigned int bisr_remap_bits = (_BIT0);
        const unsigned int bisr_done_bits = (_BIT0);

        if( ((rtd_inl(TOPBIST_cpu_bisr_remap_reg) & bisr_remap_bits) != bisr_remap_bits) &&
                ((rtd_inl(TOPBIST_cpu_bisr_done_reg) & bisr_done_bits) != bisr_done_bits) )
        {
                printf("\n\033[31m[ERR] NO GPU BISR FLOW!\033[0m\n");
                return -1;
        }
        #endif
        return 0;
}


int rtk_check_emcu_bisr(void)
{
        const unsigned int bisr_remap_bits = (_BIT0);
        const unsigned int bisr_done_bits = (_BIT0);

        if( ((rtd_inl(STB_ST_bisr_remap_reg) & bisr_remap_bits) != bisr_remap_bits) &&
                ((rtd_inl(STB_ST_bisr_done_reg) & bisr_done_bits) != bisr_done_bits) )
	{
		printf("\n\033[31m[ERR] NO EMCU BISR FLOW!\033[0m\n");
		return -1;
	}

	return 0;
}

typedef int (*bisr_fnc_t) (void);

bisr_fnc_t bisr_sequence[] = {
	rtk_check_scpu_bisr,
	rtk_check_acpu_bisr,
        rtk_check_vcpu_bisr,
        rtk_check_ve_bisr,
        //rtk_check_kme_bisr,
        rtk_check_kmc_bisr,
        rtk_check_dispd_bisr,
        rtk_check_dispi_bisr,
        rtk_check_sfg_bisr,
        rtk_check_gpu_bisr,
        rtk_check_emcu_bisr,
};

#endif // ENABLE_BISR_CHECK

int rtk_check_ip_bisr(void)
{
	int ret = 0;
#ifdef ENABLE_BISR_CHECK
	int check_bisr = 0, i;
#endif

	if(bIsSTR)
	{
		enable_console();
	}
	else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
	{
		enable_console();
	}

	if((boot_mode != BOOT_NORMAL_MODE) || (boot_power_on_mode == POWER_OFF))
	{
		printf("\n\033[31m[WARNING] NOT NORMAL BOOTING FOW,  SKIP CHECKING BISR FLOW!\033[0m\n");
		return ret;
	}

#ifdef ENABLE_BISR_CHECK
	printf("\n\033[31m[INFO] CHECK BISR FLOW!\033[0m\n");
	for ( i = 0; i < sizeof(bisr_sequence)/sizeof(bisr_sequence[0]) ; i++) {
		check_bisr = bisr_sequence[i]();
		if ( check_bisr != 0 )
			ret = check_bisr;
        }
	rtd_clearbits(TOPBIST_cpu_bisr_rstn_reg, _BIT2);
#else
	printf("\n\033[31m[WARNING] NO CHECK BISR FLOW!\033[0m\n");
#endif

	if(bIsSTR)
	{
		disable_console();
	}
	else if(!(strncmp(getenv("print"),"off",3)))// log mode is off
	{
		disable_console();
	}

	return ret;
}

