#include <common.h>
#include <command.h>
#include <system.h>
#include <fw_info.h>
#include <platform_lib/board/pcb.h>
#include <platform_lib/board/gpio.h>
#include <drivers/gpio/rtk_gpio.h>
#if defined(Config_Module_Test_Boot_TRUE)

void modt_send_byte(unsigned char data)
{
    int i = 0;
    for(i=0; i<8; i++)
    {
        IO_Set(PIN_SCL, 0);
        mdelay(10);
        IO_Set(PIN_SDA, data & 0x80);
        data = data<<1;
        mdelay(10);
        IO_Set(PIN_SCL, 1);
        mdelay(10);
    }
    return;
}

void modt_send_buf(const unsigned char *buf, unsigned int n)
{
    int i = 0;
    int retry = 0;
    int wait_count = 0;
    const unsigned char *p = NULL;
retry:
    p = buf;
    IO_Set(PIN_SDA, 0);
    IO_Set(PIN_SCL, 0);
    while(!IO_Get(PIN_SIG));

    for(i=0; i<n; i++)
    {
       modt_send_byte(*p);
       p++;
    }

    IO_Set(PIN_SCL, 1);
    IO_Set(PIN_SDA, 0); // relate receive
    mdelay(100);
#if 0
    while(1)
    {
        printf(".");
        if(!getISOGPIO(PIN_SIG))
        {
            mdelay(10);
            if(!getISOGPIO(PIN_SIG))
            {
                while(!getISOGPIO(PIN_SIG));
                retry++;
                printf("modt_send_buf fail,and retry %d\n", retry);
                goto retry;
            }
        }
    }
#endif
    while(!IO_Get(PIN_SIG))
    {
        printf(".");
        mdelay(10);
        if(IO_Get(PIN_SIG))
        {
            retry++;
            wait_count = 0;
            printf("modt_send_buf fail,and retry %d\n", retry);
            goto retry;
        }
        wait_count++;
        if(wait_count == 4)
        {
            break;
        }
    }
    printf("modt_send_buf OK\n");

    return;
}

unsigned char calculateLRC(const unsigned char *buf, unsigned int n)
{
    unsigned char checksum = 0;
    while(n>0){
        checksum += *buf++;
        n--;
    }
    return ((char) -checksum );
}

void modt_send(void)
{
    char buf[3] = {0};
    buf[0] = 0xaa;  // magic_num
    buf[1] = MODT_Test_Version;  // sw version
    buf[2] = calculateLRC((const unsigned char *)buf, 2); //LRC

    if(IO_Get(PIN_SIG))
        modt_send_buf((const unsigned char *)buf, 3);

    return;
}

void modt_init(void)
{
   printf("init ISOGPIO [9:8] => 2'b00(Socket Signal)\n");//used for module test
   IO_Set(PIN_STATUS1, 0);
   IO_Set(PIN_STATUS0, 0);
 }

void modt_post_init(void)
{
   //boot start signal
   puts("init ISOGPIO [9:8] => 2'b01(Socket Signal)\n");//used for module test
   IO_Set(PIN_STATUS1, 0);
   IO_Set(PIN_STATUS0, 1);
}

void usb_rw_test(void)
{
    int ret = 0;
#ifdef CONFIG_USB_STORAGE_RW_TEST
    extern MT_USB_TEST usb_rw_result;
#endif
#ifdef CONFIG_USB_8723_ENUM_TEST
    extern MT_USB_TEST usb_enumeration_result;
#endif
    printf("------------------ %s ------------------\n", __func__);
    // usb start to init usb and do the rw test
    ret = run_command("usb start", 0);
    if (ret) {
        printf("[USB] retry usb start.... \n");
        ret = run_command("usb start", 0);
        if (ret) {
            printf("[USB] retry usb start still fail... \n");
#ifdef CONFIG_USB_STORAGE_RW_TEST
            usb_rw_result = USB_FAIL;
#endif
#ifdef CONFIG_USB_8723_ENUM_TEST
            usb_enumeration_result = USB_FAIL;
#endif
        }
    }
#ifdef CONFIG_USB_STORAGE_RW_TEST
    if (usb_rw_result == USB_FAIL)
        ret = -1;
#endif
#ifdef CONFIG_USB_8723_ENUM_TEST
    if (usb_enumeration_result == USB_FAIL)
        ret = -1;
#endif
    if (ret) {
        printf("usb MT fail, it's fail IC, stop to test\n");
        IO_Set(PIN_STATUS1, 1);
        IO_Set(PIN_STATUS0, 1);
        while(1);
    }
}
#endif
