/*
 * armboot - Startup Code for OMAP3530/ARM Cortex CPU-core
 *
 * Copyright (c) 2004	Texas Instruments <r-woodruff2@ti.com>
 *
 * Copyright (c) 2001	Marius Gröger <mag@sysgo.de>
 * Copyright (c) 2002	Alex Züpke <azu@sysgo.de>
 * Copyright (c) 2002	Gary Jennejohn <garyj@denx.de>
 * Copyright (c) 2003	Richard Woodruff <r-woodruff2@ti.com>
 * Copyright (c) 2003	Kshitij <kshitij@ti.com>
 * Copyright (c) 2006-2008 Syed Mohammed Khasim <x0khasim@ti.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <asm-offsets.h>
#include <config.h>
#include <version.h>
#include <asm/system.h>
#include <linux/linkage.h>
#include "../../../../../src/include/project_config.h"

#ifdef Config_Boot_ATF_TRUE
#define LINUX_KERNEL_ENTRY 0x108000
#define LINUX64_KERNEL_ENTRY 0x180000
#endif

.globl _start
_start: b	reset
	ldr	pc, _undefined_instruction
	ldr	pc, _software_interrupt
	ldr	pc, _prefetch_abort
	ldr	pc, _data_abort
	ldr	pc, _not_used
	ldr	pc, _irq
	ldr	pc, _fiq
#ifdef CONFIG_SPL_BUILD
_undefined_instruction: .word _undefined_instruction
_software_interrupt:	.word _software_interrupt
_prefetch_abort:	.word _prefetch_abort
_data_abort:		.word _data_abort
_not_used:		.word _not_used
_irq:			.word _irq
_fiq:			.word _fiq
_pad:			.word 0x12345678 /* now 16*4=64 */
#else
_undefined_instruction: .word undefined_instruction
_software_interrupt:	.word software_interrupt
_prefetch_abort:	.word prefetch_abort
_data_abort:		.word data_abort
_not_used:		.word not_used
_irq:			.word irq_handler
_fiq:			.word fiq
_pad:			.word 0x12345678 /* now 16*4=64 */
#endif	/* CONFIG_SPL_BUILD */

.global _end_vect
_end_vect:

	.balignl 16,0xdeadbeef
/*************************************************************************
 *
 * Startup Code (reset vector)
 *
 * do important init only if we don't start from memory!
 * setup Memory and board specific bits prior to relocation.
 * relocate armboot to ram
 * setup stack
 *
 *************************************************************************/

.globl _TEXT_BASE
_TEXT_BASE:
	.word	CONFIG_SYS_TEXT_BASE

/*
 * These are defined in the board-specific linker script.
 */
.globl _bss_start_ofs
_bss_start_ofs:
	.word __bss_start - _start

.global	_image_copy_end_ofs
_image_copy_end_ofs:
	.word 	__image_copy_end - _start

.globl _bss_end_ofs
_bss_end_ofs:
	.word __bss_end__ - _start

.globl _end_ofs
_end_ofs:
	.word _end - _start

#ifdef CONFIG_NO_RELOCATION
.globl _stack_start
_stack_start:
	.word __stack_start - _start + CONFIG_SYS_TEXT_BASE
.globl _stack_start_str
_stack_start_str:
	.word __stack_start_str - _start + CONFIG_SYS_TEXT_BASE
.globl _irq_stack_start
_irq_stack_start:
	.word __stack_start - _start + 0x10 + CONFIG_SYS_TEXT_BASE
#endif

#ifdef CONFIG_USE_IRQ
/* IRQ stack memory (calculated at run-time) */
.globl IRQ_STACK_START
IRQ_STACK_START:
	.word	0x0badc0de

/* IRQ stack memory (calculated at run-time) */
.globl FIQ_STACK_START
FIQ_STACK_START:
	.word 0x0badc0de
#endif

/* IRQ stack memory (calculated at run-time) + 8 bytes */
.globl IRQ_STACK_START_IN
IRQ_STACK_START_IN:
	.word	CONFIG_SYS_IRQ_STACK_START_IN_ADDR

#define ENABLE_UART_DEBUG_IN_UBOOT

/*
 * the actual reset code
 */

#define REG_CRT_SUSPEND    0x18060100
#define KERNEL_RSV_MEM_MAGICNO 0x2379beef
#define REG_STR_JUMP_ADDR_reg 0x18060104

#define CONFIG_ROLLBACK_CODE
#ifdef CONFIG_ROLLBACK_CODE // rollback backup text data if QC code is too large
//#define BOOTCODE_OVER_AUX_BACKUP_DATA_ADDR  0x010f0000
#define BOOTCODE_OVER_AUX_BACKUP_DATA_ADDR	0x010f0800 // (original addr + 0x800)
#define BOOTCODE_OVER_AUX_MAGIC_ADDR		0x010e0000
#define BOOTCODE_OVER_AUX_BACKUP_SIZE_ADDR	0x010e0004
#define BOOTCODE_OVER_AUX_MAGIC_NUMBER		0x01a2a3a4
#define BOOTCODE_OVER_AUX_MAGIC_CLEAR		0x00000000
#define BOOTCODE_OVER_AUX_SELF_COPY_SIZE	0x800
#endif

reset:
#ifdef CONFIG_QC_VERIFY
#if CONFIG_QC_SCPU_PWR_CTRL
	@ -- keep core0, bypass coreX
	MRC   p15, 0, r0, c0, c0, 5
	LSR   r0, r0, #8
	AND   r0, r0, #0x3

	CMP   r0, #1
	BEQ   secondary_start

	CMP   r0, #2
	BEQ   secondary_start

	CMP   r0, #3
	BEQ   secondary_start

	@ -- prepare coreX
	ldr r1 , =0x0 @ address
	ldr r0 , =0xea007ffe @ value
	str r0 , [r1] @ instruction at 0x0 for coreX jump to 0x20000

	ldr r1 , =0x1805b828 @ address
	ldr r0 , =0x0 @ value
	str r0 , [r1] @

	ldr r1 , =0x1805b888 @ address
	ldr r0 , =0x0 @ value
	str r0 , [r1] @
#endif

	@ -- release di_sram mapping
	ldr r1 , =0x18007500
	ldr r0 , =0x0
	str r0 , [r1]
#if 0
	@ -- switch to non_secure
	MRC p15,0,r0,c1,c1,0 
	orr r0, r0, #0x05 @ [0]: SCR_NS, [2]: SCR_FIQ
	MCR p15,0,r0,c1,c1,0
#endif
#ifdef CONFIG_ROLLBACK_CODE
@ -- check magic
	LDR   r0, =BOOTCODE_OVER_AUX_MAGIC_ADDR
	LDR   r1, [r0]
	LDR   r2, =BOOTCODE_OVER_AUX_MAGIC_NUMBER
	CMP   r1, r2
	BNE   skip_rollback_code

@ -- clear magic
	LDR   r3, =BOOTCODE_OVER_AUX_MAGIC_CLEAR
	STR   r3, [r0]

@ -- start rollback
	LDR   r0, =BOOTCODE_OVER_AUX_BACKUP_SIZE_ADDR
	LDR   r1, [r0]
	LDR   r0, =BOOTCODE_OVER_AUX_BACKUP_DATA_ADDR
//	LDR   r1, =0x200000
	LDR   r2, =0x20800
	ADD   r3, r2, r1

rollback_loop:
	CMP   r3, r2
	BEQ   rollback_loop_end
	LDR   r4, [r0]
	STR   r4, [r2]
	ADD   r0, r0, #4
	ADD   r2, r2, #4
	B     rollback_loop

rollback_loop_end:
skip_rollback_code:
#endif
#endif

#ifdef Config_Boot_ATF_TRUE
       MRC   p15, 0, r0, c0, c0, 5
       LSR   r0, r0, #8
       AND   r0, r0, #0x3

       CMP   r0, #1
       BEQ   run_secondary

       CMP   r0, #2
       BEQ   run_third

       CMP   r0, #3
       BEQ   run_fourth
       B     run_master
run_secondary:
run_third:
run_fourth:
       ldr   r0, =0x520
       ldr   r2, =0x1234abcd
run_last:
       ldr   r1, [r0]
       cmp   r1, r2
       bne run_last
       mov r0, #0
       mov r1, #0
       mov r2, #0
#ifdef CONFIG_JUMP_TO_64BIT_KERNEL
       ldr r3, =#LINUX64_KERNEL_ENTRY
#else
       ldr r3, =#LINUX_KERNEL_ENTRY
#endif
       smc #0 

#endif
run_master:
#ifdef ENABLE_UART_DEBUG_IN_UBOOT
	ldr	r1 , =0x1801b100
wait_ready:
	ldr	r0, [r1, #0x14]
	and	r0, r0, #0x60    // (LSR_TEMT | LSR_THRE)
	cmp	r0, #0x60
	bne	wait_ready
	ldr	r0 , =0x00000055  /* 'U' */
	str	r0 , [r1]
#endif

#ifndef CONFIG_BSP_REALTEK
	bl	save_boot_params
#endif
	/*
	 * set the cpu to SVC32 mode
	 */
	mrs	r0, cpsr
	bic	r0, r0, #0x1f
	orr	r0, r0, #0xd3
	msr	cpsr,r0

#if !defined(CONFIG_TEGRA2)
/*
 * Setup vector:
 * (OMAP4 spl TEXT_BASE is not 32 byte aligned.
 * Continue to use ROM code vector only in OMAP4 spl)
 */
#Chuck set exception vector at 0x0
#if !(defined(CONFIG_OMAP44XX) && defined(CONFIG_SPL_BUILD))
	// we do here
	/* Set V=0 in CP15 SCTRL register - for VBAR to point to vector */
	mrc	p15, 0, r0, c1, c0, 0	@ Read CP15 SCTRL Register
	bic	r0, #CR_V		@ V = 0
	mcr	p15, 0, r0, c1, c0, 0	@ Write CP15 SCTRL Register

	/* Set vector address in CP15 VBAR register */
	ldr	r0, =_start
	mcr	p15, 0, r0, c12, c0, 0	@Set VBAR
#endif
#endif	/* !Tegra2 */

	// Add bisr for scpu1/scpu2
#ifdef ENABLE_SCPU1_BISR
	bl	bsp_scpu1_bisr
#endif

#ifdef ENABLE_SCPU2_BISR
#ifndef CONFIG_KCPU_LOAD_OPTEE
	bl	bsp_scpu2_bisr
#endif
#endif

#ifdef CONFIG_BSP_REALTEK
	bl  cpu_invalid_all
	bl  bsp_boot_init_plat		@smpen should set before M/C/I
	bl  cpu_init_cp15

	/* fixme,
	** please remove these 2 lines,
	** if no need of DECLARE_GLOBAL_DATA_PTR for gd.
	*/
	adr r8, _irq_stack_start  	@DECLARE_GLOBAL_DATA_PTR
	ldr r8, [r8]
#endif
	/* the mask ROM code should have PLL and others stable */
#ifndef CONFIG_SKIP_LOWLEVEL_INIT
	bl  cpu_invalid_all
	bl	cpu_init_cp15
	bl	cpu_init_crit
#endif

/* Set stackpointer in internal RAM to call board_init_f */
call_board_init_f:
#ifdef CONFIG_NO_RELOCATION
	#ifdef CONFIG_SYS_SUPPORT_MULTICORE
		mrc   p15, 0, r0, c0, c0, 5
		lsr     r0, r0, #0x8
		and   r0, r0, #3
		cmp   r0, #0
		beq   core0
		cmp   r0, #1
		beq   core1
	#if defined(CONFIG_RTD299O) || defined(CONFIG_RTD289X) || defined(CONFIG_RTD287O) || defined(CONFIG_RTD285X)
		cmp   r0, #2
		beq   core2
		cmp   r0, #3
		beq   core3
	#endif
	cluster_1:
		and	r0, r0, #3
		cmp	r0, #0
		beq	core2
		ldr   sp, =CONFIG_SYS_INIT_CORE3_SP_ADDR
		ldr   r0,=0x00000003
		b  next
	#if defined(CONFIG_RTD299O) || defined(CONFIG_RTD289X) || defined(CONFIG_RTD287O) || defined(CONFIG_RTD285X)
	core3:
		ldr   sp, =CONFIG_SYS_INIT_CORE3_SP_ADDR
		ldr   r0,=0x00000003
		b  secondary_boot
	#endif
	core2:
		ldr   sp, =CONFIG_SYS_INIT_CORE2_SP_ADDR
		ldr   r0,=0x00000002
		b  secondary_boot
	core1:
		ldr   sp, =CONFIG_SYS_INIT_CORE1_SP_ADDR
		ldr   r0,=0x00000001
		b  secondary_boot
	core0:
		ldr	sp, _stack_start
		ldr   r0,=0x00000000
	#else
                LDR     r0, =REG_CRT_SUSPEND
                LDR     r1, [r0]
                LDR     r0, =KERNEL_RSV_MEM_MAGICNO
                CMP     r1, r0
                BEQ resume_uboot_sp   @ FIXME: STR need to check more
normal_uboot_sp:
                ldr   sp, _stack_start
                ldr   r0,=0x00000000
                b  primary_boot
resume_uboot_sp:
                ldr   sp, _stack_start_str
                ldr   r0,=0x00000000
	#endif
#else
	ldr sp, =(CONFIG_SYS_INIT_SP_ADDR)
	ldr   r0,=0x00000000
#endif

primary_boot:
next:
#ifdef CONFIG_MAX_POWER_MEASURE
     mrc p15, 0,r6, c1, c0, 2 //Read Coprocessor Access Control Register
     orr  r6, #0xF00000
     mcr p15, 0,r6, c1, c0, 2 //Write Coprocessor Access Control Register
#endif
	bic	sp, sp, #7 /* 8-byte alignment for ABI compliance */
	bl	board_init_f
	
secondary_boot:
#ifdef CONFIG_MAX_POWER_MEASURE
      mrc p15, 0,r6, c1, c0, 2 //Read Coprocessor Access Control Register
      orr  r6, #0xF00000
      mcr p15, 0,r6, c1, c0, 2 //Write Coprocessor Access Control Register
#endif

	bic	sp, sp, #7 /* 8-byte alignment for ABI compliance */
	bl	secondary_init_r
/*------------------------------------------------------------------------------*/

/*
 * void relocate_code (addr_sp, gd, addr_moni)
 *
 * This "function" does not return, instead it continues in RAM
 * after relocating the monitor code.
 *
 */

#define REG_CRT_SUSPEND    0x18060100
#define KERNEL_RSV_MEM_MAGICNO 0x2379beef
#define REG_STR_JUMP_ADDR_reg 0x18060104

ENTRY(lowlevel_init)

	/* Enable the the VFP */
	mrc	p15, 0, r1, c1, c0, 2
	orr	r1, r1, #(0x3 << 20)
	orr	r1, r1, #(0x3 << 20)
	mcr	p15, 0, r1, c1, c0, 2
	isb
	fmrx	r1, FPEXC
	orr	r1,r1, #(1<<30)
	fmxr	FPEXC, r1

	/* Move back to caller */
	mov	pc, lr
ENDPROC(lowlevel_init)

ENTRY(relocate_code)
	mov	r4, r0	/* save addr_sp */
	mov	r5, r1	/* save addr of gd */
	mov	r6, r2	/* save addr of destination */

	/* Set up the stack						    */
stack_setup:
	mov	sp, r4

#ifdef CONFIG_STR_ENABLE
	LDR     r0, =REG_STR_JUMP_ADDR_reg
	LDR     r1, [r0]
	MOV     r0, #0
	CMP     r1, r0
	BEQ     stack_setup_ac

	LDR     r0, =REG_CRT_SUSPEND
	LDR     r1, [r0]
	LDR     r0, =KERNEL_RSV_MEM_MAGICNO
	CMP     r1, r0
	moveq   r9, #0x00000000         /* clear                            */
	BEQ     clear_bss               /* clear_bss                        */
#endif

stack_setup_ac:

	adr	r0, _start
	cmp	r0, r6
	moveq	r9, #0		/* no relocation. relocation offset(r9) = 0 */
	beq fix_rel			/* skip copy */
	mov	r1, r6			/* r1 <- scratch for copy_loop */
	ldr	r3, _image_copy_end_ofs
	add	r2, r0, r3		/* r2 <- source end address	    */

copy_loop:
	ldmia	r0!, {r9-r10}		/* copy from source address [r0]    */
	stmia	r1!, {r9-r10}		/* copy to   target address [r1]    */
	cmp	r0, r2			/* until source end address [r2]    */
	blo	copy_loop

fix_rel:
#ifndef CONFIG_SPL_BUILD
	/*
	 * fix .rel.dyn relocations
	 */
	ldr	r0, _TEXT_BASE		/* r0 <- Text base */
	sub	r9, r6, r0		/* r9 <- relocation offset */
	ldr	r10, _dynsym_start_ofs	/* r10 <- sym table ofs */
	add	r10, r10, r0		/* r10 <- sym table in FLASH */
	ldr	r2, _rel_dyn_start_ofs	/* r2 <- rel dyn start ofs */
	add	r2, r2, r0		/* r2 <- rel dyn start in FLASH */
	ldr	r3, _rel_dyn_end_ofs	/* r3 <- rel dyn end ofs */
	add	r3, r3, r0		/* r3 <- rel dyn end in FLASH */
fixloop:
	ldr	r0, [r2]		/* r0 <- location to fix up, IN FLASH! */
	add	r0, r0, r9		/* r0 <- location to fix up in RAM */
	ldr	r1, [r2, #4]
	and	r7, r1, #0xff
	cmp	r7, #23			/* relative fixup? */
	beq	fixrel
	cmp	r7, #2			/* absolute fixup? */
	beq	fixabs
	/* ignore unknown type of fixup */
	b	fixnext
fixabs:
	/* absolute fix: set location to (offset) symbol value */
	mov	r1, r1, LSR #4		/* r1 <- symbol index in .dynsym */
	add	r1, r10, r1		/* r1 <- address of symbol in table */
	ldr	r1, [r1, #4]		/* r1 <- symbol value */
	add	r1, r1, r9		/* r1 <- relocated sym addr */
	b	fixnext
fixrel:
	/* relative fix: increase location by offset */
	ldr	r1, [r0]
	add	r1, r1, r9
fixnext:
	str	r1, [r0]
	add	r2, r2, #8		/* each rel.dyn entry is 8 bytes */
	cmp	r2, r3
	blo	fixloop
	b	clear_bss
_rel_dyn_start_ofs:
	.word __rel_dyn_start - _start
_rel_dyn_end_ofs:
	.word __rel_dyn_end - _start
_dynsym_start_ofs:
	.word __dynsym_start - _start

#endif	/* #ifndef CONFIG_SPL_BUILD */

clear_bss:
#ifdef CONFIG_SPL_BUILD
	/* No relocation for SPL */
	ldr	r0, =__bss_start
	ldr	r1, =__bss_end__
#else
	ldr	r0, _bss_start_ofs
	ldr	r1, _bss_end_ofs
	mov	r4, r6			/* reloc addr */
	add	r0, r0, r4
	add	r1, r1, r4
#endif
	mov	r2, #0x00000000		/* clear			    */

clbss_l:cmp	r0, r1			/* clear loop... */
	bhs	clbss_e			/* if reached end of bss, exit */
	str	r2, [r0]
	add	r0, r0, #4
	b	clbss_l
clbss_e:

/*
 * We are done. Do not return, instead branch to second part of board
 * initialization, now running from RAM.
 */
jump_2_ram:
/*
 * If I-cache is enabled invalidate it
 */
#ifndef CONFIG_SYS_ICACHE_OFF
	mcr	p15, 0, r0, c7, c5, 0	@ invalidate icache
	mcr     p15, 0, r0, c7, c10, 4	@ DSB
	mcr     p15, 0, r0, c7, c5, 4	@ ISB
#endif
/*
 * Move vector table
 */
#if !defined(CONFIG_TEGRA2)
#if !(defined(CONFIG_OMAP44XX) && defined(CONFIG_SPL_BUILD))
	/* Set vector address in CP15 VBAR register */
	ldr     r0, =_start
	add     r0, r0, r9
	mcr     p15, 0, r0, c12, c0, 0  @Set VBAR
#endif
#endif /* !Tegra2 */

	ldr	r0, _board_init_r_ofs
	adr	r1, _start
	add	lr, r0, r1
	add	lr, lr, r9
	/* setup parameters for board_init_r */
	mov	r0, r5		/* gd_t */
	mov	r1, r6		/* dest_addr */
	/* jump to it ... */
	mov	pc, lr

_board_init_r_ofs:
	.word board_init_r - _start
ENDPROC(relocate_code)



ENTRY(relocate_code_str)
    mov r4, r0  /* save addr_sp */
    mov r5, r1  /* save addr of gd */
    mov r6, r2  /* save addr of destination */
    mov sp, r4
    mov r9, #0      /* no relocation. relocation offset(r9) = 0 */

clear_bss1:
    ldr r0, _bss_start_ofs
    ldr r1, _bss_end_ofs
    mov r4, r6          /* reloc addr */
    add r0, r0, r4
    add r1, r1, r4

    mov r2, #0x00000000     /* clear                */

clbss_l1:
    cmp r0, r1          /* clear loop... */
    bhs clbss_e     /* if reached end of bss, exit */
    str r2, [r0]
    add r0, r0, #4
    b   clbss_l1
ENDPROC(relocate_code_str)

@-------------------------------------------------
@Macro: SET_SVC32_MODE
@-------------------------------------------------
.macro set_svc32_mode
	mrs r1, cpsr
	bic r1, r1, #0x1f
	orr r1, r1, #0xd3
	msr cpsr,r1
.endm

@-------------------------------------------------
@Macro: SET_IRQ_MODE
@-------------------------------------------------
.macro set_irq_mode
    mrs r1, cpsr
    bic r1, r1, #0x1f
    orr r1, r1, #0xd2
    msr cpsr,r1
.endm

.globl set_irq_sp
set_irq_sp:
#ifdef CONFIG_USE_IRQ
    /* r0=irq stack pointer */
    set_irq_mode            @ change to IRQ mode
    mov sp, r0                  @ set IRQ_SP
    set_svc32_mode       @ change back to SVC mode
#endif
    mov pc, lr

/*************************************************************************
 * BSP related init
 *************************************************************************/
#ifdef CONFIG_BSP_REALTEK

//#ifdef A53
//#define TSGEN_BASE	0xf0010000
//#else
//#define TSGEN_BASE      0x1f300000
//#endif
//#define TSGEN_CNTFID0	(TSGEN_BASE + 0x20)
//#define TSGEN_FREQ	0x019bfcc0

//#define TSGEN_CNTCR	(TSGEN_BASE + 0x0)
//#define TSGEN_ENABLE	0x1

#ifdef CONFIG_USE_CCI_BUS
#define CONFIG_CCI_BASE                 0x180A0000
#define CONFIG_CCI_STATUS		(CONFIG_CCI_BASE + 0xc)

#define CONFIG_CCI_SLAVE0		(CONFIG_CCI_BASE + 0x4000) /* A53 */
#define CONFIG_CCI_SLAVE1		(CONFIG_CCI_BASE + 0x5000) /* A57 */

#define CCI_ENABLE_SNOOP_REQ    0x1
#define CCI_ENABLE_DVM_REQ      0x2
#define CCI_ENABLE_REQ          (CCI_ENABLE_SNOOP_REQ)
#endif

ENTRY(bsp_boot_init_plat)
#ifdef CONFIG_USE_CCI_BUS

#ifdef CONFIG_ARCH_CPU_ARMA53
	ldr	r0, =CONFIG_CCI_SLAVE0
	ldr	r1, =CCI_ENABLE_REQ | CCI_ENABLE_DVM_REQ
	str	r1, [r0]

	ldr	r0, =CONFIG_CCI_STATUS
1:	ldr	r1, [r0]
	tst	r1, #0x1
	bne	1b
#endif

#ifdef CONFIG_ARCH_CPU_ARMA57
	ldr r0, =CONFIG_CCI_SLAVE1
	ldr r1, =CCI_ENABLE_REQ | CCI_ENABLE_DVM_REQ
	str r1, [r0]

	ldr r0, =CONFIG_CCI_STATUS
2:	ldr r1, [r0]
	tst r1, #0x1
	bne 2b
#endif
#endif

#ifdef CONFIG_ARCH_CPU_ARMA57 // A53 TRM-4.3.64 L2CTLR_EL1 is writeable. However, all writes to this register are ignored.
/* change L2 to W2R3 state */
	mrc p15, 1, r0, c9, c0, 2     @ Read current control reg
	orr r0, r0, #(0x0 << 9)       @ L2 tag  RAM setup
	orr r0, r0, #(0x1 << 5)       @ L2 data RAM setup
	orr r0, r0, #(0x2 << 0)       @ L2 data RAM latency
	mcr p15, 1, r0, c9, c0, 2     @ Write reg back
#endif

	mov	pc, lr

ENDPROC(bsp_boot_init_plat)
#endif

#ifdef ENABLE_SCPU1_BISR
ENTRY(bsp_scpu1_bisr)
	@ run only for core0
	ldr   r1, =0x0000ffff	
	mrc   p15, 0, r0, c0, c0, 5
	and	  r0, r1
	cmp   r0, #0x00
	bne   __bsp_scpu1_bisr_exit

	mov		r0, #0x18000000	@ 	rtd_setbits( TOP_BIST_cpu_bisr_rstn_reg, _BIT0); 
	ldr		r1, [r0, #0xf0c]
	mov 	r2, #20 		@timeout as 20ms from AlvinChen
	orr		r1, r1, #0x01
	str		r1, [r0, #0xf0c]

__WAIT_SCPU1_BISR_DONE:
	@ init timer6 to 1ms
	ldr 	r3, =0x1801b000
	mov 	r4,#0x00
	str 	r4,[r3, #0x668]
	str 	r4,[r3, #0x648]
	ldr 	r4, =27000
	str 	r4,[r3, #0x608] 
	mov 	r4, #0x04
	str 	r4,[r3, #0x6ec]
	ldr 	r4, =0xc0000000
	str 	r4,[r3, #0x648]
					
_WAIT_SCPU1_DELAY_1ms:
	mov 	r5, #0x04
	ldr 	r4,[r3, #0x6ec]
	and 	r4,r5
	cmp 	r4, #0x04
	bne 	_WAIT_SCPU1_DELAY_1ms
	nop

	mov		r5,#0x01		
	ldr 	r1, [r0, #0xf2c]	@	rtd_inl(TOP_BIST_cpu_bisr_done_reg)	
	and 	r1,r5
	cmp 	r1, #0x01
	bne 	__CHECK_TIMEOUT_SCPU1_BISR
	nop
		
__SCPU1_SET_BISR_REMAP:
	ldr 	r1, [r0, #0xf1c]	@	rtd_setbits(TOP_BIST_cpu_bisr_remap_reg, _BIT0);
	orr 	r1, r1, #0x01	
	str 	r1, [r0, #0xf1c]
	B		__bsp_scpu1_bisr_exit

__CHECK_TIMEOUT_SCPU1_BISR:
	sub 	r2, r2, #1
	cmp 	r2, #0
	bne 	__WAIT_SCPU1_BISR_DONE
		
	@ show error log
	ldr 	r3, =0x18062300
	mov 	r4, #'['
	str 	r4,[r3]
	mov 	r4, #'B'
	str 	r4,[r3]
	mov 	r4, #'I'
	str 	r4,[r3] 
	mov 	r4, #'S'
	str 	r4,[r3]
	mov 	r4, #'R'
	str 	r4,[r3]
	mov 	r4, #']'
	str 	r4,[r3]
	mov 	r4, #'S'
	str 	r4,[r3] 
	mov 	r4, #'C'
	str 	r4,[r3] 
	mov 	r4, #'P'
	str 	r4,[r3] 
	mov 	r4, #'U'
	str 	r4,[r3] 
	mov 	r4, #'1'
	str 	r4,[r3] 
	mov 	r4, #' '
	str 	r4,[r3]
	mov 	r4, #'E'
	str 	r4,[r3]
	mov 	r4, #'R'
	str 	r4,[r3]
	mov 	r4, #'R'
	str 	r4,[r3]
__scpu1_bisr_error_loop:
	B		__scpu1_bisr_error_loop
				
__bsp_scpu1_bisr_exit:
	mov		pc, lr				@ back to my caller
ENDPROC(bsp_scpu1_bisr)
#endif

#ifdef ENABLE_SCPU2_BISR
ENTRY(bsp_scpu2_bisr)
	@ run only for core0
	ldr   r1, =0x0000ffff	
	mrc   p15, 0, r0, c0, c0, 5
	and	  r0, r1
	cmp   r0, #0x00
	bne   __bsp_scpu2_bisr_exit
	
	mov 	r0, #0x18000000 		@ start scpu bisr 
	ldr 	r1, [r0, #0xf0c]
	mov 	r2, #20					@ timeout as 20ms from AlvinChen
	orr 	r1, r1, #0x20	
	str 	r1, [r0, #0xf0c]
	
__WAIT_SCPU2_BISR_DONE:
	@ init timer6 to 1ms
	ldr 	r3, =0x1801b000
	mov 	r4,#0x00
	str 	r4,[r3, #0x668]
	str 	r4,[r3, #0x648]
	ldr 	r4, =27000
	str 	r4,[r3, #0x608] 
	mov 	r4, #0x04
	str 	r4,[r3, #0x6ec]
	ldr 	r4, =0xc0000000
	str 	r4,[r3, #0x648]
					
_WAIT_SCPU2_DELAY_1ms:
	mov 	r5, #0x04
	ldr 	r4,[r3, #0x6ec]
	and		r4,r5
	cmp 	r4, #0x04
	bne 	_WAIT_SCPU2_DELAY_1ms
	nop
	
	mov		r5,#0x20
	ldr 	r1, [r0, #0xf2c]		@	Read done flag for scpu bisr
	and 	r1,r5
	cmp 	r1, #0x20
	bne 	__CHECK_TIMEOUT_SCPU2_BISR
	nop
		
__SCPU2_SET_BISR_REMAP:
	ldr 	r1, [r0, #0xf1c]		@	Set remap after bisr done
	orr 	r1, r1, #0x20	
	str 	r1, [r0, #0xf1c]
	B		__bsp_scpu2_bisr_exit
	
__CHECK_TIMEOUT_SCPU2_BISR:
	sub 	r2, r2, #1
	cmp 	r2, #0
	bne 	__WAIT_SCPU2_BISR_DONE
	
	ldr 	r3, =0x18062300		@ show error log for bisr 
	mov 	r4, #'['
	str 	r4,[r3]
	mov 	r4, #'B'
	str 	r4,[r3]
	mov 	r4, #'I'
	str 	r4,[r3] 
	mov 	r4, #'S'
	str 	r4,[r3]
	mov 	r4, #'R'
	str 	r4,[r3]
	mov 	r4, #']'
	str 	r4,[r3]
	mov 	r4, #'S'
	str 	r4,[r3] 
	mov 	r4, #'C'
	str 	r4,[r3] 
	mov 	r4, #'P'
	str 	r4,[r3] 
	mov 	r4, #'U'
	str 	r4,[r3] 
	mov 	r4, #'0'
	str 	r4,[r3] 
	mov 	r4, #' '
	str 	r4,[r3]
	mov 	r4, #'E'
	str 	r4,[r3]
	mov 	r4, #'R'
	str 	r4,[r3]
	mov 	r4, #'R'
	str 	r4,[r3]
__bisr_error_loop:
	B		__bisr_error_loop		@ infinitely loop because bisr failure
			
__bsp_scpu2_bisr_exit:
	mov 	pc, lr					@ back to my caller
ENDPROC(bsp_scpu2_bisr)
#endif			

/*************************************************************************
 *
 * cpu_init_cp15
 *
 * Setup CP15 registers (cache, MMU, TLBs). The I-cache is turned on unless
 * CONFIG_SYS_ICACHE_OFF is defined.
 *
 *************************************************************************/
ENTRY(cpu_invalid_all)
	/*
	 * Invalidate L1 I/TLB/BTB
	 */
	mov	r0, #0			@ set up for MCR
	mcr p15, 0, r0, c8, c7, 0	@ invalidate TLBs
	/* fixme, why not to invalid I/D TLB ? */
#if 0 //def CONFIG_RTD298X  
	mcr p15, 0, r0, c8, c6, 0	@ invalidate data TLBs (DTLBIALL)
	mcr p15, 0, r0, c8, c5, 0	@ invalidate instruction TLBs (ITLBIALL)
#endif
	mcr p15, 0, r0, c7, c5, 0	@ invalidate icache
	mcr p15, 0, r0, c7, c5, 6	@ invalidate BP array
	mcr p15, 0, r0, c7, c10, 4	@ DSB
	mcr p15, 0, r0, c7, c5, 4	@ ISB
	dsb @ avoid above 2 deprecated after v8
	isb
	clrex

	/*
	 * Invalidate dcache all
	 */
	@ core0 invalid all, others invalid LOUIS
	ldr   r1, =0x0000ffff	
	mrc   p15, 0, r0, c0, c0, 5
	and	  r0, r1
	cmp   r0, #0x00
	moveq r3, #0
	beq   flush_level0

	mcr	p15, 0, r0, c7, c10, 5	@ dmb to ensure ordering with previous memory accesses
	mrc	p15, 1, r0, c0, c0, 1	@ read clidr
	ands	r3, r0, #0x7000000	@ extract loc from clidr
	mov	r3, r3, lsr #23		@ left align loc bit field
	beq	finished		@ if loc is 0, then no need to clean
flush_level0:	
	mov	r10, #0			@ start clean at cache level 0
flush_levels:
	add	r2, r10, r10, lsr #1	@ work out 3x current cache level
	mov	r1, r0, lsr r2		@ extract cache type bits from clidr
	and	r1, r1, #7		@ mask of the bits for current cache only
	cmp	r1, #2			@ see what cache we have at this level
	blt	skip			@ skip if no cache, or just i-cache
	mcr	p15, 2, r10, c0, c0, 0	@ select current cache level in cssr
	mcr 	p15, 0, r0, c7, c5, 4	@ isb to sych the new cssr&csidr
	mrc	p15, 1, r1, c0, c0, 0	@ read the new csidr
	and	r2, r1, #7		@ extract the length of the cache lines
	add	r2, r2, #4		@ add 4 (line length offset)
	ldr	r4, =0x400
	sub	r4, r4, #1
	ands	r4, r4, r1, lsr #3	@ find maximum number on the way size
	clz	r5, r4			@ find bit position of way size increment
	ldr	r7, =0x8000
	sub	r7, r7, #1
	ands	r7, r7, r1, lsr #13	@ extract max number of the index size
loop1:
	mov	r9, r4			@ create working copy of max way size
loop2:
	orr	r11, r10, r9, lsl r5	@ factor way and cache number into r11
	orr	r11, r11, r7, lsl r2	@ factor index number into r11
	mcr	p15, 0, r11, c7, c6, 2	@ invalidate by set/way
	subs	r9, r9, #1		@ decrement the way
	bge	loop2
	subs	r7, r7, #1		@ decrement the index
	bge	loop1
skip:
	add	r10, r10, #2		@ increment cache number
	cmp	r3, r10
	bgt	flush_levels
finished:
	mov	r10, #0			@ swith back to cache level 0
	mcr	p15, 2, r10, c0, c0, 0	@ select current cache level in cssr
	mcr	p15, 0, r0, c7, c10, 4	@ dsb
	mcr	p15, 0, r0, c7, c5, 4	@ isb
	dsb	@ avoid above 2 deprecated after v8
	isb
	mov	pc, lr			@ back to my caller
ENDPROC(cpu_invalid_all)

ENTRY(cpu_init_cp15)
	/*
	 * disable MMU stuff and caches
	 */
	mrc	p15, 0, r0, c1, c0, 0
	bic	r0, r0, #0x00002000	@ clear bits 13 (--V-)
	bic	r0, r0, #0x00000007	@ clear bits 2:0 (-CAM)
	orr	r0, r0, #0x00000002	@ set bit 1 (--A-) Align
	orr	r0, r0, #0x00000800	@ set bit 11 (Z---) BTB
#ifdef CONFIG_SYS_ICACHE_OFF
	bic	r0, r0, #0x00001000	@ clear bit 12 (I) I-cache
#else
	orr	r0, r0, #0x00001000	@ set bit 12 (I) I-cache
#endif
	mcr	p15, 0, r0, c1, c0, 0
	mov	pc, lr			@ back to my caller
ENDPROC(cpu_init_cp15)

#ifndef CONFIG_SKIP_LOWLEVEL_INIT
/*************************************************************************
 *
 * CPU_init_critical registers
 *
 * setup important registers
 * setup memory timing
 *
 *************************************************************************/
ENTRY(cpu_init_crit)
	/*
	 * Jump to board specific initialization...
	 * The Mask ROM will have already initialized
	 * basic memory. Go here to bump up clock rate and handle
	 * wake up conditions.
	 */
	mov	ip, lr			@ persevere link reg across call
	bl	lowlevel_init		@ go setup pll,mux,memory
	mov	lr, ip			@ restore link
	mov	pc, lr			@ back to my caller
ENDPROC(cpu_init_crit)
#endif

#if CONFIG_QC_SCPU_PWR_CTRL
ENTRY(cpu_core_pwr_dwn)
	/*
	 * @ r0: 1 or 2 or 3
	 */
	mrc p15, 0, r1, c15, c2, 7 @ CORTEX_A55_CPUPWRCTLR_EL1
	orr r1, r1, #0x1 @ CORTEX_A55_CORE_PWRDN_EN_MASK
	mcr	p15, 0, r1, c15, c2, 7 @ CORTEX_A55_CPUPWRCTLR_EL1
	add r2, r0, #0x5a @ print "[\]"
	ldr r3, =0x1801b100
	str r2, [r3]
	isb
	wfi

ENDPROC(cpu_core_pwr_dwn)
#endif

#ifdef CONFIG_QC_VERIFY
secondary_start:
	/*
	 * @ r0: 1 or 2 or 3
	 * store magic value:0x508 @0x508
	 * store magic value:0x510 @0x510
	 * store magic value:0x518 @0x518
	 */
	mov r1, #0x8
	mul r1, r1, r0
	mov r2, #0x0
	add r1, r1, #0x500
	str r2, [r1]
secondary_loop:
	ldr r3, =0x20000
	ldr r2, [r1]
	cmp r2,r3
	bne secondary_loop
	b run_master
#endif
#ifndef CONFIG_SPL_BUILD
/*
 *************************************************************************
 *
 * Interrupt handling
 *
 *************************************************************************
 */
@
@ IRQ stack frame.
@
#define S_FRAME_SIZE	72

#define S_OLD_R0	68
#define S_PSR		64
#define S_PC		60
#define S_LR		56
#define S_SP		52

#define S_IP		48
#define S_FP		44
#define S_R10		40
#define S_R9		36
#define S_R8		32
#define S_R7		28
#define S_R6		24
#define S_R5		20
#define S_R4		16
#define S_R3		12
#define S_R2		8
#define S_R1		4
#define S_R0		0

#define MODE_SVC 0x13
#define I_BIT	 0x80

/*
 * use bad_save_user_regs for abort/prefetch/undef/swi ...
 * use irq_save_user_regs / irq_restore_user_regs for IRQ/FIQ handling
 */

	.macro	bad_save_user_regs
	sub	sp, sp, #S_FRAME_SIZE		@ carve out a frame on current
						@ user stack
	stmia	sp, {r0 - r12}			@ Save user registers (now in
						@ svc mode) r0-r12
	ldr	r2, IRQ_STACK_START_IN		@ set base 2 words into abort
						@ stack
	ldmia	r2, {r2 - r3}			@ get values for "aborted" pc
						@ and cpsr (into parm regs)
	add	r0, sp, #S_FRAME_SIZE		@ grab pointer to old stack

	add	r5, sp, #S_SP
	mov	r1, lr
	stmia	r5, {r0 - r3}			@ save sp_SVC, lr_SVC, pc, cpsr
	mov	r0, sp				@ save current stack into r0
						@ (param register)
	.endm

	.macro	irq_save_user_regs
	sub	sp, sp, #S_FRAME_SIZE
	stmia	sp, {r0 - r12}			@ Calling r0-r12
	add	r8, sp, #S_PC			@ !! R8 NEEDS to be saved !!
						@ a reserved stack spot would
						@ be good.
	stmdb	r8, {sp, lr}^			@ Calling SP, LR
	str	lr, [r8, #0]			@ Save calling PC
	mrs	r6, spsr
	str	r6, [r8, #4]			@ Save CPSR
	str	r0, [r8, #8]			@ Save OLD_R0
	mov	r0, sp
	.endm

	.macro	irq_restore_user_regs
	ldmia	sp, {r0 - lr}^			@ Calling r0 - lr
	mov	r0, r0
	ldr	lr, [sp, #S_PC]			@ Get PC
	add	sp, sp, #S_FRAME_SIZE
	subs	pc, lr, #4			@ return & move spsr_svc into
						@ cpsr
	.endm

	.macro get_bad_stack
	ldr	r13, IRQ_STACK_START_IN		@ setup our mode stack (enter
						@ in banked mode)

	str	lr, [r13]			@ save caller lr in position 0
						@ of saved stack
	mrs	lr, spsr			@ get the spsr
	str	lr, [r13, #4]			@ save spsr in position 1 of
						@ saved stack

	mov	r13, #MODE_SVC			@ prepare SVC-Mode
	@ msr	spsr_c, r13
	msr	spsr, r13			@ switch modes, make sure
						@ moves will execute
	mov	lr, pc				@ capture return pc
	movs	pc, lr				@ jump to next instruction &
						@ switch modes.
	.endm

	.macro get_bad_stack_swi
	sub	r13, r13, #4			@ space on current stack for
						@ scratch reg.
	str	r0, [r13]			@ save R0's value.
	ldr	r0, IRQ_STACK_START_IN		@ get data regions start
						@ spots for abort stack
	str	lr, [r0]			@ save caller lr in position 0
						@ of saved stack
	mrs	r0, spsr			@ get the spsr
	str	lr, [r0, #4]			@ save spsr in position 1 of
						@ saved stack
	ldr	r0, [r13]			@ restore r0
	add	r13, r13, #4			@ pop stack entry
	.endm

	.macro get_irq_stack			@ setup IRQ stack
	ldr	sp, IRQ_STACK_START
	.endm

	.macro get_fiq_stack			@ setup FIQ stack
	ldr	sp, FIQ_STACK_START
	.endm

/*
 * exception handlers
 */
	.align	5
undefined_instruction:
	get_bad_stack
	bad_save_user_regs
	bl	do_undefined_instruction

	.align	5
software_interrupt:
	get_bad_stack_swi
	bad_save_user_regs
	bl	do_software_interrupt

	.align	5
prefetch_abort:
	get_bad_stack
	bad_save_user_regs
	bl	do_prefetch_abort

	.align	5
data_abort:
	get_bad_stack
	bad_save_user_regs
	bl	do_data_abort

	.align	5
not_used:
	get_bad_stack
	bad_save_user_regs
	bl	do_not_used

#ifdef CONFIG_USE_IRQ

	.align	5
irq_handler:
    sub sp, sp, #S_FRAME_SIZE
    stmia   sp, {r0 - r11}          @ Calling r0-r12
    add r8, sp, #S_PC               @ !! R8 NEEDS to be saved !!
    str lr, [r8, #0]                @ Save lr_irq
    mrs r6, spsr                    @ Get SPSR
    str r6, [r8, #4]                @ Save SPSR
    str r12,[r8, #8]                @ Save r12 for temp usage

    @Acknowledge Interrup
    ldr   r0, =0x18082000
    ldr   r1, =0xc                  @ Interrupt Acknowledge Register offset
    ldr   r2, [r0, r1]
    ldr   r3, =0x1FF                @ ACK_INTID mask
    and   r4, r2, r3                @ INTID - we need this at irq_handler_end
  #ifdef CONFIG_SYS_SUPPORT_MULTITHREAD
    cmp   r4, #16                   @Check if SGI
    bgt   do_timer
 #endif
    mov   r0, r4
    bl    handle_IPI                @do_ipi
    b     irq_handler_end
 #ifdef CONFIG_SYS_SUPPORT_MULTITHREAD
do_timer:
    cmp   r4, #29                   @Private timer ID
    bleq  arch_timer_handler
#endif
    blne  handle_IRQ                @do_irq
irq_handler_end:
    @Signal End of Interrupt
    ldr   r0, =0x18082000
    ldr   r1, =0x10                 @ End Of Interrupt Register offset
    str   r4, [r0, r1]
    cmp  r4, #2
    beq   end_int
 #ifdef CONFIG_SYS_SUPPORT_MULTITHREAD
    bl    need_to_reschedule
    cmp   r0, #0
    bne  resched
#endif

end_int:
    ldmia sp, {r0 - r11}            @ Calling r0 - lr
    ldr	  lr, [sp, #S_PC]           @ Get PC
    ldr   r12,[sp, #S_PC+4]
    msr   spsr, r12
    ldr	  r12, [sp, #S_PC+8]        @ Get PC
    add   sp, sp, #S_FRAME_SIZE
    subs  pc, lr, #4                @ return & move spsr_svc into cpsr
 #ifdef CONFIG_SYS_SUPPORT_MULTITHREAD
resched:
    cpsid aif, #0x13                @ Into the SVC mode with irq disabled/fiq disabled/abort disabled
    push   {lr}

    @ Free to go on SVC mode.
    bl    reschedule_irq
    clrex
    pop {lr}
    cpsid aif, #18
    b     irq_handler_end
#endif
	.align	5
fiq:
	get_fiq_stack
	/* someone ought to write a more effective fiq_save_user_regs */
	irq_save_user_regs
	bl	do_fiq
	irq_restore_user_regs

#else

	.align	5
irq_handler:
	get_bad_stack
	bad_save_user_regs
	bl	do_irq

	.align	5
fiq:
	get_bad_stack
	bad_save_user_regs
	bl	do_fiq

#endif /* CONFIG_USE_IRQ */
#endif /* CONFIG_SPL_BUILD */
ENTRY(call_smc)
        smc   #0
ENDPROC(call_smc)
 
