#include <common.h>
#include <command.h>
#include <fw_info.h>
#include <factorylib.h>
#include <errno.h>

#include <asm/arch/sys_flash_layout.h>
#if defined(CONFIG_SYS_RTK_EMMC_FLASH)
 #include "../drivers/mmc/rtkmmc.h"
#elif defined(CONFIG_SYS_RTK_NAND_FLASH)
 #include <linux/mtd/nand.h>		//surport nand flash driver
 #include <nand.h>
 #include <malloc.h>
#elif defined(CONFIG_SYS_RTK_SPI_FLASH)
 #include <flashdev_s.h>
 #include <parse_h.h>
#endif
#include <linux/ctype.h>
#include <asm/arch/system.h>

#ifdef CONFIG_RUN_FLASH_WRITER
#include <malloc.h>
#endif
struct layout_info
{
	u64 offset;      	/*fw image start address*/
	u64 align_size;		/*fw image align size*/
};

struct file_info
{
	uint file_type;		/*fw type eg: for  0 for linux_kernel*/
	uint file_addr;		/*ddr address*/
	uint file_size;		/*file size*/
};

struct burn_info
{
	struct file_info   fileinfo;
	struct layout_info layout;
};

struct burn_image_head
{
	uint check_sum;
	uint fw_tag;
};

struct media_info
{
	uint media_type;
	uint erase_size;
	u64  media_size;
};

struct fw_type_info
{
	int  fw_type_tag;
	char fw_type_name[128];
};

enum BL_TYPE{
	BL_TYPE_CERT=0,   		/*bootcode start*/
	BL_TYPE_PARM,
	BL_TYPE_AUX,
	BL_TYPE_HWSETTING,
	BL_TYPE_ARM_REDIECT,
	BL_TYPE_EXT_FW_CERT,
	BL_TYPE_EXT_FW,
	BL_TYPE_KBOOT,
	BL_TYPE_UBOOT,			/*bootcode end*/
	BL_TYPE_MAX,
};

struct bl_burn_info
{
	unsigned char bl_type;		/*bootcode type 0byte*/
	unsigned int  offset;		/*bootcode file offset  1~4byte*/
	unsigned int  size;		/*bootcode file size  5~8byte*/
	unsigned int  checksum;		/*bootcode file checksum 9~12byte*/
	unsigned char reserved[7]; 	/*reserved 13-19byte*/
}__attribute__((packed));

struct bl_burn_head
{
	unsigned char signature[8];     /*bootcode type 0 -7byte*/
	unsigned int  checksum;		/*bootcode type 8 -11byte*/
	unsigned int  total_bl_num;     /*bootcode type 12 -15byte*/
	unsigned char reserved[4]; 	/*reserved 16-19byte*/
}__attribute__((packed));

typedef struct T_NOCS_CONTROLLING_DATA
{
	unsigned int Market_SegmentID_EXT;
	unsigned int Mask_MSID;
	unsigned int Versioning_EXT;
	unsigned int Mask_Vers;
} NOCS_Controlling_Data;

typedef struct T_NOCS_CERTIFICATE
{
	unsigned int External_Public_RSA_Key[64];	// 2048 bit
	unsigned int Key_Rights;
	unsigned int Size_Of_SCS_Params_Area;
	unsigned int Address_Of_SCS_Param_Area;
	unsigned int External_R2_Public_RSA_Key[64];	// 2048 bit
	NOCS_Controlling_Data Controlling_Data;
	unsigned int Signature[64];
} NOCS_Certificate;

typedef struct T_NOCS_PARAMS
{
	unsigned int Key_Rights_CheckID;
	unsigned int External_Flash_Device_Params;
	unsigned int External_RAM_Params;
	unsigned int External_Boot_Specific_Params;
	unsigned int STR_Default_Region_Min_Size;
	unsigned int Address_of_Aux_Code_Area;	// patched this field as the size of kboot
	unsigned int Size_of_Aux_Code_Area;
	unsigned int Address_of_Boot_Code_Area; // patched this field as the size of uboot
	unsigned int Total_Area_Size[4];
	unsigned int Total_Area_Size_Selection;
	unsigned char Flash_Protection_Key_S[16];
	unsigned int Hw_Setting_Size;
	unsigned int Arm_Rediect_Size;
	unsigned int Ext_Fw_Cert_Size;
	unsigned int Ext_Fw_Size;
	NOCS_Controlling_Data Controlling_Data;
	unsigned int Flash_Protection_Key[4];
	unsigned int Signature[64];
} NOCS_Params;

typedef struct T_NOCS_AUX_CODE_TAIL
{
	unsigned int Key_Rights;
	NOCS_Controlling_Data Controlling_Data;
	unsigned int Signature[64];
}DTE_KBOOT_AREA_Tail;

#define TMP_DDR_ADDR				0x108000
#define TMP_DDR						"108000"
#define EMMC_MEIDA_TYPE				0
#define NAND_MEIDA_TYPE				1
#define EMMC_FW_TABLE_SIZE  		0x8000
#define EMMC_ERASE_SIZE				512
#define VERONA_SIGNATURE 			"VERONA__"
#define RSA_SIGNATURE_LENGTH 		256
#define STAR_COUNT 					30.0
#define EMMC_BOOTCODE_START			0x00020800
#define EMMC_EXT_FW_RESERVE_SIZE 	4096
#define CERT_SIZE					1024 
#define BURN_BL_IMG					0 //bootcode image
#define BURN_RESCUE_IMG				1 //rescue linux image
#define BURN_FW_IMG					2 //fw image
#define BL_SIGNATURE				"RTK_BL__"
#define SIZE_ALIGN_BOUNDARY_MORE(len, size) 	(((len - 1) & ~((size) - 1)) + size)

#define PRINT               printf

enum debug_level {
    DEBUG_LEVEL_DISABLE = 0,
    DEBUG_LEVEL_ERR,
    DEBUG_LEVEL_INFO,
    DEBUG_LEVEL_DEBUG
};

#define DEBUG_SET_LEVEL(x)   static int print_debug_level= x
#define	PRINT_ERR(fmt, args...) \
do {													\
	if (print_debug_level >= DEBUG_LEVEL_ERR) { 	\
		PRINT("%s(%d) %s: " "\033[1;31m"fmt"\033[m", __FILE__,__LINE__,__FUNCTION__,##args); \
	}													\
} while (0)

#define	PRINT_INFO(fmt, args...) \
do {													\
	if (print_debug_level >= DEBUG_LEVEL_INFO) { 	\
		PRINT("%s %d: "fmt, __FUNCTION__,__LINE__,##args); \
	}													\
} while (0)

#define	PRINT_DEBUG(fmt, args...) \
do {													\
	if (print_debug_level >= DEBUG_LEVEL_DEBUG) { 	\
		PRINT("%s(%d) %s: "fmt, __FILE__,__LINE__,__FUNCTION__,##args); \
	}													\
} while (0)

DEBUG_SET_LEVEL(DEBUG_LEVEL_INFO);
#define LAYOUT_FILE_NAME		FACTORY_HEADER_FILE_NAME"layout.txt"

extern uint get_checksum(uchar *p, uint len);
extern int sscanf(const char *buf, const char *fmt, ...);
static struct fw_type_info fw_type_info_arr[]=
{
	{FW_TYPE_KERNEL,		 "FW_KERNEL"}, 
	{FW_TYPE_AUDIO,			 "FW_AKERNEL"},
	{FW_TYPE_VIDEO,			 "FW_VKERNEL"},
	{FW_TYPE_AUDIO_2, 		 "FW_AKERNEL2"},
	{FW_TYPE_VIDEO_2,		 "FW_VKERNEL2"},
	{FW_TYPE_KCPU,			 "FW_KCPU"},
	{FW_TYPE_AUDIO_FILE,	 "FW_DALOGO"},
	{FW_TYPE_VIDEO_FILE,	 "FW_DVLOGO"},
	{FW_TYPE_IMAGE_FILE,	 "FW_DVLOGO"},
	{FW_TYPE_VIDEO_RAW_FILE, "FW_DVLOGO"},
};

static struct fw_type_info bl_type_info_arr[]=
{
	{BL_TYPE_CERT,		"cert.bin"}, 
	{BL_TYPE_PARM,		"params.bin"},
	{BL_TYPE_AUX,		"aux.bin"},
	{BL_TYPE_HWSETTING, 	"hw_setting.bin"},
	{BL_TYPE_ARM_REDIECT,	"arm_redirect.bin"},
	{BL_TYPE_EXT_FW_CERT,	"ext_fw_cert.bin"},
	{BL_TYPE_EXT_FW,	"ext_fw.bin"},
	{BL_TYPE_KBOOT,		"kboot.bin"},
	{BL_TYPE_UBOOT,		"bootcode.bin"},
};

void dump_buf_byte(char *buf, int buf_size)
{
    int loop = 0;
    printf("buf 0x%x size 0x%x", (unsigned int)buf, buf_size);
    for(loop = 0; loop < buf_size; loop++)
    {
        if(loop%16 == 0)
        {
            printf("\n%08xh:",loop); 
        }
        printf(" %02x",*buf);
        buf++;
    }
    printf("\n");	   
}

int send_file_to_tftp(char *file_path)
{
	char *tftp_argv[] = {"tftp", NULL, NULL, NULL};
	tftp_argv[1] = TMP_DDR;
	tftp_argv[2] = file_path;

	if (do_tftpb(NULL, 0, 3, tftp_argv)){
		PRINT_ERR("tftp fail ddr address %s file path %s \n",tftp_argv[1],tftp_argv[2]);
		return -ENOENT;
	}
	return 0;
}

int get_tftp_fileinfo(char *file_path,struct file_info *ddr_fileinfo)
{
	int ret = 0;
	/*send file to tftp*/
	if((ret = send_file_to_tftp(file_path)) < 0){
		return ret;
	}

	/*get file ddr address*/
	if((ddr_fileinfo->file_addr= getenv_ulong("fileaddr", 16, 0)) == 0){
		PRINT_ERR("get tftp ddr address size fail!\n");
		return -1;
	}
	
	/*get  file size */
	if((ddr_fileinfo->file_size= getenv_ulong("filesize", 16, 0)) == 0){
		PRINT_ERR("get tftp total size fail!\n");
		return -1;
	}
	return ret;
}

int get_usb_fileinfo(char *file_path,struct file_info *ddr_fileinfo)
{
	int ret = 0;
	char cmd[256] = {0};
	/*init usb*/
	if((ret = run_command( "usb start_ac", 0)) < 0){
		PRINT_ERR("init usb fail! \n");
		return ret;
	}

	/*read file content from usb*/
	sprintf(cmd, "fatload usb 0:1 %x %s", TMP_DDR_ADDR, file_path);
	if((ret = run_command(cmd, 0)) != 0){
		PRINT_ERR("read %s failed. (No such file)\n", file_path);
		return ret;
	}

	/*get file ddr address*/
	ddr_fileinfo->file_addr = TMP_DDR_ADDR;
	
	/*get  file size*/
	if((ddr_fileinfo->file_size = getenv_ulong("filesize", 16, 0)) == 0){
		PRINT_ERR("get usb total file size fail!\n");
		return -1;
	}
	return ret;
}

int get_ddr_fileinfo(int get_way, char *file_path, struct file_info *burn_fileinfo)
{
	int ret = 0;
	struct file_info ddr_fileinfo = {0};
	/*get file information*/
	if(get_way == 0){
		ret = get_tftp_fileinfo(file_path,&ddr_fileinfo);
	}else{
		ret = get_usb_fileinfo(file_path, &ddr_fileinfo);
	}
	return ret;
}

int get_bootcode_image_info(struct file_info *burn_fileinfo,
	struct file_info *ddr_fileinfo)
{
	int ret = -1;
	unsigned int  checksum = 0,offset = 0;
	struct bl_burn_head *bl_head = NULL;
	bl_head = (struct bl_burn_head *)((uchar *)ddr_fileinfo->file_addr);
	/*compare signature*/
	if(strncmp((char *)bl_head->signature,
		BL_SIGNATURE,sizeof(bl_head->signature)) != 0){
		return ret;
	}

	/*compare checksum*/
	offset = sizeof(bl_head->signature) + sizeof(bl_head->checksum);
	checksum = get_checksum((uchar *)(ddr_fileinfo->file_addr + offset),
			ddr_fileinfo->file_size - offset);
	if(checksum != bl_head->checksum){
		PRINT_INFO("check sum ok 0x%x fail 0x%x \n",
			bl_head->checksum,checksum);
	}else{
		memcpy(burn_fileinfo, ddr_fileinfo, sizeof(struct file_info));
		ret = 0;
	}
	return ret;
}

/*************************************************************
*
*
*		return value: 
*				1	:	burn rescue linux
*				0	:     non-rescue linux or parse error
**************************************************************/
int is_rescue_linux_data(void *data_buff, unsigned int data_size, unsigned int force_parse)
{
extern int rtk_get_secure_boot_type(void);
extern int decrypt_image_fpk(char *src, char *dst, uint length, unsigned int ini_key[4]);
	static int is_rescue = 0;
	static int is_check_done = 0;

	if(force_parse){
		is_rescue = 0;
		is_check_done = 0;
	}

	if(!is_check_done){
		unsigned int secure_mode;
		rescue_linux_info_t *rescue_linux_hdr_in = (rescue_linux_info_t *)TEMP_BUFFER_FOR_FLASH_DATA_ADDR;
		rescue_linux_info_t *rescue_linux_hdr_out = (rescue_linux_info_t *)(TEMP_BUFFER_FOR_FLASH_DATA_ADDR + 0x100000);

		memcpy(rescue_linux_hdr_in, data_buff, sizeof(rescue_linux_info_t));
		
		secure_mode = rtk_get_secure_boot_type();

		/* if secure mode, do decrypt header first */
		if( secure_mode == RTK_SECURE_BOOT) {
			if (decrypt_image_fpk((char *)rescue_linux_hdr_in,(char *)rescue_linux_hdr_out,
						((sizeof(rescue_linux_info_t)+(0x10-1))&(~(0x10-1))),NULL)){
				PRINT_ERR("Decrypt header error!\n");
				is_check_done++;
				return 0;
			}
		}

		//PRINT_INFO("magic = 0x%x \r\n", rescue_linux_hdr_out->dwMagicNumber);
		//PRINT_INFO("size    = 0x%x \r\n", rescue_linux_hdr_out->dwImageSize);

		if ((rescue_linux_hdr_out->dwMagicNumber == RESCUE_LINUX_HEADER_MAGICNO) && 
			(rescue_linux_hdr_out->dwImageSize == (data_size-RSA_SIGNATURE_LENGTH))){
			PRINT_INFO("Burn rescue linux data!\r\n");
			is_rescue = 1;
		}else{
			is_rescue = 0;
		}
		
		is_check_done++;
	}

	return is_rescue;
}

int get_rescue_image_info(struct file_info *burn_fileinfo,struct file_info *ddr_fileinfo)
{
	int ret = 0;
	if(is_rescue_linux_data((void *)(ddr_fileinfo->file_addr),
		ddr_fileinfo->file_size, 1) == 1){
		memcpy(burn_fileinfo, ddr_fileinfo, sizeof(struct file_info));
	}else{
		ret = -1;
	}
	return ret;
}
int verify_image_content(struct file_info *ddr_fileinfo)
{
	int ret = -1;
	uint checksum = 0;
	struct burn_image_head *image_head = NULL;
	image_head = (struct burn_image_head *)((uchar *)ddr_fileinfo->file_addr);
	/*get content checksum*/
	checksum = get_checksum((uchar *)ddr_fileinfo->file_addr + sizeof(struct burn_image_head), 
							ddr_fileinfo->file_size - sizeof(struct burn_image_head));
	PRINT_INFO("file checksum 0x%x \n",checksum);
	/*get fw tag checksum*/
	checksum += get_checksum((uchar *)&(image_head->fw_tag), sizeof(image_head->fw_tag));
	PRINT_INFO("fw tag 0x%x checksum 0x%x get checksum 0x%x \n",
		image_head->fw_tag,image_head->check_sum,checksum);
	/*compare checksum*/
	if(image_head->check_sum == checksum){
		ret = 0;
	}
	return ret;
}

void set_burn_fileinfo(struct file_info *ddr_fileinfo,struct file_info *burn_fileinfo)
{
	struct burn_image_head *image_head = NULL;
	image_head = (struct burn_image_head *)((uchar *)ddr_fileinfo->file_addr);
	burn_fileinfo->file_type = image_head->fw_tag;
	burn_fileinfo->file_addr = ddr_fileinfo->file_addr + sizeof(struct burn_image_head);
	burn_fileinfo->file_size = ddr_fileinfo->file_size - sizeof(struct burn_image_head);
	PRINT_INFO("burn file type %d file address 0x%x file size 0x%x \n",
		burn_fileinfo->file_type,burn_fileinfo->file_addr,burn_fileinfo->file_size);
}

int get_fw_image_info(struct file_info *burn_fileinfo,struct file_info *ddr_fileinfo)
{
	int ret = 0;
	/*verify image content */
	if((ret = verify_image_content(ddr_fileinfo))<0){
		return ret;
	}

	/*set burn file information*/
	set_burn_fileinfo(ddr_fileinfo,burn_fileinfo);
	return ret;
}

int get_burn_fileinfo(int get_way, char *file_path, struct file_info *burn_fileinfo,
	unsigned int *burn_type)
{
	int ret = 0;
	struct file_info ddr_fileinfo = {0};
	/*get ddr file information from tftp or usb*/
	if(get_way == 0){
		ret = get_tftp_fileinfo(file_path,&ddr_fileinfo);
	}else{
		ret = get_usb_fileinfo(file_path,&ddr_fileinfo);
	}

	if(ret != 0){
		return ret;
	}

	/*get bootcode image information*/
	if((ret = get_bootcode_image_info(burn_fileinfo,&ddr_fileinfo)) == 0){
		*burn_type = BURN_BL_IMG;
		return ret;
	}

	/*get rescue linux information*/
	if((ret = get_rescue_image_info(burn_fileinfo,&ddr_fileinfo)) == 0){
		*burn_type = BURN_RESCUE_IMG;
		return ret;
	}

	/*get fw image information*/
	if((ret = get_fw_image_info(burn_fileinfo,&ddr_fileinfo)) == 0){
		*burn_type = BURN_FW_IMG;
	}
	return ret;
}

int init_media_info(struct media_info *mediainfo)
{
	int ret = 0;
#if defined(CONFIG_SYS_RTK_NAND_FLASH)
	int dev = nand_curr_device;
	nand_info_t *nand =&nand_info[dev];		//nand_info[0];
	mediainfo->media_type = NAND_MEIDA_TYPE;
	mediainfo->media_size = nand->size;
	mediainfo->erase_size = nand->erasesize;
#elif defined( CONFIG_SYS_RTK_EMMC_FLASH)
	mediainfo->media_type = EMMC_MEIDA_TYPE;
	mediainfo->media_size = emmc_card.mmc->capacity;
	mediainfo->erase_size = EMMC_ERASE_SIZE;
#endif
	PRINT_INFO("media_type %d erase size 0x%x size 0x%llx \n",
		mediainfo->media_type,mediainfo->erase_size,mediainfo->media_size);
	return ret;
}

int check_align(u64 size,uint align_size)
{
	int ret = 0;
	ret = (size % align_size)?-1:0;
	PRINT_DEBUG("align_size 0x%x ret %d align %d  size 0x%llx\n",
		 align_size,ret,(int)(size % align_size),size);
	return ret;
}

int read_nand_to_buf(struct media_info *mediainfo,u64 offset, u64 size, uchar *buf)
{
	int ret = 0;
#ifdef CONFIG_SYS_RTK_NAND_FLASH
	int dev = nand_curr_device, loop = 0,page_num = 0;
	uint retlen = 0,left_size = 0;
	nand_info_t *nand =&nand_info[dev]; 	//nand_info[0];
	uchar *extra_buf = NULL;
	u64 read_offset = offset;
	/*check offset page size align*/
	if((ret = check_align(offset,nand->oobblock)) < 0){
		PRINT_ERR("offset not align 0x%llx error!\n",offset);
		goto end;
	}

	/*read page content*/
	page_num   = size/nand->oobblock;
	
	for(loop = 0; loop < page_num; loop++)
	{
		nand->read(nand, read_offset, nand->oobblock, &retlen,buf);
		if (retlen != nand->oobblock)
		{
			PRINT_ERR("offset 0x%llx size 0x%x fail! \n",read_offset,nand->oobblock);
			ret = -1;
			goto end;
		}
		read_offset += nand->oobblock;
		buf += nand->oobblock;
	}

	/*read left page*/
	left_size = size%nand->oobblock;
	if(left_size != 0){
		if((extra_buf = (uchar *)malloc(sizeof(uchar)*nand->oobblock))== NULL){
			PRINT_ERR("malloc size 0x%x error!\n",nand->oobblock);
			ret = -1;
			goto end;
		}
		nand->read(nand, read_offset,nand->oobblock, &retlen, extra_buf);
		if (retlen != nand->oobblock)
		{
			PRINT_ERR("offset 0x%llx size 0x%x fail! \n",read_offset,nand->oobblock);
			ret = -1;
			goto end;
		}
		memcpy(buf,extra_buf,left_size);
	}
end:
	if(extra_buf != NULL){
		free(extra_buf);
	}
#endif
	return ret;
}

int read_emmc_to_buf(struct media_info *mediainfo,u64 offset, u64 size, uchar *buf)
{
	int ret = 0;
#ifdef CONFIG_SYS_RTK_EMMC_FLASH
	uint blk_cnt = 0,left_size = 0;
	uchar *extra_buf = NULL;
	u64 read_offset = offset;
	if((ret = check_align(read_offset,mediainfo->erase_size)) < 0){
		PRINT_ERR("offset not align 0x%llx error!\n",read_offset);
		ret = -1;
		goto end;
	}
	blk_cnt = rtk_eMMC_read(read_offset/mediainfo->erase_size,size,(unsigned int *)buf);
	if(blk_cnt != (size/mediainfo->erase_size)){
		PRINT_ERR("read  blk_cnt%d  read cnt %d offset 0x%llx size 0x%llx error !\n",
			blk_cnt,(int)(size/mediainfo->erase_size),read_offset,size);
		ret = -1;
		goto end;
	}
	
	left_size = (size%mediainfo->erase_size);
	if(left_size !=0 ){
		if((extra_buf = (uchar *)malloc(sizeof(uchar)*mediainfo->erase_size))== NULL){
			PRINT_ERR("malloc size 0x%x error!\n",mediainfo->erase_size);
			ret = -1;
			goto end;
		}
		read_offset += blk_cnt*mediainfo->erase_size;
		blk_cnt = rtk_eMMC_read(read_offset/mediainfo->erase_size,mediainfo->erase_size,
				(unsigned int *)extra_buf);
		if (blk_cnt != 1)
		{
			PRINT_ERR("offset 0x%llx size 0x%x fail!",read_offset,mediainfo->erase_size);
			ret = -1;
			goto end;
		}
		memcpy(buf,extra_buf,left_size);
	}
end:
	if(extra_buf != NULL){
		free(extra_buf);
	}
#endif
	return ret;
}

int read_media_to_buf(struct media_info *mediainfo,u64 offset, u64 size, uchar *buf)
{
	int ret = 0;
	if(mediainfo->media_type == NAND_MEIDA_TYPE){
		ret = read_nand_to_buf(mediainfo,offset,size,buf);
	}else if(mediainfo->media_type == EMMC_MEIDA_TYPE){
		ret = read_emmc_to_buf(mediainfo,offset,size,buf);
	}
	return ret;
}

int is_fwtable_loc(struct media_info *mediainfo,char * fw_table_buf,
	struct layout_info *layout)
{
	int ret = -1;
	fw_desc_table_v1_t *fw_dec_table = NULL;
	/*get fwtable content*/
	if(read_media_to_buf(mediainfo,layout->offset,layout->align_size,(unsigned char *)fw_table_buf) < 0){
		PRINT_ERR("get fw table content fail! \n");
		return ret;
	}

	/*check signature*/
	fw_dec_table = (fw_desc_table_v1_t *)fw_table_buf;
	if(strncmp((char *)fw_dec_table->signature,VERONA_SIGNATURE,sizeof(fw_dec_table->signature)) != 0){
		PRINT_DEBUG("offset 0x%llx align_size 0x%llx signature check fail OK signature is %s!!\n",
				layout->offset,layout->align_size,VERONA_SIGNATURE);
	}else{
		//dump_buf_byte(fw_dec_table,sizeof(fw_desc_table_v1_t));
		ret = 0;
	}
	return ret;
}

void get_default_fwtable_layout(struct media_info *mediainfo,struct layout_info *fwtable_layout)
{
	if(mediainfo->media_type == NAND_MEIDA_TYPE){
#ifdef CONFIG_SYS_RTK_NAND_FLASH
		fwtable_layout->offset 		= SYS_PARAM_FW_TABLE_START;
		fwtable_layout->align_size	= mediainfo->erase_size;
#endif
	}else if(mediainfo->media_type ==  EMMC_MEIDA_TYPE){
		fwtable_layout->offset      = SYS_PARAM_FW_TABLE_START;
		fwtable_layout->align_size	= EMMC_FW_TABLE_SIZE;
	}
	PRINT_INFO("default fwtable layout start 0x%llx size 0x%llx \n",
		fwtable_layout->offset,fwtable_layout->align_size);
}

char *get_fwtable_from_media(struct media_info *mediainfo,struct layout_info *fwtable_layout)
{
	char * fw_table_buf = NULL;
	struct layout_info pre_layout = {0},cur_layout = {0};
	if((fw_table_buf = (char *)malloc(sizeof(char)*fwtable_layout->align_size)) == NULL){
		PRINT_ERR("malloc size 0x%llx fail!\n",fwtable_layout->align_size);
		return NULL;
	}
	/*use default fwtable layout*/
	memcpy((char *)&cur_layout, (char *)fwtable_layout,sizeof(struct layout_info));
	memcpy((char *)&pre_layout, (char *)fwtable_layout,sizeof(struct layout_info));
	
	/*get fwtable content and fw table*/
	while((cur_layout.offset < mediainfo->media_size)
		&&(pre_layout.offset  > pre_layout.align_size)){
		PRINT_DEBUG("cur start 0x%llx offset 0x%llx \n",cur_layout.offset,cur_layout.align_size);
		/*check fwtable correct or not -->serach add mode*/
		if((is_fwtable_loc(mediainfo,fw_table_buf,&cur_layout) == 0)){
			PRINT_INFO("use fwtable layout offset 0x%llx size 0x%llx default:offset 0x%llx size 0x%llx\n",
				cur_layout.offset,cur_layout.align_size,fwtable_layout->offset,fwtable_layout->align_size);
			fwtable_layout->offset = cur_layout.offset;
			fwtable_layout->align_size = cur_layout.align_size;
			break;
		}
		cur_layout.offset += cur_layout.align_size;

		/*check fwtable correct or not -->serach decrease mode*/
		pre_layout.offset = pre_layout.offset - pre_layout.align_size;
		PRINT_DEBUG("pre start 0x%llx offset 0x%llx \n",pre_layout.offset,pre_layout.align_size);
		if((is_fwtable_loc(mediainfo,fw_table_buf,&pre_layout) == 0)){
			PRINT_INFO("use fwtable layout offset 0x%llx size 0x%llx default:offset 0x%llx size 0x%llx\n",
				pre_layout.offset,pre_layout.align_size,fwtable_layout->offset,fwtable_layout->align_size);
			fwtable_layout->offset = pre_layout.offset;
			fwtable_layout->align_size = pre_layout.align_size;
			break;
		}
	}

	if((cur_layout.offset >= mediainfo->media_size)||(pre_layout.offset<=pre_layout.align_size)){
		PRINT_ERR("can't find fwtable content!\n");
		free(fw_table_buf);
		fw_table_buf =  NULL;
	}
	return fw_table_buf;
}

char * get_fwtable_info(struct media_info *mediainfo,struct layout_info *fwtable_layout)
{
	char * fw_table_buf =  NULL;
	/*get default fwtable layout*/
	get_default_fwtable_layout(mediainfo,fwtable_layout);
	
	/*get fwtable buffer and use fwtable layout*/
	fw_table_buf = get_fwtable_from_media(mediainfo,fwtable_layout);
	return fw_table_buf;	
}	

int get_fwinfo_from_fwtable(char *fw_table_buf,struct burn_info * burninfo,
		int *need_update_fwtable)
{
	int fw_exist = -1;
	uint loop = 0,offset = 0,fw_num = 0,fw_checksum = 0,checksum = 0;
	fw_desc_table_v1_t *fw_dec_table = NULL;
	fw_desc_entry_v1_t *fw_entry = NULL;
	fw_dec_table = (fw_desc_table_v1_t *)fw_table_buf;
	/*get total fw image number*/
	if(fw_dec_table->version == FW_DESC_TABLE_V1_T_VERSION_1){
		fw_num = fw_dec_table->fw_list_len/sizeof(fw_desc_entry_v1_t);
	}else if(fw_dec_table->version == FW_DESC_TABLE_V1_T_VERSION_21){
		fw_num = fw_dec_table->fw_list_len/sizeof(fw_desc_entry_v21_t);
	}
	offset = sizeof(fw_desc_table_v1_t) + fw_dec_table->part_list_len;
	PRINT_INFO("version 0x%x fw_num %d fw_list_len %d v1 size %d v21 size %d offset %d\n",
		fw_dec_table->version,fw_num,fw_dec_table->fw_list_len,
		sizeof(fw_desc_entry_v1_t),sizeof(fw_desc_entry_v21_t),offset);
	for(loop=0; loop < fw_num;loop++)
	{
		/*get fw entry*/
		if(fw_dec_table->version == FW_DESC_TABLE_V1_T_VERSION_1){
			fw_entry = (fw_desc_entry_v1_t *)(fw_table_buf+offset+
					sizeof(fw_desc_entry_v1_t)*loop);
		}else if(fw_dec_table->version == FW_DESC_TABLE_V1_T_VERSION_21){
			fw_entry = (fw_desc_entry_v1_t *)(fw_table_buf+offset+
					sizeof(fw_desc_entry_v21_t)*loop);
		}

		/*get  fw image layout by fw type and update fwtable information or not*/
		if(fw_entry->type == burninfo->fileinfo.file_type){
			PRINT_DEBUG("fw info:offset 0x%x length 0x%x align size 0x%x checksum 0x%x\n",
				fw_entry->offset,fw_entry->length,fw_entry->paddings,fw_entry->checksum);
			if(burninfo->fileinfo.file_size > fw_entry->paddings){
				PRINT_ERR("fw size 0x%x out of max size 0x%x!!! \n",burninfo->fileinfo.file_size,fw_entry->paddings);
			}else{
				burninfo->layout.offset = fw_entry->offset;
				burninfo->layout.align_size = fw_entry->paddings;
				fw_checksum = get_checksum((uchar *)burninfo->fileinfo.file_addr, burninfo->fileinfo.file_size);
				/*update fw entry in fwtable-->if fw image checksum or size change*/
				if((burninfo->fileinfo.file_size != fw_entry->length)||
					(fw_checksum != fw_entry->checksum)){
					*need_update_fwtable = 1;
					fw_entry->length   = burninfo->fileinfo.file_size;
					fw_entry->checksum = fw_checksum;
					checksum = get_checksum((uchar*)fw_table_buf +offsetof(fw_desc_table_v1_t, version),
						sizeof(fw_desc_table_v1_t) -
						offsetof(fw_desc_table_v1_t, version) +
						fw_dec_table->part_list_len +fw_dec_table->fw_list_len);
					fw_dec_table->checksum = checksum;
					PRINT_INFO("change fw length 0x%x checksum 0x%x fwtable checksum 0x%x!\n",
						fw_entry->length,fw_checksum,fw_dec_table->checksum);
				}else{
					PRINT_INFO("not need change fwtable content!\n");
				}
				PRINT_INFO("find fw entry need update fwatble flag %d \n",*need_update_fwtable);
				fw_exist = 0;
			}
			break;
		}
	}

	if(fw_exist != 0){
		PRINT_ERR("can't find index %d in fwtable or out of max fw size!\n",burninfo->fileinfo.file_type);
	}
	return fw_exist;
}

int my_memcmp(const uchar *pa, const uchar *pb, int size)
{
	int i = 0, error_count = 0;
	for (i = 0; i < size; i++, pa++, pb++)
	{
		if (*pa != *pb){
			error_count++;
			PRINT_ERR("byte %d size %d error_count %d src 0x%x dst 0x%x \n",
				i,size,error_count,*pa,*pb);
		}
	}
	return error_count;
}

int write_page_to_nand(int start_page, int page_num,uchar *buf)
{
	int ret = 0;
#ifdef CONFIG_SYS_RTK_NAND_FLASH
	int loop = 0, dev = nand_curr_device;
	nand_info_t *nand =&nand_info[dev]; 	//nand_info[0];
	uint retlen = 0;
	uchar *compare_buf = NULL;
	u64 offset = 0;
	if((compare_buf = (uchar *)malloc(nand->oobblock*sizeof(uchar))) == NULL){
		PRINT_ERR("malloc size size 0x%x fail! \n",nand->oobblock);
		return -1;
	}
	offset = start_page*nand->oobblock;
	for(loop = 0; loop < page_num; loop++)
	{
		nand->write(nand, offset, nand->oobblock, &retlen,buf);
		if (retlen !=  nand->oobblock){
			PRINT_ERR("offset 0x%llx size 0x%x fail! \n",offset,nand->oobblock);
			ret = -1;
			goto end;
		}
		nand->read(nand,offset,nand->oobblock,&retlen,compare_buf);
		if (retlen !=  nand->oobblock){
			ret = -1;
			goto end;
		}
		if(my_memcmp(buf,compare_buf,nand->oobblock)!=0){
			PRINT_ERR("compare offset 0x%llx size 0x%x error!\n",offset,nand->oobblock);
			ret = -1;
			goto end;
		}
		offset += nand->oobblock;
		buf += nand->oobblock;
	}
end:
	if(compare_buf != NULL){
		free(compare_buf);
	}
#endif
	return ret;
}

int write_buf_to_nand(struct media_info *mediainfo,u64 offset, u64 size, uchar *buf)
{
	int ret = 0;
#ifdef CONFIG_SYS_RTK_NAND_FLASH
	int dev = nand_curr_device,page_num = 0,page_start = 0;
	uchar *extra_buf = NULL;
	struct erase_info instr;
	uint retlen = 0;
	nand_info_t *nand =&nand_info[dev];

	/*check nand block offset align*/
	if((ret = check_align(offset,mediainfo->erase_size))< 0){
		PRINT_ERR("offset size 0x%llx not align size 0x%x fail! \n",offset,mediainfo->erase_size);
		return ret;
	}

	if(check_align(size,mediainfo->erase_size) < 0){
		if((extra_buf = (uchar *)malloc(mediainfo->erase_size*sizeof(uchar))) == NULL){
			PRINT_ERR("malloc size size 0x%x fail! \n",mediainfo->erase_size);
			ret = -1;
			goto end;
		}
		nand->read(nand,((offset+size)/mediainfo->erase_size)*mediainfo->erase_size,
			mediainfo->erase_size,&retlen,extra_buf);
		if (retlen !=  nand->erasesize){
			ret = -1;
			goto end;
		}
		memcpy(extra_buf,buf+(size/mediainfo->erase_size)*mediainfo->erase_size,size%mediainfo->erase_size);
	}

	/*erase block*/
	instr.mtd = nand;
	instr.addr = offset;
	instr.len = ((size+nand->erasesize-1)/nand->erasesize)*nand->erasesize;
	nand->erase(nand,&instr);

	/*write block*/
	page_start = offset/nand->oobblock;
	page_num   = (size/nand->erasesize)*(nand->erasesize/nand->oobblock);
	if((ret = write_page_to_nand(page_start,page_num,buf)) < 0){
		goto end;
	}
	
	/*write extra block*/
	if(check_align(size,mediainfo->erase_size) < 0){
		if((ret = write_page_to_nand(page_start+page_num,
				(nand->erasesize/nand->oobblock),extra_buf)) < 0){
			goto end;
		}	
	}
end:
	if(extra_buf != NULL){
		free(extra_buf);
	}
#endif
	return ret;
}

void display_progress(unsigned int len, unsigned int length)
{
	int i = 0, starcount = 0;
	if((length < 2)||(len < 1024)){
		return;
	}
	
	if(len == 0){
		printf("\n");
	}
	printf("\r");
	printf("%3d |", (int) (100.0*len/length));
	starcount = (int) (STAR_COUNT*len/length);
	for(i=0;i<starcount;i++) 
		printf("*");
	for(i=0;i<STAR_COUNT-starcount;i++)
			printf(" ");
	if(len > 1024){
                printf("| %u KB", len>>10);
        }else{
                printf("| %u B", len);
        }
	if(len == length) 
		printf("\n");
}

int write_block_to_emmc(struct media_info *mediainfo,int start_block, int total_block,uchar *buf)
{
	int ret = 0;
#ifdef CONFIG_SYS_RTK_EMMC_FLASH
	int loop = 0,write_blk = 0,read_blk = 0;
	uchar *compare_buf =  NULL;
	/*malloc compare buffer*/
	if((compare_buf = (uchar *)malloc(sizeof(uchar)*mediainfo->erase_size)) == NULL){
		PRINT_ERR("malloc size %d fail! \n",mediainfo->erase_size);
		return -1;
	}

	for(loop = 0; loop < total_block; loop++)
	{
		write_blk = rtk_eMMC_write(start_block+loop,mediainfo->erase_size,(unsigned int *)buf);
		if(write_blk != 1){
			PRINT_ERR("write block %d size 0x%x error!\n",(start_block+loop),mediainfo->erase_size);
			ret = -1;
			goto end;
		}
		read_blk = rtk_eMMC_read(start_block+loop,mediainfo->erase_size,(unsigned int *)compare_buf);
		if(read_blk != 1){
			PRINT_ERR("read block %d size 0x%x error!\n",(start_block+loop),mediainfo->erase_size);
			ret = -1;
			goto end;
		}
		if(my_memcmp(buf,compare_buf,mediainfo->erase_size)!=0){
			PRINT_ERR("compare block %d size 0x%x error!\n",(start_block+loop),mediainfo->erase_size);
			ret = -1;
			goto end;
		}else if(loop %64 == 0){
			display_progress(loop*mediainfo->erase_size,total_block*mediainfo->erase_size);
		}
		buf += mediainfo->erase_size;
	}
	display_progress(total_block*mediainfo->erase_size,total_block*mediainfo->erase_size);
end:
	if(compare_buf != NULL){
		free(compare_buf);
	}
#endif
	return ret;
}

int write_buf_to_mmc(struct media_info *mediainfo,u64 offset,u64 size,uchar *buf)
{
	int ret = 0;
#ifdef CONFIG_SYS_RTK_EMMC_FLASH
	int read_blk = 0;
	uint total_block = size/mediainfo->erase_size;
	uchar *extra_buf = NULL;
	/*check emmc offset  erase align*/
	if((ret = check_align(offset,mediainfo->erase_size))< 0){
		return ret;
	}

	/*write align block to emmc*/
	if((ret = write_block_to_emmc(mediainfo,offset/mediainfo->erase_size,
			total_block,buf)) < 0){
		goto end;
	}

	/*check emmc size erase align-->if not,get extra buffer to emmc*/
	if(check_align(size,mediainfo->erase_size) < 0){
		if((extra_buf = (uchar *)malloc(sizeof(uchar)*mediainfo->erase_size)) == NULL){
			PRINT_ERR("malloc size %d fail! \n",mediainfo->erase_size);
			ret = -1;
			goto end;
		}
		/*get extra buffer content*/
		memset(extra_buf,0xFF,sizeof(uchar)*mediainfo->erase_size);
		read_blk = rtk_eMMC_read((offset+size)/mediainfo->erase_size,
				mediainfo->erase_size,(unsigned int *)extra_buf);
		if(read_blk != 1){
			PRINT_ERR("read offset 0x%llx size 0x%x error!\n",(offset+size),mediainfo->erase_size);
			ret = -1;
			goto end;
		}
		memcpy(extra_buf,buf+total_block*mediainfo->erase_size,size%mediainfo->erase_size);

		/*write extra buffer content*/
		if((ret = write_block_to_emmc(mediainfo,offset/mediainfo->erase_size+total_block,
				1,extra_buf)) < 0){
			goto end;
		}
	}

end:
	if(extra_buf != NULL){
		free(extra_buf);
	}
#endif
	return ret;
}

int write_buf_to_media(struct media_info *mediainfo,u64 offset,u64 size,uchar *buf)
{
	int ret = 0;
	if(mediainfo->media_type == NAND_MEIDA_TYPE){
		ret = write_buf_to_nand(mediainfo,offset,size,buf);
	}else if(mediainfo->media_type == EMMC_MEIDA_TYPE){
		ret = write_buf_to_mmc(mediainfo,offset,size,buf);
	}
	if(ret == 0){
		PRINT_INFO("write media offset 0x%llx size 0x%llx successful!\n",offset,size);
	}
	return ret;
}

int get_fw_name_by_tag(char * fw_name,int fw_tag,
	struct fw_type_info *type_info_arry, unsigned int arry_len)
{
	int ret = -1;
	unsigned int loop = 0;
	for(loop = 0; loop < arry_len; loop++)
	{
		if(type_info_arry[loop].fw_type_tag == fw_tag){	
			break;
		}
	}
	
	if(loop != arry_len){
		sprintf(fw_name,"%s",type_info_arry[loop].fw_type_name);
		PRINT_INFO("fw tag %d fw name %s !\n",fw_tag,fw_name);
		ret = 0;
	}else{
		PRINT_ERR("can't find fw tag %d!\n",fw_tag);
	}
	return ret;
}

int get_line_from_buf(char *buf, int buf_len,char *line)
{
	int ret = -1, loop = 0;
	char tmp_byte = 0;
	for(loop = 0; loop < buf_len; loop++)
	{
		tmp_byte = buf[loop];
		if((tmp_byte ==  '\n') ||(tmp_byte == '\r')){
			strncpy(line,buf, loop+1);
			ret = 0;
			break;
		}
	}
	return ret;
}

int get_dig_num(u32 dig)
{
	int num = 0;
	char size_byte[16] = {0};
	snprintf(size_byte,sizeof(size_byte)/sizeof(size_byte[0]),
			"%x",dig);
	num = strlen(size_byte);
	return num;
}

int get_fw_entry_info(char *key,struct burn_info *burninfo,int *need_update
	,char * update_key)
{
	int ret = 0;
	u32 mem_offset = 0,offset = 0,file_size = 0,size_pos = 0,before_len = 0,after_len = 0;
	char size_byte[16] = {0};
	char *find_str = NULL,*next_str = NULL;
	sscanf(key, "\" target=%x offset=%x size=%x", 
		&(mem_offset),&(offset),&(file_size));
	PRINT_INFO("key is %s len:%d mem_offset %x offset %x file_size %x\n",
		key,strlen(key),mem_offset,offset,file_size);
	/*change size= content*/
	if( burninfo->fileinfo.file_size!= file_size){	
		if(((find_str = strstr(key,"size="))!= NULL)&&
			((next_str = strstr(key," type="))!= NULL)){
			size_pos   = find_str + strlen("size=")- key;
			before_len = next_str - (find_str + strlen("size="));
			after_len  = get_dig_num(burninfo->fileinfo.file_size); 
			PRINT_DEBUG("find str %s next str %s size_pos 0x%x before_len %d after_len %d \n",
				find_str,next_str,size_pos,before_len,after_len);
			
			snprintf(size_byte,sizeof(size_byte)/sizeof(size_byte[0]),
						"%x",burninfo->fileinfo.file_size);
			if(before_len != after_len){
				PRINT_INFO("change buffer len before %d after %d \n",before_len, after_len);
				memcpy(update_key,key,size_pos); /* 0~size_pos-1*/
				memcpy(update_key+size_pos,size_byte,after_len); /*size_pos~size_pos+after_len-1*/
				memcpy(update_key+size_pos+after_len,next_str,
					strlen(key)-(size_pos+before_len)); /*size_pos+after_len~size_pos+after_len+left_len*/
				*need_update = 2;
			}else{
				memcpy(key +size_pos,size_byte,after_len);
				*need_update = 1;
			}
		}
		PRINT_INFO("need change flag %d after key %s len %d before size 0x%x after 0x%x \n", 
			*need_update,key,strlen(key),file_size,burninfo->fileinfo.file_size);
	}else{
		PRINT_INFO("no need change  fw entry!\n");
	}
	return ret;
}

char * get_new_file_buf(char *file_buf, int change_offset,
	char *change_buf,int chang_buf_size, char *left_buf,
	int left_len, char *new_buf_len)
{
	char *buf =  NULL;
	int buf_len = 0;
	buf_len = change_offset + chang_buf_size + left_len;
	*new_buf_len = buf_len;
	PRINT_DEBUG("change_offset 0x%x chang_buf_size 0x%x left_len 0x%x total len 0x%x\n",
		change_offset,chang_buf_size,left_len,buf_len);
	if((buf = (char *)malloc(sizeof(char)*buf_len)) == NULL){
		PRINT_ERR("malloc buf len 0x%x error !\n",buf_len);
		return buf;
	}
	memcpy(buf,file_buf,change_offset);
	memcpy(buf+change_offset,change_buf,chang_buf_size);
	memcpy(buf+change_offset+chang_buf_size,left_buf,left_len);
	return buf;
}

int get_boottable_content(char *file_name,char **file_buf,int *file_length,
		struct burn_info *burninfo,int *need_update)
{
	int ret = 0,tmp_file_len = 0;
	char fw_name[512] = {0},line[1024] = {0},key[512] = {0},ch_key[512] ={0};
	char *tmp_buf = NULL,*key_value = NULL;
	
	/*get fw name by fw tag eg:FW_KERNEL*/
	if((ret = get_fw_name_by_tag(fw_name,burninfo->fileinfo.file_type,
		fw_type_info_arr,sizeof(fw_type_info_arr)/sizeof(fw_type_info_arr[0]))) <0){
		return ret;
	}
	
	/*get layout.txt content buffer and length*/
	if (factory_read(file_name, file_buf, file_length)){
		PRINT_INFO("can't find file:%s!\n", file_name);
		return ret;
	}
	tmp_file_len = *file_length;
	tmp_buf = *file_buf;
	
	/*get  line content  eg:#define FW_KERNEL " target=deaddead offset=xxx size=xxx type=bin "*/
	while((tmp_file_len > 0)&&(get_line_from_buf(tmp_buf,tmp_file_len,line) == 0))
	{
		/*special case-->layout.txt data lost*/
		if(strlen(line) == 0){
			PRINT_INFO("line content may error! \n");
			break;
		}
	
		/*get key content,eg:FW_KERNEL*/
		sscanf(line,"#define  %s",key);
		PRINT_DEBUG("line %s len %d key is %s len %d \n",line,strlen(line),key,strlen(key));

		/*key match fw_name eg:FW_KERNEL */
		if(strstr(key,fw_name)){
			/*get key value content eg:" target=deaddead offset=xxx size=xxx type=bin "*/
			key_value = line+strlen("#define ")+strlen(key)+1;
			if((ret = get_fw_entry_info(key_value,burninfo,need_update,ch_key)) == 0){
				if( *need_update != 0){
					PRINT_DEBUG("need update factory key value %s\n",key_value);
					if(strlen(ch_key) == 0){
						memcpy(tmp_buf+(strlen(line)- strlen(key_value)),
							key_value,strlen(key_value));
					}else{
						PRINT_DEBUG("change key value %s len %d \n",ch_key,strlen(ch_key));
						/*get new file buf and file length*/
						if((*file_buf = get_new_file_buf(*file_buf,
								tmp_buf - *file_buf +(strlen(line)- strlen(key_value)),
								ch_key,strlen(ch_key),tmp_buf+strlen(line),
								(tmp_file_len - strlen(line)),(char *)file_length)) == NULL){
							ret = -1;
						}
					}
				}
			}
			break;
		}
		
		tmp_buf += strlen(line);
		tmp_file_len -= strlen(line);
		memset(line,0,sizeof(line)/sizeof(line[0]));
		PRINT_DEBUG("file_buf 0x%x dst_length %d!\n", (unsigned int)tmp_buf,tmp_file_len);
	}
	return ret;
}

int update_boot_table(struct burn_info *burninfo)
{
	int ret = -1,file_length = 0,need_update = 0;
	char *file_buf;
	
	/*get layout.txt file buffer content*/
	if((ret = get_boottable_content(LAYOUT_FILE_NAME,&file_buf,
			&file_length,burninfo,&need_update)) < 0){
		return ret;
	}

	/*update layout.txt file buffer content to media*/
	if(need_update){
		if((ret = factory_write(LAYOUT_FILE_NAME, file_buf, file_length)!=0)) {
			PRINT_ERR("factory_write %s len %d error!\n",LAYOUT_FILE_NAME,file_length);
			return ret;
		}

		if((ret = factory_save())!=0) {
			PRINT_ERR("factory_save error!\n");
			return ret;
		}

		/*need_update -->malloc new buffer,need free*/
		if(need_update == 2){
			free(file_buf);
		}
	}
	return ret;
}

int get_emmc_cert_info(struct media_info *mediainfo,NOCS_Certificate* cert_info)
{
	int ret = 0;
	unsigned char * cer_buf = NULL;
	if((cer_buf = (unsigned char *)malloc(sizeof(unsigned char)*CERT_SIZE)) == NULL){
		PRINT_ERR("malloc size 0x%x fail! \n",CERT_SIZE);
		return -1;
	}
	
	/*get cert information*/
	if((ret = read_media_to_buf(mediainfo,EMMC_BOOTCODE_START,
		CERT_SIZE,(unsigned char *)cer_buf)) < 0){
		PRINT_ERR("get cert content fail! \n");
	}else{
		memcpy((unsigned char *)cert_info,cer_buf,sizeof(NOCS_Certificate));
	}
	
	if(cer_buf!= NULL){
		free(cer_buf);
	}
	return ret;
}

int get_cert_info(struct media_info *mediainfo,NOCS_Certificate* cert_info)
{
	int ret = 0;
	if(mediainfo->media_type == EMMC_MEIDA_TYPE){
		ret = get_emmc_cert_info(mediainfo,cert_info);
	}
	return ret;
}

int get_emmc_param_info(struct media_info *mediainfo,
	NOCS_Certificate* cert_info,NOCS_Params *param_info)
{
	int ret = 0;
	unsigned char * param_buf = NULL;
	unsigned param_offset = 0;
	if((param_buf = (unsigned char *)malloc(sizeof(unsigned char)*
			cert_info->Size_Of_SCS_Params_Area)) == NULL){
		PRINT_ERR("malloc size 0x%x fail! \n",cert_info->Size_Of_SCS_Params_Area);
		return -1;
	}
	
	/*get param information*/
	param_offset = EMMC_BOOTCODE_START + SIZE_ALIGN_BOUNDARY_MORE(CERT_SIZE,EMMC_ERASE_SIZE);
	if((ret = read_media_to_buf(mediainfo,param_offset,
		cert_info->Size_Of_SCS_Params_Area,(unsigned char *)param_buf)) < 0){
		PRINT_ERR("get param content fail! \n");
	}else{
		memcpy((unsigned char *)param_info,param_buf,sizeof(NOCS_Params));
	}
	
	if(param_buf!= NULL){
		free(param_buf);
	}
	return ret;
}

int get_param_info(struct media_info *mediainfo,
	NOCS_Certificate* cert_info,NOCS_Params *param_info)
{
	int ret = 0;
	if(mediainfo->media_type == EMMC_MEIDA_TYPE){
		ret = get_emmc_param_info(mediainfo,cert_info,param_info);
	}
	return ret;
}
void  get_emmc_bl_layout(struct layout_info *layout_array,
	NOCS_Certificate* cert_info,NOCS_Params *param_info)
{
	unsigned int loop = 0;
	layout_array[BL_TYPE_CERT].align_size = SIZE_ALIGN_BOUNDARY_MORE(CERT_SIZE,EMMC_ERASE_SIZE);
	layout_array[BL_TYPE_PARM].align_size = SIZE_ALIGN_BOUNDARY_MORE(cert_info->Size_Of_SCS_Params_Area,
					EMMC_ERASE_SIZE);
	layout_array[BL_TYPE_AUX].align_size = SIZE_ALIGN_BOUNDARY_MORE(param_info->Size_of_Aux_Code_Area,
					EMMC_ERASE_SIZE);
	layout_array[BL_TYPE_HWSETTING].align_size = SIZE_ALIGN_BOUNDARY_MORE(param_info->Hw_Setting_Size,
					EMMC_ERASE_SIZE);
	layout_array[BL_TYPE_ARM_REDIECT].align_size = SIZE_ALIGN_BOUNDARY_MORE(param_info->Arm_Rediect_Size,
					EMMC_ERASE_SIZE);
	layout_array[BL_TYPE_EXT_FW_CERT].align_size = SIZE_ALIGN_BOUNDARY_MORE(param_info->Ext_Fw_Cert_Size,
					EMMC_ERASE_SIZE);
	layout_array[BL_TYPE_EXT_FW].align_size = SIZE_ALIGN_BOUNDARY_MORE(param_info->Ext_Fw_Size,
					EMMC_ERASE_SIZE)+ EMMC_EXT_FW_RESERVE_SIZE;
	layout_array[BL_TYPE_KBOOT].align_size = param_info->Address_of_Aux_Code_Area == 0? 0:
		SIZE_ALIGN_BOUNDARY_MORE((param_info->Address_of_Aux_Code_Area+
			sizeof(DTE_KBOOT_AREA_Tail)),EMMC_ERASE_SIZE);
	layout_array[BL_TYPE_UBOOT].align_size = SIZE_ALIGN_BOUNDARY_MORE(param_info->External_Boot_Specific_Params,
					EMMC_ERASE_SIZE);
	/*get layout offset*/
	layout_array[BL_TYPE_CERT].offset = EMMC_BOOTCODE_START;
	for(loop = BL_TYPE_PARM; loop < BL_TYPE_MAX; loop++)
	{
		layout_array[loop].offset += (layout_array[loop - 1].offset + 
				layout_array[loop - 1].align_size);
	}
}

int get_bl_layout(struct media_info *mediainfo,struct layout_info *layout_array)
{
	int ret = 0;
	NOCS_Certificate cert_info = {0};
	NOCS_Params param_info = {0};
	/*get cert info(include param.bin size) */
	if((ret = get_cert_info(mediainfo,&cert_info)) < 0){
		return ret;
	}

	/*get param info(include auxcode/hwsetting/arm_redicet/ext_fw_cert/ext_fw/kboot/uboot size)*/
	if((ret = get_param_info(mediainfo,&cert_info,&param_info)) < 0){
		return ret;
	}

	/*get emmc bootloader layout*/
	if(mediainfo->media_type == EMMC_MEIDA_TYPE){
		get_emmc_bl_layout(layout_array,&cert_info,&param_info);
	}
	return ret;
}
int burn_bootcode_image(struct media_info *mediainfo, struct burn_info * burninfo)
{
	int ret = 0;
	unsigned int loop = 0,bl_checksum = 0,offset = 0;
	char bl_name[128] = {0};
	struct layout_info layout_array[BL_TYPE_MAX] = {0},*cur_layout = NULL;
	struct bl_burn_head *bl_head = NULL;
	struct bl_burn_info *bl_info = NULL;
	struct file_info    *fileinfo = &(burninfo->fileinfo);
	/*get bootcode layout*/
	if((ret = get_bl_layout(mediainfo,layout_array)) < 0){
		return ret;
	}

	/*get bootloader head information*/
	bl_head = (struct bl_burn_head *)((unsigned char *)fileinfo->file_addr);
	offset =  sizeof(struct bl_burn_head);
	PRINT_INFO("22 burn total image %d \n",bl_head->total_bl_num);
	for(loop = 0; loop < bl_head->total_bl_num; loop++)
	{
		/*get boot image information*/
		bl_info = (struct bl_burn_info *)((unsigned char *)fileinfo->file_addr+offset);
		if(bl_info->bl_type > BL_TYPE_MAX){
			ret = -1;
			PRINT_ERR("can't find bl type %d offset 0x%x\n",
				bl_info->bl_type,bl_info->offset);
			break;
		}
		memset(bl_name,0,sizeof(bl_name));
		if((ret = get_fw_name_by_tag(bl_name,bl_info->bl_type,bl_type_info_arr,
			sizeof(bl_type_info_arr)/sizeof(bl_type_info_arr[0]))) < 0){
			PRINT_ERR("can't find bl type %d \n",bl_info->bl_type);
			break;
		}
		
		/*get burn image layout*/
		cur_layout = &(layout_array[bl_info->bl_type]);

		/*skip layout size == 0 or file size  == 0 
		   eg:burn image have kboot but run image not have not update*/
		if((cur_layout->align_size == 0)||(bl_info->size == 0)){
			PRINT_INFO("can't find bl type %d name %s align size 0x%llx size 0x%x\n",
				bl_info->bl_type,bl_name,cur_layout->align_size,bl_info->size);
			continue;
		}

		/*compare burn image check sum*/
		bl_checksum = get_checksum((uchar *)(fileinfo->file_addr + bl_info->offset),
			bl_info->size);
		if(bl_info->checksum != bl_checksum){
			ret = -1;
			PRINT_ERR("bl type %d bl_name %s ok checksum 0x%x dd checksum 0x%x\n",
				bl_info->bl_type,bl_name,bl_info->checksum,bl_checksum);
			break;
		}

		/*burn bootloader image to media*/
		PRINT_INFO("burn bl type %d image name %s offset 0x%llx size 0x%x align_size 0x%llx\n",
			bl_info->bl_type,bl_name,cur_layout->offset,bl_info->size,cur_layout->align_size);
		if((ret = write_buf_to_media(mediainfo,cur_layout->offset,bl_info->size,
			(uchar *)(fileinfo->file_addr + bl_info->offset))) != 0){
			break;
		}

		/*update offset*/
		offset += (bl_info->size + sizeof(struct bl_burn_info));
	}
	return ret;
}

int burn_rescue_image(struct media_info *mediainfo,struct burn_info * burninfo)
{
	int ret = 0;
	ret = write_buf_to_media(mediainfo,SYS_PARAM_RESCUE_START,
		burninfo->fileinfo.file_size, (uchar *)(burninfo->fileinfo.file_addr));
	return ret;
}

int burn_fw_image(struct media_info *mediainfo,struct burn_info * burninfo)
{
	int ret = 0,need_update_fwtable = 0;
	struct layout_info fwtable_layout = {0};
	char * fw_table_buf = NULL;
	/*get fwtable content buffer */
	if((fw_table_buf = get_fwtable_info(mediainfo,&fwtable_layout)) == NULL){
		goto end;
	}

	/*get fw information from fwtable content and update fwtable information*/
	if((ret = get_fwinfo_from_fwtable(fw_table_buf,burninfo,&need_update_fwtable)) != 0){
		goto end;
	}

	/*burn fw image to media*/
	if((ret = write_buf_to_media(mediainfo,burninfo->layout.offset,burninfo->fileinfo.file_size,
			(uchar *)(burninfo->fileinfo.file_addr))) != 0){
		goto end;
	}

	/*update fwtable if fw size or fw content change*/
	if(need_update_fwtable == 1){
		if((ret = write_buf_to_media(mediainfo,fwtable_layout.offset,
			fwtable_layout.align_size,(uchar *)fw_table_buf))!= 0){
			goto end;
		}
	}

	/*update factory layout.txt if fw size change*/
	ret = update_boot_table(burninfo);
end:
	if(fw_table_buf != NULL){
		free(fw_table_buf);
	}
	return ret;
}

int fw_upgrade(int get_way,char *file_path)
{
	int ret = 0;
	struct burn_info  burninfo = {0};
	struct media_info mediainfo = {0};
	unsigned int  burn_type	= 0;
	/*get file information from tftp or usb*/
	if((ret = get_burn_fileinfo(get_way,file_path,&(burninfo.fileinfo),
		&burn_type))!= 0){
		return ret;
	}

	/*init media information*/
	if((ret = init_media_info(&mediainfo)) != 0){
		return ret;
	}

	/*burn image*/
	if(burn_type == BURN_BL_IMG){
		ret = burn_bootcode_image(&mediainfo,&burninfo);
	}else if(burn_type == BURN_RESCUE_IMG){
		ret = burn_rescue_image(&mediainfo,&burninfo);
	}else if(burn_type == BURN_FW_IMG){
		ret = burn_fw_image(&mediainfo,&burninfo);
	}
	return ret;
}
int do_fw_upgrade(cmd_tbl_t * cmdtp, int flag, int argc, char * const argv[])
{
    /* at least two arguments please */
	int ret = 0;
	char *cmd = NULL;
    if (argc < 2){
        PRINT_ERR("arguments %d > 2 eg: arg0:swu or swuu arg1:file_path \n",argc);
        goto usage;
    }
	cmd = argv[0];
	if(strcmp(cmd, "swuu") == 0) {
		ret = fw_upgrade(1,argv[1]);
	}else if(strcmp(cmd, "swu") == 0){
		ret = fw_upgrade(0,argv[1]);
	}else{
		goto usage;
	}
	
	if(ret == 0){
		PRINT_INFO("*********update fw name %s successful!********* \n",argv[1]);
	}else{
		PRINT_ERR("@@@@@@@@@@update fw name %s fail!@@@@@@@@@@ \n",argv[1]);
	}
	return 0;
	
usage:
	return CMD_RET_USAGE;
}

U_BOOT_CMD(
    swu, CONFIG_SYS_MAXARGS, 1, do_fw_upgrade,
    "FW upgrade sub-system",
    "swu fw_name          - fw_name:upgrade file name upgrade by tftp\n"
);

U_BOOT_CMD(
    swuu, CONFIG_SYS_MAXARGS, 1, do_fw_upgrade,
    "FW upgrade sub-system",
    "swuu fw_name         - fw_name:upgrade file name upgrade by usb\n"
);
