#include <common.h>
#include <command.h>
#include <io.h>

#define MENU_LEVEL_MAX (10) //at most 10 level menus
#define MENU_NAME_LEN (16) //max length for menu name show in path
#define BOOT_PARAM_LEN_MAX (256)
#define FRED_START "\033[1;31;40m"
#define FGREEN_START "\033[1;32;40m"
#define FYELLOW_START "\033[1;33;40m"
#define FBLUE_START "\033[1;34;40m"
#define FCOLOR_END "\033[0m"

#define DBG_DFT(format, ...) do{ printf(format, ##__VA_ARGS__); }while(0)
#define DBG_RED(format, ...) do{ printf(FRED_START format FCOLOR_END, ##__VA_ARGS__); }while(0)
#define DBG_GREEN(format, ...) do{ printf(FGREEN_START format FCOLOR_END, ##__VA_ARGS__); }while(0)
#define DBG_YELLOW(format, ...) do{ printf(FYELLOW_START format FCOLOR_END, ##__VA_ARGS__); }while(0)
#define DBG_BLUE(format, ...) do{ printf(FBLUE_START format FCOLOR_END, ##__VA_ARGS__); }while(0)

/*
menu format:
        menu index '0' is the whole menu name and description
        menu index '1' to 'end-1' should be sub-menu or item
        menu index 'end' should be all 'NULL' as {NULL, NULL, NULL, NULL}
*/

typedef struct setmenu
{
        char * name;//name
        char * desc;//description
        int (*func)(void);//name
        struct setmenu * father;//father menu
} setmenu;

typedef struct menu_state
{
        setmenu * menu;
        int level;
        char path[MENU_LEVEL_MAX][MENU_NAME_LEN];
} menu_state;

menu_state state;

int go_dcmt_top(void);
int go_sb2_top(void);
int go_armwrapper_top(void);
int go_hwm_top(void);
int go_up_menu(void);
extern int gets(char *s);
extern void *malloc(size_t size);
extern void free(void *ptr);
extern char *getenv(const char *name);
extern int setenv(const char *varname, const char *varvalue);

/****COMMON FUNC AREA****/

setmenu top_menu[] =
{
        {"root", "******** Welcome to hw-monitor set menu ********", NULL, NULL}, //title
        {"DCMT", "DCMT setting", go_dcmt_top, NULL},
        {"SB2", "SB2 setting", go_sb2_top, NULL},
        {"ARM WRAPPER", "ARM WRAPPER setting", go_armwrapper_top, NULL},
        {"HW MONITOR", "HW MONITOR setting", go_hwm_top, NULL},
        {"exit", "go exit", go_up_menu, NULL},
        {NULL, NULL, NULL, NULL},
};

void prepare_go_ahead(void)
{
        if(state.level > MENU_LEVEL_MAX)
        {
                return;
        }

        snprintf(state.path[state.level],MENU_NAME_LEN,state.menu[0].name);
        state.level++;
}

void prepare_go_back(void)
{
        if(state.level > 0)
        {
                state.level--;
        }
}

int go_up_menu(void)
{
        prepare_go_back();
        state.menu = state.menu[0].father;

        return 0;
}

void show_now_path(void)
{
        int i = 0;
        printf(FYELLOW_START);
        for(i = 0;i < state.level;++i)
        {
                if(i != 0)
                {
                        printf(" -> ");
                }
                else
                {
                        printf("menu path:");
                }
                printf("%s",state.path[i]);
        }
        printf("\n" FCOLOR_END);
}

int go_exit(void)
{
        printf("******** bye-bye! ********\n");
        return 0;
}

void show_menu(setmenu * p)
{
        int i = 0;

        printf("%s\n", p[0].desc);//print title
        i = 1;
        while(p[i].name != NULL)//print items
        {
                printf("[%d] %-16s : %s\n", i, p[i].name, p[i].desc);
                i++;
        }
}

int get_menu_len(setmenu * p)
{
        int i = 0;
        while(p[i].name != NULL)//print items
        {
                i++;
        }

        return i-1;//don't count p[0](p[0] is a title item)
}

void dump_buffer_data(char * buf,int len)
{
        int i = 0;
        for(i = 0;i < len;++i)
        {
                if(i % 8 == 0)
                {
                        printf("\n%4d:",i);
                }
                printf(" %02x(%c) ",buf[i],buf[i]);
        }
}

void menu_wait(setmenu * p)
{
        int index = 0;
        char buf[64] = {0};
        int len;

        show_now_path();
        show_menu(p);
        len = get_menu_len(p);

        index = 0;
        while((index <= 0) || (index > len))
        {
                printf("\ninput chooice('ESC'+'ENTER' to go back):");
                gets(buf);
                //dump_buffer_data(buf,4);
                if(buf[0] == '\e') //if use ESC, go back
                {
                        printf("\n");
                        go_up_menu();
                        break;
                }
                index = simple_strtoul(buf, NULL, 16);
                printf("\n");
                //printf("\nget chooice:%d\n", index);
        }

        if(p[index].func != NULL)
        {
                (*(p[index].func))();
        }
}

void show_boot_param(char * key)
{
        char * cmdline = NULL;

        cmdline = getenv(key);
        if(cmdline != NULL)
        {
                printf("[%s]:%s\n",key,cmdline);
        }
        else
        {
                printf("[%s] is NULL!\n",key);
        }
}

int update_boot_param(char * key,char * buf)
{
        char cmd[128];

        memset(cmd,0,sizeof(cmd));

        snprintf(cmd,sizeof(cmd),"env set %s %s",key,buf);

        if (run_command(cmd,0)!=0)
        {
                printf("%s failed.  env set failed.\n",cmd);
                return -1;
        }

        if (run_command("saveenv",0)!=0)
        {
                printf("%s failed.  saveenv failed.\n",cmd);
                return -1;
        }

        return 0;
}


int add_boot_param(char * key,char * buf)
{
        char * cmdline = NULL;
        char cmd[BOOT_PARAM_LEN_MAX];

        memset(cmd,0,sizeof(cmd));

        cmdline = getenv(key);

        if(!cmdline)
        {
                snprintf(cmd,sizeof(cmd),"env set %s %s",key,buf);
        }
        else
        {
                snprintf(cmd,sizeof(cmd),"env set %s %s,%s",key,cmdline,buf);
        }

        if (run_command(cmd,0)!=0)
        {
                printf("%s failed.  env set failed.\n",cmd);
                return -1;
        }

        if (run_command("saveenv",0)!=0)
        {
                printf("%s failed.  saveenv failed.\n",cmd);
                return -1;
        }

        return 0;
}

int clear_boot_param(char * key)
{
        char cmd[128];

        memset(cmd,0,sizeof(cmd));

        snprintf(cmd,sizeof(cmd),"env set %s ",key);

        if (run_command(cmd,0)!=0)
        {
                printf("%s failed.  env set failed.\n",cmd);
                return -1;
        }

        if (run_command("saveenv",0)!=0)
        {
                printf("%s failed.  saveenv failed.\n",cmd);
                return -1;
        }

        return 0;
}

/****DCMT LOGIC AREA****/
#define DCMT_ENV_MEM_SET        "dcmt_set"
#define DCMT_ENV_MEM_UNSET      "dcmt_unset"
#define DCMT_ENV_DBG_LOG        "dcmt_log"
#define DCMT_ENV_KEEP_GO        "dcmt_keepgo"
#define DCMT_ENV_DRIVER_DISABLE "dcmt_disable"

#define DCMT_CMD_BUF_LEN (256)

char dcmt_cmd_buf[DCMT_CMD_BUF_LEN];

int go_dcmt_mem_set(void)
{
        char* buf = dcmt_cmd_buf;
        char * cmdline = NULL;
        char * tmp;

        printf("\n******** dcmt set flow ********\n");
        memset(buf,0,DCMT_CMD_BUF_LEN);
        cmdline = getenv(DCMT_ENV_MEM_SET);
        DBG_YELLOW("now mem set param is : %s\n",cmdline);
        printf("you can input command to set dcmt monitor.\n"
                "input example: " FYELLOW_START "set 0x108000,0x700000 ip:CPUS(w) mode:1 mem:7\n" FCOLOR_END
                "you can use ',' to split 2 settings as:"  FYELLOW_START "set 0x108000,0x700000 ip:CPUS(w),set 0x800000,0x900000 ip:CPUS(w)\n" FCOLOR_END
                "if you use key words 'none', param will be cleard!\n");
        printf("\nadd set cmd(append):");
        gets(buf);

        //change ' ' to '#' avoid param recognize error
        for(tmp = buf;*tmp != '\0';tmp++)
        {
                if(*tmp == ' ')
                {
                        *tmp = '#';
                }
        }

        DBG_YELLOW("\nyour chooice is:%s\n",buf);

        if(strcmp(buf,"none") == 0)
        {
                clear_boot_param(DCMT_ENV_MEM_SET);
        }
        else
        {
                add_boot_param(DCMT_ENV_MEM_SET,buf);
        }

        cmdline = getenv(DCMT_ENV_MEM_SET);
        DBG_YELLOW("\nupdate mem set param to : %s\n",cmdline);

        return 0;
}

int go_dcmt_mem_unset(void)
{
        char* buf = dcmt_cmd_buf;
        char * cmdline = NULL;

        printf("\n******** dcmt unset flow ********\n");
        memset(buf,0,DCMT_CMD_BUF_LEN);
        cmdline = getenv(DCMT_ENV_MEM_UNSET);
        DBG_YELLOW("now mem unset param is : %s\n",cmdline);
        printf("you can input a mem number or monitor-type to unset DCMT monitor.\n"
                "input option : <number>,'all','over-range','kcpu','acpu','vcpu','s-ker','kasan','none'.\n"
                "input example: " FYELLOW_START "0,1,all,acpu,over-range \n" FCOLOR_END
                "if you use key words 'none', param will be cleard!\n");
        printf("\nupdate unset option(update):");
        gets(buf);
        DBG_YELLOW("\nyour chooice is:%s\n",buf);

        if(strcmp(buf,"none") == 0)
        {
                clear_boot_param(DCMT_ENV_MEM_UNSET);
        }
        else
        {
                update_boot_param(DCMT_ENV_MEM_UNSET,buf);
        }

        cmdline = getenv(DCMT_ENV_MEM_UNSET);
        DBG_YELLOW("\nupdate mem unset param to : %s\n",cmdline);

        return 0;
}

int go_dcmt_dbg_log(void)
{
        char* buf = dcmt_cmd_buf;
        char * cmdline = NULL;

        printf("\n******** dcmt log control flow ********\n");
        memset(buf,0,DCMT_CMD_BUF_LEN);
        cmdline = getenv(DCMT_ENV_DBG_LOG);
        DBG_YELLOW("now debug log param is : %s\n",cmdline);
        printf("you can input a option to enable some DCMT debug log.\n"
                "input option : 'all','init','set','isr','err','cmd','tip','addr','dbg'.\n"
                "input example: " FYELLOW_START "addr,tip \n" FCOLOR_END
                "if you use key words 'none', param will be cleard!\n");
        printf("\nupdate debug log option(update):");
        gets(buf);
        DBG_YELLOW("\nyour chooice is:%s\n",buf);

        if(strcmp(buf,"none") == 0)
        {
                clear_boot_param(DCMT_ENV_DBG_LOG);
        }
        else
        {
                update_boot_param(DCMT_ENV_DBG_LOG,buf);
        }

        cmdline = getenv(DCMT_ENV_DBG_LOG);
        DBG_YELLOW("\nupdate debug log param to : %s\n",cmdline);

        return 0;
}

int go_dcmt_keep_going(void)
{
        char* buf = dcmt_cmd_buf;
        char * cmdline = NULL;

        printf("\n******** dcmt interrupt keep going flow ********\n");
        memset(buf,0,DCMT_CMD_BUF_LEN);
        cmdline = getenv(DCMT_ENV_KEEP_GO);
        DBG_YELLOW("DCMT interrupt keep going param is : %s\n",cmdline);
        printf("\noption '0':silence mode, '1':simple mode, '2':normal mode, '3':panic mode \n"
               "if you use key words 'none', param will be cleard!\n");
        printf("\ninput keep going option(update):");
        gets(buf);
        DBG_YELLOW("\nyour chooice is:%s\n",buf);

        if(strcmp(buf,"none") == 0)
        {
                clear_boot_param(DCMT_ENV_KEEP_GO);
        }
        else
        {
                update_boot_param(DCMT_ENV_KEEP_GO,buf);
        }

        cmdline = getenv(DCMT_ENV_KEEP_GO);
        DBG_YELLOW("\nupdate interrupt keep going param to : %s\n",cmdline);

        return 0;
}

int go_dcmt_driver_disable(void)
{
        char* buf = dcmt_cmd_buf;
        char * cmdline = NULL;

        printf("\n******** dcmt driver disable flow ********\n");
        memset(buf,0,DCMT_CMD_BUF_LEN);
        cmdline = getenv(DCMT_ENV_DRIVER_DISABLE);
        DBG_YELLOW("DCMT driver disable param is : %s\n",cmdline);
        printf("\ndo you want to disable DCMT driver?(Yes/No):");
        gets(buf);
        DBG_YELLOW("\nyour chooice is:%s\n",buf);

        if((strcmp(buf,"y") == 0)||(strcmp(buf,"Y") == 0)||(strcmp(buf,"yes") == 0)||(strcmp(buf,"YES") == 0)||(strcmp(buf,"Yes") == 0))
        {
                update_boot_param(DCMT_ENV_DRIVER_DISABLE,"y");
        }
        else if((strcmp(buf,"n") == 0)||(strcmp(buf,"N") == 0)||(strcmp(buf,"no") == 0)||(strcmp(buf,"NO") == 0)||(strcmp(buf,"No") == 0))
        {
                clear_boot_param(DCMT_ENV_DRIVER_DISABLE);
        }

        cmdline = getenv(DCMT_ENV_DRIVER_DISABLE);
        DBG_YELLOW("\nupdate disable DCMT driver param to : %s\n",cmdline);

        return 0;
}

int go_dcmt_param_default(void)
{
        char* buf = dcmt_cmd_buf;

        printf("\n******** dcmt default param flow ********\n");
        memset(buf,0,DCMT_CMD_BUF_LEN);
        DBG_YELLOW("dcmt boot param is as below:\n");
        show_boot_param(DCMT_ENV_KEEP_GO);
        show_boot_param(DCMT_ENV_MEM_SET);
        show_boot_param(DCMT_ENV_MEM_UNSET);
        show_boot_param(DCMT_ENV_DBG_LOG);
        show_boot_param(DCMT_ENV_DRIVER_DISABLE);
        printf("\ndo you want to clear them all?(Yes/No):");
        gets(buf);
        DBG_YELLOW("\nyour chooice is:%s\n",buf);

        if((strcmp(buf,"y") == 0)||(strcmp(buf,"Y") == 0)||(strcmp(buf,"yes") == 0)||(strcmp(buf,"YES") == 0)||(strcmp(buf,"Yes") == 0))
        {
                clear_boot_param(DCMT_ENV_KEEP_GO);
                clear_boot_param(DCMT_ENV_MEM_SET);
                clear_boot_param(DCMT_ENV_MEM_UNSET);
                clear_boot_param(DCMT_ENV_DBG_LOG);
                clear_boot_param(DCMT_ENV_DRIVER_DISABLE);

                DBG_YELLOW("\nupdate dcmt boot param is as below:\n");
                show_boot_param(DCMT_ENV_KEEP_GO);
                show_boot_param(DCMT_ENV_MEM_SET);
                show_boot_param(DCMT_ENV_MEM_UNSET);
                show_boot_param(DCMT_ENV_DBG_LOG);
                show_boot_param(DCMT_ENV_DRIVER_DISABLE);
        }
        else
        {
                printf("param not changed...\n");
        }

        return 0;
}

int go_dcmt_param_show(void)
{
        char* buf = dcmt_cmd_buf;

        //cmdline=getenv(str);
        memset(buf,0,DCMT_CMD_BUF_LEN);
        show_boot_param(DCMT_ENV_KEEP_GO);
        show_boot_param(DCMT_ENV_MEM_SET);
        show_boot_param(DCMT_ENV_MEM_UNSET);
        show_boot_param(DCMT_ENV_DBG_LOG);
        show_boot_param(DCMT_ENV_DRIVER_DISABLE);

        //run_command(cmd,0); //cmd = "env set %s %x"
        printf("\npress 'ENTER' to continue...\n");
        gets(buf);

        return 0;
}

setmenu dcmt_root_menu[] =
{
        {"dcmt", "******** Welcome to dcmt set menu ********", NULL, top_menu},
        {"mem set", "set DCMT to monitor a region", go_dcmt_mem_set, dcmt_root_menu},
        {"mem unset", "unset a region which DCMT monitor by default", go_dcmt_mem_unset, dcmt_root_menu},
        {"keep go", "disable/enable kernel panic when trash happened", go_dcmt_keep_going, dcmt_root_menu},
        {"driver disable", "disable DCMT driver in kernel", go_dcmt_driver_disable, dcmt_root_menu},
        {"dbg log", "enable/disable DCMT debug log", go_dcmt_dbg_log, dcmt_root_menu},
        {"show boot param", "show bootcode param for DCMT driver", go_dcmt_param_show, dcmt_root_menu},
        {"default", "restore bootcode param as default setting", go_dcmt_param_default, dcmt_root_menu},
        {"go back", "go to up menu", go_up_menu, dcmt_root_menu},
        {NULL, NULL, NULL, NULL},
};

int go_dcmt_top(void)
{
        state.menu = dcmt_root_menu;
        prepare_go_ahead();
        return 0;
}

int append_bootargs(char * str)
{
        char *commandline = getenv("bootargs");
        char *tmp_cmdline = NULL;
        char *value;
        int  len          = 0;

        if(str == NULL)
        {
                return -1;
        }

        value = getenv(str);

        if (value == NULL)
        {
                return -1;
        }

        len = strlen(commandline) + strlen(str) + strlen(value) + 3;

        tmp_cmdline = (char *)malloc(len);
        if (!tmp_cmdline)
        {
                printf("[ERR] %s: Malloc failed\n", __func__);
                return -1;
        }

        memset(tmp_cmdline, 0, len);

        sprintf(tmp_cmdline, "%s %s=%s", commandline, str,value);

        setenv("bootargs", tmp_cmdline);

        free(tmp_cmdline);

        return 0;
}

int rtk_plat_boot_prep_dcmt(void)
{
        append_bootargs(DCMT_ENV_MEM_SET);
        append_bootargs(DCMT_ENV_MEM_UNSET);
        append_bootargs(DCMT_ENV_KEEP_GO);
        append_bootargs(DCMT_ENV_DBG_LOG);
        append_bootargs(DCMT_ENV_DRIVER_DISABLE);

        return 0;
}

/****SB2 LOGIC AREA****/
int go_sb2_top(void)
{
        prepare_go_ahead();
        return 0;
}

/****ARM_WRAPPER LOGIC AREA****/
int go_armwrapper_top(void)
{
        prepare_go_ahead();
        return 0;
}

/****HW MONITOR LOGIC AREA****/
#define HWM_ENV_RANGE_DFT        "hwm_range"
#define HWM_CMD_BUF_LEN (256)

char hwm_cmd_buf[HWM_CMD_BUF_LEN];

int go_hwm_range_dft(void)
{
        char* buf = hwm_cmd_buf;
        char * cmdline = NULL;

        printf("\n******** HW monitor default range monitor ********\n");
        memset(buf,0,HWM_CMD_BUF_LEN);
        cmdline = getenv(HWM_ENV_RANGE_DFT);
        DBG_YELLOW("HW monitor default range param is : %s\n",cmdline);
        printf("\ndo you want HW monitor to monitor default range?(Yes/No/Default):");
        gets(buf);
        DBG_YELLOW("\nyour chooice is:%s\n",buf);

        if((strcmp(buf,"y") == 0)||(strcmp(buf,"Y") == 0)||(strcmp(buf,"yes") == 0)||(strcmp(buf,"YES") == 0)||(strcmp(buf,"Yes") == 0))
        {
                update_boot_param(HWM_ENV_RANGE_DFT,"y");
                DBG_YELLOW("\nupdate HW monitor default range param as below:\n");
                show_boot_param(HWM_ENV_RANGE_DFT);
        }
        else if((strcmp(buf,"n") == 0)||(strcmp(buf,"N") == 0)||(strcmp(buf,"no") == 0)||(strcmp(buf,"NO") == 0)||(strcmp(buf,"No") == 0))
        {
                update_boot_param(HWM_ENV_RANGE_DFT,"n");
        }
        else if((strcmp(buf,"d") == 0)||(strcmp(buf,"D") == 0)||(strcmp(buf,"default") == 0)||(strcmp(buf,"DEFAULT") == 0)||(strcmp(buf,"Default") == 0))
        {
                clear_boot_param(HWM_ENV_RANGE_DFT);
        }
        else
        {
                printf("param not changed...\n");
        }

        cmdline = getenv(HWM_ENV_RANGE_DFT);
        DBG_YELLOW("\nupdate disable DCMT driver param to : %s\n",cmdline);

        return 0;
}

int go_hwm_param_default(void)
{
        char* buf = hwm_cmd_buf;

        printf("\n******** hwm default param flow ********\n");
        memset(buf,0,HWM_CMD_BUF_LEN);
        DBG_YELLOW("hwm boot param is as below:\n");
        show_boot_param(HWM_ENV_RANGE_DFT);
        printf("\ndo you want to clear them all?(Yes/No):");
        gets(buf);
        DBG_YELLOW("\nyour chooice is:%s\n",buf);

        if((strcmp(buf,"y") == 0)||(strcmp(buf,"Y") == 0)||(strcmp(buf,"yes") == 0)||(strcmp(buf,"YES") == 0)||(strcmp(buf,"Yes") == 0))
        {
                clear_boot_param(HWM_ENV_RANGE_DFT);

                DBG_YELLOW("\nupdate dcmt boot param is as below:\n");
                show_boot_param(HWM_ENV_RANGE_DFT);
        }
        else
        {
                printf("param not changed...\n");
        }

        return 0;
}

int go_hwm_param_show(void)
{
        char* buf = hwm_cmd_buf;

        //cmdline=getenv(str);
        memset(buf,0,HWM_CMD_BUF_LEN);
        show_boot_param(HWM_ENV_RANGE_DFT);

        //run_command(cmd,0); //cmd = "env set %s %x"
        printf("\npress 'ENTER' to continue...\n");
        gets(buf);

        return 0;
}

setmenu hwm_root_menu[] =
{
        {"hwm", "******** Welcome to HW monitor set menu ********", NULL, top_menu},
        {"default range", "set DCMT/ARM_WRAPPER/SB2_DBG to monitor default region", go_hwm_range_dft, hwm_root_menu},
        {"show boot param", "show bootcode param for HW monitor", go_hwm_param_show, hwm_root_menu},
        {"default", "restore bootcode param as default setting", go_hwm_param_default, hwm_root_menu},
        {"go back", "go to up menu", go_up_menu, hwm_root_menu},
        {NULL, NULL, NULL, NULL},
};

int go_hwm_top(void)
{
        state.menu = hwm_root_menu;
        prepare_go_ahead();
        return 0;
}

int rtk_plat_boot_prep_hwm(void)
{
        append_bootargs(HWM_ENV_RANGE_DFT);

        return 0;
}

/****MAIN FRAME AREA****/
int hwmonitor_set (cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
        state.menu = top_menu;
        prepare_go_ahead();
        while(state.menu != NULL)
        {
                menu_wait(state.menu);
        }
        go_exit();

        return 0;
}

U_BOOT_CMD(
        hwmonitor,   1,      0,      hwmonitor_set,
        "hw monitor setting",
        ""
);

