#include <common.h>
#include <asm/arch/system.h>
#include "flash_spi_list.h"
#include "flashdev_s.h"
#include <asm/arch/sys_flash_layout.h>
#include "../../../src/include/flash/spi_mgr.h"
#include <linux/compiler.h>
#include <malloc.h>
//__maybe_unused
//#define CONFIG_MTD_SFC_DEBUG
#ifdef CONFIG_MTD_SFC_DEBUG
#define SFC_MSG_DEBUG(fmt, args...) printf(fmt, ##args)
#else
#define SFC_MSG_DEBUG(fmt, args...)     /* empty macro */
#endif

void  mask_register(const unsigned int reg_addr, const unsigned int AND_filter, const unsigned int OR_filter)
{
	rtd_outl(reg_addr, (rtd_inl(reg_addr) & AND_filter) | OR_filter);
}


#define CONFIG_MTD_SFC_PRINTK_INFO
#ifdef CONFIG_MTD_SFC_PRINTK_INFO
#define SFC_MSG_INFO(fmt, args...) printf("[SFC_LOG][INFO] " fmt, ## args)
#else
#define SFC_MSG_INFO(fmt, args...)     /* empty macro */
#endif

#define SFC_MSG_ERR(fmt, args...) printf("[SFC_LOG][ERROR] " fmt, ## args)

#define MD_READ_WRITE
#ifdef MD_READ_WRITE
static int md_rw_flag = 1;
#else
static int md_rw_flag = 0;
#endif

/* must keep sync with dvrmain/uboot/kernel */
#define CONFIG_RTK_SPI_WP_SOFT
#ifdef CONFIG_RTK_SPI_WP_SOFT
//#define CONFIG_RTK_SPI_WP_HW  // if config this, the quad mode must disable
#endif

/************************************************************************
 *  Defines SFC registers
 ************************************************************************/
#define SPI_BASE_ADDR   0xb8200000      /* NOR flash start addess */
#define MD_PP_DATA_SIZE_SHIFT   8               /* 256 bytes */
#define MD_PP_DATA_SIZE         (1 << MD_PP_DATA_SIZE_SHIFT)

#define REG_SFC_BASE     0x1801a800

#define REG_SFC_OPCODE          (REG_SFC_BASE + 0x00)
#define REG_SFC_CTL                     (REG_SFC_BASE + 0x04)
#define REG_SFC_SCK                     (REG_SFC_BASE + 0x08)
#define REG_SFC_CE                      (REG_SFC_BASE + 0x0c)
#define REG_SFC_WP                      (REG_SFC_BASE + 0x10)
#define REG_SFC_POS_LATCH       (REG_SFC_BASE + 0x14)
#define REG_SFC_WAIT_WR         (REG_SFC_BASE + 0x18)
#define REG_SFC_EN_WR           (REG_SFC_BASE + 0x1c)
#define REG_SFC_FAST_RD         (REG_SFC_BASE + 0x20)
#define REG_SFC_SCK_TAP         (REG_SFC_BASE + 0x24)
#define REG_SFC_32MB_EN          (REG_SFC_BASE + 0x28)
#define REG_SPI_NAND_0          (REG_SFC_BASE + 0x2c)
#define SB2_CTRL_REG                 0x1801A960

/* REG_SFC_OPCODE */
#define DUAL_QUAL_MODE_OFFSET 9
#define DUAL_QUAL_MODE_MASE 0xF
#define FAT_MODE_OFFSET (8)
#define WRITE_BYTES_MASK  (0xf)
#define CMD_MASK  0xff
#define NORMAL_MODE    ((0b0000&DUAL_QUAL_MODE_MASE)<<DUAL_QUAL_MODE_OFFSET)
#define DUAL_READ_MODE ((0b0001&DUAL_QUAL_MODE_MASE)<<DUAL_QUAL_MODE_OFFSET)
#define QUAD_READ_MODE ((0b0101&DUAL_QUAL_MODE_MASE)<<DUAL_QUAL_MODE_OFFSET)

#define CMD_READ_ID                 0x9f
#define CMD_READ_STATUS1            0x05
#define CMD_READ_STATUS2            0x35
#define CMD_READ_STATUS3            0x15
#define CMD_WRITE_ENABLE            0x06
#define CMD_WRITE_STATUS1           0X01
#define CMD_WRITE_STATUS2           0X31
#define CMD_WRITE_STATUS3           0X11
#define CMD_PAGE_PROGRAM            0x02
#define CMD_READ_NORMAL             0x03
#define CMD_READ_NORMAL_FAST        0x0b
#define CMD_READ_DUAL_FAST          0x3b
#define CMD_READ_QUAD_FAST          0x6b
#define CMD_ERASE_BLOCK_64K         0xd8
#define CMD_ERASE_BLOCK_32K         0x52
#define CMD_ERASE_SECTOR_4K         0x20

#define CMD_ENTER_4BYTES_ADDR_MODE	0xB7
#define CMD_EXIT_4BYTES_ADDR_MODE	0xE9

#define CMD_SOFTWARE_DIE_SELECT		0xC2

#define STATUS_QE_S9                (1<<0x01)

/* REG_SFC_CTL  */
#define DATA_EN_OFFSET     (4)
#define ADDR_EN_OFFSET     (3)
#define DMY_CNT_OFFSET     (0)
#define DMY_CNT_MASK       (0x7)

/* REG_SFC_SCK */
#define MODE_OFFSET         8
#define FDIV_MASK           0xff
#define FDIV_OFFSET         0

/* REG_SFC_EN_WR */
#define WT_PROG_EN_OFFSET   (8)
#define WR_EN_OP_OFFSET     (0)
#define WR_EN_OP_MASK     (0xff)

/* REG_SFC_CE */
#define TDT_OFFSET          (16)
#define TDT_MASK           (0xffff)
#define PHCNT_OFFSET         (8)
#define PHCNT_MASK         (0xff)
#define PLCNT_OFFSET          0
#define PLCNT_MASK          (0xff)


/* REG_SFC_POS_LATCH */
#define POS_LATCH_OFFSET  0
#define POS_LATCH_FALL    0
#define POS_LATCH_RISE    1
#define DELAY_SEL_OFFSET  1
#define DELAY_SEL_MASK    0x07
#define DELAY_SEL_4    0x4



/* REG_SFC_WAIT_WR */
#define WT_PROG_DONE_OFFSET   (8)
#define RDSR_OP_OFFSET     (0)
#define RDSR_OP_MASK     (0Xff)

/* REG_SFC_WP */
#define HW_WP_N_OFFSET     0
#define HW_WP_N_MASK       0x01
#define HW_WP_N_ENABLE     0
#define HW_WP_N_DISABLE    1
#define HW_WP_EN1_OFFSET   1

/* SPI STATUS REGISTER */
#define WIP_OFFSET 0
#define WEL_OFFSET 1
#define STATUS_WIP                      0x01
/************************************************************************
 *  Defines Move Data Engine registers
 ************************************************************************/
#define REG_MD_FDMA_DDR_SADDR   (0x1800b088)
#define REG_MD_FDMA_FL_SADDR    (0x1800b08c)
#define REG_MD_FDMA_CTRL2               (0x1800b090)
#define REG_MD_FDMA_CTRL1               (0x1800b094)
#define REG_MD_SMQ_INT_STATUS   (0x1800b004)

/* REG_MD_FDMA_CTRL2 */
#define FORWART_MAPPINT_OFFSET   29
#define FDMA_SWAP_OFFSET         28
#define SWAP_VALUE                1
#define FDMA_MAX_XFER_OFFSET     26
#define SET_DMA_LEN               3
#define SET_DMA_LEN_512           2
#define SET_DMA_LEN_256           1

#define FDMA_DIR_OFFSET          25
#define FLASH_2_DDR               0
#define DDR_2_FLASH               1
#define FDMA_LEN_MASK      0x1ffffff
#define FDMA_LEN_OFFSET           0

/* REG_MD_FDMA_CTRL1 */
#define WRITE_EN1_OFFSET 1
#define FDMA_GO_OFFSET 0

/* REG_MD_SMQ_INT_STATUS */
#define FDMA_DONE_INIT_STATUS_OFFSET 4

/************************************************************************
 *  Define misc
 ************************************************************************/
#define SFC_DELAY 100
#define SFC_DELAY_TW 5000           // write status register cycle time
#define MD_PP_DATA_SIZE_SHIFT   8       /* 256 bytes */
#define MD_PP_DATA_SIZE         (1 << MD_PP_DATA_SIZE_SHIFT)
#define FOUR_BYTES_MASK  (0x3)
#define ERASE_DATA_TIMEOUT 25000 // 2500ms
#define WRITE_DATA_TIMEOUT 100   // 5ms
#define SET_FEATURE_TIMEOUT 1000   // 50ms

#define SFC_WP_MASK_BIT2_TO_BIT7 0xfc
#define SFC_WP_MASK_BIT6         0x40
#define SFC_WP_MASK_BIT2         0x04

#define BOOTCODE_SIZE_OFFSET            28 //bootocde size offset in sysinfo.bin

/************************************************************************
 ************************************************************************
 ************************************************************************/
static spinor_info_t *spi_dev = NULL;

void sfc_debug(void)
{
        SFC_MSG_INFO("\n%x: %x\n", REG_SFC_OPCODE, rtd_inl(REG_SFC_OPCODE));
        SFC_MSG_INFO("%x: %x\n", REG_SFC_CTL,    rtd_inl(REG_SFC_CTL));
        SFC_MSG_INFO("%x: %x\n", REG_SFC_SCK,    rtd_inl(REG_SFC_SCK));
        SFC_MSG_INFO("%x: %x\n", REG_SFC_CE,     rtd_inl(REG_SFC_CE));
        SFC_MSG_INFO("%x: %x\n", REG_SFC_WP,     rtd_inl(REG_SFC_WP));
        SFC_MSG_INFO("%x: %x\n", REG_SFC_POS_LATCH, rtd_inl(REG_SFC_POS_LATCH));
        SFC_MSG_INFO("%x: %x\n", REG_SFC_WAIT_WR, rtd_inl(REG_SFC_WAIT_WR));
        SFC_MSG_INFO("%x: %x\n", REG_SFC_EN_WR,   rtd_inl(REG_SFC_EN_WR));
        //SFC_MSG_INFO("%x: %x\n", REG_SFC_FAST_RD,   rtd_inl(REG_SFC_FAST_RD));
        //SFC_MSG_INFO("%x: %x\n", REG_SFC_SCK_TAP,   rtd_inl(REG_SFC_SCK_TAP));

}

void sfc_debug_md(void)
{
        SFC_MSG_INFO("\nREG_MD_FDMA_DDR_SADDR %x: %x\n", REG_MD_FDMA_DDR_SADDR, rtd_inl(REG_MD_FDMA_DDR_SADDR));
        SFC_MSG_INFO("REG_MD_FDMA_FL_SADDR %x: %x\n", REG_MD_FDMA_FL_SADDR,    rtd_inl(REG_MD_FDMA_FL_SADDR));
        SFC_MSG_INFO("REG_MD_FDMA_CTRL2 %x: %x\n", REG_MD_FDMA_CTRL2,    rtd_inl(REG_MD_FDMA_CTRL2));
        SFC_MSG_INFO("REG_MD_FDMA_CTRL1 %x: %x\n", REG_MD_FDMA_CTRL1,     rtd_inl(REG_MD_FDMA_CTRL1));
        SFC_MSG_INFO("REG_MD_SMQ_INT_STATUS %x: %x\n", REG_MD_SMQ_INT_STATUS,     rtd_inl(REG_MD_SMQ_INT_STATUS));
}

unsigned int swap_endian(unsigned int input)
{
        unsigned int output;

        output = (input & 0xff000000) >> 24 |
                 (input & 0x00ff0000) >> 8 |
                 (input & 0x0000ff00) << 8 |
                 (input & 0x000000ff) << 24;

        return output;
}

static void sfc_enable_write(void)
{
        __maybe_unused unsigned char tmp;
        rtd_outl(REG_SFC_OPCODE, CMD_WRITE_ENABLE & CMD_MASK);
        rtd_outl(REG_SFC_CTL, 0x00); // data_en = 0, addr_en=0 , dmy_en=0
        asm("DSB");
        udelay(SFC_DELAY);
        tmp = rtd_inb(SPI_BASE_ADDR);
	tmp = 0;
        return;
}

static void sfc_get_feature(unsigned char rdsr, unsigned char *data)
{
        asm("DSB");
        rtd_outl(REG_SFC_OPCODE, rdsr & CMD_MASK);
        rtd_outl(REG_SFC_CTL, 1 << DATA_EN_OFFSET);
        asm("DSB");

        udelay(SFC_DELAY);
        *data = rtd_inb(SPI_BASE_ADDR);
        asm("DSB");
        return;
}

static int sfc_set_feature(unsigned char wrsr, unsigned char data)
{
        unsigned char tmp;
        int ret = 0, retry = 0;

        sfc_enable_write();
        rtd_outl(REG_SFC_OPCODE, wrsr & CMD_MASK);
        rtd_outl(REG_SFC_CTL, 1 << DATA_EN_OFFSET);
        asm("DSB");
        udelay(SFC_DELAY);
        rtd_outb(SPI_BASE_ADDR, data);

        do {
                udelay(SFC_DELAY);
                sfc_get_feature(CMD_READ_STATUS1, &tmp);
        } while ((tmp & STATUS_WIP) && (++retry < SET_FEATURE_TIMEOUT));

        if (retry == SET_FEATURE_TIMEOUT) {
                SFC_MSG_ERR("%s wait timet out  0x%x\n", __FUNCTION__, tmp);
                ret = -1;
        }
        return ret;
}

static int sfc_set_feature_16bit(unsigned char wrsr, unsigned short data)
{
        unsigned char tmp;
        int ret = 0, retry = 0;

        sfc_enable_write();
        rtd_outl(REG_SFC_OPCODE, wrsr & CMD_MASK);
        rtd_outl(REG_SFC_CTL, 1 << DATA_EN_OFFSET);
        asm("DSB");
        udelay(SFC_DELAY);
        rtd_outw(SPI_BASE_ADDR, data);

        do {
                udelay(SFC_DELAY);
                sfc_get_feature(CMD_READ_STATUS1, &tmp);
        } while ((tmp & STATUS_WIP) && (++retry < SET_FEATURE_TIMEOUT));

        if (retry == SET_FEATURE_TIMEOUT) {
                SFC_MSG_ERR("%s wait timet out  0x%x\n", __FUNCTION__, tmp);
                ret = -1;
        }
        return ret;
}

static int sfc_wait_erase_done(void)
{
        int ret = 0, retry = 0;
        unsigned char data = 0;

        do {
                udelay(SFC_DELAY);
                sfc_get_feature(CMD_READ_STATUS1, &data);
        } while ((data & STATUS_WIP) && (++retry < ERASE_DATA_TIMEOUT));

        if (retry == ERASE_DATA_TIMEOUT) {
                SFC_MSG_ERR("%s wait timet out  0x%x\n", __FUNCTION__, data);
                ret = -1;
        }
        return ret;
}

static int sfc_wait_write_done(void)
{
        int ret = 0, retry = 0;
        unsigned char data = 0;

        do {
                udelay(SFC_DELAY);
                sfc_get_feature(CMD_READ_STATUS1, &data);
        } while ((data & STATUS_WIP) && (++retry < WRITE_DATA_TIMEOUT));

        if (retry == WRITE_DATA_TIMEOUT) {
                SFC_MSG_ERR("%s wait timet out  0x%x\n", __FUNCTION__, data);
                ret = -1;
        }
        return ret;
}

static __maybe_unused int sfc_wait_set_feature_done(void)
{
        int ret = 0, retry = 0;
        unsigned char data = 0;

        do {
                udelay(SFC_DELAY);
                sfc_get_feature(CMD_READ_STATUS1, &data);
        } while ((data & STATUS_WIP) && (++retry < SET_FEATURE_TIMEOUT));

        if (retry == SET_FEATURE_TIMEOUT) {
                SFC_MSG_ERR("%s wait timet out  0x%x\n", __FUNCTION__, data);
                ret = -1;
        }
        printf("***11retry %d\n", retry);
        return ret;
}

// return 1 : quad enable  return 0: quad disable
static int sfc_check_quad(spinor_info_t *spinor_info)
{
        unsigned char data = 0;

        sfc_get_feature(spinor_info->cmd_quad_rdsr, &data);
        if(data & spinor_info->cmd_quad_bit)
                return 1;
        else
                return 0;
}
static int sfc_enable_quad(spinor_info_t *spinor_info)
{
        unsigned char data = 0;
        unsigned char tmp = 0;
        unsigned short data_16bit = 0xffff;
        int ret = 0;

        if(spinor_info->id == WINBOND_W32Q64_64Mbit) {
                sfc_get_feature(CMD_READ_STATUS1, &data);
                data_16bit = data;
                sfc_get_feature(spinor_info->cmd_quad_rdsr, &data);
                data = data | spinor_info->cmd_quad_bit;
                data_16bit = ((data << 8) | data_16bit);
                ret = sfc_set_feature_16bit(spinor_info->cmd_quad_wrsr, data_16bit);
                if(ret < 0) {
                        return ret;
                }
                sfc_get_feature(spinor_info->cmd_quad_rdsr, &tmp);
                SFC_MSG_DEBUG("%s tmp:0x%x data:0x%x\n", __FUNCTION__, tmp, data);
                if(tmp != data) {
                        SFC_MSG_DEBUG("%s 0x%x != 0x%x\n", __FUNCTION__, tmp, data);
                        return -1;
                }
        } else {
                sfc_get_feature(spinor_info->cmd_quad_rdsr, &data);
                data = data | spinor_info->cmd_quad_bit;
                ret = sfc_set_feature(spinor_info->cmd_quad_wrsr, data);
                if(ret < 0) {
                        return ret;
                }
                sfc_get_feature(spinor_info->cmd_quad_rdsr, &tmp);
                SFC_MSG_DEBUG("%s tmp:0x%x data:0x%x\n", __FUNCTION__, tmp, data);

                if(tmp != data) {
                        SFC_MSG_DEBUG("%s 0x%x != 0x%x\n", __FUNCTION__, tmp, data);
                        return -1;
                }
        }
        return ret;
}

static __maybe_unused void sfc_disable_quad(spinor_info_t *spinor_info)
{
        unsigned char data = 0;
	  __maybe_unused int ret = 0;
	  
        sfc_get_feature(spinor_info->cmd_quad_rdsr, &data);
        data = data & (~spinor_info->cmd_quad_bit);
        ret = sfc_set_feature(spinor_info->cmd_quad_wrsr, data);
        return;
}

static void switch_read_mode(spinor_info_t *spinor_info)
{
        switch(spinor_info->read_mode) {
                case eMODE_READ_NORMAL:
                        rtd_outl(REG_SFC_OPCODE, (CMD_READ_NORMAL & CMD_MASK) | NORMAL_MODE); //switch flash to read mode
                        rtd_outl(REG_SFC_CTL, (1 << DATA_EN_OFFSET) | (1 << ADDR_EN_OFFSET)); //data_en=1 addr_en=1 dmy_en=0
                        break;
                case eMODE_READ_NORMAL_FAST:
                        rtd_outl(REG_SFC_OPCODE, (CMD_READ_NORMAL_FAST & CMD_MASK) | NORMAL_MODE); //switch flash to read mode
                        rtd_outl(REG_SFC_CTL, (1 << DATA_EN_OFFSET) | (1 << ADDR_EN_OFFSET) | ((1 & DMY_CNT_MASK) << DMY_CNT_OFFSET)); //data_en=1 addr_en=1 dmy_en=1
                        break;
                case eMODE_READ_DUAL_FAST:
                        rtd_outl(REG_SFC_OPCODE, (CMD_READ_DUAL_FAST & CMD_MASK) | DUAL_READ_MODE ); //switch flash to read mode
                        rtd_outl(REG_SFC_CTL, (1 << DATA_EN_OFFSET) | (1 << ADDR_EN_OFFSET) | ((1 & DMY_CNT_MASK) << DMY_CNT_OFFSET)); //data_en=1 addr_en=1 dmy_en=1
                        break;
                case eMODE_READ_QUAD_FAST:
                        rtd_outl(REG_SFC_OPCODE, (CMD_READ_QUAD_FAST & CMD_MASK) | QUAD_READ_MODE); //switch flash to read mode
                        rtd_outl(REG_SFC_CTL, (1 << DATA_EN_OFFSET) | (1 << ADDR_EN_OFFSET) | ((1 & DMY_CNT_MASK) << DMY_CNT_OFFSET)); //data_en=1 addr_en=1 dmy_en=1
                        break;
                default:
                        rtd_outl(REG_SFC_OPCODE, (CMD_READ_NORMAL & CMD_MASK) | NORMAL_MODE); //switch flash to read mode
                        rtd_outl(REG_SFC_CTL, (1 << DATA_EN_OFFSET) | (1 << ADDR_EN_OFFSET)); //data_en=1 addr_en=1 dmy_en=0
                        break;
        }

        // disable auto
        //rtd_outl(REG_SFC_EN_WR, (0<<WT_PROG_DONE_OFFSET) | (0x0&WR_EN_OP_MASK));
        //rtd_outl(REG_SFC_WAIT_WR, (0<<WT_PROG_EN_OFFSET) | (0x0&RDSR_OP_MASK));
        asm("DSB");
}

static void switch_write_mode(spinor_info_t *spinor_info)
{
        SFC_MSG_DEBUG("%s_%d\n", __FUNCTION__, __LINE__);
        // auto
        //rtd_outl(REG_SFC_EN_WR, (1<<WT_PROG_EN_OFFSET) | (CMD_WRITE_ENABLE & WR_EN_OP_MASK));
        //rtd_outl(REG_SFC_WAIT_WR, (1<<WT_PROG_DONE_OFFSET) | (CMD_READ_STATUS1 & RDSR_OP_MASK));

        if(spinor_info->write_mode == eMODE_WRITE_NORMAL) {
                rtd_outl(REG_SFC_OPCODE, CMD_PAGE_PROGRAM & CMD_MASK);
                rtd_outl(REG_SFC_CTL, (1 << DATA_EN_OFFSET | 1 << ADDR_EN_OFFSET)); // data_en=1 addr_en=1 dmy_en=0
                asm("DSB");
        } else if(spinor_info->write_mode == eMODE_WRITE_QUAD_FAST) {
                SFC_MSG_ERR("the spi dirver not surport write quad mode\n");

                /*SFC_MSG_DEBUG("%s_%d\n", __FUNCTION__, __LINE__);
                sfc_enable_quad();
                SFC_MSG_DEBUG("%s_%d\n", __FUNCTION__, __LINE__);
                rtd_outl(REG_SFC_OPCODE, (CMD_READ_QUAD_FAST&CMD_MASK) | QUAD_READ_MODE); //switch flash to read mode
                    rtd_outl(REG_SFC_CTL, (1<<DATA_EN_OFFSET)|(1<<ADDR_EN_OFFSET) | ((1&DMY_CNT_MASK)<<DMY_CNT_OFFSET)); //data_en=1 addr_en=1 dmy_en=1
                SFC_MSG_DEBUG("%s_%d\n", __FUNCTION__, __LINE__);
                */
        }
}

static void switch_erase_mode(spinor_info_t *spinor_info)
{
        SFC_MSG_DEBUG("%s_%d\n", __FUNCTION__, __LINE__);
        // clean auto WEL and wait WIP
        //rtd_outl(REG_SFC_EN_WR, (0<<WT_PROG_DONE_OFFSET) | (0x0&WR_EN_OP_MASK));
        //rtd_outl(REG_SFC_WAIT_WR, (0<<WT_PROG_EN_OFFSET) | (0x0&RDSR_OP_MASK));
        asm("DSB");

        //if(spinor_info->read_mode == eMODE_READ_QUAD_FAST)
        // {
        //     sfc_disable_quad(spinor_info);
        // }

        return;
}

static void sfc_enable_4bytes_addr_mode(void)
{
	__maybe_unused unsigned char tmp = 0;

	/*enable spi device 4bytes address mode*/
	rtd_outl(REG_SFC_OPCODE, CMD_ENTER_4BYTES_ADDR_MODE & CMD_MASK);
	rtd_outl(REG_SFC_CTL, (0 << DATA_EN_OFFSET) | (0 << ADDR_EN_OFFSET) | (0 & DMY_CNT_MASK));
	asm("DSB");
	udelay(SFC_DELAY);
	/*trigger cmd*/
	tmp = rtd_inb(SPI_BASE_ADDR);

	/*enable spi controler 4bytes address mode*/
	rtd_outl(REG_SFC_32MB_EN, 0x1);
}

static void sfc_disable_4bytes_addr_mode(void)
{
	__maybe_unused unsigned char tmp = 0;

	/*disable spi device 4bytes address mode*/
	rtd_outl(REG_SFC_OPCODE, CMD_EXIT_4BYTES_ADDR_MODE & CMD_MASK);
	rtd_outl(REG_SFC_CTL, (0 << DATA_EN_OFFSET) | (0 << ADDR_EN_OFFSET) | (0 & DMY_CNT_MASK));
	asm("DSB");
	udelay(SFC_DELAY);
	/*trigger cmd*/
	tmp = rtd_inb(SPI_BASE_ADDR);

	/*disable spi controler 4bytes address mode*/
	rtd_outl(REG_SFC_32MB_EN, 0x0);
}

static void sfc_die_select(unsigned char die_id)
{
	/*select die*/
	rtd_outl(REG_SFC_OPCODE, CMD_SOFTWARE_DIE_SELECT & CMD_MASK);
	rtd_outl(REG_SFC_CTL, (1 << DATA_EN_OFFSET) | (0 << ADDR_EN_OFFSET) | (0 & DMY_CNT_MASK));
	asm("DSB");
	udelay(SFC_DELAY);

	rtd_outl(SPI_BASE_ADDR, die_id);
}

static void sfc_read_id(unsigned int* id)
{
        rtd_outl(REG_SFC_OPCODE, CMD_READ_ID & CMD_MASK);
        //rtd_outl(REG_SFC_OPCODE, 0x9f);
        rtd_outl(REG_SFC_CTL, (1 << DATA_EN_OFFSET) | (0 << ADDR_EN_OFFSET) | (0 & DMY_CNT_MASK)); /* dataen = 1, adren = 0, dmycnt = 0 */
        //rtd_outl(REG_SFC_CTL, 0x00000010); /* dataen = 1, adren = 0, dmycnt = 0 */
        asm("DSB");
        udelay(SFC_DELAY);
        *id = rtd_inl(SPI_BASE_ADDR);
        sync();
        *id = swap_endian(*id);
        (*id) = (*id) >> 8;
        return;
}

static int sfc_match_chip(spinor_info_t **dev, unsigned int chip_id)
{
        int i = 0;

        for (i = 0; i < DEV_SIZE_S; i++) {
                if (chip_id != s_device[i].id) // find the match flash brand
                        continue;

                memcpy(*dev, &s_device[i], sizeof(spinor_info_t));
                break;
        }

        if (i == DEV_SIZE_S) {
                SFC_MSG_ERR("[SPINOR-INFO]Cannot find any match ID!!!\n");
                *dev =  NULL;
                return -1;
        }
        return 0;
}

static int sfc_lock(spinor_info_t *spinor_info)
{
        unsigned char status1 = 0xFF, status2 = 0xFF, status3 = 0xFF;
        unsigned short status_16bit = 0xffff;

        switch(spinor_info->id) {
                case GD25Q128C_128Mbit:
                case WINBOND_W25Q128FV_128Mbit:
                        sfc_get_feature(CMD_READ_STATUS1, &status1);
                        sfc_get_feature(CMD_READ_STATUS2, &status2);
                        sfc_get_feature(CMD_READ_STATUS3, &status3);
                        SFC_MSG_INFO("old wp status: 0x%x 0x%x 0x%x\n", status1, status2, status3);

                        if(((status1 & SFC_WP_MASK_BIT2_TO_BIT7) != 0x0c) \
                            || ((status2 & SFC_WP_MASK_BIT6) != SFC_WP_MASK_BIT6)) {
                                status1 = (status1 & (~SFC_WP_MASK_BIT2_TO_BIT7)) | 0x0c;
                                status2 = status2 | SFC_WP_MASK_BIT6;
                                if((-1 == sfc_set_feature(CMD_WRITE_STATUS1, status1)) \
                                    || (-1 == sfc_set_feature(CMD_WRITE_STATUS2, status2))) {
                                        SFC_MSG_ERR("sfc_lock fail\n");
                                        return -1;
                                }
                                SFC_MSG_INFO("new wp status: 0x%x 0x%x\n", status1, status2);
                        }
                        break;

                case GD25Q64C_64Mbit:
                        sfc_get_feature(CMD_READ_STATUS1, &status1);
                        sfc_get_feature(CMD_READ_STATUS2, &status2);
                        sfc_get_feature(CMD_READ_STATUS3, &status3);
                        SFC_MSG_INFO("old wp status: 0x%x 0x%x 0x%x\n", status1, status2, status3);

                        if(((status1 & SFC_WP_MASK_BIT2_TO_BIT7) != 0x10) \
                            || ((status2 & SFC_WP_MASK_BIT6) != SFC_WP_MASK_BIT6)) {
                                status1 = (status1 & (~SFC_WP_MASK_BIT2_TO_BIT7)) | 0x10;
                                status2 = status2 | SFC_WP_MASK_BIT6;
                                if((-1 == sfc_set_feature(CMD_WRITE_STATUS1, status1)) \
                                    || (-1 == sfc_set_feature(CMD_WRITE_STATUS2, status2))) {
                                        SFC_MSG_ERR("sfc_lock fail\n");
                                        return -1;
                                }
                                SFC_MSG_INFO("new wp status: 0x%x 0x%x\n", status1, status2);
                        }
                        break;

                case WINBOND_W32Q64_64Mbit:
                        sfc_get_feature(CMD_READ_STATUS1, &status1);
                        sfc_get_feature(CMD_READ_STATUS2, &status2);
                        SFC_MSG_INFO("old wp status: 0x%x 0x%x\n", status1, status2);
                        if(((status1 & SFC_WP_MASK_BIT2_TO_BIT7) != 0x10) \
                            || ((status2 & SFC_WP_MASK_BIT6) != SFC_WP_MASK_BIT6)) {
                                status1 = (status1 & (~SFC_WP_MASK_BIT2_TO_BIT7)) | 0x10;
                                status2 = status2 | SFC_WP_MASK_BIT6;
                                status_16bit = ((status2 << 8) | status1);
                                if(-1 == sfc_set_feature_16bit(CMD_WRITE_STATUS1, status_16bit)) {
                                        SFC_MSG_ERR("sfc_lock fail\n");
                                        return -1;
                                }
                                SFC_MSG_INFO("new wp status: 0x%x 0x%x\n", status1, status2);
                        }
                        break;

                default:
                        break;
        }

        return 0;
}

static int sfc_unlock(spinor_info_t *spinor_info)
{
        unsigned char status1 = 0xFF, status2 = 0xFF, status3 = 0xFF;
        unsigned short status_16bit = 0xffff;
        switch(spinor_info->id) {
                case GD25Q128C_128Mbit:
                case WINBOND_W25Q128FV_128Mbit:
                case WINBOND_W25Q256FV_256Mbit:
                case WINBOND_W25M512JV_512Mbit:
                        sfc_get_feature(CMD_READ_STATUS1, &status1);
                        sfc_get_feature(CMD_READ_STATUS2, &status2);
                        sfc_get_feature(CMD_READ_STATUS3, &status3);
                        SFC_MSG_INFO("old wp status: 0x%x 0x%x 0x%x\n", status1, status2, status3);
                        if(((status1 & SFC_WP_MASK_BIT2_TO_BIT7)) \
                            || (status2 & SFC_WP_MASK_BIT6)) {
                                status1 = (status1 & (~SFC_WP_MASK_BIT2_TO_BIT7));
                                status2 = status2 & (~SFC_WP_MASK_BIT6);
                                if((-1 == sfc_set_feature(CMD_WRITE_STATUS1, status1)) \
                                    || (-1 == sfc_set_feature(CMD_WRITE_STATUS2, status2))) {
                                        SFC_MSG_ERR("sfc_lock fail\n");
                                        return -1;
                                }
                                SFC_MSG_INFO("new wp status: 0x%x 0x%x\n", status1, status2);
                        }
                        break;

                case GD25Q64C_64Mbit:
                        sfc_get_feature(CMD_READ_STATUS1, &status1);
                        sfc_get_feature(CMD_READ_STATUS2, &status2);
                        sfc_get_feature(CMD_READ_STATUS3, &status3);
                        SFC_MSG_INFO("old wp status: 0x%x 0x%x 0x%x\n", status1, status2, status3);
                        if(((status1 & SFC_WP_MASK_BIT2_TO_BIT7)) \
                            || (status2 & SFC_WP_MASK_BIT6)) {
                                status1 = (status1 & (~SFC_WP_MASK_BIT2_TO_BIT7));
                                status2 = status2 & (~SFC_WP_MASK_BIT6);
                                if((-1 == sfc_set_feature(CMD_WRITE_STATUS1, status1)) \
                                    || (-1 == sfc_set_feature(CMD_WRITE_STATUS2, status2))) {
                                        SFC_MSG_ERR("sfc_lock fail\n");
                                        return -1;
                                }
                                SFC_MSG_INFO("new wp status: 0x%x 0x%x\n", status1, status2);
                        }
                        break;

                case WINBOND_W32Q64_64Mbit:
                        sfc_get_feature(CMD_READ_STATUS1, &status1);
                        sfc_get_feature(CMD_READ_STATUS2, &status2);
                        SFC_MSG_INFO("old wp status: 0x%x 0x%x \n", status1, status2);
                        if(((status1 & SFC_WP_MASK_BIT2_TO_BIT7)) \
                            || (status2 & SFC_WP_MASK_BIT6)) {
                                status1 = (status1 & (~SFC_WP_MASK_BIT2_TO_BIT7));
                                status2 = status2 & (~SFC_WP_MASK_BIT6);
                                status_16bit = ((status2 << 8) | status1);
                                if(-1 == sfc_set_feature_16bit(CMD_WRITE_STATUS1, status_16bit)) {
                                        SFC_MSG_ERR("sfc_lock fail\n");
                                        return -1;
                                }
                                SFC_MSG_INFO("new wp status: 0x%x 0x%x\n", status1, status2);
                        }
                        break;

                default:
                        break;

        }

        return 0;
}

static void sfc_init_early(void)
{
        rtd_outl(REG_SFC_SCK, ((0xB & FDIV_MASK) << FDIV_OFFSET));  // SCK  500M/12 = 42M
        rtd_maskl(REG_SFC_CE, 0xffff00ff, (0x19 & PHCNT_MASK) << PHCNT_OFFSET); 	/*set phcnt value to 0x19*/
        rtd_outl(REG_SFC_POS_LATCH, (POS_LATCH_FALL<< POS_LATCH_OFFSET) | ((0x4 & DELAY_SEL_MASK) << DELAY_SEL_OFFSET));
		
        // auto
        rtd_outl(REG_SFC_EN_WR, (1 << WT_PROG_EN_OFFSET) | (CMD_WRITE_ENABLE & WR_EN_OP_MASK));
        rtd_outl(REG_SFC_WAIT_WR, (1 << WT_PROG_DONE_OFFSET) | (CMD_READ_STATUS1 & RDSR_OP_MASK));

        return;
}

static int sfc_init(spinor_info_t *spinor_info)
{
        int ret = 0;

        if(spinor_info->multi_die_support){
                int i = 0;
                int die_num = spinor_info->size/spinor_info->per_die_size;
                unsigned char status3 = 0;
		
                for( i = 0; i < die_num; i++){
                        sfc_die_select((unsigned char)i);
                        sfc_get_feature(CMD_READ_STATUS3, &status3);

                        if((status3 & 0x60) != 0x20){
                        	sfc_unlock(spinor_info);
						
                        	status3 =  status3 | 0x60;
                        	status3 =  status3 & 0xBF;
                        	sfc_set_feature(CMD_WRITE_STATUS3, status3);
                        }
                }
				
                sfc_die_select(0);
        }


        if((spinor_info->read_mode == eMODE_READ_QUAD_FAST || spinor_info->write_mode == eMODE_WRITE_QUAD_FAST)) {
                if((spinor_info->cmd_quad_rdsr == EMPTY_VALUE) || (spinor_info->cmd_quad_wrsr == EMPTY_VALUE) || (spinor_info->cmd_quad_bit == EMPTY_VALUE)) {
                        SFC_MSG_ERR("quad mode set error, please check s_device[]\n");
                        return -1;
                }

                if(sfc_check_quad(spinor_info) == 0) {
                        SFC_MSG_INFO("enabel quad mode\n");
                        ret = sfc_enable_quad(spinor_info);
                        if(ret == -1) {
                                SFC_MSG_ERR("enabel quad mode fail\n");
                                return ret;
                        }
                }
        }

#ifdef CONFIG_RTK_SPI_WP_HW
        if(sfc_check_quad(spinor_info) == 1) {
                SFC_MSG_ERR("the WP_HW is conflict with quad mode, only chose the one\n");
                return -1;
        }
#endif

#ifdef CONFIG_RTK_SPI_WP_SOFT
        sfc_lock(spi_dev);
#endif

        //rtd_outl(REG_SFC_EN_WR, (0<<WT_PROG_DONE_OFFSET) | (0x0&WR_EN_OP_MASK));
        //rtd_outl(REG_SFC_WAIT_WR, (0<<WT_PROG_EN_OFFSET) | (0x0&RDSR_OP_MASK));
        return ret;
}

static int sfc_identify(spinor_info_t **dev)
{
        unsigned int id = 0;
        unsigned int retry = 2;

#if defined(Config_Target_Board_MERLIN2) || defined(Config_Target_Board_RTD284X)
        rtd_outl(SB2_CTRL_REG, rtd_inl(SB2_CTRL_REG) & 0xfffffffe); // release spi flash
#endif

        // REG_SFC_POS_LATCH :9 come from dora
        //rtd_outl(REG_SFC_POS_LATCH, (POS_LATCH_RISE << POS_LATCH_OFFSET) | (DELAY_SEL_4 << DELAY_SEL_OFFSET));
        *dev = malloc(sizeof(spinor_info_t));

        sfc_init_early();
        
RETRY:
        sfc_read_id(&id);
        //sfc_debug();

        SFC_MSG_INFO("spi nor id: 0x%x\n", id);
        if(sfc_match_chip(dev, id) == -1) {
                if(retry) {
                        retry--;
                        SFC_MSG_INFO("[SPINOR-INFO] do_identify_s retry %d!!!\n", retry);
                        goto RETRY;
                } else {
                        SFC_MSG_ERR("[SPINOR-INFO] do_identify_s fail!!!\n");
                        free(*dev);
                        *dev = NULL;
                        return -1;
                }
        }

        SFC_MSG_INFO("do_identify_s success!!!\n");

        return 0;
}

inline size_t sfc_memcpy(unsigned int dst, unsigned int src, size_t len)
{
        unsigned int i;
        for (i = 0; i < len; i++) {
                rtd_outb((dst + i), rtd_inb(src + i));
        }

        return len;
}

static void sfc_read_bytes(spinor_info_t *spinor_info, unsigned int from, unsigned int len, unsigned int *retlen, unsigned char *buf)
{
        if(spinor_info->address_4bytes_enable)
                sfc_enable_4bytes_addr_mode();

        switch_read_mode(spinor_info);

        *retlen = sfc_memcpy((unsigned int)buf, (SPI_BASE_ADDR + (unsigned int)from), len);

        if(spinor_info->address_4bytes_enable)
                sfc_disable_4bytes_addr_mode();

        return;
}

static int sfc_write_bytes(spinor_info_t *spinor_info, unsigned int to, unsigned int len, unsigned int *retlen, unsigned char *buf)
{
        int ret = 0;

        if(spinor_info->address_4bytes_enable)
                sfc_enable_4bytes_addr_mode();
		
        switch_write_mode(spinor_info);
        *retlen = sfc_memcpy((SPI_BASE_ADDR + (unsigned int)to), (unsigned int)buf, len);

        ret = sfc_wait_write_done();

        if(spinor_info->address_4bytes_enable)
                sfc_disable_4bytes_addr_mode();
        return ret;
}

static void sfc_read_md_ops(spinor_info_t *spinor_info, unsigned int from, unsigned int len, unsigned int *retlen, unsigned char *buf)
{
        //SFC_MSG_INFO("md-read from 0x%x len 0x%x buf 0x%x!\r\n", from, len, (unsigned int)buf);
		
        if(spinor_info->address_4bytes_enable)
                sfc_enable_4bytes_addr_mode();
        switch_read_mode(spinor_info);

        SFC_MSG_DEBUG("%s_0x%x_from:0x%x  buf_0x%x len_0x%x\n", __FUNCTION__, __LINE__, from, (unsigned int)buf, (unsigned int)len);
        SFC_MSG_DEBUG("(unsigned long)buf %x, (unsigned longt)len %x\n", (unsigned long)buf, (unsigned long)len);
        rtd_outl(REG_MD_FDMA_DDR_SADDR, (unsigned int)buf);
        rtd_outl(REG_MD_FDMA_FL_SADDR, (unsigned int)from);
        //rtd_outl(REG_MD_FDMA_CTRL2, (1 << FORWART_MAPPINT_OFFSET) | (SET_DMA_LEN << FDMA_MAX_XFER_OFFSET) | ((len&FDMA_LEN_MASK)<<FDMA_LEN_OFFSET) | (FLASH_2_DDR<<FDMA_DIR_OFFSET)); // forward map, no_swap, DMA LEN, FLASH to DDR
        rtd_outl(REG_MD_FDMA_CTRL2, (1 << FORWART_MAPPINT_OFFSET) | (SET_DMA_LEN_256 << FDMA_MAX_XFER_OFFSET) | ((len & FDMA_LEN_MASK) << FDMA_LEN_OFFSET) | (FLASH_2_DDR << FDMA_DIR_OFFSET)); // forward map, no_swap, DMA LEN, FLASH to DDR
        flush_dcache_range((unsigned long)buf, (unsigned long)buf + len);
        rtd_outl(REG_MD_FDMA_CTRL1, (1 << WRITE_EN1_OFFSET) | (1 << FDMA_GO_OFFSET));
        asm("DSB");
        while(rtd_inl(REG_MD_FDMA_CTRL1) & (1 << FDMA_GO_OFFSET)); // wait for MD done its operation
        rtd_outl(REG_MD_SMQ_INT_STATUS, (1 << FDMA_DONE_INIT_STATUS_OFFSET));  // clear flash done bit
        *retlen = len;
        if(spinor_info->address_4bytes_enable)
                sfc_disable_4bytes_addr_mode();

        //sfc_debug_md();
		
        return;
}

static void sfc_read_md(spinor_info_t *spinor_info, unsigned int from, unsigned int len, unsigned int *retlen, unsigned char *buf)
{
	unsigned int unit = 0x3C00 /* 768bytes * 20 */, cnt = 0;
	int read_byte = 0, total_read_byte = 0;
	unsigned int offset = from;
	unsigned char *read_buf = (unsigned char *)buf;

	if(offset % unit){
		read_byte = unit - (offset % unit);
		if(read_byte > len)
			read_byte = len;

		sfc_read_bytes(spinor_info, offset, read_byte, retlen, read_buf);

		read_buf += read_byte;
		total_read_byte += read_byte;
		offset += read_byte;
	}

	cnt = (len - total_read_byte)/unit;
	if(cnt){
		unsigned int loop = 0;

		for(loop = 0; loop < cnt; loop++){
			read_byte = unit;

			sfc_read_md_ops(spinor_info, offset, read_byte, retlen, read_buf);

			read_buf += read_byte;
			total_read_byte += read_byte;
			offset += read_byte;
		}
	}

	if(len-total_read_byte){
		read_byte = len- total_read_byte;

		sfc_read_bytes(spinor_info, offset, read_byte, retlen, read_buf);

		read_buf += read_byte;
		total_read_byte += read_byte;
		offset += read_byte;
	}

	*retlen = len;

	return;
}

static int sfc_write_md(spinor_info_t* spinor_info, unsigned int to, unsigned int len, unsigned int *retlen, unsigned char *buf)
{
        int ret = 0;

        if(spinor_info->address_4bytes_enable)
                sfc_enable_4bytes_addr_mode();
        switch_write_mode(spinor_info);

        // set DDR and flash addr
        rtd_outl(REG_MD_FDMA_DDR_SADDR, (unsigned int)buf);
        rtd_outl(REG_MD_FDMA_FL_SADDR, (unsigned int)to);
        rtd_outl(REG_MD_FDMA_CTRL2, (1 << FORWART_MAPPINT_OFFSET) | (SET_DMA_LEN_512 << FDMA_MAX_XFER_OFFSET) | ((len & FDMA_LEN_MASK) << FDMA_LEN_OFFSET) | (DDR_2_FLASH << FDMA_DIR_OFFSET)); // forward map, swap, DMA LEN, FLASH to DDR

        flush_dcache_range((unsigned long)buf, (unsigned long)(buf + len));
        asm("DSB");

        rtd_outl(REG_MD_FDMA_CTRL1, (1 << WRITE_EN1_OFFSET) | (1 << FDMA_GO_OFFSET));
        while(rtd_inl(REG_MD_FDMA_CTRL1) & (1 << FDMA_GO_OFFSET)); // wait for MD done its operation
        rtd_outl(REG_MD_SMQ_INT_STATUS, (1 << FDMA_DONE_INIT_STATUS_OFFSET));  // clear flash done bit

        *retlen = len;
        ret = sfc_wait_write_done();
        if(spinor_info->address_4bytes_enable)
                sfc_disable_4bytes_addr_mode();
        return ret;
}

static __maybe_unused int sfc_verify_after_write(spinor_info_t *spinor_info, unsigned char * to_verify, unsigned int len_verify, unsigned char *buf_verify)
{
        size_t i = 0;
        to_verify += SPI_BASE_ADDR;
		
        if(spinor_info->address_4bytes_enable)
                sfc_enable_4bytes_addr_mode();
		
        switch_read_mode(spinor_info);
        mdelay(10);
        //printf("\nverify: (0x%x, 0x%x, 0x%x)\n", (unsigned int)to_verify, (unsigned int)buf_verify, len_verify);
        do {
                if (*to_verify != *buf_verify) {
                        SFC_MSG_ERR("0x%08x -> 0x%02x,  0x%08x -> 0x%02x\n", (unsigned int)to_verify, *to_verify, (unsigned int)buf_verify, *buf_verify);
                        goto fail;
                }
                to_verify++;
                buf_verify++;
                asm("DSB");
        } while (++i < len_verify);

        if(spinor_info->address_4bytes_enable)
                sfc_disable_4bytes_addr_mode();

        return 0;

fail:
        if(spinor_info->address_4bytes_enable)
                sfc_disable_4bytes_addr_mode();

        return -1;
}


// the from and len must align with rtk_spi_nor_get_align
static int sfc_erase_flash(spinor_info_t *spinor_info, unsigned int from, unsigned int len)
{
        __maybe_unused unsigned char tmp;
        unsigned int size;
        unsigned int erase_addr = SPI_BASE_ADDR + from;
        unsigned int erase_size = len;
        unsigned int cmd_erase = 0;
        int ret = 0;

        if(spinor_info->min_erase_size == 0) {
                SFC_MSG_ERR("the spi nor min_erase_size is not init\n");
                return -1;
        }

        if(len % spinor_info->min_erase_size != 0) {
                SFC_MSG_ERR("the erase size is not align\n");
                return -1;
        }

        if((erase_addr & (spinor_info->min_erase_size - 1)) != 0) {
                SFC_MSG_ERR("the erase addr is not align\n");
                return -1;
        }

        if(spinor_info->address_4bytes_enable)
                sfc_enable_4bytes_addr_mode();
        switch_erase_mode(spinor_info);

        while(erase_size > 0) {
                //sfc_enable_write();

                if ((erase_size >= ERASE_SZIE_64K) && (((unsigned int)erase_addr & (ERASE_SZIE_64K - 1)) == 0) && (spinor_info->erase_capacity & ERASE_64K)) {
                        //printf("(64k)");
                        cmd_erase = CMD_ERASE_BLOCK_64K;
                        size = ERASE_SZIE_64K;
                } else if ((erase_size >= ERASE_SZIE_32K) && (((unsigned int)erase_addr & (ERASE_SZIE_32K - 1)) == 0) && (spinor_info->erase_capacity & ERASE_32K)) {
                        //printf("(32k)");
                        cmd_erase = CMD_ERASE_BLOCK_32K;
                        size = ERASE_SZIE_32K;
                } else if ((erase_size >= ERASE_SZIE_4K) && (((unsigned int)erase_addr & (ERASE_SZIE_4K - 1)) == 0) && (spinor_info->erase_capacity & ERASE_4K)) {
                        //printf("(4k)");
                        cmd_erase = CMD_ERASE_SECTOR_4K;
                        size = ERASE_SZIE_4K;
                }

                if(cmd_erase) {
                        rtd_outl(REG_SFC_OPCODE, cmd_erase & CMD_MASK);
                        rtd_outl(REG_SFC_CTL, 0x08);
                        rtd_outl(REG_SFC_CTL, (1 << ADDR_EN_OFFSET)); // data_en=0 addr_en=1 dmy_en=0
                        asm("DSB");
                        tmp = rtd_inb(erase_addr);
                        udelay(SFC_DELAY);
                        cmd_erase = 0;
                } else {
                        SFC_MSG_ERR("the erase size is not aligned,please use rtk_spi_nor_get_align()\n");
                        goto fail;
                }

                ret = sfc_wait_erase_done();
                if(ret < 0) {
                        goto fail;
                }

                size = (erase_size > size) ? size : erase_size ;
                erase_size -= size;
                erase_addr += size;
        }

        if(spinor_info->address_4bytes_enable)
                sfc_disable_4bytes_addr_mode();
        return 0;
fail:
        if(spinor_info->address_4bytes_enable)
                sfc_disable_4bytes_addr_mode();
        return -1;
}

static int sfc_check_pin_mux(void)
{
        volatile unsigned int regval1;
        volatile unsigned int regval2;

        /*gpio_73_ps  SPI_DO_src0 <I/O>,<HS>[31:28] = 0x2*/
        mask_register(0xb8000850, 0x0fffffff, (2 << 28));

        /*gpio_74_ps SPI_WP_src0 <I/O>,<HS>[23:20] = 0x2*/
        mask_register(0xb8000850, 0xff0fffff, (2 << 20));
	
        /*gpio_75_ps SPI_DI_src0 <I/O>,<HS>[15:12] = 0x2*/
        mask_register(0xb8000850, 0xffff0fff, (2 << 12));

        /*gpio_76_ps SPI_CS_N <O>,<HS>[7:4] = 0x2*/
        mask_register(0xb8000850, 0xffffff0f, (2 << 4));
	
        /*gpio_77_ps SPI_HOLD_src0 <I/O>,<HS>[31:28] = 0x2*/
        mask_register(0xb8000854, 0x0fffffff, (2 << 28));
	
        /*gpio_78_ps  SPI_SCLK <O>,<HS>[23:20] = 0x2*/
        mask_register(0xb8000854, 0xff0fffff, (2 << 20));

        /*set clock drivering strength to 8mA*/
        rtd_maskl(0xb8000854,0xFFFDFFFF,0x00020000);//8mA

        regval1 = rtd_inl(0xb8000850);
        
        regval2 = rtd_inl(0xb8000854);

        if(((regval1 & 0xf0f0f0f0) != 0x20202020) || ((regval2 & 0xf0f00000) != 0x20200000)) {
                printf("the spi pinmux error!!: regval1:%x regval2:%x\n", regval1, regval2);
                return -1;
        }

        return 0;
}

#ifdef CONFIG_RTK_SPI_WP_HW
// wp high: sfc_hw_wp_set(HW_WP_N_DISABLE);
// wp low:  sfc_hw_wp_set(HW_WP_N_ENABLE);
void sfc_hw_wp_set(unsigned char status)
{
        unsigned char tmp;
        rtd_outl(REG_SFC_WP,  rtd_inl(REG_SFC_WP) & (~HW_WP_N_MASK) | (status << HW_WP_N_OFFSET) | (1 << HW_WP_EN1_OFFSET));
        sfc_get_feature(CMD_READ_STATUS1, &tmp);  // for HW_WP_ENABLE is accept when next cmd

        return;
}
#endif

static int rtk_spi_nor_read_operation(unsigned int from, unsigned int len, unsigned char* buf)
{
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;
        unsigned int rlen = 0;
        unsigned int len_to_read = 0;
        unsigned int len_done_read = 0;
        unsigned int len_verify = len;

        if(spinor_info == NULL) {
                SFC_MSG_ERR("the spi is not init\n");
                return -1;
        }

        if(buf == NULL)
                return -1;

        if(md_rw_flag == 0) {
                len_to_read = len;
                sfc_read_bytes(spinor_info, from, len_to_read, &rlen, buf);
                len_done_read += rlen;
        } else {
                len_to_read = len;
                if(len_to_read < 0x300) { // if data len smaller than 16, not need md
                        sfc_read_bytes(spinor_info, from, len_to_read, &rlen, buf);
                } else {
                        sfc_read_md(spinor_info, from, len_to_read, &rlen, buf);
                }
                len_done_read = rlen;
        }


        if(len_done_read != len_verify) {
                SFC_MSG_ERR("Error, len_done_read:0x%x != len_verify:0x%x!\n", len_done_read, len_verify);
                return -1;
        }
        return 0;
}

/************************************************************************
 *                          rtk_spi_nor_read
 *  Description :
 *  implement the flash write
 *  Parameters :
 *  Return values :
 ************************************************************************/
int rtk_spi_nor_read(unsigned int from, unsigned int len, unsigned char* buf)
{
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;
        int ret = 0;

        if(spinor_info == NULL) {
                SFC_MSG_ERR("the spi is not init\n");
                return -1;
        }

        if(spinor_info->multi_die_support){
                 if(from >= spinor_info->per_die_size){
                        sfc_die_select(0x1);

                        from -= spinor_info->per_die_size;
                }else{
                        sfc_die_select(0x0);

                        if((from + len) > spinor_info->per_die_size){
                                unsigned int read_len = spinor_info->per_die_size - from;
                                
                                ret = rtk_spi_nor_read_operation(from,read_len,buf);
                                if(ret < 0){
                                        SFC_MSG_ERR("%s_%d: read fail\n",__func__, __LINE__);
                                        return ret;
                                }

                                from = 0x0;
                                len -= read_len;
                                buf += read_len;
                                sfc_die_select(0x1);
                        }
                }
        }

        ret = rtk_spi_nor_read_operation(from,len,buf);
        if(spinor_info->multi_die_support)
        	sfc_die_select(0x0);
		
        return ret;
}

int rtk_spi_nor_write_md(unsigned int to, unsigned int len, unsigned char *buf)
{
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;
        unsigned int wlen = 0;
        unsigned int len_to_write = 0;
        unsigned int len_done_write = 0;
        __maybe_unused unsigned char *to_verify = (unsigned char *)to;
        __maybe_unused unsigned int len_verify = len;
        __maybe_unused unsigned char *buf_verify = buf;
        int ret = 0;

        if(spinor_info == NULL) {
                SFC_MSG_ERR("the spi is not init\n");
                return -1;
        }
        if(buf == NULL)
                return -1;

        // 1st pass: write data which smaller than a single page (256 bytes) to make destination address within 256-bytes alignment
        if((len > 0) && (((to + MD_PP_DATA_SIZE - 1) & (~(MD_PP_DATA_SIZE - 1))) - to) > 0) {
                len_to_write = ((to + MD_PP_DATA_SIZE - 1) & (~(MD_PP_DATA_SIZE - 1))) - to;
                if(len_to_write > len)
                        len_to_write = len;

                if(len_to_write < WRITE_BYTES_MASK) { // if data len smaller than 16, not need md
                        SFC_MSG_DEBUG("write 1st pass use cpu\n");
                        ret = sfc_write_bytes(spinor_info, to, len_to_write, &wlen, buf);
                        if(ret < 0) {
                                goto to_error;
                        }

                } else {
                        SFC_MSG_DEBUG("write 1st pass use md\n");
                        ret = sfc_write_md(spinor_info, to, len_to_write, &wlen, buf);
                        if(ret < 0) {
                                goto to_error;
                        }
                }
                to += wlen;
                buf += wlen;
                len -= wlen;
                len_done_write += wlen;
        }

        // 2rd pass: write data in page unit (256 bytes)
        while(len >= MD_PP_DATA_SIZE) {
                SFC_MSG_DEBUG("write 2rd pass use md\n");
                len_to_write = MD_PP_DATA_SIZE;
                ret = sfc_write_md(spinor_info, to, len_to_write, &wlen, buf);
                if(ret < 0) {
                        goto to_error;
                }
                to += wlen;
                buf += wlen;
                len -= wlen;
                len_done_write += wlen;
        }

        // 3th pass: write data which smaller than a single page (256 bytes)
        if(len > 0) {
                len_to_write = len;
                if(len_to_write < WRITE_BYTES_MASK) { // if data len smaller than 16, not need md
                        SFC_MSG_DEBUG("write 3st pass use cpu\n");
                        ret = sfc_write_bytes(spinor_info, to, len_to_write, &wlen, buf);
                        if(ret < 0) {
                                goto to_error;
                        }
                } else {
                        SFC_MSG_DEBUG("write 3th pass use md\n");
                        ret = sfc_write_md(spinor_info, to, len_to_write, &wlen, buf);
                        if(ret < 0) {
                                goto to_error;
                        }
                }
                to += wlen;
                buf += wlen;
                len -= wlen;
                len_done_write += wlen;
        }

        if(len_done_write != len_verify) {
                SFC_MSG_ERR("Error, len_done_write:0x%x != len_verify:0x%x!\n", len_done_write, len_verify);
                return -1;
        }

#if 0
do_verify:
        ret = sfc_verify_after_write(spinor_info, to_verify, len_verify, buf_verify);
        if(ret < 0) {
                printf("spi nor verify fail after write\n");
                return -1;
        }
#endif

to_error:
        return ret;
}

int rtk_spi_nor_write_cpu(unsigned int to, unsigned int len, unsigned char *buf)
{
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;
        unsigned int wlen = 0;
        unsigned int len_to_write = 0;
        unsigned int len_done_write = 0;
        __maybe_unused unsigned char *to_verify = (unsigned char *)to;
        __maybe_unused unsigned int len_verify = len;
        __maybe_unused unsigned char *buf_verify = buf;
        int ret = 0;
        if(spinor_info == NULL) {
                SFC_MSG_ERR("the spi is not init\n");
                return -1;
        }
        if(buf == NULL)
                return -1;

        // 1st pass: write data which smaller than a single page (256 bytes) to make destination address within 256-bytes alignment
        if((len > 0) && (((to + MD_PP_DATA_SIZE - 1) & (~(MD_PP_DATA_SIZE - 1))) - to) > 0) {
                len_to_write = ((to + MD_PP_DATA_SIZE - 1) & (~(MD_PP_DATA_SIZE - 1))) - to;
                if(len_to_write > len)
                        len_to_write = len;

                SFC_MSG_DEBUG("write 1st pass use cpu\n");
                ret = sfc_write_bytes(spinor_info, to, len_to_write, &wlen, buf);
                if(ret < 0) {
                        goto to_error;
                }
                to += wlen;
                buf += wlen;
                len -= wlen;
                len_done_write += wlen;
        }

        // 2rd pass: write data in page unit (256 bytes)
        while(len >= MD_PP_DATA_SIZE) {
                SFC_MSG_DEBUG("write 2rd pass use md\n");
                len_to_write = MD_PP_DATA_SIZE;
                ret = sfc_write_bytes(spinor_info, to, len_to_write, &wlen, buf);
                if(ret < 0) {
                        goto to_error;
                }
                to += wlen;
                buf += wlen;
                len -= wlen;
                len_done_write += wlen;
        }

        // 3th pass: write data which smaller than a single page (256 bytes)
        if(len > 0) {
                len_to_write = len;
                SFC_MSG_DEBUG("write 3st pass use cpu\n");
                ret = sfc_write_bytes(spinor_info, to, len_to_write, &wlen, buf);
                if(ret < 0) {
                        goto to_error;
                }
                to += wlen;
                buf += wlen;
                len -= wlen;
                len_done_write += wlen;
        }

        if(len_done_write != len_verify) {
                SFC_MSG_ERR("Error, len_done_write:0x%x != len_verify:0x%x!\n", len_done_write, len_verify);
                return -1;
        }

#if 0
do_verify:
        ret = sfc_verify_after_write(spinor_info, to_verify, len_verify, buf_verify);
        if(ret < 0) {
                printf("spi nor verify fail after write\n");
                return -1;
        }
#endif

to_error:
        return 0;
}

static int rtk_spi_nor_write_operation(unsigned int to, unsigned int len, unsigned char *buf)
{        
        int ret = 0;
        
        if(md_rw_flag == 0)
                ret = rtk_spi_nor_write_cpu(to, len, buf);
        else
                ret = rtk_spi_nor_write_md(to, len, buf);
        
        return ret;
}

int rtk_spi_nor_write(unsigned int to, unsigned int len, unsigned char *buf)
{
        spinor_info_t *spinor_info =  (spinor_info_t *)spi_dev;
        int ret = 0;

        if(spinor_info == NULL) {
                SFC_MSG_ERR("the spi is not init\n");
                return -1;
        }
        
        if(spinor_info->multi_die_support){
                if(to >= spinor_info->per_die_size){
                        sfc_die_select(0x1);

                        to -= spinor_info->per_die_size;
                }else{
                        sfc_die_select(0x0);
                        
                        if((to + len) > spinor_info->per_die_size){
                                unsigned int write_len = spinor_info->per_die_size - to;
                                
                                ret = rtk_spi_nor_write_operation(to, write_len, buf);                             
                                if(ret < 0){
                                        SFC_MSG_ERR("%s_%d: write fail\n",__func__, __LINE__);
                                        return ret;
                                }

                                to = 0x0;
                                len -= write_len;
                                buf += write_len;
                                sfc_die_select(0x1);
                        }
                }
        }          

        ret =  rtk_spi_nor_write_operation(to, len, buf);
        if(spinor_info->multi_die_support)
        	sfc_die_select(0x0);

        return ret;
}

int rtk_spi_nor_ioctl(char type)
{
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;
        if(spinor_info == NULL) {
                SFC_MSG_ERR("spi_nor is not init\n");
                return -1;
        }

        switch(type) {
                case SFC_LOCK:
                        sfc_lock(spinor_info);
                        break;
                case SFC_UNLOCK:
                        sfc_unlock(spinor_info);
                        break;
                default:
                        break;

        }
        return 0;
}

static int rtk_spi_nor_erase_operation(unsigned int from, unsigned int len)
{
        int ret = 0;
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;
        if(spinor_info == NULL) {
                SFC_MSG_ERR("spi_nor is not init\n");
                return -1;
        }
        ret = sfc_erase_flash(spinor_info, from, len);
        return ret;
}

// the from and len must align with rtk_spi_nor_get_align
int rtk_spi_nor_erase(unsigned int from, unsigned int len)
{
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;
        int ret = 0;
		
        if(spinor_info == NULL) {
                SFC_MSG_ERR("spi_nor is not init\n");
                return -1;
        }
        
        if(spinor_info->multi_die_support){
                 if(from >= spinor_info->per_die_size){
                        sfc_die_select(0x1);

                        from -= spinor_info->per_die_size;
                }else{
                        sfc_die_select(0x0);

                        if((from + len) > spinor_info->per_die_size){
                                unsigned int erase_len = spinor_info->per_die_size - from;
                                
                                ret = rtk_spi_nor_erase_operation(from,erase_len);
                                if(ret < 0){
                                        SFC_MSG_ERR("%s_%d: erase fail\n",__func__, __LINE__);
                                        return ret;
                                }

                                from = 0x0;
                                len -= erase_len;
                                sfc_die_select(0x1);
                        }
                }

        }
        
        ret = rtk_spi_nor_erase_operation(from,len);
        if(spinor_info->multi_die_support)
        	sfc_die_select(0x0);

        return ret;
}

int rtk_spi_nor_get_align(void)
{
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;
        if(spinor_info == NULL) {
                SFC_MSG_ERR("spi_nor is not init\n");
                return -1;
        }
        return spinor_info->min_erase_size;
}

int rtk_spi_nor_get_size(void)
{
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;
        if(spinor_info == NULL) {
                SFC_MSG_ERR("spi_nor is not init\n");
                return -1;
        }

        return spinor_info->size;
}

int rtk_spi_nor_init(void)
{
        int ret;

        if(sfc_check_pin_mux() < 0)
                return -1;

        ret = sfc_identify(&spi_dev);
        if(ret < 0 || spi_dev == NULL) {
                return -1;
        }
        SFC_MSG_INFO("SPI FLASH   =  %s   \n", spi_dev->string);
        ret = sfc_init(spi_dev);
        if(ret < 0) {
                return -1;
        }
        
        if(spi_dev->read_mode == eMODE_READ_NORMAL) {
                SFC_MSG_INFO("support eMODE_READ_NORMAL\n");
        } else if(spi_dev->read_mode == eMODE_READ_NORMAL_FAST) {
                SFC_MSG_INFO("support eMODE_READ_NORMAL_FAST\n");
        } else if(spi_dev->read_mode == eMODE_READ_DUAL_FAST) {
                SFC_MSG_INFO("support eMODE_READ_DUAL_FAST\n");
        } else if(spi_dev->read_mode == eMODE_READ_QUAD_FAST) {
                SFC_MSG_INFO("support eMODE_READ_QUAD_FAST\n");
        }

	 if(spi_dev->address_4bytes_enable){
                SFC_MSG_INFO("support 4Bytes address feature!\r\n");
	 }
	 
        if(spi_dev->multi_die_support){
                SFC_MSG_INFO("support multi die feature, total_size = 0x%x die_size = 0x%x total_die_counts = 0x%x!\r\n", \
                        spi_dev->size, spi_dev->per_die_size, spi_dev->size/spi_dev->per_die_size);
        }
		
        return 0;
}

#define TEST_ADDR 0X0070000
#define TEST_SIZE 0X1000
int debug_read_write(void)
{
        unsigned char write_buf[4096] = {0};
        unsigned char read_buf[4096] = {0xff};
        int i = 0;
        __maybe_unused int erase_size = 0;

        write_buf[0] = 0x55;
        write_buf[1] = 0x22;
        write_buf[2] = 0x44;
        write_buf[3] = 0x33;
        write_buf[4] = 0x66;
        write_buf[5] = 0x55;
        write_buf[6] = 0x44;
        write_buf[7] = 0x33;
        write_buf[8] = 0x22;
        write_buf[9] = 0x11;
        write_buf[10] = 0x00;

        write_buf[2096 - 16] = 0xff;
        write_buf[2096 - 15] = 0xee;
        write_buf[2096 - 14] = 0xdd;
        write_buf[2096 - 13] = 0xcc;

        rtk_spi_nor_init();

        printf("\n\n************read before erase*****\n");
        rtk_spi_nor_read(TEST_ADDR, TEST_SIZE, read_buf);
        for(i = 0; i < 16; i++) {
                if(i % 16 == 0)
                        printf("\n");
                printf("0x%x\t", read_buf[i]);
        }

        rtk_spi_nor_erase(TEST_ADDR, TEST_SIZE);
        rtk_spi_nor_read(TEST_ADDR, TEST_SIZE, read_buf);
        printf("\n\n************read after erase*****\n");
        for(i = 0; i < 16; i++) {
                if(i % 16 == 0)
                        printf("\n");
                printf("0x%x\t", read_buf[i]);
        }

        udelay(SFC_DELAY);

        rtk_spi_nor_write(TEST_ADDR, TEST_SIZE, write_buf);
        rtk_spi_nor_read(TEST_ADDR, TEST_SIZE, read_buf);
        printf("\n\n************read after write*****\n");
        for(i = 0; i < 16; i++) {
                if(i % 16 == 0)
                        printf("\n");
                printf("0x%x\t", read_buf[i]);
        }

        printf("\n\n************read bootcode*********\n");
        rtk_spi_nor_read(0x00012800, 0X400, read_buf);
        for(i = 0; i < 16; i++) {
                if(i % 16 == 0)
                        printf("\n");
                printf("0x%x\t", read_buf[i]);
        }

        return 0;
}

static void fill_buffer(unsigned char * buffer, unsigned int size)
{
        int i = 0;

        printf("%s_%d: 0x%p, 0x%x", __func__, __LINE__, buffer, size);
        for(i = 0; i < size; i = i + 4) {
                *(buffer++) = 0x1a;
                *(buffer++) = 0x2b;
                *(buffer++) = 0x3c;
                *(buffer++) = 0x4d;
        }
}

static void fill_buffer_erase(unsigned char * buffer, unsigned int size)
{
        int i = 0;

        printf("%s_%d: 0x%p, 0x%x", __func__, __LINE__, buffer, size);
        for(i = 0; i < size; i++) {
                *(buffer++) = 0xff;
        }
}


// rtkspi test 0x100000 0x1000
int rtk_spi_test(unsigned int addr_test, unsigned int size_test)
{
        spinor_info_t *spinor_info = (spinor_info_t *)spi_dev;

        unsigned int test_times = 0;
        unsigned int addr_erase = addr_test ;
        unsigned int size_erase = size_test;
        unsigned char * buffer_erase = NULL;

        unsigned int addr = addr_test;
        unsigned int size = size_test;
        unsigned char * buffer = NULL;
        unsigned char * buffer_bak = NULL;

        unsigned long time_start = 0;
        unsigned long time_end = 0;

        if((addr_test & (spinor_info->min_erase_size - 1)) || (size_test & (spinor_info->min_erase_size - 1))) {
                printf("%s_%d: addr_test(0x%x) and size_test(0x%x) must align %x \n", __func__, __LINE__, addr_test, size_test, spinor_info->min_erase_size);
        }

        if(size_test <= 0x10000) {
                buffer_erase = malloc(size_erase);
                buffer = malloc(size);
                buffer_bak = malloc(size_erase);
        } else {
                buffer_erase = (unsigned char *)0x01500000;
                buffer = (unsigned char *)0x01700000;
                buffer_bak =  (unsigned char *)0x01900000;
        }

        if(buffer ==  NULL || buffer_bak == NULL || buffer_bak == NULL) {
                printf("%s_%d: malloc fail: buffer_erase: 0x%p buffer: 0x%p buffer_bak: 0x%p\n", __func__, __LINE__, buffer_erase, buffer, buffer_bak);
                goto error;
        } else {
                printf("%s_%d: malloc ok: buffer_erase: 0x%p buffer: 0x%p buffer_bak: 0x%p\n", __func__, __LINE__, buffer_erase, buffer, buffer_bak);
        }

        printf("addr_test:0x%x size_test: %x\n", addr_test, size_test);
        fill_buffer(buffer, size);
        fill_buffer_erase(buffer_erase, size_erase);
        while(1) {
                printf("erase start\n");
                time_start = get_boot_time_ms();
                if(rtk_spi_nor_erase(addr_erase, size_erase) != 0) {
                        printf("rtk_spi_nor_erase error\n");
                        break;
                } else {
                        time_end = get_boot_time_ms();
                        printf("erase time: %ld\n", time_end - time_start);
                        if(rtk_spi_nor_read(addr_erase, size_erase, buffer_bak) != 0) {
                                printf("erase rtk_spi_nor_read  error\n");
                                break;
                        } else {
                                if(memcmp(buffer_erase, buffer_bak, size_erase) == 0) {
                                        //printf("%d\n", __LINE__);
                                } else {
                                        printf("erase memcmp error:time: %d\n", test_times);
                                        break;
                                }
                        }

                        time_start = get_boot_time_ms();
                        if(rtk_spi_nor_write(addr, size, buffer) != 0) {
                                printf("rtk_spi_nor_write error\n");
                                break;
                        } else {
                                time_end = get_boot_time_ms();
                                printf("write time: %ld\n", time_end - time_start);
                                time_start = get_boot_time_ms();
                                if(rtk_spi_nor_read(addr, size, buffer_bak) != 0) {
                                        printf("write rtk_spi_nor_read error\n");
                                        break;
                                } else {
                                        time_end = get_boot_time_ms();
                                        printf("read time: %ld\n", time_end - time_start);
                                        //printf("%d\n", __LINE__);
                                        if(memcmp(buffer, buffer_bak, size) == 0) {
                                                printf("test_time %d OK\n", test_times);
                                                test_times++;
                                        } else {
                                                printf("write memcmp error: times: %d\n", test_times);
                                                break;
                                        }
                                }
                        }
                }
        }

        printf("test_time %d FAIL\n", test_times);

error:
        if(buffer !=  NULL)
                free(buffer);
        if(buffer_bak != NULL)
                free(buffer_bak);
        if(buffer_erase != NULL)
                free(buffer_erase);

        return 0;
}

typedef struct spi_data_header {
        unsigned int checksum;
        unsigned int spi_addr;
        unsigned int image_size;
        unsigned int reserve;
} spi_data_header_t;

unsigned int check_sum(unsigned char* buffer, unsigned int size)
{
        unsigned int i = 0;
        unsigned int sum = 0;

        for(i = 0; i < size; i++) {
                sum += buffer[i];
        }

        return sum;
}

static unsigned int align_to_boundary(unsigned int size, unsigned int block_size)
{
        return ((size % block_size) ? (size / block_size + 1) * block_size : (size / block_size) * block_size);
}

int do_burn_spi_data(void)
{
        spi_data_header_t burn_header;
        unsigned int image_load_ddr_addr = TEMP_BUFFER_FOR_FLASH_DATA_ADDR;
        unsigned int image_read_ddr_addr = ENCRYPTED_LINUX_KERNEL_ADDR;
        unsigned int image_size = 0;
        signed int ret;
        unsigned int sum = 0;
        char cmd[256];

        memset(cmd, 0, sizeof(cmd));
        memset(&burn_header, 0, sizeof(spi_data_header_t));

        // download image file to image_load_ddr_addr
        snprintf(cmd, sizeof(cmd), "loady %x",  image_load_ddr_addr);
        printf("%s...\n", cmd);
        ret = run_command(cmd, 0);
        if (ret != 0) {
                SFC_MSG_INFO("[INFO %s] %s: failed\n", __FUNCTION__, cmd);
                goto FAIL;
        }
        image_size = getenv_ulong("filesize", 16, 0);

        // check load size
        memcpy((void *)&burn_header, (void *)image_load_ddr_addr, sizeof(spi_data_header_t ));
        if(image_size != (burn_header.image_size + sizeof(spi_data_header_t))) {
                SFC_MSG_INFO("%s: load image size is error\n", __FUNCTION__);
                goto FAIL;
        }

        // checksum: skip spi_data_header_t.checksum
        sum = check_sum((unsigned char *)(image_load_ddr_addr + 4), image_size - 4);
        if(sum != burn_header.checksum) {
                SFC_MSG_INFO("the image checksum error\n");
                goto FAIL;
        }

        unsigned int align  = rtk_spi_nor_get_align();
        if(align == 0) {
                SFC_MSG_INFO("get align size error\n");
                goto FAIL;
        }

        unsigned int write_addr = burn_header.spi_addr;
        unsigned int write_size = burn_header.image_size;
        unsigned int erase_size = align_to_boundary(write_size, align);
        unsigned int data_load_ddr_addr = image_load_ddr_addr + sizeof(spi_data_header_t);
        unsigned int data_read_ddr_addr = image_read_ddr_addr +  sizeof(spi_data_header_t);
        memcpy((void *)image_read_ddr_addr, (void*)&burn_header, sizeof(spi_data_header_t ));

        // erase
        ret = rtk_spi_nor_erase(write_addr, erase_size);
        if(ret < 0) {
                SFC_MSG_INFO("%s: erase fail\n", __FUNCTION__);
                goto FAIL;
        }

        // write
        ret = rtk_spi_nor_write(write_addr, write_size, (unsigned char *)data_load_ddr_addr);
        if(ret < 0) {
                SFC_MSG_INFO("%s burn data write fail\n", __FUNCTION__);
                goto FAIL;
        }

        // read
        ret = rtk_spi_nor_read(write_addr, write_size, (unsigned char *)data_read_ddr_addr);
        if(ret < 0) {
                SFC_MSG_INFO("%s burn data read fail\n", __FUNCTION__);
                goto FAIL;
        }

        // verify: skip spi_data_header_t.checksum
        sum = check_sum((unsigned char *)(image_read_ddr_addr + 4), image_size - 4);
        if(sum != burn_header.checksum) {
                SFC_MSG_INFO("%s burn data verify fail\n", __FUNCTION__);
                goto FAIL;
        }

        SFC_MSG_INFO("%s: burn spi data Success\n", __FUNCTION__);

        return 0;

FAIL:
        SFC_MSG_ERR("%s: burn spi data fail\n", __FUNCTION__);
        return ret;
}

int do_rtk_spi(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
        int ret = CMD_RET_USAGE;
        unsigned long mem_addr;
        unsigned long spi_addr;
        unsigned long byte_size;

        if( argc == 5 ) {
                mem_addr   = simple_strtoul(argv[2], NULL, 16);
                spi_addr   = simple_strtoul(argv[3], NULL, 16);
                byte_size  = simple_strtoul(argv[4], NULL, 16);
                if( strncmp( argv[1], "read", 4) == 0 ) {
                        ret = rtk_spi_nor_read(spi_addr, byte_size, (unsigned char*)mem_addr);
                        return (ret ? 0 : 1);
                } else if( strncmp( argv[1], "write", 5 ) == 0 ) {
                        ret = rtk_spi_nor_write(spi_addr, byte_size, (unsigned char*)mem_addr);
                        return (ret ? 0 : 1);
                }
        } else if(argc == 4) {
                spi_addr   = simple_strtoul(argv[2], NULL, 16);
                byte_size  = simple_strtoul(argv[3], NULL, 16);
                if( strncmp( argv[1], "erase", 5 ) == 0 ) {
                        ret = rtk_spi_nor_erase(spi_addr, byte_size);
                        return (ret ? 0 : 1);
                }
                if( strncmp( argv[1], "test_cpu", 8 ) == 0 ) {
                        md_rw_flag = 0;
                        ret = rtk_spi_test(spi_addr, byte_size);
                        return (ret ? 0 : 1);
                }
                if( strncmp( argv[1], "test_md", 8 ) == 0 ) {
                        md_rw_flag = 1;
                        ret = rtk_spi_test(spi_addr, byte_size);
                        return (ret ? 0 : 1);
                }
	 } else if( argc == 3 ){
                if( strncmp( argv[1], "select", 6 ) == 0 ) {
                        unsigned long die_id = simple_strtoul(argv[2], NULL, 16);
                        sfc_die_select((unsigned char)die_id);
                        SFC_MSG_INFO("Select die%d done!\r\n",  (unsigned char)die_id);
                        return 0;
                }
        } else if( argc == 2 ) {
                if( strncmp( argv[1], "init", 4 ) == 0 ) {
                        ret = rtk_spi_nor_init();
                        return (ret ? 0 : 1);
                } else if (strcmp(argv[1], "rtice") == 0) {
                        return do_burn_spi_data();
                } else if( strncmp( argv[1], "unlock", 4 ) == 0 ) {
                        ret = rtk_spi_nor_ioctl(SFC_UNLOCK);
                        return (ret ? 0 : 1);
                } else if( strncmp( argv[1], "lock", 4 ) == 0 ) {
                        ret = rtk_spi_nor_ioctl(SFC_LOCK);
                        return (ret ? 0 : 1);
                }
        }

        return  ret;
}

U_BOOT_CMD(
        rtkspi, 5, 0, do_rtk_spi,
        "RTK SPI NOR direct function",
        ": all argument in HEX type.\n"
        "rtkspi read  DDR_addr SPI_addr size\n"
        "rtkspi write DDR_addr SPI_addr size\n"
        "rtkspi erase SPI_addr size\n"
        "rtkspi init : execute re-initial flow\n"
        "rtkspi test_cpu : test driver with cpu\n"
        "rtkspi test_md : test driver with md\n"
        "rtkspi unlock : cancle write protect\n"
        "rtkspi lock : enable write protect\n"
        "rtkspi rtice : for mac5 bring up\n"
        "rtkspi select die_id : select die\n"
        "\n"
);
