#include <common.h>
#include <malloc.h>
#include <errno.h>
#include <linux/compat.h>
#include <linux/types.h>
#include <linux/bitops.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>
#include <mtd/mtd-abi.h>
#include "rtk_spi_nand.h"
#include "rtk_spi_nand_base.h"
#include <asm/arch/sys_flash_layout.h>

DEBUG_SN_SET_LEVEL(DEBUG_LEVEL_INFO);


#define NOT_ALIGNED(mtd, x) (x & (mtd->oobblock-1)) != 0

spi_nand_info_t spi_nand_info[CONFIG_SYS_MAX_SPI_NAND_DEVICE];

static struct spi_nand_chip spi_nand_chip_info[CONFIG_SYS_MAX_SPI_NAND_DEVICE];
static char   spi_dev_name[CONFIG_SYS_MAX_SPI_NAND_DEVICE][8];
int    spi_nand_curr_device = -1;

unsigned int g_sn_BootcodeSize = 0;
unsigned int g_sn_Factory_param_size = 0;
unsigned int g_sn_Factory_ro_param_size = 0;

extern int sn_get_free_block(struct mtd_info *mtd,int bad_log_block);
extern int sn_get_log2phy_block(struct mtd_info *mtd,int log_block);
extern int sn_get_log2phy_page(struct mtd_info *mtd,int log_page);
extern  int sn_init_table(struct mtd_info *mtd);
extern int sn_init_bbt_table(struct mtd_info *mtd);
extern int sn_remap_erase_fail_block(struct mtd_info *mtd,int log_block,int bad_block);
extern int sn_remap_write_fail_page(struct mtd_info *mtd,int log_block,int fail_page,
	const u_char * page_content);
extern int sn_remap_read_fail_page(struct mtd_info *mtd,int log_block,int bad_block);

int spi_nand_erase(struct mtd_info *mtd, struct erase_info *instr)
{
    int ret = 0, erase_block_start = 0,erase_block_num = 0,block_id = 0,phy_block= 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv; 
    instr->fail_addr  = 0xffffffff;
    instr->state      = MTD_ERASING;
    erase_block_start = (instr->addr)/this->block_size;
    erase_block_num   = (instr->len + this->block_size - 1)/this->block_size;
    while(block_id < erase_block_num)
    {
        phy_block = sn_get_log2phy_block(mtd,(erase_block_start + block_id));
        if(phy_block > this->total_block_num)
        {
            SN_ERR("%s log block %d get remap failed\n",__FUNCTION__, (erase_block_start + block_id));
            ret = -1;
            break;
        }
        SN_DEBUG("%s log block %d phy block %d \n",__FUNCTION__,(erase_block_start + block_id),phy_block);
        if(this->erase_block(mtd, phy_block) != 0)
        {
            if((ret = sn_remap_erase_fail_block(mtd,(erase_block_start + block_id),
				 phy_block)) != 0){
                break;
            }
        }
        if(ret == 0)
        {
            block_id++; 
        }
    }
    instr->state = MTD_ERASE_DONE;
    return ret;       
}

int spi_nand_write_ecc (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen,
			const u_char * buf, const u_char *oob_buf,struct nand_oobinfo *oobsel)
{
    int ret = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv; 
    unsigned int page_start = 0, page_num = 0, page_id = 0,page_offset=0;
    unsigned int phy_page = 0, phy_block = 0,log_block = 0;

    if ((to + len) > mtd->size) 
    {
        SN_ERR ("%s: Attempt write beyond end of device\n",__FUNCTION__);
        *retlen = 0;
        return -EINVAL;
    }
	
    if (NOT_ALIGNED(mtd, to)||NOT_ALIGNED(mtd, len)) 
    {
        SN_ERR ("%s: Attempt to write not page aligned data\n",__FUNCTION__);
        return -EINVAL;
    }
	if(*retlen)
		*retlen = 0;
	  
    page_start = to/this->page_size;
    page_num   = len/this->page_size;
    
    while(page_id < page_num)
    {
        /*get physical page*/
        log_block   = (page_start + page_id)/this->block_per_page;
        page_offset = (page_start + page_id)%this->block_per_page;
        phy_block   =  sn_get_log2phy_block(mtd,log_block);
        if(phy_block > this->total_block_num)
        {
            SN_ERR("%s log block %d get remap failed\n",__FUNCTION__, log_block);
            ret = -1;
            break;
        }
        phy_page  =  phy_block*this->block_per_page + page_offset; 

       	SN_DEBUG("%s logical page_start %d page_num %d page_offset %d phy_page %d\n",
			__FUNCTION__,page_start,page_num, page_offset,phy_page);

        /*write to spi nand*/
        if(this->write_ecc_page(mtd, phy_page, buf,oob_buf) != 0)
        {
            /*using remap block*/
            if((ret = sn_remap_write_fail_page(mtd,log_block,phy_page,buf))!= 0)
            {
                break;
            }       
        }
   
        if(oob_buf != NULL)
        {
            oob_buf += this->oob_size;
        }
        buf += this->page_size;
       *retlen = *retlen + this->page_size;
        page_id++;	     
    }   
    return ret;
}

static int spi_nand_write (struct mtd_info *mtd, loff_t to, size_t len, size_t *retlen, const u_char *buf)
{
    int ret = 0;
    ret = spi_nand_write_ecc (mtd, to, len, retlen, buf, NULL,NULL);
    return ret;
}

static int spi_nand_write_oob (struct mtd_info *mtd, loff_t to, struct mtd_oob_ops *ops)//for 2.6.34 YAFFS-->mtd
{
    int ret = 0;
    ret =  spi_nand_write_ecc (mtd,to, ops->len, &ops->retlen,ops->datbuf, ops->oobbuf,NULL);
    return ret;
}

static int spi_nand_read_ecc (struct mtd_info *mtd, loff_t from, size_t len,
			size_t *retlen, u_char *buf, u_char *oob_buf, struct nand_oobinfo *oobsel)
{
    int ret = 0,retry_count = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv; 
    unsigned int page_start = 0, page_num = 0, page_id = 0,page_offset=0;
    unsigned int phy_page = 0, phy_block = 0,log_block = 0;
    if ((from + len) > mtd->size) 
    {
        SN_ERR ("%s: Attempt len %d from %lu read beyond end of device\n", __FUNCTION__,len,from);
        *retlen = 0;
        return -EINVAL;
    }

    if (NOT_ALIGNED (mtd, from) || NOT_ALIGNED(mtd, len)) 
    {
        SN_ERR ("%s: Attempt to read not page aligned data\n",__FUNCTION__);
        *retlen = 0;
        return -EINVAL;
    }

	if(*retlen)
		*retlen = 0;
	  
    page_start = from/this->page_size;
    page_num   = len/this->page_size;
    
    while(page_id < page_num)
    {
        /*get physical page*/
        log_block   = (page_start + page_id)/this->block_per_page;
        page_offset = (page_start + page_id)%this->block_per_page;
        phy_block   =  sn_get_log2phy_block(mtd,log_block);
        if(phy_block > this->total_block_num)
        {
            SN_ERR("%s log block %d get remap failed\n",__FUNCTION__, log_block);
            ret = -1;
            break;
        }
        phy_page  =  phy_block*this->block_per_page+ page_offset; 
        SN_DEBUG("%s logical page_start %d page_num %d page_offset %d phy_page %d\n",
			__FUNCTION__,page_start,page_num, page_offset,phy_page);
    	  
        for(retry_count = 0;retry_count < READ_RETRY_COUNT;retry_count++)
        {
            if((ret = this->read_ecc_page (mtd, phy_page,buf,oob_buf))== 0)
            {
                break;
            } 
            SN_INFO("%s retry_count %d page_start %d page_num %d  phy_page %d \n",
					__FUNCTION__,retry_count,page_start,page_num,phy_page);
        }
		if(ret != 0){
          	sn_remap_read_fail_page(mtd,log_block,phy_page);
        }
        if(ret == 0)
        {
            if(oob_buf != NULL)
            {
                oob_buf += this->oob_size;
            }
            buf += this->page_size;
            *retlen = *retlen + this->page_size;
            page_id++;
        }
        else
        {
            SN_ERR("[%s]log_block %d,phy_page %d error: data_len %d != len %d\n",
				__FUNCTION__,log_block,phy_page,*retlen , len);
    	    break;
        }      
    }   
    return ret;
}

static int spi_nand_read (struct mtd_info *mtd, loff_t from, size_t len, size_t * retlen, u_char * buf)
{
    int ret=0;
    unsigned int rd_len = 0, tmp_len = 0;
    tmp_len = rd_len = len;
    // align read length to page boundary
    if (rd_len & (mtd->oobblock - 1))
    {
        rd_len = (rd_len / mtd->oobblock + 1) * mtd->oobblock;
    }

    len = rd_len;
    ret = spi_nand_read_ecc(mtd, from, len, retlen, buf, NULL,NULL);

    if(*retlen == len)
        *retlen=tmp_len;
		
    return ret;
}

static int spi_nand_read_oob (struct mtd_info *mtd, loff_t from, struct mtd_oob_ops *ops)
{
    int ret = 0;
    mtd->oobinfo.useecc = ops->mode;
    ret = spi_nand_read_ecc(mtd, from, ops->len, &ops->retlen,ops->datbuf, ops->oobbuf,NULL);
    return ret;
}

static int spi_nand_rescue_image_read(struct mtd_info *mtd, unsigned char* dst_buf, unsigned int from, 
	unsigned int length, unsigned int page_count, const unsigned char magic)
{
	struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
	int res = 0;
	unsigned int count = 0,retry = 0x100,spare_tags = 0,current_page=0;
	int data_len = 0;
    SN_DEBUG("%s len 0x%x page_count 0x%x magic 0x%x \n",__FUNCTION__,length,page_count,magic);

    if ((from + length) > mtd->size) {
        SN_ERR("%s Attempt read beyond end of device\n",__FUNCTION__);
        return -EINVAL;
    }

    if (NOT_ALIGNED (mtd, from)) {
        SN_ERR("%s: Attempt to read not page aligned data\n",__FUNCTION__);
        return -EINVAL;
    }

    current_page = from/this->page_size;
    if (length % this->page_size){
        length = (length / this->page_size + 1) * this->page_size;
	}

    while (count < length) {
        memset(this->g_oobbuf, 0xff, this->oob_size);
        res = this->read_ecc_page (mtd, current_page, &dst_buf[data_len], this->g_oobbuf);
        if (res) {	// ECC error happened
jump_to_next_blk:
            if (--retry) {
                // jump to the same offset of next block
                current_page += this->block_per_page;
                continue;
            }
            else	// retry too many times, return failed
                return -1;
        }

		// check spare area to see if this correct page
        spare_tags = ((page_count & 0xffff) << 16) | magic;
		SN_DEBUG("%s spare_tags 0x%x current_page %d page_count %d \n",__FUNCTION__,spare_tags,current_page,page_count);
		if (spare_tags != ((unsigned int *)this->g_oobbuf)[this->magic_tag_offset]) {
			// maybe page counter is not correct, or magic number not match, or just an empty page (it's not the page we want anyway)
			goto jump_to_next_blk;
		}

		// read ok, go to next page
        current_page++;
        data_len += this->page_size;
        count += this->page_size;
        page_count++;	// remember to increment page counter in spare area
	}

	return 0;
}

static int spi_nand_scan_block_tag(struct mtd_info *mtd, unsigned int from, unsigned int depth, 
	unsigned int page_count,unsigned char block_tag)
{
	struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
	int res = 0,stop_block = 0,block_id =0, page = 0,tag_block = 0;
	unsigned int spare_tags = 0;

    spare_tags = ((page_count & 0xffff) << 16)|block_tag;
    stop_block = from + depth;
	
    for(block_id = from; block_id < stop_block; block_id++ )
    {
        page = block_id*this->block_per_page;
        memset(this->g_oobbuf, 0xff, this->oob_size);
        res = this->read_oob(mtd, page,this->g_oobbuf,this->oob_size);
        if(res < 0){
            //read fail, jump next block
            continue;
        }else{
            //check spare value
            if(spare_tags == ((unsigned int *)this->g_oobbuf)[this->magic_tag_offset]){
                tag_block = block_id;
                break;
			}
        }
    }

    return tag_block;
}

int spi_nand_read_page(struct mtd_info *mtd,int start_page,int page_num, char *buf_content)
{
    int ret = 0,loop = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv; 
    SN_DEBUG("%s start_page %d page_num %d !\n",__FUNCTION__,start_page,page_num); 
    for(loop = 0; loop < page_num; loop++)
    {
        if((ret = this->read_ecc_page(mtd,start_page,buf_content,NULL))!=0)
        {
            SN_ERR("%s read page %d error !\n",__FUNCTION__,start_page ); 
            break;
        }
        start_page++;
        buf_content += (1<<this->page_shift);
    }
    return ret;
}

int spi_nand_read_random(struct mtd_info *mtd,char *buf,
	int start_addr, int length,int raw)
{
    int ret = 0,page_num = 0,loop = 0,size = length;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv; 
    int left = start_addr%this->page_size,start_page = 0,phy_page = 0;
    char * tmp_buf = (char *)malloc(sizeof(char)*this->page_size);
    SN_DEBUG("%s start_addr 0x%x length 0x%x left 0x%x raw %d\n",__FUNCTION__,start_addr,length,left,raw);
    if(tmp_buf!=NULL)
    {
        if(left != 0)
        {
            start_page = start_addr/this->page_size;
            size  = length>(this->page_size - left)?(this->page_size - left):length;
            phy_page = (raw == 0)?(sn_get_log2phy_page(mtd,start_page)):start_page;   
            SN_DEBUG("%s start_page %d left-size 0x%x  phy_page %d\n",
               __FUNCTION__,start_page,size,phy_page); 
            if((ret = spi_nand_read_page(mtd,phy_page,1,tmp_buf))!=0)
            {
                goto out;
            }
            memcpy(buf,tmp_buf+left,size);
            start_addr += size;
            buf    += size;
            length -=size;
        }
    
        start_page = start_addr/this->page_size;
        page_num   = length/this->page_size;
        SN_DEBUG("%s start_addr 0x%x start_page %d length 0x%x page_num %d\n",
			     __FUNCTION__,start_addr,start_page,length,page_num); 
        for(loop = 0; loop < page_num ; loop++)
        {
            phy_page = (raw == 0)?(sn_get_log2phy_page(mtd,start_page)):start_page; 
            if((ret = spi_nand_read_page(mtd,phy_page,1,tmp_buf))!=0)
            {
                goto out;
            }
            memcpy(buf,tmp_buf,this->page_size);
            buf += this->page_size;
            length -= this->page_size;
            start_page++;
        }
        if(length != 0)
        {
            phy_page = (raw == 0)?(sn_get_log2phy_page(mtd,start_page)):start_page; 
            SN_DEBUG("%s start_page %d remain-size 0x%x  phy_page %d length 0x%x\n",
               __FUNCTION__,start_page,size,phy_page,length); 
            if((ret = spi_nand_read_page(mtd,phy_page,1,tmp_buf))!=0)
            {
                goto out;
            }
            memcpy(buf,tmp_buf,length); 
        }
        free(tmp_buf);
    }
out:
    return ret;
}

int spi_nand_write_block(struct mtd_info *mtd,int start_page,
	char *buf_content, int  block_tag)
{
    int ret = 0,loop = 0; 
    unsigned int spare_value = 0, page_counter = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv; 
    memset(this->g_oobbuf,0xff,this->oob_size);
    page_counter = (block_tag>>16)& 0xffff;
    SN_DEBUG("%s start_page %d block_tag 0x%x page_counter 0x%x!\n",__FUNCTION__,start_page,block_tag,page_counter);
    for(loop = 0; loop < this->block_per_page; loop++)
    {
        spare_value  = (block_tag & 0x0000ffff) | ((page_counter & 0xffff) << 16);
        *(unsigned int*)(&this->g_oobbuf[this->magic_tag_offset]) = spare_value;
        if((ret = this->write_ecc_page(mtd,start_page,buf_content,this->g_oobbuf))!=0)
        {
            SN_DEBUG("%s write page %d error !\n",__FUNCTION__,start_page ); 
            break;
        }
        start_page++;
        page_counter++;
        buf_content += (1<<this->page_shift);
    }
    return ret;
}

int spi_nand_write_random(struct mtd_info *mtd,char *buf,
	int start_addr, int length,int raw)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv; 
    int ret = 0,start_block =0,block_num = 0,phy_page = 0,size = length;
    int loop = 0,left = start_addr%this->block_size;
    char * tmp_buf = (char *)malloc(sizeof(char)*this->block_size);
    SN_DEBUG("%s start_addr 0x%x length 0x%x left 0x%x raw %d\n",__FUNCTION__,start_addr,length,left,raw); 
    if(tmp_buf!=NULL)
    {
        memset(tmp_buf,0xFF,this->block_size);
        if(left != 0)
        {
            start_block = start_addr/this->block_size;
            size =  length>(this->block_size-left)?(this->block_size-left):length;
            phy_page = (raw == 0)?(sn_get_log2phy_page(mtd,start_block*this->block_per_page))
                              :start_block*this->block_per_page;
            SN_DEBUG("%s start_block  %d size 0x%x  phy_page %d\n",
               __FUNCTION__,start_block,size,phy_page); 
            if((ret = spi_nand_read_page(mtd,phy_page,this->block_per_page,tmp_buf))!=0)
            {
                goto out;
            }
            if((ret = this->erase_block(mtd, phy_page*this->block_per_page))!=0)
            {
                goto out;
            }
            memcpy(tmp_buf+left,buf,size);
            if((ret = spi_nand_write_block(mtd,phy_page,tmp_buf,0xff))!= 0)
            {
                goto out;
            }
            buf += size;
            start_addr += size;
            length -= size;
        }

        start_block = start_addr/this->block_size;
        block_num   = length/this->block_size;   
        SN_DEBUG("%s start_block %d start_addr 0x%x,length 0x%x block_num %d \n",
			  __FUNCTION__,start_block,start_addr,length,block_num); 
        for(loop = 0; loop < block_num; loop++)
        {
            phy_page = (raw == 0)?(sn_get_log2phy_page(mtd,start_block*this->block_per_page))
                              :start_block*this->block_per_page;
            memcpy(tmp_buf,buf,this->block_size);
            if((ret =this->erase_block (mtd, phy_page*this->block_per_page))!=0)
            {
                goto out;
            }
            if((ret = spi_nand_write_block(mtd,phy_page,tmp_buf,0xff))!= 0)
            {
                goto out;
            }
            buf += this->block_size;
            length -= this->block_size;
            start_block++;
        }   
    
        if(length != 0)
        {
            phy_page = (raw == 0)?(sn_get_log2phy_page(mtd,start_block*this->block_per_page))
                              :start_block*this->block_per_page;
            memset(tmp_buf,0xFF,this->block_size);
            SN_DEBUG("%s start_block  %d size 0x%x phy_page %d\n",
               __FUNCTION__,start_block,length,phy_page); 
            if((ret = spi_nand_read_page(mtd,phy_page,this->block_per_page,tmp_buf))!=0)
            {
                goto out;
            }
            if((ret = this->erase_block (mtd,phy_page*this->block_per_page))!=0)
            {
                goto out;
            }
            memcpy(tmp_buf,buf,length);
            if((ret = spi_nand_write_block(mtd,phy_page,tmp_buf,0xff))!= 0)
            {
                goto out;
            }  
        }
        free(tmp_buf);
    }
	else
	{
	    SN_ERR("%s malloc size 0x%x failed!\n",__FUNCTION__,this->block_size); 
	}
    
out:
    return ret;
} 

int is_good_block(struct mtd_info *mtd,unsigned int phy_block)
{
    int ret = 0,page_addr = 0,loop = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    unsigned char block_status_p1 = 0;
    page_addr = phy_block*this->block_per_page;
    if((ret = this->read_oob(mtd, page_addr, this->g_oobbuf,this->oob_size))!=0)
    {
        SN_ERR("%s: read_oob page=%d failed\n", __FUNCTION__, page_addr);
    }
    else
    {
        block_status_p1 = this->g_oobbuf[this->bad_tag_offset];
        if(phy_block < this->bootcode_end_phy_block)
        {
            block_status_p1 = this->g_oobbuf[this->magic_tag_offset];
            for(loop = 0; loop< SPECIAL_SPI_BLOCK_TAG_N; loop++)
            {
                if(block_status_p1 == Special_SPI_BlockTag[loop])
                {
                    break;
                }
            }
            if(loop >= SPECIAL_SPI_BLOCK_TAG_N)
            {
                SN_ERR("Bootcode/Secure-store/Factory area: block tag error!!! block_status_p1 = [0x%x], block_no = [0x%x]\n",block_status_p1,phy_block);
                ret = 1;		//error block tag, bad block
            }else{
                //check special block
                if(phy_block == 0x0)
                {
                    if (( block_status_p1 != BLOCK_CLEAN)&&(block_status_p1 != BLOCK_OTHER_DATA))
                    {
                        SN_ERR("Profile area: block tag error!!! block_status_p1 = [0x%x], block_no = [0x%x]\n",block_status_p1,phy_block);
                        ret = 1;			//error block tag, bad block
                    }
                }
                else if(phy_block < IBT_SCAN_START_BLOCK){
					if((block_status_p1 != BLOCK_BBT)&&(block_status_p1 != BLOCK_IBT)
						&&(block_status_p1 != BLOCK_CLEAN)){
						SN_ERR("IBT-BBT area: block tag error!!! block_status_p1 = [0x%x], "
							"block_no = [0x%x]\n",block_status_p1,phy_block);
                        ret = 1;		//error block tag, bad block
					}
				}
            }
        }else{
            if((block_status_p1 == 0xff)||(block_status_p1 == BLOCK_FACTORY_SETTING))
            {
                 ret = 0;
            }
			else
			{
			    SN_ERR("block tag error!!! block_status_p1 = [%x], block_no = [0x%x]\n",block_status_p1,phy_block);
                ret = 1;		//error block tag, bad block
			}
        }
    }   
    return ret;
}

int spi_nand_block_is_bad(struct mtd_info *mtd,loff_t ofs)
{
    int ret = 0,phy_block = 0;
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv; 
    phy_block = ofs/this->block_size;
    ret = is_good_block(mtd,phy_block);
    return ret;
}

static int spi_nand_block_status(struct mtd_info *mtd,  loff_t ofs)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv; 
    unsigned int page_addr = 0;
    unsigned char block_status_p1 = 0;
	
    page_addr = ((int) ofs) >> this->page_shift;
    memset(this->g_oobbuf, 0xff, this->oob_size);
    if(this->read_oob(mtd, page_addr, this->g_oobbuf,this->oob_size)!=0)
    {
        SN_ERR("%s: read_oob page=%d failed\n", __FUNCTION__, page_addr);
    }
    else
    {
        block_status_p1 = this->g_oobbuf[this->magic_tag_offset];
    } 
    return block_status_p1;
}

void printf_spi_nand_info(struct mtd_info *mtd)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    printf("---------------spi nandflash %s id 0x%x infomation--------------\n",mtd->PartNum,this->id);
    printf("block_size 0x%x block_shift: %lu\n", this->block_size,(unsigned long)this->block_shift);
    printf("page_size 0x%x page_shift: %lu\n", this->page_size,(unsigned long)this->page_shift);
    printf("oob_size %d oob_shift: 0x%x\n",this->oob_size, this->oob_shift);
    printf("total block num %d page num: %d\n",this->total_block_num, this->total_page_num);
    printf("block per page %d pagemask: %lu\n",this->block_per_page, (unsigned long)this->pagemask);  
    printf("bad block offset %d magic offset %d \n",this->bad_tag_offset,this->magic_tag_offset);
	printf("support cache mode %d (0:normal 1:dual 2:quad 3:quad_IO)\n", this->support_cache_mode);
}

int rtk_spi_nand_scan(struct mtd_info *mtd)
{
    struct spi_nand_chip *this = (struct spi_nand_chip *)mtd->priv;
    unsigned int i = 0,spi_nand_type_id = 0;
    int rtk_lookup_table_flag =0;

    mtd->name = "rtk_spi_nand";
    SN_INFO("[%s] \n", mtd->name);

    /*get spi nand id*/  
    this->read_id(mtd, &spi_nand_type_id);
    SN_INFO("READ ID:0x%x \n",spi_nand_type_id);
 
    for (i = 0; spi_nand_device[i].name; i++)
    {
        if (spi_nand_device[i].id==spi_nand_type_id)
        {
            mtd->PartNum             = spi_nand_device[i].name;
            this->id                 = spi_nand_type_id;
            this->device_size        = spi_nand_device[i].size;
            this->block_size         = spi_nand_device[i].blockSize;
            this->page_size          = spi_nand_device[i].pageSize;
            this->oob_size           = spi_nand_device[i].oobSize;
            this->bad_tag_offset     = spi_nand_device[i].bad_tag_offset - spi_nand_device[i].pageSize;
            this->magic_tag_offset   = spi_nand_device[i].magic_tag_offset-spi_nand_device[i].pageSize;
#if defined(CONFIG_RTD293X_Q128)
            if(spi_nand_device[i].support_cache_mode >= CACHE_4X)
            {
                this->support_cache_mode = CACHE_2X; //128 pin not have wp and hold pin
            }
#else 
            this->support_cache_mode = spi_nand_device[i].support_cache_mode;
#endif
            this->cache_bit_offset   = spi_nand_device[i].cache_bit_offset;
            rtk_lookup_table_flag    = 1;
            SN_INFO("nand part=%s, id=%x, device_size=%lu page_size=%d\n",
            spi_nand_device[i].name, spi_nand_device[i].id, (unsigned long)spi_nand_device[i].size, spi_nand_device[i].pageSize);
            break;
        }   	
    }

    if ( !rtk_lookup_table_flag )
    {
        SN_ERR("Warning: Lookup spi nand Table do not have this spi nand !!\n");
        return -1;
    }
        
    this->block_shift      = generic_ffs(this->block_size)-1;
    this->page_shift       = generic_ffs(this->page_size )-1;
    this->oob_shift        = generic_ffs(this->oob_size)-1;
    this->total_block_num  = this->device_size>>this->block_shift;
    this->total_page_num   = this->device_size>>this->page_shift;
    this->block_per_page   = this->block_size/this->page_size;
    this->pagemask         = (this->device_size >> this->page_shift) - 1;
        
    mtd->size              = this->device_size;
    mtd->oobsize           = this->oob_size;
    mtd->writesize         = this->page_size;
    mtd->oobblock          = mtd->writesize;
    mtd->erasesize         = this->block_size;
    mtd->eccsize           = 512;
  
    this->g_databuf = kmalloc( this->page_size, GFP_KERNEL );
    if ( !this->g_databuf )
    {
        SN_ERR("%s: Error, no enough memory for g_databuf\n",__FUNCTION__);
        return -ENOMEM;
    }
    memset(this->g_databuf, 0xff, this->page_size);
    this->g_oobbuf = kmalloc(this->oob_size, GFP_KERNEL );
    if ( !this->g_oobbuf )
    {
        SN_ERR("%s: Error, no enough memory for g_oobbuf\n",__FUNCTION__);
        return -ENOMEM;
    }
    memset(this->g_oobbuf, 0xff, this->oob_size);
    printf_spi_nand_info(mtd);
    return 0;
}

int spi_nand_mtd_init (struct mtd_info *mtd)
{
    int ret = 0; 
    set_spi_nand_pin_mux();
    sfc_reset();
	
    if (rtk_spi_nand_scan (mtd) < 0 || mtd->size == 0)
    {
        SN_ERR("%s: Error, cannot do nand_scan(on-board)\n", __FUNCTION__);
        ret = -ENODEV;
    }
    else
    {
        sfc_unlock_flash();
        sfc_set_cachemode(mtd);
        ret = sn_init_table(mtd);   
    }
    return ret;
}

void spi_nand_function_init(struct mtd_info *mtd)
{
    struct spi_nand_chip *this  = (struct spi_nand_chip *)mtd->priv;
    this->read_id               = rtk_spi_nand_read_id;
    this->read_ecc_page         = rtk_spi_read_ecc_page;
    this->read_oob              = rtk_spi_read_oob;
    this->write_ecc_page        = rtk_spi_write_ecc_page;
    this->write_oob	            = rtk_spi_write_oob;
    this->erase_block           = rtk_spi_erase_block;
    this->is_good_block         = is_good_block;
    this->suspend               = NULL;
    this->resume                = NULL;
    this->sync                  = NULL;   
    mtd->type                   = MTD_NANDFLASH;
    mtd->flags                  = MTD_CAP_NANDFLASH;
    mtd->ecctype                = MTD_ECC_NONE;
    mtd->erase                  = spi_nand_erase;
    mtd->read                   = spi_nand_read;
    mtd->write                  = spi_nand_write;
    mtd->read_ecc               = spi_nand_read_ecc;
    mtd->write_ecc              = spi_nand_write_ecc;
    mtd->read_oob               = spi_nand_read_oob;
    mtd->write_oob              = spi_nand_write_oob;
    mtd->block_isbad            = spi_nand_block_is_bad;
    mtd->block_status_get       = spi_nand_block_status;
    mtd->reload_bbt             = sn_init_bbt_table;
    mtd->rescue_image_read      = spi_nand_rescue_image_read;
    mtd->scan_block_tag         = spi_nand_scan_block_tag;
    mtd->point                  = NULL;
    mtd->unpoint                = NULL;
    mtd->sync                   = NULL;
    mtd->lock                   = NULL;
    mtd->unlock                 = NULL;
    mtd->block_markbad          = NULL;
    mtd->set_secu_nand_prot_reg = NULL;
}

int board_spi_nand_init(struct mtd_info *mtd)
{
    int rc = 0;
    if((mtd==NULL)||(mtd->priv == NULL))
    {
        SN_ERR(" error: board nand init failed!\n");
        rc = -1;
        return rc;
    }
    
    //default value
    g_sn_BootcodeSize          = SYS_PARAM_BOOTCODE_SIZE;
    g_sn_Factory_param_size    = SYS_PARAM_FACTORY_SIZE;
    g_sn_Factory_ro_param_size = SYS_PARAM_FACTORY_RO_SIZE;

    spi_nand_function_init(mtd);
    
    if((rc = spi_nand_mtd_init(mtd)) == 0)
    {
        SN_INFO("Realtek SPI Nand Flash Driver is successfully installing.\n");
    }   
    return rc;
}

int spi_nand_register(int devnum)
{
    struct mtd_info *mtd;
    if (devnum >= CONFIG_SYS_MAX_SPI_NAND_DEVICE)
        return -EINVAL;

    mtd      = &spi_nand_info[devnum];
    sprintf(spi_dev_name[devnum], "spi_nand%d", devnum);
    mtd->name = spi_dev_name[devnum];

#ifdef CONFIG_MTD_DEVICE
	/*
	 * Add MTD device so that we can reference it later
	 * via the mtdcore infrastructure (e.g. ubi).
	 */
    add_mtd_device(mtd);
#endif

    if (spi_nand_curr_device == -1)
        spi_nand_curr_device = devnum;

    SN_INFO("%s current device %d \n",__FUNCTION__,spi_nand_curr_device);
    return 0;
}

void spi_nand_init_chip(int index)
{
    struct mtd_info *mtd           = &spi_nand_info[index];
    struct spi_nand_chip *spi_nand = &spi_nand_chip_info[index];
    mtd->priv = spi_nand;

    if (board_spi_nand_init(mtd))
        return;
    spi_nand_register(index);
}

void spi_nand_init(void)
{
    int loop = 0;
    SN_INFO("\n[merlin3]spi nand initialize starting.....\n");
    for (loop = 0; loop < CONFIG_SYS_MAX_SPI_NAND_DEVICE; loop++)
    {
        spi_nand_init_chip(loop);
    }
}
