#include <common.h>
#include <malloc.h>
#include <linux/compat.h>
#include <asm/io.h>
#include  <pcmcia/rtk_pcmcia.h>

#if 0   
/*old timing parameters*/
static unsigned long pcmcia_amtc0 = 0x00130304;
static unsigned long pcmcia_amtc1 = 0x00110002;
static unsigned long pcmcia_iomtc = 0x08020602;
static unsigned long pcmcia_matc0 = 0x00130002;
static unsigned long pcmcia_matc1 = 0x0013000a;
#elif 0  
/*optimal timing parameters */
static unsigned long pcmcia_amtc0 = 0x000b0303;
static unsigned long pcmcia_amtc1 = 0x00090002;
static unsigned long pcmcia_iomtc = 0x06020502;
static unsigned long pcmcia_matc0 = 0x00110002;
static unsigned long pcmcia_matc1 = 0x00110008;
#else 
/*for saving inpack pin timing parameters*/
static unsigned long pcmcia_amtc0 = 0x000b0303;
static unsigned long pcmcia_amtc1 = 0x00090002;
static unsigned long pcmcia_iomtc = 0x06040502;
static unsigned long pcmcia_matc0 = 0x00110002;
static unsigned long pcmcia_matc1 = 0x00110008;
#endif
#define PCMCIA_GPIO_INDEX 0x4B
static int rtk_pcmcia_init = 0;
static int g_auto_pinmux_enable = 0; /*1: auto pin mux, 0: no auto pin mux*/
static int g_fifo_func_use_hw_mode = 0; /*1: HW mode, 0: SW mode*/

extern int setGPIO(int GPIO_NUM, int value);
//#define MAC3_VERIFY
#define CONFIG_PCMCIA_FIFO_MODE
static int inline venus_pcmcia_start_transfer(int is_fifo_mode)
{
	int times = 0;
	if (!(GET_PCMCIA_STS() & PC_PRES1))
	{
        		printf("access pcmcia failed - no card exists!!!\n");
        		return PCMCIA_NODEV;
	}
    
	if (!(GET_PCMCIA_CTRL() & PC_PCR1_OE))
	{
		printf("access pcmcia failed - please enable port first!!!\n");
		return PCMCIA_FAIL;
	}            
	if(g_auto_pinmux_enable) {
		if(is_fifo_mode && g_fifo_func_use_hw_mode) {
			SET_PCMCIA_FIFO_CTRL(GET_PCMCIA_FIFO_CTRL()
				& (~PC_PIN_SHARE_SEL_SOFT_MODE)); /*Let PCMCIA be HW mode */
			SET_TP_OUT_CTRL_REG(GET_TP_OUT_CTRL_REG()
				|TP_OUT_SHARE_HW_MODE | TP_OUT_SHARE_EN);

/*Hw mode no need to polling pause_int*/
#if 0			
			times = 5000000; /*timeout for 500ms.*/
			while(!(GET_TP_OUT_INT_REG() & TP_PAUSE_INT)) {
				if (times-- < 0) {
					PCMCIA_WARNING("hw mode share enable pin timeout\n");
					SET_TP_OUT_CTRL_REG(GET_TP_OUT_CTRL_REG()
						| TP_OUT_SHARE_EN(0));/*disable auto share*/
					return PCMCIA_FAIL;
				}
				ndelay(100);
			}
#endif			
			
			
		} else {
			SET_PCMCIA_FIFO_CTRL(GET_PCMCIA_FIFO_CTRL()
				| PC_PIN_SHARE_SEL_SOFT_MODE); /*Let PCMCIA be SW mode */
			SET_TP_OUT_CTRL_REG((GET_TP_OUT_CTRL_REG()
				| TP_OUT_SHARE_EN) & (~TP_OUT_SHARE_HW_MODE));
			times = 5000000; /*timeout for 500ms.*/
			while(!(GET_TP_OUT_INT_REG() & TP_PAUSE_INT)) {
				if (times-- < 0) {
					PCMCIA_WARNING("sw mode share enable pin timeout\n");
					SET_TP_OUT_CTRL_REG(GET_TP_OUT_CTRL_REG()
						& (~TP_OUT_SHARE_EN));/*disable auto share*/
					return PCMCIA_FAIL;
				}
				ndelay(100);
			}
			CLEAR_TP_PUASE_INT();
			SET_TP_OUT_CTRL_REG(GET_TP_OUT_CTRL_REG()
				| TP_OUT_DATA_SEL_PCMCIA);/*sw mode switch pin mux from tp*/
		}
	}
	return PCMCIA_OK;
}    

static int inline venus_pcmcia_stop_transfer(int is_fifo_mode)
{
	if(g_auto_pinmux_enable) {
		if(is_fifo_mode && g_fifo_func_use_hw_mode) {
/*Hw mode no need to clear out_share_en*/			
#if 0			
			SET_TP_OUT_CTRL_REG(GET_TP_OUT_CTRL_REG()
				| TP_OUT_SHARE_EN(0));/*disable auto share*/
#endif
		}
		else {
			SET_TP_OUT_CTRL_REG(GET_TP_OUT_CTRL_REG()
				& (~TP_OUT_DATA_SEL_PCMCIA));/*sw mode switch pin mux from tp*/
			SET_TP_OUT_CTRL_REG(GET_TP_OUT_CTRL_REG()
				& (~TP_OUT_SHARE_EN));/*disable auto share*/
		}
	}
	return PCMCIA_OK;
}

static int inline venus_pcmcia_do_command(unsigned long cmd, unsigned long ctrl)
{                      
	unsigned long event = 0;
	int wait_time = 200;    
	ctrl &= ~PC_ACCESS_INT;
	ctrl |= PC_CE1_CARD1;  //PC_CE1_CARD1
                        
	SET_PCMCIA_STS(PC_ACCESS_INT);        
    
	SET_PCMCIA_CTRL(ctrl);    
    
	SET_PCMCIA_CMDFF(cmd);
                                             
	while(wait_time--)    
	{
		event=GET_PCMCIA_STS();
		if (event & PC_ACCESS_INT)
			break;
		ndelay(100);    
	}
    
	if (event & PC_APFI)                    
		return PCMCIA_OK;            
    
	if (event & PC_AFI)                
	{
		printf("access pcmcia failed - Access Fault!!!\n");                    
	}
	else
	{
		printf("access pcmcia failed - Access Time Out!!!\n");        
		//venus_pcmcia_phy_reset();     
	}

	SET_PCMCIA_STS(PC_ACCESS_INT);  
    
    return PCMCIA_FAIL;    
}

int venus_pcmcia_read( unsigned short addr, unsigned char* p_data, unsigned short len, unsigned long flags)
{        
	int ret = PCMCIA_OK;  
	unsigned char attr = (flags & PCMCIA_FLAGS_ATTR) ? 1 : 0;    
	unsigned long ctrl =  GET_PCMCIA_CTRL();    
	//printf("rd %s %s, addr = %d, len=%d\n", (flags & PCMCIA_FLAGS_ATTR) ? "attr" : "io", (flags & PCMCIA_FLAGS_FIFO) ? "port" : "mem", addr, len);                    
	ret = venus_pcmcia_start_transfer(0);
	if(ret != PCMCIA_OK)
	{
		printf("pcmcia read operation is failed!!\n");
		return ret;
	}
    	while(len && ret ==PCMCIA_OK)
	{        
        		ret = venus_pcmcia_do_command(PC_CT_READ | PC_AT(attr) | PC_PA(addr), ctrl); 
        		if (ret!=PCMCIA_OK)
            			break;
            
        		*p_data =(unsigned char)(GET_PCMCIA_CMDFF() & 0xFF);    
                
        		//printf("%s MEM[%d] = %02x\n", (attr) ? "ATTR" : "IO", addr, *p_data);
        		p_data++;
        		if ((flags & PCMCIA_FLAGS_FIFO)==0)  // sequential read
            			addr += (attr) ? 2 : 1; 
        		len--;        
    	}         
	venus_pcmcia_stop_transfer(0);        
    	return ret;        
}

int venus_pcmcia_write(unsigned short addr, unsigned char* p_data, unsigned short len, unsigned long flags)
{
	int ret = PCMCIA_OK;
	unsigned char attr = (flags & PCMCIA_FLAGS_ATTR) ? 1 : 0;    
	unsigned long ctrl =  GET_PCMCIA_CTRL();  
	//printf("wr %s %s, addr = %d, len=%d\n", (flags & PCMCIA_FLAGS_ATTR) ? "attr" : "io", (flags & PCMCIA_FLAGS_FIFO) ? "mem" : "port", addr, len);                

	ret = venus_pcmcia_start_transfer(0);
    	while(len && ret ==PCMCIA_OK)
	{
        		ret = venus_pcmcia_do_command(PC_CT_WRITE | PC_AT(attr) | PC_PA(addr) | PC_DF(*p_data++), ctrl);
        		if (ret!=PCMCIA_OK)
            			break;
            		//printf("venus_pcmcia_write sucess\n");
        		if ((flags & PCMCIA_FLAGS_FIFO)==0)  // sequential read
            			addr += (attr) ? 2 : 1; 
        		len--;
	}
    	venus_pcmcia_stop_transfer(0);
    	return ret;
}

int venus_pcmcia_fifo_read(unsigned short addr,
			unsigned char *p_data, unsigned short len, unsigned long flags)
{
	int ret = PCMCIA_OK;
	unsigned long fifo_ctrl;
	unsigned long ctrl =  GET_PCMCIA_CTRL();
	unsigned char attr = (flags & PCMCIA_FLAGS_ATTR) ? 1 : 0;
	unsigned long cmd = PC_CT_READ | PC_AT(attr) | PC_PA(addr);
    	unsigned int times = 0;
	unsigned long event = 0;
	/*PCMCIA_DBG("rd %s %s, addr = %d, len=%d\n", (flags & PCMCIA_FLAGS_ATTR) ? "attr" : "io",
		(flags & PCMCIA_FLAGS_FIFO) ? "port" : "mem", addr, len);*/
	if (len == 0)
		return ret;

	ret = venus_pcmcia_start_transfer(1);
	if (ret != PCMCIA_OK) {
		PCMCIA_WARNING("%s pcmcia read operation is failed!!\n" , __FUNCTION__ );
		return ret;
	}
	
	ctrl &= ~(PC_ACCESS_INT |  PC_FIFO_R_EN | PC_FIFO_WR_OVER_EN );
	ctrl |= PC_CE1_CARD1;
	SET_PCMCIA_CTRL(ctrl);
	SET_PCMCIA_STS(PC_ACCESS_INT | PC_FIFO_R_IE | PC_FIFO_WR_OVER_IE);
	if(g_auto_pinmux_enable && g_fifo_func_use_hw_mode) {
		SET_PCMCIA_NUM_TOTAL(len);
	}
	fifo_ctrl = GET_PCMCIA_FIFO_CTRL();
	if(g_auto_pinmux_enable)
		fifo_ctrl = fifo_ctrl & PC_PIN_SHARE_SEL_SOFT_MODE;
	else
		fifo_ctrl = PC_PIN_SHARE_SEL_SOFT_MODE;
	if (attr)
		fifo_ctrl = fifo_ctrl | PC_ADDR_STEP(1) | PC_WR_NUM(len)
			| PC_FIFO_R_TRIG(0) | PC_FIFO_WR_MODE_ADDRESS_INCREASE
			| PC_FIFO_EN(1);
	else
		fifo_ctrl = fifo_ctrl | PC_ADDR_STEP(0) | PC_WR_NUM(len)
			| PC_FIFO_R_TRIG(0) | PC_FIFO_WR_MODE_ADDRESS_NOCHANGE
			| PC_FIFO_EN(1);
	SET_PCMCIA_FIFO_CTRL(fifo_ctrl);		
	SET_PCMCIA_CMDFF(cmd);

	while (len > 0 && ret == PCMCIA_OK) {
		times = 5000 ;
		event = GET_PCMCIA_STS();
		while ( !( event & PC_FIFO_R_IE )  && times-- > 0){
			ndelay(100);
			event = GET_PCMCIA_STS();
		}
		if (event & PC_FIFO_R_IE) {
			unsigned short fifo_len = PC_FIFO_LEN(GET_PCMCIA_FIFO_STATUS());
			SET_PCMCIA_STS(PC_FIFO_R_IE);
			while (fifo_len) {
				if (len && fifo_len) {
					//ndelay(100); // wait data stable , maybe can take out or less
					*p_data = GET_PCMCIA_FIFO_DATA();
					p_data++;
					len--;
					fifo_len--;
				}
				ret = PCMCIA_OK;
			}
			event = GET_PCMCIA_STS();
			if ((event & PC_FIFO_ERRO_IE)) {
				ret = PCMCIA_FAIL;
				PCMCIA_WARNING("%s access error!!! \n" , __FUNCTION__ );
			}

			if ( event & PC_FIFO_WR_OVER_IE ) {

				SET_PCMCIA_STS(PC_FIFO_WR_OVER_IE);
			}

		} else {
			PCMCIA_WARNING("%s access pcmcia failed - Access Time Out!!!\n" , __FUNCTION__ );
			ret = PCMCIA_FAIL;
		}
	}
	venus_pcmcia_stop_transfer(1);
	return ret;
}

int venus_pcmcia_fifo_write(unsigned short addr,
			unsigned char *p_data, unsigned short len, unsigned long flags)
{
	int ret = PCMCIA_OK;
	unsigned long fifo_ctrl;
	unsigned char attr = (flags & PCMCIA_FLAGS_ATTR) ? 1 : 0;
	unsigned long ctrl =  GET_PCMCIA_CTRL();
	unsigned long cmd = PC_CT_WRITE | PC_AT(attr) | PC_PA(addr);
	unsigned long event = 0;
    	unsigned long times = 0;

	/*PCMCIA_WARNING("wr %s %s, addr = %d, len=%d\n", (flags & PCMCIA_FLAGS_ATTR) ? "attr" : "io",
		(flags & PCMCIA_FLAGS_FIFO) ? "port" : "mem", addr, len);*/
	if (p_data == NULL || len <= 0)
		return ret;
	ret = venus_pcmcia_start_transfer(1);
	if (ret != PCMCIA_OK) {
		PCMCIA_WARNING("%s pcmcia read operation is failed!!\n" , __FUNCTION__ );
		return ret;
	}

	ctrl &= ~(PC_ACCESS_INT | PC_FIFO_W_EN | PC_FIFO_WR_OVER_EN);
	ctrl |= PC_CE1_CARD1;
	SET_PCMCIA_STS(PC_ACCESS_INT | PC_FIFO_W_IE | PC_FIFO_WR_OVER_IE);
	SET_PCMCIA_CTRL(ctrl);
	if(g_auto_pinmux_enable && g_fifo_func_use_hw_mode) {
		SET_PCMCIA_NUM_TOTAL(len);
	}
	fifo_ctrl = GET_PCMCIA_FIFO_CTRL();
	if(g_auto_pinmux_enable)
		fifo_ctrl = fifo_ctrl & PC_PIN_SHARE_SEL_SOFT_MODE;
	else
		fifo_ctrl = PC_PIN_SHARE_SEL_SOFT_MODE;
	if (attr)
		fifo_ctrl = fifo_ctrl | PC_ADDR_STEP(1) | PC_WR_NUM(len)
			| PC_FIFO_R_TRIG(0) | PC_FIFO_WR_MODE_ADDRESS_INCREASE
			| PC_FIFO_EN(1);
	else
		fifo_ctrl = fifo_ctrl | PC_ADDR_STEP(0) | PC_WR_NUM(len)
			| PC_FIFO_R_TRIG(0) | PC_FIFO_WR_MODE_ADDRESS_NOCHANGE
			| PC_FIFO_EN(1);
	SET_PCMCIA_FIFO_CTRL(fifo_ctrl);

	while (len && ret == PCMCIA_OK) {

		ret = PCMCIA_OK;
		times = 5000 ;
		while ( PC_FIFO_FULL(GET_PCMCIA_FIFO_STATUS()) && times-->0)
			ndelay(100);

		if (!PC_FIFO_FULL(event)) {
			while (len) {
				if (PC_FIFO_FULL(GET_PCMCIA_FIFO_STATUS()))
					break;

				SET_PCMCIA_FIFO_DATA( cmd , (*p_data) ) ;
				SET_PCMCIA_CMDFF(cmd);
				len--;
				p_data++;
				ret = PCMCIA_OK;
			}
		} else {
			PCMCIA_WARNING("%s access pcmcia failed - Access Time Out!!!\n" ,
				__FUNCTION__);
			ret = PCMCIA_FAIL;
		}
	}
	if (ret == PCMCIA_OK) {

		times = 5000 ;
		event = GET_PCMCIA_STS();
		while ( !( ( event & PC_FIFO_WR_OVER_IE ) || ( event & PC_FIFO_ERRO_IE ) )  && times-- > 0){
			ndelay(100);
			event = GET_PCMCIA_STS();
		}

		if (event & PC_FIFO_WR_OVER_IE) {
			/*done*/
		} else if (event & PC_FIFO_ERRO_IE) {
			PCMCIA_WARNING("%s,GET_PCMCIA_STS() = 0x%08x!!\n", __FUNCTION__,GET_PCMCIA_STS());
			ret = PCMCIA_FAIL;
		}
	}else{
		PCMCIA_WARNING("%s access error!\n" , __FUNCTION__);
	}

	venus_pcmcia_stop_transfer(1);
	return ret;
}

void venus_pcmcia_clock_enable(unsigned char On)
{     
	if (On)
	{                
		rtd_outl(PCMCIA_CLOCK_ENABLE, SYS_CLKEN2_clken_pc_mask);     // disable clock
        
		udelay(10);
		rtd_outl(PCMCIA_SOFT_RESET, SYS_SRST2_rstn_pc_mask);          // assert pcmcia
        
		udelay(10);
        
		rtd_outl(PCMCIA_SOFT_RESET, SYS_SRST2_rstn_pc_mask | SYS_CLK_RST_WRTIE_DATA);           // deassert pcmcia
        
		udelay(10);
           
		rtd_outl(PCMCIA_CLOCK_ENABLE, SYS_CLKEN2_clken_pc_mask | SYS_CLK_RST_WRTIE_DATA);
	}
	else    
		rtd_outl(PCMCIA_CLOCK_ENABLE, SYS_CLKEN2_clken_pc_mask);     // disable clock         
}

int venus_pcmcia_card_reset(unsigned long CardResetTimeout)
{   
	int ret = PCMCIA_FAIL;                
	unsigned long timeout;                 
	if (!(GET_PCMCIA_STS() & PC_PRES1))
	{ 
		printf("hardware reset pcmcia failed - no card detected\n");        
		return PCMCIA_NODEV;
	}
          
	// convert from ms to jiffes
	printf("reset pcmcia with %lu ms timeout\n", CardResetTimeout);      
	              
	SET_PCMCIA_STS(PC_PII1);                
	SET_PCMCIA_CTRL(GET_PCMCIA_CTRL() | PC_PCR1);
        
	timeout = CardResetTimeout/10;           // timeout = 3 sec        
    
	while ((GET_PCMCIA_CTRL() & PC_PCR1) && timeout-- > 0)                                
		msleep(10);            

	if ((GET_PCMCIA_CTRL() & PC_PCR1))
	{
		printf("hardware reset pcmcia failed,timeout\n");
		goto end_proc;
	}
    
	msleep(1);    

	//PC card asserts READY within 10us of RESET if it will not be ready for access after 20ms
	while ((GET_PCMCIA_STS() & PC_PII1) && timeout-- > 0)    
		msleep(10);            

	if((GET_PCMCIA_STS() & PC_PII1))
	{
		ret = PCMCIA_FAIL;  
		printf("hardware reset pcmcia failed - wait PII timeout\n");
		goto end_proc;
	}
	
	SET_PCMCIA_STS(PC_PII1);  
	printf("hardware reset pcmcia successed \n");
	ret = PCMCIA_OK;
    
end_proc:    
    	return ret;
}

void venus_pcmcia_phy_reset(void)
{       
	int try_cnt = 0;
	printf("venus_pcmcia_phy_reset\n");
    
	SET_PCMCIA_CTRL(GET_PCMCIA_CTRL() | PC_PSR);    
    
	printf("wait PSR done\n");
    
	while(GET_PCMCIA_CTRL() & PC_PSR && try_cnt++<100) 
	{
		printf("wait PSR ready : %lx=%lx\n", PCMCIA_REG_CTRL, GET_PCMCIA_CTRL());
		msleep(1);             
	}
	printf("PSR done\n");
    
	if (GET_PCMCIA_CTRL() & PC_PSR)
	{
		printf("Wait PSR timeout\n");
		printf("%lx=%lx\n", PCMCIA_CLOCK_ENABLE, rtd_inl(PCMCIA_CLOCK_ENABLE));
		printf("%lx=%lx\n", PCMCIA_SOFT_RESET, rtd_inl(PCMCIA_SOFT_RESET));
	}    
}

#if 0
static void pcmcia_pinmux_set(void)
{

#ifdef MAC3_VERIFY
	PCMCIA_WARNING("\nbefore setting start\n");
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000118, rtd_inl(0xb8000118));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000108, rtd_inl(0xb8000108));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000118, rtd_inl(0xb8000118));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000108, rtd_inl(0xb8000108));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000118, rtd_inl(0xb8000118));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000108, rtd_inl(0xb8000108));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8060208, rtd_inl(0xb8060208));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8061100, rtd_inl(0xb8061100));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8061108, rtd_inl(0xb8061108));

	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000884, rtd_inl(0xb8000884));	/*    Addr 0     GPIO130 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000880, rtd_inl(0xb8000880));	/*      Addr 1     GPIO128 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000880, rtd_inl(0xb8000880));	/*      Addr 2     GPIO126 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800087C, rtd_inl(0xb800087C));	/*      Addr 3     GPIO124 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800087C, rtd_inl(0xb800087C));	/*      Addr 4     GPIO122 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*      Addr 5     GPIO120 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*      Addr 6     GPIO118 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000874, rtd_inl(0xb8000874));	/*      Addr 7     GPIO116 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000864, rtd_inl(0xb8000864));	/*      Addr 8     GPIO101 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000864, rtd_inl(0xb8000864));	/*      Addr 9     GPIO99 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*      Addr 10    GPIO94 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*      Addr 11    GPIO97 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000874, rtd_inl(0xb8000874));	/*    Addr 12    GPIO114 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000868, rtd_inl(0xb8000868));	/*      Addr 13    GPIO103 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000868, rtd_inl(0xb8000868));	/*      Addr 14    GPIO105 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000884, rtd_inl(0xb8000884));	/*      Data 0     GPIO132 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000888, rtd_inl(0xb8000888));	/*      Data 1     GPIO134 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000888, rtd_inl(0xb8000888));	/*      Data 2     GPIO136 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000854, rtd_inl(0xb8000854));	/*      Data 3     GPIO83 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000854, rtd_inl(0xb8000854));	/*      Data 4     GPIO85 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000858, rtd_inl(0xb8000858));	/*      Data 5     GPIO87 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000858, rtd_inl(0xb8000858));	/*    Data 6     GPIO89 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800085c, rtd_inl(0xb800085c));	/*      Data 7     GPIO91 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800086c, rtd_inl(0xb800086c));	/*    WE_N       GPIO107 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000850, rtd_inl(0xb8000850));	/*      IREQ0_N    GPIO81 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*      OE_N       GPIO95 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800085c, rtd_inl(0xb800085c));	/*      CE1_N      GPIO93 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*      RST0       GPIO119 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*    WAIT0_N    GPIO121 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800087c, rtd_inl(0xb800087c));	/*      INPACK0_N  GPIO123 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800087c, rtd_inl(0xb800087c));	/*    REG_N      GPIO61 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*    IORD_N     GPIO96 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000864, rtd_inl(0xb8000864));	/*      IOWR_N     GPIO98 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000854, rtd_inl(0xb8000854));	/*      CD0_N      GPIO82 */

	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B724, rtd_inl(0xb801B724));

	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B70C, rtd_inl(0xb801B70C));	/*  attribute memory timing control */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B728, rtd_inl(0xb801B728));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B710, rtd_inl(0xb801B710));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B714, rtd_inl(0xb801B714));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B72C, rtd_inl(0xb801B72C));
	

	PCMCIA_WARNING("\nbefore setting end\n");


	rtd_outl(0xb8000118, rtd_inl(0xb8000118)  | 0x00010001);
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000118, rtd_inl(0xb8000118));
	rtd_outl(0xb8000108, rtd_inl(0xb8000118)  | 0x00010001);
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000108, rtd_inl(0xb8000108));
	rtd_outl(0xb8000118, rtd_inl(0xb8000118)  | 0x00010000);
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000118, rtd_inl(0xb8000118));
	rtd_outl(0xb8000108, rtd_inl(0xb8000118)  | 0x00010000);
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000108, rtd_inl(0xb8000108));
	rtd_outl(0xb8000118, rtd_inl(0xb8000118)  | 0x00010001);
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000118, rtd_inl(0xb8000118));
	rtd_outl(0xb8000108, rtd_inl(0xb8000118)  | 0x00010001);
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000108, rtd_inl(0xb8000108));
	
	rtd_outl(0xb8060208, (rtd_inl(0xb8060208) & 0xFFFFFF0F) | 0x000000F0);	/*       power enable     ST_GPIO_11 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8060208, rtd_inl(0xb8060208));
	rtd_outl(0xb8061100, (rtd_inl(0xb8061100) & 0xFFFFF7FF) | 0x00000800);
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8061100, rtd_inl(0xb8061100));
	rtd_outl(0xb8061108, (rtd_inl(0xb8061108) & 0xFFFFF7FF) | 0x00000800);
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8061108, rtd_inl(0xb8061108));
	/*   fifo mode */
	/*    rtd_outl(0xb801B71C, ( rtd_inl(0xb801B71C) & 0x0)   );  */
	/*    printk("reg 0x%X = 0x%08X\n", 0xb801B71C  , rtd_inl(0xb801B71C)   ); */
	rtd_outl(0xb8000884, (rtd_inl(0xb8000884) & 0x0FFFFFFF) | 0x10000000);	/*    Addr 0     GPIO130 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000884, rtd_inl(0xb8000884));	/*    Addr 0     GPIO130 */

	rtd_outl(0xb8000880, (rtd_inl(0xb8000880) & 0xFFFF0FFF) | 0x00001000);	/*      Addr 1     GPIO128 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000880, rtd_inl(0xb8000880));	/*      Addr 1     GPIO128 */

	rtd_outl(0xb8000880, (rtd_inl(0xb8000880) & 0x0FFFFFFF) | 0x10000000);	/*      Addr 2     GPIO126 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000880, rtd_inl(0xb8000880));	/*      Addr 2     GPIO126 */

	rtd_outl(0xb800087C, (rtd_inl(0xb800087C) & 0xFFFF0FFF) | 0x00001000);	/*      Addr 3     GPIO124 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb800087C, rtd_inl(0xb800087C));	/*      Addr 3     GPIO124 */

	rtd_outl(0xb800087C, (rtd_inl(0xb800087C) & 0x0FFFFFFF) | 0x10000000);	/*      Addr 4     GPIO122 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb800087C, rtd_inl(0xb800087C));	/*      Addr 4     GPIO122 */

	rtd_outl(0xb8000878, (rtd_inl(0xb8000878) & 0xFFFF0FFF) | 0x00001000);	/*      Addr 5     GPIO120 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*      Addr 5     GPIO120 */

	rtd_outl(0xb8000878, (rtd_inl(0xb8000878) & 0x0FFFFFFF) | 0x10000000);	/*      Addr 6     GPIO118 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*      Addr 6     GPIO118 */

	rtd_outl(0xb8000874, (rtd_inl(0xb8000874) & 0xFFFF0FFF) | 0x00001000);	/*      Addr 7     GPIO116 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000874, rtd_inl(0xb8000874));	/*      Addr 7     GPIO116 */

	rtd_outl(0xb8000864, (rtd_inl(0xb8000864) & 0xFFFFFF0F) | 0x00000010);	/*      Addr 8     GPIO101 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000864, rtd_inl(0xb8000864));	/*      Addr 8     GPIO101 */

	rtd_outl(0xb8000864, (rtd_inl(0xb8000864) & 0xFF0FFFFF) | 0x00100000);	/*      Addr 9     GPIO99 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000864, rtd_inl(0xb8000864));	/*      Addr 9     GPIO99 */

	rtd_outl(0xb8000860, (rtd_inl(0xb8000860) & 0x0FFFFFFF) | 0x10000000);	/*      Addr 10    GPIO94 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*      Addr 10    GPIO94 */

	rtd_outl(0xb8000860, (rtd_inl(0xb8000860) & 0xFFFFFF0F) | 0x00000010);	/*      Addr 11    GPIO97 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*      Addr 11    GPIO97 */

	rtd_outl(0xb8000874, (rtd_inl(0xb8000874) & 0x0FFFFFFF) | 0x10000000);	/*    Addr 12    GPIO114 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000874, rtd_inl(0xb8000874));	/*    Addr 12    GPIO114 */

	rtd_outl(0xb8000868, (rtd_inl(0xb8000868) & 0xFF0FFFFF) | 0x00100000);	/*      Addr 13    GPIO103 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000868, rtd_inl(0xb8000868));	/*      Addr 13    GPIO103 */

	rtd_outl(0xb8000868, (rtd_inl(0xb8000868) & 0xFFFFFF0F) | 0x00000010);	/*      Addr 14    GPIO105 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000868, rtd_inl(0xb8000868));	/*      Addr 14    GPIO105 */

	rtd_outl(0xb8000884, (rtd_inl(0xb8000884) & 0xFFFF0FFF) | 0x00001000);	/*      Data 0     GPIO132 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000884, rtd_inl(0xb8000884));	/*      Data 0     GPIO132 */
	
	rtd_outl(0xb8000888, (rtd_inl(0xb8000888) & 0x0FFFFFFF) | 0x10000000);	/*      Data 1     GPIO134 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000888, rtd_inl(0xb8000888));	/*      Data 1     GPIO134 */

	rtd_outl(0xb8000888, (rtd_inl(0xb8000888) & 0xFFFF0FFF) | 0x00001000);	/*      Data 2     GPIO136*/
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000888, rtd_inl(0xb8000888));	/*      Data 2     GPIO136 */

	rtd_outl(0xb8000854, (rtd_inl(0xb8000854) & 0xFF0FFFFF) | 0x00100000);	/*      Data 3     GPIO83 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000854, rtd_inl(0xb8000854));	/*      Data 3     GPIO83 */
	
	rtd_outl(0xb8000854, (rtd_inl(0xb8000854) & 0xFFFFFF0F) | 0x00000010);	/*      Data 4     GPIO85 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000854, rtd_inl(0xb8000854));	/*      Data 4     GPIO85 */

	rtd_outl(0xb8000858, (rtd_inl(0xb8000858) & 0xFF0FFFFF) | 0x00100000);	/*      Data 5     GPIO87 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000858, rtd_inl(0xb8000858));	/*      Data 5     GPIO87 */

	rtd_outl(0xb8000858, (rtd_inl(0xb8000858) & 0xFFFFFF0F) | 0x00000010);	/*    Data 6     GPIO89 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000858, rtd_inl(0xb8000858));	/*    Data 6     GPIO89 */
	
	rtd_outl(0xb800085c, (rtd_inl(0xb800085c) & 0xFF0FFFFF) | 0x00100000);	/*      Data 7     GPIO91 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb800085c, rtd_inl(0xb800085c));	/*      Data 7     GPIO91 */
	
	rtd_outl(0xb800086c, (rtd_inl(0xb800086c) & 0xFF0FFFFF) | 0x00100000);	/*    WE_N       GPIO107 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb800086c, rtd_inl(0xb800086c));	/*    WE_N       GPIO107 */
	
	rtd_outl(0xb8000850, (rtd_inl(0xb8000850) & 0xFFFFFF0F) | 0x00000010);	/*      IREQ0_N    GPIO81 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000850, rtd_inl(0xb8000850));	/*      IREQ0_N    GPIO81 */

	rtd_outl(0xb8000860, (rtd_inl(0xb8000860) & 0xFF0FFFFF) | 0x00100000);	/*      OE_N       GPIO95 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*      OE_N       GPIO95 */
	
	rtd_outl(0xb800085c, (rtd_inl(0xb800085c) & 0xFFFFFF0F) | 0x00000010);	/*      CE1_N      GPIO93 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb800085c, rtd_inl(0xb800085c));	/*      CE1_N      GPIO93 */

	rtd_outl(0xb8000878, (rtd_inl(0xb8000878) & 0xFF0FFFFF) | 0x00100000);	/*      RST0       GPIO119 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*      RST0       GPIO119 */
	
	rtd_outl(0xb8000878, (rtd_inl(0xb8000878) & 0xFFFFFF0F) | 0x00000010);	/*    WAIT0_N    GPIO121 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*    WAIT0_N    GPIO121 */
	
	rtd_outl(0xb800087c, (rtd_inl(0xb800087c) & 0xFF0FFFFF) | 0x00100000);	/*      INPACK0_N  GPIO123 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb800087c, rtd_inl(0xb800087c));	/*      INPACK0_N  GPIO123 */

	rtd_outl(0xb800087c, (rtd_inl(0xb800087c) & 0xFFFFFF0F) | 0x00000010);	/*       REG_N      GPIO61 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb800087c, rtd_inl(0xb800087c));	/*    REG_N      GPIO61 */

	rtd_outl(0xb8000860, (rtd_inl(0xb8000860) & 0xFFFF0FFF) | 0x00001000);	/*    IORD_N     GPIO96 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*    IORD_N     GPIO96 */

	rtd_outl(0xb8000864, (rtd_inl(0xb8000864) & 0x0FFFFFFF) | 0x10000000);	/*      IOWR_N     GPIO98 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000864, rtd_inl(0xb8000864));	/*      IOWR_N     GPIO98 */

	rtd_outl(0xb8000854, (rtd_inl(0xb8000854) & 0x0FFFFFFF) | 0x10000000);	/*      CD0_N      GPIO82 */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb8000854, rtd_inl(0xb8000854));	/*      CD0_N      GPIO82 */

	rtd_outl(0xb801B724, (rtd_inl(0xb801B724) & 0xFFFFFFC0) | 0x00000034);	/* wait 0/1 , inpack */
	PCMCIA_DGB("reg 0x%X = 0x%08X\n", 0xb801B724, rtd_inl(0xb801B724));
#if 0
	rtd_outl(0xb801B70C, (rtd_inl(0xb801B70C) & 0x0) | 0x2cd20203);	/*  attribute memory timing control */
	rtd_outl(0xb801B710, (rtd_inl(0xb801B710) & 0x0) | 0x00060252);
	rtd_outl(0xb801B714, (rtd_inl(0xb801B714) & 0x0) | 0x11201108);
#endif
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B70C, rtd_inl(0xb801B70C));	/*  attribute memory timing control */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B728, rtd_inl(0xb801B728));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B710, rtd_inl(0xb801B710));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B714, rtd_inl(0xb801B714));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B72C, rtd_inl(0xb801B72C));


	PCMCIA_WARNING("\nafter setting start\n");
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000118, rtd_inl(0xb8000118));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000108, rtd_inl(0xb8000108));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000118, rtd_inl(0xb8000118));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000108, rtd_inl(0xb8000108));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000118, rtd_inl(0xb8000118));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000108, rtd_inl(0xb8000108));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8060208, rtd_inl(0xb8060208));/*?????*/
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8061100, rtd_inl(0xb8061100));/*?????*/
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8061108, rtd_inl(0xb8061108));/*?????*/

	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000884, rtd_inl(0xb8000884));	/*    Addr 0     GPIO130 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000880, rtd_inl(0xb8000880));	/*      Addr 1     GPIO128 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000880, rtd_inl(0xb8000880));	/*      Addr 2     GPIO126 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800087C, rtd_inl(0xb800087C));	/*      Addr 3     GPIO124 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800087C, rtd_inl(0xb800087C));	/*      Addr 4     GPIO122 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*      Addr 5     GPIO120 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*      Addr 6     GPIO118 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000874, rtd_inl(0xb8000874));	/*      Addr 7     GPIO116 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000864, rtd_inl(0xb8000864));	/*      Addr 8     GPIO101 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000864, rtd_inl(0xb8000864));	/*      Addr 9     GPIO99 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*      Addr 10    GPIO94 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*      Addr 11    GPIO97 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000874, rtd_inl(0xb8000874));	/*    Addr 12    GPIO114 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000868, rtd_inl(0xb8000868));	/*      Addr 13    GPIO103 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000868, rtd_inl(0xb8000868));	/*      Addr 14    GPIO105 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000884, rtd_inl(0xb8000884));	/*      Data 0     GPIO132 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000888, rtd_inl(0xb8000888));	/*      Data 1     GPIO134 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000888, rtd_inl(0xb8000888));	/*      Data 2     GPIO136 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000854, rtd_inl(0xb8000854));	/*      Data 3     GPIO83 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000854, rtd_inl(0xb8000854));	/*      Data 4     GPIO85 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000858, rtd_inl(0xb8000858));	/*      Data 5     GPIO87 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000858, rtd_inl(0xb8000858));	/*    Data 6     GPIO89 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800085c, rtd_inl(0xb800085c));	/*      Data 7     GPIO91 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800086c, rtd_inl(0xb800086c));	/*    WE_N       GPIO107 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000850, rtd_inl(0xb8000850));	/*      IREQ0_N    GPIO81 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*      OE_N       GPIO95 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800085c, rtd_inl(0xb800085c));	/*      CE1_N      GPIO93 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*      RST0       GPIO119 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000878, rtd_inl(0xb8000878));	/*    WAIT0_N    GPIO121 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800087c, rtd_inl(0xb800087c));	/*      INPACK0_N  GPIO123 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb800087c, rtd_inl(0xb800087c));	/*    REG_N      GPIO61 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000860, rtd_inl(0xb8000860));	/*    IORD_N     GPIO96 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000864, rtd_inl(0xb8000864));	/*      IOWR_N     GPIO98 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb8000854, rtd_inl(0xb8000854));	/*      CD0_N      GPIO82 */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B724, rtd_inl(0xb801B724));

	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B70C, rtd_inl(0xb801B70C));	/*  attribute memory timing control */
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B728, rtd_inl(0xb801B728));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B710, rtd_inl(0xb801B710));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B714, rtd_inl(0xb801B714));
	PCMCIA_WARNING("reg 0x%X = 0x%08X\n", 0xb801B72C, rtd_inl(0xb801B72C));

	PCMCIA_WARNING("\nafter setting end\n");

#endif /*
 */
}
#endif
static inline int venus_pcmcia_card_enable(unsigned char On)
{                
	if (On)
	{
        	if (!(GET_PCMCIA_STS() & PC_PRES1))
    		{
            			printf("enable card failed, no card detected\n");        
            			return PCMCIA_FAIL;
        	}	
		setGPIO(PCMCIA_GPIO_INDEX, 1);
		udelay(1000 * 2);
		
		SET_PCMCIA_CTRL((GET_PCMCIA_CTRL() | PC_PCR1_OE) & (~(PC_PII1 | PC_CARD_DETECT1_INT))); //enable card
	}
	else
	{
		setGPIO(PCMCIA_GPIO_INDEX, 0);
		SET_PCMCIA_CTRL(GET_PCMCIA_CTRL() & ~PC_PCR1_OE);//disable pc card
	}
    	return PCMCIA_OK;
}

//functions for call
inline int venus_pcmcia_enable(unsigned char on)
{
	int ret = venus_pcmcia_card_enable(on);
	if(ret == PCMCIA_OK)
		msleep(100);// will wait 100ms to let pcmcia stable;
	return ret;
}
int venus_pcmcia_init()
{
       if(!rtk_pcmcia_init)
       {
               rtk_pcmcia_init = 1;
               venus_pcmcia_clock_enable(1);       //  enable PCMCIA clock
               venus_pcmcia_phy_reset();            //reset pcmcia physically
               SET_PCMCIA_AMTC_0(pcmcia_amtc0);
               SET_PCMCIA_AMTC_1(pcmcia_amtc1);
               SET_PCMCIA_IOMTC(pcmcia_iomtc);
               SET_PCMCIA_MATC_0(pcmcia_matc0);
               SET_PCMCIA_MATC_1(pcmcia_matc1);
		  rtd_outl(PCMCIA_REG_PIN_SEL, rtd_inl(PCMCIA_REG_PIN_SEL) | PC_INPACK0_DIS(1) | PC_WAIT0_DIS(0));
       }
       //pcmcia_pinmux_set();               //no need, because bootcode has set pin mux before
       return 0;
}

int venus_pcmcia_finit()
{
	return 0;
}

inline int venus_pcmcia_read_attr_mem(unsigned short addr, char *pVal, int len)
{            
	if(pVal && len >= 0)
	{
#ifdef CONFIG_PCMCIA_FIFO_MODE
		return venus_pcmcia_fifo_read(addr << 1, pVal, len, PCMCIA_FLAGS_ATTR);
#else	
		return venus_pcmcia_read(addr << 1, pVal, len, PCMCIA_FLAGS_ATTR);
#endif
	}
	return PCMCIA_FAIL;
}   

inline int venus_pcmcia_write_attr_mem(unsigned short addr, char *pVal, int len)
{  
	if(pVal && len >= 0)
	{
#ifdef CONFIG_PCMCIA_FIFO_MODE
		return venus_pcmcia_fifo_write(addr << 1, pVal, len, PCMCIA_FLAGS_ATTR);
#else	
		return venus_pcmcia_write(addr << 1, pVal, len, PCMCIA_FLAGS_ATTR);
#endif
	}
	return PCMCIA_FAIL;	
}   

inline int venus_pcmcia_read_io_mem(unsigned short addr, char *pVal, int len)
{            
	if(pVal && len >= 0)
	{
#ifdef CONFIG_PCMCIA_FIFO_MODE
		return venus_pcmcia_fifo_read(addr, pVal, len, PCMCIA_FLAGS_IO | PCMCIA_FLAGS_FIFO);
#else	
		return venus_pcmcia_read(addr, pVal, len, PCMCIA_FLAGS_IO | PCMCIA_FLAGS_FIFO);
#endif
	}
	return PCMCIA_FAIL;
}   

inline int venus_pcmcia_write_io_mem(unsigned short addr, char *pVal, int len)
{  
	if(pVal && len >= 0)
	{
#ifdef CONFIG_PCMCIA_FIFO_MODE
		return venus_pcmcia_fifo_write(addr, pVal, len, PCMCIA_FLAGS_IO | PCMCIA_FLAGS_FIFO);
#else
		return venus_pcmcia_write(addr, pVal, len, PCMCIA_FLAGS_IO | PCMCIA_FLAGS_FIFO);
#endif
	}
	return PCMCIA_FAIL;	
}   







//Test case

void pcmcia_read_cis_test()
{
	int i = 0;
	char buf[256]={0x00};
	memset(buf,0, 256);
	unsigned long flags = PCMCIA_FLAGS_ATTR;
	venus_pcmcia_init();
	if(venus_pcmcia_enable(1) != PCMCIA_OK)
		return;
	venus_pcmcia_card_reset(3000);
	msleep(300);
	while(1)
	{
		printf("************read cis start*********************\n");
		for(i = 0; i < 128; i++)
		{
			venus_pcmcia_read_attr_mem(i, &buf[i], 1);	
		}
		printf("************CIS SPACE*********************");
		for(i = 0; i < 128; i++)
		{
			if(i % 16 == 0)
				printf("\n0x%0.8x: ",i);
			printf("%0.2x ",buf[i]);
		}	
		printf("\n************read cis start*********************\n");
		msleep(1);
	}
	venus_pcmcia_finit();
}


/* 
 *
 *Test write and read io memory
 */
void pcmcia_read_write_io_test()
{
	int i = 0;
	char buf[256]={0x00};
	memset(buf,0, 256);
	unsigned long flags = PCMCIA_FLAGS_ATTR;
	venus_pcmcia_init();
	if(venus_pcmcia_enable(1) != PCMCIA_OK)
		return;	
	venus_pcmcia_card_reset(3000);
	msleep(300);
	venus_pcmcia_read_attr_mem(0x110, buf, 1);
	printf("Before config:Attribute memory addr 0x220 =%d\n",buf[0]);
	buf[0]=0x0f;
	venus_pcmcia_write_attr_mem(0x110, buf, 1);
	buf[0] = 0x0;
	venus_pcmcia_read_attr_mem(0x110, buf, 1);
	printf("After config:Attribute memory addr 0x220 =%d\n",buf[0]);
	buf[0] = 0x08;
	venus_pcmcia_write_io_mem(0x01, buf, 1);	
	while(i++ < 128)
	{
		venus_pcmcia_read_io_mem(0x01, buf, 1);	
		printf("11:%0.2x \n",buf[0]);
		if(buf[0] & 0x40)
			break;
		msleep(10);
	}
	buf[0] = 0x00;
	venus_pcmcia_write_io_mem(0x01, buf, 1);	
		
	buf[0] = 0x04;
	venus_pcmcia_write_io_mem(0x01, buf, 1);	
	i = 0;
	while(i++ < 128)
	{
		venus_pcmcia_read_io_mem(0x01, buf, 1);	
		printf("22:%0.2x \n",buf[0]);
		if(buf[0] & 0x80)
			break;
		msleep(10);
	}
	venus_pcmcia_read_io_mem(0x02, buf, 1);	
	printf("33:%0.2x \n",buf[0]);
	venus_pcmcia_read_io_mem(0x03, buf, 1);	
	printf("44:%0.2x \n",buf[0]);
	
	venus_pcmcia_finit();
}

/* 
 *
 *Test write attribute memory
 */
void pcmcia_config_io_mode()
{
	int i = 0;
	char buf[2]={0x00};
	memset(buf,0, 2);
	unsigned long flags = PCMCIA_FLAGS_ATTR;
	venus_pcmcia_init();
	if(venus_pcmcia_enable(1) != PCMCIA_OK)
		return;	
	//while(1)
	{
		venus_pcmcia_read_attr_mem(0x110, buf, 1);
		printf("Before config:Attribute memory addr 0x220 =%d\n",buf[0]);
		buf[1]=0x0f;
		venus_pcmcia_write_attr_mem(0x110, buf+1, 1);
		venus_pcmcia_read_attr_mem(0x110, buf, 1);
		printf("After config:Attribute memory addr 0x220 =%d\n",buf[0]);
		msleep(1);
	}
	venus_pcmcia_finit();
}

/*auto_pinmux_en:1: auto pin mux, 0: no auto pin mux*/
/*hw_mode_en:1: HW mode, 0: SW mode*/
void pcmcia_fpga_test(int auto_pinmux_en, int hw_mode_en)
{
	int i = 0;
	char buf[256];
	static int is_card_enable_and_rst = 0;
	memset(buf,0, 256);
	g_auto_pinmux_enable = auto_pinmux_en; 
	g_fifo_func_use_hw_mode = hw_mode_en; 
	venus_pcmcia_init();
	if(!is_card_enable_and_rst) {
		if(venus_pcmcia_enable(1) != PCMCIA_OK)
			return;
		venus_pcmcia_card_reset(3000);
		is_card_enable_and_rst = 1;
		msleep(300);
	}
	
	printf("************read cis start*********************\n");
	venus_pcmcia_read_attr_mem(0x0, buf, 256);	
	printf("************CIS SPACE*********************");
	for(i = 0; i < 256; i++) {
		if(i % 16 == 0)
			printf("\n0x%0.8x: ",i);
		printf("%0.2x ",buf[i]);
	}	
	printf("\n************read cis start*********************\n");
	
	venus_pcmcia_finit();
}

