/*
 * (C) Copyright 2013
 * EJ Hsu <ejhsu@realtek.com>
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

/*
 * Branch according to exception level
 */

#include "gic.h"

.macro	switch_el, xreg, el3_label, el2_label, el1_label
	mrs	\xreg, CurrentEL
	cmp	\xreg, 0xc
	b.eq	\el3_label
	cmp	\xreg, 0x8
	b.eq	\el2_label
	cmp	\xreg, 0x4
	b.eq	\el1_label
.endm

/*
 * Branch if CPU ID of current processor is 0.
 */
.macro	branch_if_cpuid_0, xreg1, xreg2, master_label
	/* NOTE: MPIDR handling will be erroneous on multi-cluster machines */
	mrs	\xreg1, mpidr_el1
	lsr	\xreg1, \xreg1, #0x8
	mov	\xreg2, #0xf
	and	\xreg1, \xreg1, \xreg2
	cbz	\xreg1, \master_label
.endm

.macro	branch_if_master, xreg1, xreg2, master_label
	/* NOTE: MPIDR handling will be erroneous on multi-cluster machines */
	mrs	\xreg1, mpidr_el1
	lsr	\xreg1, \xreg1, #0x8
	mov	\xreg2, #0xf
	and	\xreg1, \xreg1, \xreg2
	cbz	\xreg1, \master_label
.endm

.macro armv8_switch_to_el1_m, xreg1, xreg2, xreg3
	mov	\xreg1, #3 << 20
	msr	cpacr_el1, \xreg1	/* Enable FP/SIMD at EL1 */

	/* SCTLR_EL1 initialization
	 *
	 * setting RES1 bits (29,28,23,22,20,11) to 1
	 * and RES0 bits (31,30,27,21,17,13,10,6) +
	 * UCI,EE,EOE,WXN,nTWE,nTWI,UCT,DZE,I,UMA,SED,ITD,
	 * CP15BEN,SA0,SA,C,A,M to 0
	 */
	mov	\xreg1, #0x0838
	movk	\xreg1, #0x00c5, lsl #16
	msr	sctlr_el1, \xreg1

	/* Return to the EL1_SP1 mode from EL3 */
	mov	\xreg1, sp
	msr	sp_el1, \xreg1		/* Migrate SP */
	mrs	\xreg1, vbar_el3
	msr	vbar_el1, \xreg1	/* Migrate VBAR */
	msr	spsr_el3, \xreg2
	msr	elr_el3, \xreg3
	eret
.endm

.macro  branch_if_a57_core, xreg, a57_label
        mrs     \xreg, midr_el1
        lsr     \xreg, \xreg, #4
        and     \xreg, \xreg, #0x00000FFF
        cmp     \xreg, #0xD07           /* Cortex-A57 MPCore processor. */
        b.eq    \a57_label
.endm

/*
 * Branch if current processor is a Cortex-A53 core.
 */
.macro  branch_if_a53_core, xreg, a53_label
        mrs     \xreg, midr_el1
        lsr     \xreg, \xreg, #4
        and     \xreg, \xreg, #0x00000FFF
        cmp     \xreg, #0xD03           /* Cortex-A53 MPCore processor. */
        b.eq    \a53_label
.endm

#if defined(CONFIG_GICV3)
.macro gic_wait_for_interrupt_m xreg1
0 :	wfi
	mrs     \xreg1, ICC_IAR1_EL1
	msr     ICC_EOIR1_EL1, \xreg1
	cbnz    \xreg1, 0b
.endm
#elif defined(CONFIG_GICV2)
.macro gic_wait_for_interrupt_m xreg1, wreg2
0 :	wfi
	ldr     \wreg2, [\xreg1, GICC_AIAR]
	str     \wreg2, [\xreg1, GICC_AEOIR]
	and	\wreg2, \wreg2, #0x3ff
	cbnz    \wreg2, 0b
.endm
#endif
