#include <drivers/gpio/roku_drvGPIO.h>
#include <common.h>
#include <asm/arch/io.h>

static RT_GPIO_REGINFO MiscGpioRegInfo[] =
{
    {MIS_GP0DIR, MIS_GP0DATI, MIS_GP0DATO},
    {MIS_GP1DIR, MIS_GP1DATI, MIS_GP1DATO},
    {MIS_GP2DIR, MIS_GP2DATI, MIS_GP2DATO},
    {MIS_GP3DIR, MIS_GP3DATI, MIS_GP3DATO},
};
static RT_GPIO_REGINFO IsoGpioRegInfo[] =
{
    {ISO_GPDIR, ISO_GPDATI, ISO_GPDATO},
};

#define MISC_GPIO_REG_GROUP_NUM  sizeof(MiscGpioRegInfo)/sizeof(RT_GPIO_REGINFO)
#define ISO_GPIO_REG_GROUP_NUM   sizeof(IsoGpioRegInfo)/sizeof(RT_GPIO_REGINFO)

static void iowrite_reg_bit(volatile unsigned int regAddr, int bit, int value)
{
    int regValue;

    regValue = rtd_inl(regAddr);
    if(value)
        regValue = regValue | (0x01 << bit);
    else
        regValue = regValue & (~(0x01 << bit));
    rtd_outl(regAddr, regValue);
}

static int ioread_reg_bit(volatile unsigned int regAddr, int bit)
{
    int regValue;
    regValue = rtd_inl(regAddr);
    if(regValue & (0x01 << bit)){
        return 1;
    }
    else{
        return 0;
    }
}

/*
 * Initializes gpio driver.
 */
void PAL_Gpio_Init(void)
{
    return;
}

/*
 * Sets the gpio as HIGH.
 */
void PAL_Gpio_Set_High(int gpio)
{
    rt_gpio_output_roku(getGPIOnumber(gpio),1, getGPIOgroup(gpio),0);
    return;
}

/*
 * Sets the gpio as LOW.
 */
void PAL_Gpio_Set_Low(int gpio)
{
    rt_gpio_output_roku(getGPIOnumber(gpio),0, getGPIOgroup(gpio),0);
    return;
}

/*
 * Configures gpio as input
 */
void PAL_Gpio_Set_Input(int gpio)
{
    rt_gpio_input_roku(getGPIOnumber(gpio), getGPIOgroup(gpio),1);
    return;
}

/*
 * Configures gpio as output
 */
void PAL_Gpio_Set_Output(int gpio)
{
    rt_gpio_output_roku(getGPIOnumber(gpio),0, getGPIOgroup(gpio),1);
    return;
}

/*
 * Returns, the current configuration of a gpio (input or output)
 *     1: Gpio is an input.
 *     0: Gpio is an output.
 */
int PAL_Gpio_Get_InOut(int gpio)
{
    return rt_gpio_getlevel_roku(getGPIOnumber(gpio), getGPIOgroup(gpio));
}

/*
 * Returns current level of a given gpio.
 *  1: Gipo is high.
 *  0: Gpio is low.
 */
int PAL_Gpio_Get_Level(int gpio)
{
    return rt_gpio_input_roku(getGPIOnumber(gpio), getGPIOgroup(gpio),0);
}

/**  Align kernel GPIO base address
#define MIS_GPIO_OFST 0 // MISC Group : Bidirectional Pads (Max: 200)
#define MIS_GPI_OFST 200 // MISC Group : Input only Pads (Max: 100)
#define MIS_GPO_OFST 300 // MISC Group : Out only Pads (Max: 100)
#define ISO_GPIO_OFST 400 // ISO Group : Bidirectional Pads (Max: 100)
#define ISO_GPI_OFST 500 // ISO Group : Input only Pads (Max: 100)
#define ISO_GPO_OFST 600 // ISO Group : Output only Pads (Max: 100)
 */

int getGPIOgroup(int gpio_num) {

    int group = gpio_num/100;

    switch(group) {

        case 0:  return MISC_GPIO; break;
        case 2:  return MISC_GPI;  break;
        case 3:  return MISC_GPO;  break;
        case 4:  return ISO_GPIO; break;
        case 5:  return ISO_GPI;  break;
        case 6:  return ISO_GPO;  break;
        default:
            return -1;
    }
}

int getGPIOnumber(int gpio_num) {
    int number = gpio_num%100;
    return number;
}

/*
 * type=0 set for out value
 * type=1 set for out direction
 */
int rt_gpio_output_roku(int gpioNum, int value, RT_GPIO_TYPE rtGpioType,int type)
{
    int bitOffset;
    int regInfoIndex;

    regInfoIndex = gpioNum/32;
    bitOffset    = gpioNum%32;

    gpioDebug("set.pin:%d,I:%d,bit:%d,value:%d.Type:%d\n", gpioNum,regInfoIndex,bitOffset,value,rtGpioType);

    switch(rtGpioType)
    {
        case MISC_GPIO:
            if(regInfoIndex >= MISC_GPIO_REG_GROUP_NUM || gpioNum >= MISC_GPIO_MAX_NUM){
                gpioErr("%s misc num:%d invalid,max misc num=%d\n",__func__, gpioNum, MISC_GPIO_MAX_NUM);
                return -1;
            }

            if (type==0)
                iowrite_reg_bit(MiscGpioRegInfo[regInfoIndex].gpioDatO, bitOffset, value);
            else
                iowrite_reg_bit(MiscGpioRegInfo[regInfoIndex].gpioDir, bitOffset, GPIO_DIR_OUT);

            break;
        case ISO_GPIO:
            if(regInfoIndex >= ISO_GPIO_REG_GROUP_NUM || gpioNum >= ISO_GPIO_MAX_NUM){
                gpioErr("%s iso num:%d invalid,max iso num=%d\n",__func__, gpioNum, ISO_GPIO_MAX_NUM);
                return -1;
            }

            if (type==0)
                iowrite_reg_bit(IsoGpioRegInfo[regInfoIndex].gpioDatO, bitOffset, value);
            else
                iowrite_reg_bit(IsoGpioRegInfo[regInfoIndex].gpioDir, bitOffset, GPIO_DIR_OUT);

            break;
        default:
            break;
    }
    return 0;
}

/*
 * type=0 get for in value
 * type=1 set for in direction
 */

int rt_gpio_input_roku(int gpioNum, RT_GPIO_TYPE rtGpioType,int type)
{
    int bitOffset;
    int regInfoIndex;
    int retValue = -1;

    regInfoIndex = gpioNum/32;
    bitOffset    = gpioNum%32;

    gpioDebug("get.pin:%d,I:%d,bit:%d.Type:%d\n", gpioNum,regInfoIndex,bitOffset,rtGpioType);

    switch(rtGpioType)
    {
        case MISC_GPIO:
            if(regInfoIndex >= MISC_GPIO_REG_GROUP_NUM || gpioNum >= MISC_GPIO_MAX_NUM){
                gpioErr("%s misc num:%d invalid,max misc num=%d\n",__func__, gpioNum, MISC_GPIO_MAX_NUM);
                return -1;
            }
            if (type==0)
                retValue = ioread_reg_bit(MiscGpioRegInfo[regInfoIndex].gpioDatI, bitOffset);
            else
                iowrite_reg_bit(MiscGpioRegInfo[regInfoIndex].gpioDir, bitOffset, GPIO_DIR_IN);
            break;
        case ISO_GPIO:
            if(regInfoIndex >= ISO_GPIO_REG_GROUP_NUM || gpioNum >= ISO_GPIO_MAX_NUM){
                gpioErr("%s iso num:%d invalid,max iso num=%d\n",__func__, gpioNum, ISO_GPIO_MAX_NUM);
                return -1;
            }
            if (type==0)
                retValue = ioread_reg_bit(IsoGpioRegInfo[regInfoIndex].gpioDatI, bitOffset);
            else
                iowrite_reg_bit(IsoGpioRegInfo[regInfoIndex].gpioDir, bitOffset, GPIO_DIR_IN);
            break;
        default:
            break;
    }
    return retValue;
}

int rt_gpio_getlevel_roku(int gpioNum, RT_GPIO_TYPE rtGpioType)
{
    int bitOffset;
    int regInfoIndex;
    int value = -1;

    regInfoIndex = gpioNum/32;
    bitOffset    = gpioNum%32;

    gpioDebug("get.pin:%d,I:%d,bit:%d.Type:%d\n", gpioNum,regInfoIndex,bitOffset,rtGpioType);

    switch(rtGpioType)
    {
        case MISC_GPIO:
            if(regInfoIndex >= MISC_GPIO_REG_GROUP_NUM || gpioNum >= MISC_GPIO_MAX_NUM){
                gpioErr("%s misc num:%d invalid,max misc num=%d\n",__func__, gpioNum, MISC_GPIO_MAX_NUM);
                return -1;
            }
                value = ioread_reg_bit(MiscGpioRegInfo[regInfoIndex].gpioDir, bitOffset);
                break;
        case ISO_GPIO:
            if(regInfoIndex >= ISO_GPIO_REG_GROUP_NUM || gpioNum >= ISO_GPIO_MAX_NUM){
                gpioErr("%s iso num:%d invalid,max iso num=%d\n",__func__, gpioNum, ISO_GPIO_MAX_NUM);
                return -1;
            }
                value = ioread_reg_bit(IsoGpioRegInfo[regInfoIndex].gpioDir, bitOffset);
                break;
        default:
            break;
    }

    if (value) //output
        return 0;
    else       //input
        return 1;
}

