#include <drivers/mcp/rtk_mcp_rsa_hw_fun.h>
#include <common.h>
#include <command.h>
#include <asm/arch/system.h>

#define REG_RSA_X       0xB806E000
#define REG_RSA_Y       0xB806E100
#define REG_RSA_M       0xB806E200
#define REG_RSA_A       0xB806E300
#define REG_RSA_CTRL    0xB806E400
#define REG_RSA_CTRL1   0xB806E410
#define REG_ECC_P       0xB806E300
#define REG_ECC_P_W12   0xB806E330
#define REG_ECC_V       0xB806E368
#define REG_ECC_V_W12   0xB806E398
#define REG_ECC_U       0xB806E334
#define REG_ECC_R       0xB806E39C
#define REG_ECC_CTRL    0xB806E414


extern int BI_chkmsbbit (BI *bi);
extern int BI_is_bit_set (BI *bi, int n);

//-----------------------------------------------------------------------------
//----------------------------------------------------------------------------
//For HW RSA utility function===========================================
//-----------------------------------------------------------------------------
//----------------------------------------------------------------------------
#define TYPE_HW_MONT	0
#define TYPE_HW_IMONT	1
static void hwsu_get_data (void *addr, BI *bi, UINT8 type)
{
	int i, j, max_idx;
	unsigned int *reg = (unsigned int *) addr;

	if (type == TYPE_HW_MONT)
		max_idx = 63; //Max: 64 * 32 = 2048 bit
	else
		max_idx = 12; //Max: 13 * 32 = 416 bit

	/* Skip MSB zero */
	i = max_idx;

	while (i && (rtd_inl((unsigned int)(reg + i)) == 0)) {
        //printf("i=%d\n",i);
		i--;
	}

	/* Fill BI */
	bi->m_nLength = i + 1;
	j = 1;
	for (; i >= 0; i--) {
		bi->m_ulValue[bi->m_nLength - j] = rtd_inl((unsigned int)(reg + i));
		j++;
	}

	return;
}
//-----------------------------------------------------------------------------
static void hwsu_set_data (void *addr, BI *bi, UINT8 type)
{
	int i, max_idx;
	unsigned int *reg = (unsigned int *) addr;

	if (type == TYPE_HW_MONT)
		max_idx = 63; //Max: 64 * 32 = 2048 bit
	else
		max_idx = 12; //Max: 13 * 32 = 416 bit

	for (i = 0; i < bi->m_nLength; i++)
		rtd_outl((unsigned int)(reg + i), bi->m_ulValue[i]);

	/* Handle rest bytes (clear to 0) */
	for (; i <= max_idx; i++)
		rtd_outl((unsigned int)(reg + i), 0x0);

	return;
}
//-----------------------------------------------------------------------------
static void hwsu_set_bitnum (unsigned int bsize)
{
	rtd_outl((REG_RSA_CTRL1), (bsize >> 2) - 1);
}
//-----------------------------------------------------------------------------
static void hwsu_chk_mont_engine_ready (void)
{
	while (rtd_inl(REG_RSA_CTRL) == 1);
}
//-----------------------------------------------------------------------------
static void hwsu_mont_engine_go (void)
{
	rtd_outl((REG_RSA_CTRL), 1);
}
//-----------------------------------------------------------------------------
static void hwsu_mont (BI *resl, BI *a, BI *b, BI *m, unsigned int bsize)
{

	hwsu_chk_mont_engine_ready();

	hwsu_set_data((void *)REG_RSA_X, a, TYPE_HW_MONT);
	hwsu_set_data((void *)REG_RSA_Y, b, TYPE_HW_MONT);
	hwsu_set_data((void *)REG_RSA_M, m, TYPE_HW_MONT);

	hwsu_set_bitnum(bsize);

	hwsu_mont_engine_go();

	hwsu_chk_mont_engine_ready();


	hwsu_get_data((void *)REG_RSA_A, resl, TYPE_HW_MONT);

	return;
}

//-----------------------------------------------------------------------------
#if 0
static void hwsu_mul_mod (BI *resl, BI *r2, BI *x, BI *y, BI *m, UINT32 bsize)
{
	hwsu_mont (resl, x, y, m, bsize);
	hwsu_mont (resl, resl, r2, m, bsize);

	return;
}
#endif
//resl = base ^ exp % mod
void hwsu_mont_exp(BI *resl, BI *r2, BI *base, BI *exp, BI *mod, UINT32 bsize)
{
    int i = 0;
    BI * temp_1 = move_p(1);
    //resl = move_p(1);
    hwsu_mont(resl,resl,r2,mod,bsize);
    hwsu_mont(base,base,r2,mod,bsize);
    for ( i = BI_chkmsbbit (exp); i >= 0; i-- )
    {
        hwsu_mont(resl,resl,resl,mod,bsize);
        if ( BI_is_bit_set (exp, i) == 1 )
        {
            hwsu_mont(resl,resl,base,mod,bsize);
        }
    }
    hwsu_mont(resl,resl,temp_1,mod,bsize);
}


//-----------------------------------------------------------------------------
//End of HW RSA utility funciton=========================================

