
/*
 *
 *  mcp.c
 *
 *  Implementation of hardware CP module
 *
 *
 *
 **/
#include <common.h>
#include <asm/arch/system.h>
//#include <asm/arch/rbus/otp_reg.h>
#include <asm/arch/rsa_key_2048_big.h>
#include <asm/arch/rsa_key_2048_little.h>
#include <asm/arch/error_type.h>
#include <drivers/mcp/rtk_mcp_scpu.h>
#include <malloc.h>
#include <asm/tomcrypt.h>
extern const struct ltc_hash_descriptor sha256_desc ;

extern void rtk_hexdump( const char * str, unsigned char * pcBuf, unsigned int length );
unsigned int mcp_dscpt_addr;
const unsigned int Kh_key_default[4] = { SECURE_KH_KEY0, SECURE_KH_KEY1, SECURE_KH_KEY2, SECURE_KH_KEY3 };

unsigned int alloc_mcp_descriptor_buffer(MEM_ALLOC_TYPE cache_type)
{
    unsigned int alloc_mem;
    alloc_mem = (unsigned int)MCP_MALLOC(CP_DSCPT_POOL_SIZE, cache_type);
    mcp_dscpt_addr = alloc_mem;
    return  alloc_mem;
}

void free_mcp_descriptor_buffer(unsigned int alloc_mem, MEM_ALLOC_TYPE cache_type)
{
    MCP_FREE((void*)alloc_mem, CP_DSCPT_POOL_SIZE, cache_type);
}


t_mcp_descriptor *alloc_mcp_descriptor(void)
{
    
    //static unsigned int mcp_dscpt_addr;
    t_mcp_descriptor *current;

#ifdef MCP_DEBUG
    printf("===> mcp_dscpt_addr: 0x%x\n", mcp_dscpt_addr);
#endif
#ifdef IS_FPGA
    // check if pool is full (descriptor address overflow)
    if ((mcp_dscpt_addr + sizeof(t_mcp_descriptor) > CP_DSCPT_POOL_MAX_ADDR)  || (mcp_dscpt_addr == 0)) {
        // wrap-around (allocate from base)
        mcp_dscpt_addr = CP_DSCPT_POOL_BASE_ADDR;
        // reset pool
        memset((char *)CP_DSCPT_POOL_BASE_ADDR, 0, CP_DSCPT_POOL_SIZE);
    }
#endif
    if((mcp_dscpt_addr % CP_DESCRIPTOR_HW_ALIGN) != 0)
            mcp_dscpt_addr = mcp_dscpt_addr + CP_DESCRIPTOR_HW_ALIGN - (mcp_dscpt_addr % CP_DESCRIPTOR_HW_ALIGN);

    current = (t_mcp_descriptor *)mcp_dscpt_addr;
    memset(current, 0, sizeof(t_mcp_descriptor));
    
    mcp_dscpt_addr += CP_DESCRIPTOR_HW_ALIGN; // move forward

#ifdef MCP_DEBUG
    printf("%s\n\tmcp_dscpt_addr: 0x%x, return dscpt: 0x%x\n", __FUNCTION__, mcp_dscpt_addr, (unsigned int) current);
#endif

    return current;

}

extern void lowlevel_init(void);
int SW_SHA256_hash(unsigned char * src_addr, unsigned int length, unsigned char *dst_addr, unsigned int iv[8])
{
       unsigned int ret = 0;
 	hash_state md;
	lowlevel_init();
	ret = sha256_init(&md);
	if(ret)
		printf("SW sha256 init fail\n");
	sha256_process(&md,src_addr,length);
	ret = sha256_done(&md, dst_addr);
	printf("sw sha256 ret %x \n",ret);
	return 0;
}

int Verify_SHA256_hash( unsigned char * src_addr, unsigned int length, unsigned char * ref_sha256, unsigned int do_recovery )
{
    unsigned int ret = 0;
    unsigned char * hash1; // hash value calculated by CP engine
    unsigned char * hash2; // hash value recovery from RSA signature
    unsigned int sys_rsa_key_address;
    unsigned int signature_key_address[256];

#if 0
    sys_rsa_key_address = OTP_DATA;
#else
#if 1 // little endian
    //sys_rsa_key_address = (unsigned int)rsa_key_2048_little;
    sys_rsa_key_address = 0; //use string as input
#else // big endian
    sys_rsa_key_address = (unsigned int)rsa_key_2048_big;
#endif
#endif
    // ref_sha256 may be not 4-byte aligned, so we use local array
    memcpy((void *)signature_key_address, (void *)ref_sha256, sizeof(signature_key_address));
//    display_boottime_log("SHA1");
    hash1 = (unsigned char *)SECURE_IMAGE2HASH_BUF; // temp use this address (use malloc is better?)
    hash2 = (unsigned char *)signature_key_address;
#ifdef CONFIG_RTK_SWSHA256
    SW_SHA256_hash((unsigned char *)src_addr, length, hash1, NULL);
#else
    int padding_len = 0;
    // if length is un-aligned, do the padding
    padding_len = SHA256_padding((char *)src_addr + length, length);
//  rtk_hexdump("hash 256 padding", src_addr + length, padding_len);
    length += padding_len;



    mcp_dscpt_addr = 0;


    flush_dcache_all();

    ret = SHA256_hash((unsigned char *)src_addr, length, hash1, NULL);
#endif
//display_boottime_log("SHA2");
    if( ret ) {
        printf("[ERR] %s: caculate hash1 fail(%d)\n", __FUNCTION__, ret );
        return -1;
    }

    if( do_recovery ) {
        //rtk_hexdump("signature form ref_sha256", ref_sha256, 256 );
        //rtk_hexdump("RSA key", (unsigned char *)sys_rsa_key_address, 256 );
        ret = do_RSA((unsigned int)signature_key_address, sys_rsa_key_address, SECURE_SIGN2HASH_BUF );
        if( ret == 0 ) {
            printf("[ERR] %s: do_RSA return 0\n", __FUNCTION__ );
            return -3;
        }
        hash2 = (unsigned char *)ret;
    }

    ret = memcmp(hash1, hash2, 32);
    if( ret ) {
        printf("[ERR] %s: hash value not match\n", __FUNCTION__ );
        rtk_hexdump("hash 1", hash1, 32 );
        rtk_hexdump("hash 2", hash2, 32 );
        return -2;
    }

    return 0;
}

void enable_SCPU_RSA (void)
{
    rtd_outl(S_RSA_SEL,1);
    while(1)
    {
        if(rtd_inl(S_RSA_ST)&0x1)
        {
            break;
        }
    }
}

