#include <drivers/pwm/roku_drvPWM.h>
#include <linux/types.h>
#include <platform_lib/board/pcb_mgr.h>
#include <asm/arch/io.h>
#include <common.h>
#include <drivers/gpio/rtk_gpio.h>

static RTK_PWM_INFO PwmROKUInfoList;
static int init_roku=0;
static char *pin_name="PIN_BL_ADJ";
/*
 * All this functions are required to control the PWM used to control the brigtness of the Power LED.
 * PM = Power Mode
 */

void PAL_PM_PWM_Init(void) {

    PCB_PIN_TYPE_T pin_type;
    int i = 0;

    PCB_ENUM_T *ppcb_enum = pcb_enum;
    int iPcbPwmCounter = 0;

    if (init_roku==0) {
    /*parse the iso pwm pin*/
    do {
        if ((ppcb_enum == NULL) || (ppcb_enum[i].name == NULL) || (ppcb_enum[i].name[0] == 0))
            break;

        if (memcmp(ppcb_enum[i].name, pin_name,strlen(pin_name)) == 0) {//hard code to parse
            pin_type = GET_PIN_TYPE(ppcb_enum[i].info);
            if ((pin_type == PCB_PIN_TYPE_ISO_PWM) || (pin_type == PCB_PIN_TYPE_PWM)) {
                PwmROKUInfoList.pinInfo.val = ppcb_enum[i].info;
                PwmROKUInfoList.pinInfo.invert = GET_PWM_INVERT(ppcb_enum[i].info);
                PwmROKUInfoList.name = ppcb_enum[i].name;
                PwmROKUInfoList.mode = PWM_MODE;

                iPcbPwmCounter++;
            }
        }
        i++;
    } while (1);

    /*init pin value*/
    int invert = GET_PWM_INVERT(PwmROKUInfoList.pinInfo.val);
    int freq  = GET_PWM_FREQ(PwmROKUInfoList.pinInfo.val);
    int duty_max = GET_PWM_DUTY_MAX(PwmROKUInfoList.pinInfo.val);
    int duty  = GET_PWM_INIT_DUTY(PwmROKUInfoList.pinInfo.val);
    int index = GET_PIN_INDEX(PwmROKUInfoList.pinInfo.val);
    int clk_src = GET_PWM_CLK_SRC(PwmROKUInfoList.pinInfo.val);
    RTK_PWM_TYPE type =getPWMPinType(PwmROKUInfoList.pinInfo.val);

    pwm_init(index, type, freq, duty_max, duty, clk_src, invert);

    init_roku=1;

    } else {
        return;
    }
}

/*
 * Enable the PWM.
 */
void PAL_PM_PWM_Enable(void)
{
    PAL_PM_PWM_Init();
    RTK_PWM_TYPE type = 0;
    type =getPWMPinType(PwmROKUInfoList.pinInfo.val);

    int index = GET_PIN_INDEX(PwmROKUInfoList.pinInfo.val);
    pwm_ctrl_RBUS pwm_ctrl_reg;
    unsigned int registerAddress = 0;

    if (PWM_ISO == type && index < ISO_PWM_MAX) {
        registerAddress = pwm_iso_phy[index].p_reg_map->ISO_CTRL;
    } else if (PWM_MISC == type && index < MISC_PWM_MAX) {
        registerAddress = pwm_mis_phy[index].p_reg_map->MISC_CTRL;
    } else {
        printf("%s parameter fail register address 0x%X !\n", __func__, registerAddress);
        printf("%s %d  but  ISO_PWM_MAX =%d !\n", __func__, index, ISO_PWM_MAX);
        printf("%s %d  but MISC_PWM_MAX =%d !\n", __func__, index, MISC_PWM_MAX);
        return ;
    }
    pwm_ctrl_reg.regValue = rtd_inl(registerAddress);
    if (PWM_ISO == type)
        pwm_ctrl_reg.i_pwm0_en = 1;
    else
        pwm_ctrl_reg.m_pwm0_en = 1;

    rtd_outl(registerAddress, pwm_ctrl_reg.regValue);
}

/*
 * Sets the period of the PWM. The total period of the PWM signal.
 * Value is in nano seconds.
   NOTICE !!! the periodPWM range is 100000-1000000000 due to normal HZ is from 1 to 10K HZ
 */
void PAL_PM_PWM_Period(u_int periodPWM)
{
    PAL_PM_PWM_Init();
    int freq = 1000000000/ periodPWM;//due to nano seconds
    int current_totalcnt = 0;
    int current_duty = 0;
    int duty = 0;
    int totalcnt = 0;
    int duty_max = 255;
    unsigned short N = 0, M = 0;
    unsigned int DutySetAddress = 0;
    unsigned int TimingCtrlAddress = 0;
    pwm_duty_set_RBUS     pwm_duty_set_reg;
    pwm_timing_ctrl_RBUS  pwm_timing_ctrl_reg;
    int clk_src = GET_PWM_CLK_SRC(PwmROKUInfoList.pinInfo.val);

    rtk_hw_get_divider(freq,duty_max, &totalcnt, &M, &N, clk_src);

    RTK_PWM_TYPE type = 0;
    type =getPWMPinType(PwmROKUInfoList.pinInfo.val);

    int index = GET_PIN_INDEX(PwmROKUInfoList.pinInfo.val);

    if( PWM_ISO == type  && index < ISO_PWM_MAX ) {
            DutySetAddress = pwm_iso_phy[index].p_reg_map->ISO_DUTY_SET;
            TimingCtrlAddress = pwm_iso_phy[index].p_reg_map->ISO_TIMING_CTRL;
            printf("==> DutySetAddress:%x\n",DutySetAddress);
            printf("==> TimingCtrlAddress:%x\n",TimingCtrlAddress);
        } else if( PWM_MISC == type && index < MISC_PWM_MAX ) {
            DutySetAddress = pwm_mis_phy[index].p_reg_map->MISC_DUTY_SET;
            TimingCtrlAddress = pwm_mis_phy[index].p_reg_map->MISC_TIMING_CTRL;
        }else{
            printf("%s parameter fail register address 0x%X !\n",  __func__ , TimingCtrlAddress );
            printf("%s %d  but  ISO_PWM_MAX =%d !\n", __func__ , index , ISO_PWM_MAX );
            printf("%s %d  but MISC_PWM_MAX =%d !\n",  __func__ , index , MISC_PWM_MAX );
    }

    pwm_timing_ctrl_reg.regValue =rtd_inl(TimingCtrlAddress);
    if (PWM_ISO == type){
        pwm_timing_ctrl_reg.i_pwm0_m = M;
        pwm_timing_ctrl_reg.i_pwm0_n = N;
    }else{
        pwm_timing_ctrl_reg.m_pwm0_m = M;
        pwm_timing_ctrl_reg.m_pwm0_n = N;
    }

    rtd_outl(TimingCtrlAddress,pwm_timing_ctrl_reg.regValue);

    pwm_duty_set_reg.regValue = rtd_inl(DutySetAddress);
    if (PWM_ISO == type){
        current_duty = pwm_duty_set_reg.i_pwm0_dut;
        current_totalcnt = pwm_duty_set_reg.i_pwm0_totalcnt;
        duty = get_pwm_duty_base_change(current_duty,current_totalcnt,totalcnt);
        pwm_duty_set_reg.i_pwm0_dut = duty;
        pwm_duty_set_reg.i_pwm0_totalcnt = totalcnt;
    }
    else{
        current_duty = pwm_duty_set_reg.m_pwm0_dut;
        current_totalcnt = pwm_duty_set_reg.m_pwm0_totalcnt;
        duty = get_pwm_duty_base_change(current_duty,current_totalcnt,totalcnt);
        pwm_duty_set_reg.m_pwm0_dut = duty;
        pwm_duty_set_reg.m_pwm0_totalcnt = totalcnt;
    }
    rtd_outl(DutySetAddress,pwm_duty_set_reg.regValue);

    return;

}

/*
 * Sets the duty cycle of the PWM.
 * NOTICE!!!!! dutyPWM value is in percentage from 0-100
 */
void PAL_PM_PWM_DutyCycle(u_int dutyPWM)
{
    PAL_PM_PWM_Init();

    RTK_PWM_TYPE type = 0;
    type =getPWMPinType(PwmROKUInfoList.pinInfo.val);

    if(set_pwm_duty_roku(type,GET_PIN_INDEX(PwmROKUInfoList.pinInfo.val),dutyPWM) < 0)


    return;
}

/*
 * Sets the polarity of the PWM.
 *     1: Sets the polarity as "normal"
 *     0: Sets the polarity as "inversed"
 */
void PAL_PM_PWM_Polarity(int polarity)
{
    PAL_PM_PWM_Init();
    RTK_PWM_TYPE type = 0;
    type =getPWMPinType(PwmROKUInfoList.pinInfo.val);
    int index = GET_PIN_INDEX(PwmROKUInfoList.pinInfo.val);
    pwm_ctrl_RBUS pwm_ctrl_reg;
    unsigned int registerAddress = 0;

    if( PWM_ISO == type  && index < ISO_PWM_MAX ) {
        registerAddress = pwm_iso_phy[index].p_reg_map->ISO_CTRL;
    } else if( PWM_MISC == type && index < MISC_PWM_MAX ) {
        registerAddress = pwm_mis_phy[index].p_reg_map->MISC_CTRL;
    }else{
        printf("%s parameter fail register address 0x%X !\n",  __func__ , registerAddress );
        printf("%s %d  but  ISO_PWM_MAX =%d !\n", __func__ , index , ISO_PWM_MAX );
        printf("%s %d  but MISC_PWM_MAX =%d !\n",  __func__ , index , MISC_PWM_MAX );
        return;
    }

    pwm_ctrl_reg.regValue = rtd_inl(registerAddress);

    if (polarity == 0 || polarity == 1){
        if( PWM_ISO == type)
            pwm_ctrl_reg.i_pwm0l = polarity;
        else
            pwm_ctrl_reg.m_pwm0l = polarity;

        rtd_outl( registerAddress , pwm_ctrl_reg.regValue);
    }
    else
        printf("set pwm polarity=%d invalid!!\n",polarity);

    return;
}


/*
 * On other platforms we are using this function, but not sure at this point why is required, we call it with 0x00.
 * Keep it as it is if not required in your platform.
 */
void PAL_PM_PWM_Div(u_char u8DivPWM)
{
    return;
}

int set_pwm_duty_roku(RTK_PWM_TYPE type, int index, int duty)
{
    pwm_duty_set_RBUS pwm_duty_set_reg;
    int totalcnt = 0;
    unsigned int registerAddress = 0;

    if( duty > 101  || duty < -1 ){
        printf("set pwm duty=%d  failed!!\n\n", duty );
        return -1;
    }else
        printf("set pwm duty=%d  success!!\n\n", duty );


    if( PWM_ISO == type  && index < ISO_PWM_MAX ) {
        registerAddress = pwm_iso_phy[index].p_reg_map->ISO_DUTY_SET;
    } else if( PWM_MISC == type && index < MISC_PWM_MAX ) {
        registerAddress = pwm_mis_phy[index].p_reg_map->MISC_DUTY_SET;
    }else{
        printf("%s parameter fail register address 0x%X !\n",  __func__ , registerAddress );
        printf("%s %d  but  ISO_PWM_MAX =%d !\n", __func__ , index , ISO_PWM_MAX );
        printf("%s %d  but MISC_PWM_MAX =%d !\n",  __func__ , index , MISC_PWM_MAX );
        return -1;
    }

    pwm_duty_set_reg.regValue = rtd_inl( registerAddress );
    if( PWM_ISO == type){
        totalcnt = pwm_duty_set_reg.i_pwm0_totalcnt;
        pwm_duty_set_reg.i_pwm0_dut=totalcnt*duty/100;
    }else{
        totalcnt = pwm_duty_set_reg.m_pwm0_totalcnt;
        pwm_duty_set_reg.m_pwm0_dut=totalcnt*duty/100;
    }
    rtd_outl( registerAddress , pwm_duty_set_reg.regValue);

    return 0;
}

int get_pwm_duty_base_change(int duty,int duty_max,int totalcnt)
{
    return (duty*totalcnt)/duty_max;
}

RTK_PWM_TYPE getPWMPinType(unsigned long long val)
{
    RTK_PWM_TYPE type = 0;
    PCB_PIN_TYPE_T pin_type;
    pin_type = GET_PIN_TYPE(val);
    if (pin_type == PCB_PIN_TYPE_PWM)
        type = PWM_MISC;
    else if (pin_type == PCB_PIN_TYPE_ISO_PWM)
        type = PWM_ISO;
    return type;
}

