/*
* SAMSUNG EXYNOS5 USB HOST XHCI Controller
*
* Copyright (C) 2012 Samsung Electronics Co.Ltd
*  Vivek Gautam <gautam.vivek@samsung.com>
*  Vikas Sajjan <vikas.sajjan@samsung.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
* MA 02110-1301 USA
*/

/*
* This file is a conglomeration for DWC3-init sequence and further
* exynos5 specific PHY-init sequence.
*/

#include <common.h>
#include <malloc.h>
#include <usb.h>
#include <watchdog.h>
#include <asm-generic/errno.h>
#include <linux/compat.h>
#include <rbus/sys_reg_reg.h>
#include <rbus/usb3_top_reg.h>
#include "xhci.h"
#include "xhci_reg.h"
#include "xhci-plat-config.h"

#define XHCI_REG_BASE   HMAC_HOST_xhci_ctrl_start_reg
#define USB_PHY_DELAY   100

/*#define xhci_dbg(fmt, args...)           printf("[xhci] dbg, " fmt, ## args)*/
#define xhci_dbg(fmt, args...)
#define xhci_info(fmt, args...)          printf("[xhci] info, " fmt, ## args)
#define xhci_warn(fmt, args...)          printf("[xhci] warn, " fmt, ## args)


#define ScalerTimer_DelayXms(x) udelay(x*1000)

//-----------------------------------------------------------------
//  U2 Phy Setting API
//-----------------------------------------------------------------*/

#if 0
static const unsigned long usb2_phy_config_reg[] =
{
    HMAC_HOST_xhci_USB2PHYCFG0,
    HMAC_HOST_xhci_USB2PHYCFG1,
    HMAC_HOST_xhci_USB2PHYCFG2,
};
#endif
static const unsigned long usb2_phy_vendor_ctrl_reg[] =
{
    HMAC_HOST_xhci_GUSB2PHYACC0,
    HMAC_HOST_xhci_GUSB2PHYACC1,
    HMAC_HOST_xhci_GUSB2PHYACC2,
};
#define VSTS_DONE           (1 << 24)
#define IS_VSTS_DONE(p)     (p & VSTS_DONE)

/*-----------------------------------------------
* Func : _get_usb2_phy
*
* Desc : read usb2 phy register
*
* Parm : port1     : port index (1~3)
*        phy_reg   : phy register address
*
* Retn : others : phy register value
*        -1 : failed
*-----------------------------------------------*/
static int _get_usb2_phy(
    int                     port1,
    int                     phy_reg
    )
{
    int phy_value = 0;
    int addr = 0;

    if (port1 > 4 || port1 < 1)
    {
        xhci_warn("get_usb2_phy failed, invalid port %d\n", port1);
        return -1;
    }

    if (phy_reg < 0xe0 || phy_reg > 0xf7)
    {
        xhci_warn("get_usb2_phy failed, invalid phy_reg 0x%x\n", phy_reg);
        return -1;
    }

    addr = usb2_phy_vendor_ctrl_reg[port1-1];

    do {
        // write register
        rtd_outl(addr, 0x02000000 | (phy_reg & 0x0f) << 8);     // write phy reg bits[3:0]
        udelay(USB_PHY_DELAY);
        rtd_outl(addr, 0x02000000 | (phy_reg & 0xd0) << 4);     // write phy reg bits[7:4]
        udelay(USB_PHY_DELAY);

        // read out register
        phy_value = rtd_inl(addr);
    } while (!IS_VSTS_DONE(phy_value));

    phy_value &= 0xff;

    xhci_dbg("get_usb2_phy successed, port %d phy_reg 0x%.2x = 0x%x\n",
        port1, phy_reg, phy_value);

    return phy_value;
}


/*-----------------------------------------------
* Func : _set_usb2_phy
*
* Desc : set usb2 phy register
*
* Parm : port1     : port index (1~3)
*        phy_reg   : phy register address
*        phy_value : phy register value
*
* Retn : 0 : successed,
*       -1 : failed
*-----------------------------------------------*/
static int _set_usb2_phy(
    int             port1,
    int             phy_reg,
    int             phy_value
    )
{
    unsigned long addr= 0;

    if (port1 > 4 || port1 < 1)
    {
        xhci_warn("set_usb2_phy failed, invalid port - %d\n", port1);
        return -1;
    }

    if (phy_reg < 0xe0 || phy_reg > 0xf7)
    {
        xhci_warn("set_usb2_phy failed, invalid phy_reg 0x%x\n", phy_reg);
        return -1;
    }

    addr = usb2_phy_vendor_ctrl_reg[port1 -1];

    do {
        // write value
        rtd_outl(USB3_TOP_VSTATUS0_OUT_reg, phy_value << ((port1 - 1) * 8));
        xhci_dbg("[u2_phy] rtd_outl(0x%08x,0x%08x)\n",
                    USB3_TOP_VSTATUS0_OUT_reg, phy_value << ((port1 - 1) * 8));

        // write address
        rtd_outl(addr, 0x02000000 | (phy_reg & 0x0f) << 8);     // Write High
        xhci_dbg("[u2_phy] rtd_outl(0x%08x,0x%08x)\n", addr, 0x02000000 | (phy_reg & 0x0f) << 8);
        rtd_outl(addr, 0x02000000 | (phy_reg & 0xf0) << 4);
        xhci_dbg("[u2_phy] rtd_outl(0x%08x,0x%08x)\n", addr, 0x02000000 | (phy_reg & 0x0f) << 4);
        udelay(USB_PHY_DELAY);

    } while (!IS_VSTS_DONE(rtd_inl(addr)));

    return 0;
}



/*-----------------------------------------------
* Func : _usb2_load_phy_setting
*
* Desc : load usb2 phy setting
*
* Parm : p_reg : register list
*        n_reg : register value
*
* Retn : N/A
*-----------------------------------------------*/
static int _usb2_load_phy_setting(
    U2_PHY_REGISTER*    p_reg,
    unsigned int        n_reg
    )
{
    int err = 0;

    while(n_reg--)
    {
        _set_usb2_phy(p_reg->port1, p_reg->addr, p_reg->val);
        if (err)
            break;
        p_reg++;
    }

	return err;
}



/*-----------------------------------------------
* Func : p3_mode_control
*
* Desc : Enable/Disable p3 mode
*
* Parm : enable : Enable = 1 / Disable = 0
*
* Retn : N/A
*-----------------------------------------------*/
void p3_mode_control(int enable)
{
    if (enable)
        rtd_maskl(0xb80582c0, ~(1<<17), 1<<17);
    else
        rtd_maskl(0xb80582c0, ~(1<<17), 0);
}



/*-----------------------------------------------
* Func : USB_CLK_OnOff *
*
* Desc : On/Off USB Clock
*
* Parm : clken :  clock enable = 1 / clock disable = 0
*        private_data : pointer of data
*
* Retn : N/A
*-----------------------------------------------*/
int USB_CLK_OnOff(enum CRT_CLK clken, void *private_data)
{
    xhci_info("--------------- [%s] start ---------------\n", __func__);
#define USB_CRT_DELAY   100
    if (clken == CLK_ON)
    {
        xhci_info("clken on \n");
        /*---------------------------------------------------
        * shutdown sequence :
        *  1. mac clock off
        *  2. mac reset off
        *  3. phy reset off
        *---------------------------------------------------*/

        // clock off : USB MAC
        rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                    SYS_REG_SYS_CLKEN2_clken_usb3_mask |
                    SYS_REG_SYS_CLKEN2_write_data(0));

        udelay(USB_CRT_DELAY);

        // Reset MAC
        rtd_outl(SYS_REG_SYS_SRST2_reg,
                    SYS_REG_SYS_SRST2_rstn_usb3_mask |
                    SYS_REG_SYS_SRST2_write_data(0));

        udelay(USB_CRT_DELAY);

        // Reset PHY
        rtd_outl(SYS_REG_SYS_SRST3_reg,
                    SYS_REG_SYS_SRST3_rstn_usb2_phy_p0_mask |
                    SYS_REG_SYS_SRST3_rstn_usb2_phy_p1_mask |
                    SYS_REG_SYS_SRST3_rstn_usb2_phy_p2_mask |
                    SYS_REG_SYS_SRST3_rstn_usb3_phy_p0_mask |
                    SYS_REG_SYS_SRST3_rstn_usb3_mdio_p0_mask |
                    SYS_REG_SYS_SRST3_write_data(0));

        udelay(USB_CRT_DELAY);

        /*---------------------------------------------------
        * init sequence :
        *  1. phy reset release
        *  2. mac clock on
        *  3. mac clock off
        *  4. mac reset on
        *  5. mac clock on
        *---------------------------------------------------*/
        // Release USB Phy Reset : USB Phy Reset Ready
        rtd_outl(SYS_REG_SYS_SRST3_reg,
                    SYS_REG_SYS_SRST3_rstn_usb2_phy_p0_mask |
                    SYS_REG_SYS_SRST3_rstn_usb2_phy_p1_mask |
                    SYS_REG_SYS_SRST3_rstn_usb2_phy_p2_mask |
                    SYS_REG_SYS_SRST3_rstn_usb3_phy_p0_mask |
                    SYS_REG_SYS_SRST3_rstn_usb3_mdio_p0_mask |
                    SYS_REG_SYS_SRST3_write_data(1));

        udelay(USB_CRT_DELAY);

        // clock on : MAC
        rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                    SYS_REG_SYS_CLKEN2_clken_usb3_mask |
                    SYS_REG_SYS_CLKEN2_write_data(1));

        udelay(USB_CRT_DELAY);

        // clock off : MAC
        rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                    SYS_REG_SYS_CLKEN2_clken_usb3_mask |
                    SYS_REG_SYS_CLKEN2_write_data(0));

        udelay(USB_CRT_DELAY);

        // Relase USB MAC Reset: USB MAC Reset Ready
        rtd_outl(SYS_REG_SYS_SRST2_reg,
                    SYS_REG_SYS_SRST2_rstn_usb3_mask |
                    SYS_REG_SYS_SRST2_write_data(1));

        udelay(USB_CRT_DELAY);

        // clock on : MAC
        rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                    SYS_REG_SYS_CLKEN2_clken_usb3_mask |
                    SYS_REG_SYS_CLKEN2_write_data(1));
        udelay(USB_CRT_DELAY);

    } else {

        /*---------------------------------------------------
        * shutdown sequence :
        *  1. mac clock off
        *  2. mac reset off
        *  3. phy reset off
        *---------------------------------------------------*/
        // clock off : USB MAC
        rtd_outl(SYS_REG_SYS_CLKEN2_reg,
                    SYS_REG_SYS_CLKEN2_clken_usb3_mask |
                    SYS_REG_SYS_CLKEN2_write_data(0));

        udelay(USB_CRT_DELAY);

        // Reset MAC
        rtd_outl(SYS_REG_SYS_SRST2_reg,
                    SYS_REG_SYS_SRST2_rstn_usb3_mask |
                    SYS_REG_SYS_SRST2_write_data(0));

        udelay(USB_CRT_DELAY);

        // Reset PHY
        rtd_outl(SYS_REG_SYS_SRST3_reg,
                    SYS_REG_SYS_SRST3_rstn_usb2_phy_p0_mask |
                    SYS_REG_SYS_SRST3_rstn_usb2_phy_p1_mask |
                    SYS_REG_SYS_SRST3_rstn_usb2_phy_p2_mask |
                    SYS_REG_SYS_SRST3_rstn_usb3_phy_p0_mask |
                    SYS_REG_SYS_SRST3_rstn_usb3_mdio_p0_mask |
                    SYS_REG_SYS_SRST3_write_data(0));

        udelay(USB_CRT_DELAY);
    }
    xhci_info("--------------- [%s] complete ---------------\n", __func__);
    return CRT_SUCCESS;
}


static int xhci_usb2_load_phy_params(void)
{
    xhci_info("--------------- [%s] ---------------\n", __func__);
    return _usb2_load_phy_setting(u2_phy_config.p_reg_table, u2_phy_config.n_reg);
}

int get_xhci_usb2_phy_reg(unsigned char phy, unsigned char page, unsigned char addr)
{
    int ret;

    if (phy < 1 || phy > 3) {
        xhci_warn("set_u2_phy_reg failed, invalid phy id - %d\n", phy);
        return -1;
    }

    if (page > 3) {
        xhci_warn("set_u2_phy_reg failed, invalid page - %d\n", page);
        return -1;
    }

    _set_usb2_phy(phy, 0xF4, (page<<5) | 0x9b);  // bit [6:5] : page select = page
    ret = _get_usb2_phy(phy, addr);
    _set_usb2_phy(phy, 0xF4, 0x9b);  // bit [6:5] : page select = 0

    return ret;
}


int set_xhci_usb2_phy_reg(unsigned char phy, unsigned char page, unsigned char addr, unsigned char val)
{
    if (phy < 1 || phy > 3) {
        xhci_warn("set_u2_phy_reg failed, invalid phy id - %d\n", phy);
        return -1;
    }

    if (page > 3) {
        xhci_warn("set_u2_phy_reg failed, invalid page - %d\n", page);
        return -1;
    }

    _set_usb2_phy(phy, 0xF4, (page<<5) | 0x9b);  // bit [6:5] : page select = page
    _set_usb2_phy(phy, addr, val);
    _set_usb2_phy(phy, 0xF4, 0x9b);  // bit [6:5] : page select = 0
    return 0;
}


static int xhci_usb2_set_squelch_from_otp(void)
{
#define GET_PORT_OFFSET(port)              (8 + (port * 4))
#define GET_PORT_OTP_VAL(val, port)        ((val & (0xf << GET_PORT_OFFSET(port))) >> GET_PORT_OFFSET(port))
#ifndef EFUSE_DATAO_182_reg
#define EFUSE_DATAO_182_reg 0xB80382D8
#endif
    int portnum = 0, ret = 0;
    unsigned int squelch_otp = rtd_inl(EFUSE_DATAO_182_reg);
    unsigned int val;

    xhci_info("squelch_otp=0x%x, \n", squelch_otp);

    //[19:16]=port3, [15:12]=port2, [11:8]=port1, [7]=done
    if (squelch_otp & (0x1 << 7)) {
        for (portnum = 0; portnum < U2_PHY_CONFIG_COUNT; portnum++) {
            val = get_xhci_usb2_phy_reg(portnum + 1, 0, 0xe7);
            val &= ~0xf0;
            val |= (GET_PORT_OTP_VAL(squelch_otp, portnum) << 4);
            ret = set_xhci_usb2_phy_reg(portnum + 1, 0, 0xe7, val);
            if (ret)
                goto out;
            xhci_info("set port(%d) squelch level to 0x%x \n",
                    (portnum + 1),
                    (get_xhci_usb2_phy_reg(portnum + 1, 0, 0xe7) & 0xf0) >> 4);
        }
    }

out:
    return ret;
#undef GET_PORT_OFFSET
#undef GET_PORT_OTP_VAL
}


static int xhci_usb2_set_disconnect_from_otp(void)
{
#define GET_PORT_OFFSET(port)              (16 + (port * 4))
#define GET_PORT_OTP_VAL(val, port)        ((val & (0xf << GET_PORT_OFFSET(port))) >> GET_PORT_OFFSET(port))
#ifndef EFUSE_DATAO_200_reg
#define EFUSE_DATAO_200_reg 0xB8038320
#endif
    int portnum = 0, ret = 0;
    unsigned int disconnect_otp = rtd_inl(EFUSE_DATAO_200_reg);
    unsigned int val;

    xhci_info("disconnect_otp=0x%x \n", disconnect_otp);

    //[27:24]=port3, [23:20]=port2, [19:16]=port1
    if (disconnect_otp & 0x1) {
        for (portnum = 0; portnum < U2_PHY_CONFIG_COUNT; portnum++) {
            val = get_xhci_usb2_phy_reg(portnum + 1, 0, 0xe4);
            val &= ~0xf0;
            val |= (GET_PORT_OTP_VAL(disconnect_otp, portnum) << 4);
            ret = set_xhci_usb2_phy_reg(portnum + 1, 0, 0xe4, val);
            if (ret)
                goto out;
            xhci_info("set port(%d) disconect level to 0x%x \n",
                    (portnum + 1),
                    (get_xhci_usb2_phy_reg(portnum + 1, 0, 0xe4) & 0xf0) >> 4);
        }
    }

out:
    return ret;
#undef GET_PORT_OFFSET
#undef GET_PORT_OTP_VAL
}


static inline int xhci_usb2_set_phy_from_otp(void)
{
    return (xhci_usb2_set_disconnect_from_otp() ||
            xhci_usb2_set_squelch_from_otp());
}

//-----------------------------------------------------------------
//  MDIO API
//-----------------------------------------------------------------


/*-----------------------------------------------
* Func : mdio_write
*
* Desc : write mdio control
*
* Parm : phy_addr : phy register
*        reg_addr : register addr
*        val : mdio control value
*
* Retn : N/A
*-----------------------------------------------*/
static void mdio_write(
    unsigned char           phy_addr,
    unsigned char           reg_addr,
    unsigned short          val
    )
{
    xhci_dbg("mdio_wr(pg_%d, reg_%02x)=0x%04x\n", phy_addr, reg_addr, val);

    rtd_outl(USB3_TOP_MDIO_CTR_reg,
                USB3_TOP_MDIO_CTR_data(val) |
                USB3_TOP_MDIO_CTR_phy_addr(phy_addr) |
                USB3_TOP_MDIO_CTR_phy_reg_addr(reg_addr) |
                USB3_TOP_MDIO_CTR_mclk_rate(3) |      // clk_sys/4
                USB3_TOP_MDIO_CTR_mdio_rdwr(1));

    do {
        udelay(USB_PHY_DELAY);
    } while((rtd_inl(USB3_TOP_MDIO_CTR_reg) & USB3_TOP_MDIO_CTR_mdio_busy_mask)); // wait busy done
}

/*-----------------------------------------------
* Func : mdio_read
*
* Desc : read mdio control
*
* Parm : phy_addr : phy register
*        reg_addr : register addr
*
* Retn : N/A
*-----------------------------------------------*/
static unsigned short  __attribute__((unused)) mdio_read(
    unsigned char       phy_addr,
    unsigned char       reg_addr
    )
{
    unsigned long ctrl = USB3_TOP_MDIO_CTR_phy_addr(phy_addr) |
                        USB3_TOP_MDIO_CTR_phy_reg_addr(reg_addr) |
                        USB3_TOP_MDIO_CTR_mclk_rate(3) |
                        USB3_TOP_MDIO_CTR_mdio_rdwr(0);

    //pr_debug("mdio_ctrl=%08lx\n", ctrl);
    rtd_outl(USB3_TOP_MDIO_CTR_reg, ctrl);
    do {
        udelay(USB_PHY_DELAY);
    } while((rtd_inl(USB3_TOP_MDIO_CTR_reg) & USB3_TOP_MDIO_CTR_mdio_busy_mask)); // wait busy done
    return rtd_inl(USB3_TOP_MDIO_CTR_reg) >> 16;
}

/*-----------------------------------------------
* Func : xhci_usb3_phy_init
*
* Desc : init usb 3 phy
*
* Parm : N/A
*
* Retn : N/A
*-----------------------------------------------*/
void xhci_usb3_phy_init(void)
{
    //int i;

    xhci_info("--------------- [%s] start ---------------\n", __func__);

    // USB3_ANA_PHY open
    rtd_outl(USB3_TOP_ANA_PHY0_reg, USB3_TOP_ANA_PHY0_usb3_mbias_en_p0(1) |
                                    USB3_TOP_ANA_PHY0_usb3_bg_en_p0(1) |
                                    USB3_TOP_ANA_PHY0_usb3_pow_gating_p0(1));
    udelay(USB_PHY_DELAY);

    udelay(400);

        //u3Rxterm_force_val=0,//Rxterm_force_en=0
    rtd_outl(USB3_TOP_USB_TMP_2_reg, USB3_TOP_USB_TMP_2_usb3_eco_option1(1));
    rtd_inl(USB3_TOP_USB_TMP_2_reg);

    xhci_info("--------------- [%s] complete ---------------\n", __func__);
}

static void xhci_usb3_load_phy_params(void) {
    int i;

    xhci_info("--------------- [%s] start ---------------\n", __func__);

	rtd_outl(USB3_TOP_USB3_PHY_PIPE_reg, USB3_TOP_USB3_PHY_PIPE_reset_pipe3_p0(1));  // pipe reset u3 phy
		udelay(400);

    for (i = 0; i < U3_PHY_SETTING_SIZE; i++)
    {
        mdio_write(xhci_u3_phy_config[i].page,
                xhci_u3_phy_config[i].reg,
                xhci_u3_phy_config[i].val);
    }

	rtd_outl(USB3_TOP_USB3_PHY_PIPE_reg, USB3_TOP_USB3_PHY_PIPE_reset_pipe3_p0(0));  // pipe release u3 phy
    udelay(400);

    // Toggle. DC CALIBARION  "1 --> 0 --> 1". will do one time of calibration
    mdio_write(0x0, 0x09, 0x367c);
    mdio_write(0x0, 0x09, 0x347c);
    mdio_write(0x0, 0x09, 0x367c);

	
    xhci_info("--------------- [%s] complete ---------------\n", __func__);
}

/*-----------------------------------------------
* Func : xhci_usb_mac_init
*
* Desc : init usb mac
*
* Parm : N/A
*
* Retn : N/A
*-----------------------------------------------*/
void xhci_usb_mac_init(void)
{
    xhci_info("--------------- [%s] start ---------------\n", __func__);
    // UTMI reset of U2 P0~P2
    rtd_outl(USB3_TOP_USB2_PHY_UTMI_reg,
                USB3_TOP_USB2_PHY_UTMI_reset_utmi_p0_mask |
                USB3_TOP_USB2_PHY_UTMI_reset_utmi_p1_mask |
                USB3_TOP_USB2_PHY_UTMI_reset_utmi_p2_mask);
    udelay(USB_PHY_DELAY);

    // UTMI release reset
    rtd_outl(USB3_TOP_USB2_PHY_UTMI_reg, 0);
    udelay(USB_PHY_DELAY);

    //USB3.0 avoid no UTMI clock
    rtd_outl(USB3_TOP_USB_TMP_1_reg, USB3_TOP_USB_TMP_1_test_reg_1(1));
    udelay(USB_PHY_DELAY);

    // disable Dbus Multi-request. Do not mark this!!!
    xhci_dbg("disable multi request, bit[1]\n");
    rtd_outl(USB3_TOP_WRAP_CTR_reg , rtd_inl(USB3_TOP_WRAP_CTR_reg) | (0x1 << 1));


    // Disable auto-suspend
    //2. set the Synopsys data book GUSB2PHYCFG[6]=0 Suspend USB2.0 HS/FS/LS PHY (SusPHY)
    rtd_outl(HMAC_HOST_xhci_USB2PHYCFG0, 0x00001408);
    rtd_outl(HMAC_HOST_xhci_USB2PHYCFG1, 0x00001408);
    rtd_outl(HMAC_HOST_xhci_USB2PHYCFG2, 0x00001408);
    rtd_outl(HMAC_HOST_xhci_USB2PHYCFG3, 0x00001408);
    udelay(USB_PHY_DELAY);

    /////////////////usb 3.0 p3 mode -> p2 mode/////
    rtd_outl(0xb80582c0,0x010C0002);    //P3 mode->p2 mode

    xhci_info("--------------- [%s] finish ---------------\n", __func__);
}


static void enable_vbus_power(void)
{
#ifdef Config_Board_TV006
    // USB_CTL2(MISC_GPIO_BIT_49), USB_CTL3(MISC_GPIO_BIT_48)
    rtd_setbits(0xb8000838, (0xf << 28) | (0xf << 20));  // set pin share to MISC GPIO
    rtd_setbits(0xb801bd1c, (0x3 << 16));                // set GPO value
    rtd_setbits(0xb801bd04, (0x3 << 16));                // set IO direction to GPO
#endif
}

int xhci_hcd_init(int index, struct xhci_hccr **hccr, struct xhci_hcor **hcor)
{
    xhci_info("--------------- [%s] start ---------------\n", __func__);

    USB_CLK_OnOff(CLK_ON, NULL);
    //-------------------------------------------------------------------
    // Do initial
    //-------------------------------------------------------------------
    xhci_usb_mac_init();            // init USB mac
    udelay(USB_PHY_DELAY);

    xhci_usb3_load_phy_params();
    udelay(USB_PHY_DELAY);

    xhci_usb3_phy_init();           // init U3 Phy
    udelay(USB_PHY_DELAY);

    xhci_usb2_load_phy_params();
    udelay(USB_PHY_DELAY);

    xhci_usb2_set_phy_from_otp();
    udelay(USB_PHY_DELAY);

    ////////////let usb 3.0 R Term turn on////////////
    rtd_outl(0xb805A088, 0x5);        //Rxterm_force_en=1
    rtd_outl(0xb80582c0, 0x010E0002);  //P2 mode->p3 mode
    rtd_outl(0xb8050020, 0x00000002);
    rtd_outl(0xb8050058, 0x00000040);    //set support device no.
    rtd_outl(0xb8050020, 0x00000001);    //XHCI run
    udelay(USB_PHY_DELAY);

    //-------------------------------------------------------------------
    // Setup XHCI register space
    //-------------------------------------------------------------------
    *hccr = (struct xhci_hccr *) XHCI_REG_BASE;                 // XHCI Capability registper space
    *hcor = (struct xhci_hcor *)((uint32_t) *hccr               // XHCI Operational registper space
                + HC_LENGTH(xhci_readl(&(*hccr)->cr_capbase)));

    xhci_dbg("xhci: init hccr %x and hcor %x hc_length %d\n",
        (uint32_t)*hccr, (uint32_t)*hcor,
        (uint32_t)HC_LENGTH(xhci_readl(&(*hccr)->cr_capbase)));

    enable_vbus_power(); /* Enable USB vbus power */

    xhci_info("mdelay(100) for HW ready......\n");
    mdelay(1);
    return 0;
}


void xhci_hcd_stop(int index)
{
    return;
}


